

#include "fld_conf_id.h"


#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/mkdev.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <limits.h>



#if FLD_OS_NAME_V_IRIX
#       if FLD_OS_VERSION_V_6_5
#               if FLD_OS_RELEASE_V_30


// NOTE: POSIX Signals
//
//  In IRIX, 3 different signal handling semantics
//  are supported:
//
//   1. System V
//   2. BSD
//   3. POSIX
//
//  In fld, POSIX semantics are used because it allows
//  a signal handler to be defined just once and is
//  highly portable.


#include <signal.h>


#		endif	// FLD_OS_RELEASE_V_30
#	endif	// FLD_OS_VERSION_V_6_5
#endif	// FLD_OS_NAME_V_IRIX


#include <errno.h>
#include <string.h>


#include "fld_common.h"
#include "fld.h"


// NOTE: Macro definitions


#define FLD_DAEMON_NAME "fld"


// NOTE: Global variables


// I/O transfer buffer
static char *io_data;


// I/O transfer size
static fld_uint64_t io_data_size;


// Kernel page shift
static fld_uint32_t page_shift;


// Kernel maxdmasz
static fld_int32_t fld_maxdmasz;


// NOTE: Maximum transfer size
//
//  The maximum transfer size depends on the kernel
//  and must be calculated in runtime.


// Maximum transfer size per I/O (in bytes)
static fld_uint32_t FLD_MAX_TRANSFER_SIZE;


// I/O request structure
static struct fld_io_req_s io_req;


// Daemon started switch
static fld_uint32_t daemon_started;


// NOTE: Private function prototypes


void fld_ioreq_print(struct fld_io_req_s *);
int fld_dir_clean(void);
int fld_io_open(char *,
	        fld_int32_t,
	        mode_t,
	        fld_uint32_t,
	        fld_uint32_t,
	        fld_int32_t *,
	        fld_int32_t *);
int fld_io_lseek(fld_int32_t,
		 fld_uint64_t,
		 fld_int32_t,
	         fld_uint32_t,
	         fld_uint32_t,
		 fld_int32_t *);
int fld_io_read(fld_int32_t,
	        char *,
	        fld_uint64_t,
                fld_uint32_t,
                fld_uint32_t,
	        fld_uint64_t *,
	        fld_int32_t *);
int fld_io_write(fld_int32_t,
		 char *,
		 fld_uint64_t,
		 fld_uint32_t,
		 fld_uint32_t,
        	 fld_uint64_t *,
		 fld_int32_t *);
int fld_io_close(fld_int32_t,
	         fld_uint32_t,
	         fld_uint32_t,
		 fld_int32_t *);
int fld_io_transfer(char *,
		    fld_int32_t,
		    mode_t,
		    enum fld_ioctl_e,
		    void *,
		    fld_uint32_t *,
		    fld_int32_t *,
		    fld_int32_t *);
int fld_io_alloc_transfer_buffer(void);
int fld_io_process(void);
int fld_io_start(void);
void fld_sigurg_handler(int,
                        int,
                        struct sigcontext *);
void fld_sighup_handler(int,
                        int,
                        struct sigcontext *);


// NOTE: Private functions


void fld_ioreq_print(struct fld_io_req_s *io_req)	// -- IN
{
        fld_uint32_t control_index;
        char *fld_io_string[] =
        {
                "FLD_IO_NULL",
                "FLD_IO_READ",
                "FLD_IO_READ_OK",
                "FLD_IO_READ_INCMPL",
                "FLD_IO_READ_EOF",
                "FLD_IO_READ_NOK",
                "FLD_IO_WRITE",
                "FLD_IO_WRITE_OK",
                "FLD_IO_WRITE_INCMPL",
                "FLD_IO_WRITE_EOF",
                "FLD_IO_WRITE_NOK",
                "FLD_IO_ERROR"
        };


        // NOTE: Arguments


        // IN: Check for I/O request
        if (io_req == NULL)
        {
                return; // ERROR: No I/O request
        }


	// NOTE: Processing


        // Get control field string
        control_index = 0;
        switch (io_req->control)
        {
                case FLD_IO_NULL:

                        control_index = 0;
                        break;

                case FLD_IO_READ:

                        control_index = 1;
                        break;

                case FLD_IO_READ_OK:

                        control_index = 2;
                        break;

                case FLD_IO_READ_INCMPL:

                        control_index = 3;
                        break;

                case FLD_IO_READ_EOF:

                        control_index = 4;
                        break;

                case FLD_IO_READ_NOK:

                        control_index = 5;
                        break;

                case FLD_IO_WRITE:

                        control_index = 6;
                        break;

                case FLD_IO_WRITE_OK:

                        control_index = 7;
                        break;

                case FLD_IO_WRITE_INCMPL:

                        control_index = 8;
                        break;

                case FLD_IO_WRITE_EOF:

                        control_index = 9;
                        break;

                case FLD_IO_WRITE_NOK:

                        control_index = 10;
                        break;

                default:

                        control_index = 11;
                        break;
        }


        FLD_UPRINTF(("[fld_ioreq_print] control = %s\n", fld_io_string[control_index]))
        FLD_UPRINTF(("[fld_ioreq_print] offset = %llu (bytes)\n", io_req->offset))
        FLD_UPRINTF(("[fld_ioreq_print] size = %llu (bytes)\n", io_req->size))
        FLD_UPRINTF(("[fld_ioreq_print] dev = %u\n", io_req->dev))
        FLD_UPRINTF(("[fld_ioreq_print] |\n"))
        FLD_UPRINTF(("[fld_ioreq_print] +- major = %u\n", major(io_req->dev)))
        FLD_UPRINTF(("[fld_ioreq_print]  |\n"))
        FLD_UPRINTF(("[fld_ioreq_print]  +- minor = %u\n", minor(io_req->dev)))
        FLD_UPRINTF(("[fld_ioreq_print] file = \"%s\"\n", io_req->file))
        FLD_UPRINTF(("[fld_ioreq_print] data = 0x%p\n", io_req->data))
        FLD_UPRINTF(("[fld_ioreq_print] data_size = %llu\n", io_req->data_size))
}


int fld_dir_clean(void)
{
	DIR *dev_loop;
	char *entry_name;
	char part_name[PATH_MAX];
	char del_path[PATH_MAX];
	dirent_t *dir_entry;
	struct stat stats_path;


	// NOTE: Initialization


	// Set /dev/loop/ delete path prefix
	del_path[0] = '/';
	del_path[1] = 'd';
	del_path[2] = 'e';
	del_path[3] = 'v';
	del_path[4] = '/';
	del_path[5] = 'l';
	del_path[6] = 'o';
	del_path[7] = 'o';
	del_path[8] = 'p';
	del_path[9] = '/';
	del_path[10] = '\0';


	// NOTE: Processing


	// Open /dev/loop
	dev_loop = opendir("/dev/loop");


	if (dev_loop == NULL)
	{
		// Write to console (ERROR)
		syslog(LOG_ERR,
		       "ERROR: Couldn't open /dev/loop directory\n");


		return 0;	// ERROR: Couldn't open /dev/loop directory
	}


	// Loop through /dev/loop entries
	dir_entry = readdir(dev_loop);
	while (dir_entry != NULL)
	{
		// Get entry name
		entry_name = dir_entry->d_name;


		// Skip . and ..
		if ((entry_name[0] == '.') &&
		    ((entry_name[1] == '\0') ||
		     ((entry_name[1] == '.') &&
		      (entry_name[2] == '\0'))))
		{
		}
		else
		{
			FLD_DEBUG_UPRINTF(("[fld_dir_clean] dir_entry->d_name = \"%s\"\n", entry_name))


			// Terminate directory path
			del_path[10] = '\0';


			// Check maximum length
			if (strlen(entry_name) > (PATH_MAX - 10 - 1))
			{
				// Get partial entry name
				strcat(part_name, del_path);
				strncat(part_name, entry_name, 10);


				// Write to console (WARNING)
				syslog(LOG_WARNING,
				       "WARNING: Directory entry name overflow %s\n",
				       part_name);
			}
			else
			{
				// Add to delete path
				strcat(del_path, entry_name);


				FLD_DEBUG_UPRINTF(("[fld_dir_clean] del_path = \"%s\"\n", del_path))


				// Get stats of delete path
				if (lstat(del_path, &stats_path) != -1)
				{
					// Check for directory entry
					if (S_ISDIR(stats_path.st_mode) != 1)
					{
						FLD_DEBUG_UPRINTF(("[fld_dir_clean] Delete entry = \"%s\"\n", del_path))


						/*
						// Delete entry
						if (unlink(del_path) == -1)
						{
							// Write to console (WARNING)
							syslog(LOG_WARNING,
							       "WARNING: Couldn't delete entry %s\n",
							       del_path);
						}
						*/
					}
					else
					{
						FLD_DEBUG_UPRINTF(("[fld_dir_clean] Skip directory entry = \"%s\"\n", del_path))
					}
				}
			}
		}


		// Get next entry
		dir_entry = readdir(dev_loop);
	}


	// Close /dev/loop directory
	closedir(dev_loop);


	return 1;	// SUCCESS
}


int fld_io_open(char *open_path,		// -- IN
	        fld_int32_t open_flags,		// -- IN
	        mode_t open_modes,		// -- IN
	        fld_uint32_t rep_num,		// -- IN
	        fld_uint32_t rep_delay,		// -- IN
	        fld_int32_t *open_fd,		// -- OUT
	        fld_int32_t *open_errno)	// -- OUT
{
	fld_int32_t file_desc;
	fld_uint32_t path_open;
	fld_uint32_t rep_index;


	// NOTE: Arguments


	// IN: Check for open path
	if (open_path == NULL)
	{
		return 0;	// ERROR: No open path
	}


	// OUT: Check for open file descriptor
	if (open_fd == NULL)
	{
		return -1;	// ERROR: No open file descriptor
	}


	// OUT: Check for open errno
	if (open_errno == NULL)
	{
		return -2;	// ERROR: No open errno
	}


	// NOTE: Processing


	// Check for repetition number
	if (rep_num == 0)
	{
		return 2;	// SUCCESS: No repetition
	}


	// Repeat open
	path_open = 0;
	for (rep_index = 0; rep_index < rep_num; rep_index++)
	{
		// Delay open
		if (rep_index > 0)
		{
			// Sleep for the requested number of seconds
			sleep(rep_delay);
		}


		// Open path
		file_desc = open(open_path,
				 open_flags,
				 open_modes);
		if (file_desc == -1)
		{
			// Write to console (WARNING)
		}
		else
		{
			// Set path open
			path_open = 1;


			// Break repetition loop
			break;
		}
	}


	// Check for open path
	if (path_open != 1)
	{
		return -3;	// ERROR: Couldn't open path
	}


	// Set open file descriptor
	*open_fd = file_desc;


	return 1;	// SUCCESS
}


int fld_io_lseek(fld_int32_t file_desc,		// -- IN
		 fld_uint64_t offset,		// -- IN
		 fld_int32_t whence,		// -- IN
	         fld_uint32_t rep_num,		// -- IN
	         fld_uint32_t rep_delay,	// -- IN
		 fld_int32_t *lseek_errno)	// -- OUT
{
	fld_uint32_t file_seek;
	fld_uint32_t rep_index;


	// NOTE: Arguments


	// IN: Check for file descriptor
	if (file_desc < 1)
	{
		return 0;	// ERROR: No file descriptor
	}


	// OUT: Check for lseek errno
	if (lseek_errno == NULL)
	{
		return -2;	// ERROR: No lseek errno
	}


	// NOTE: Processing


	// Check for repetition number
	if (rep_num == 0)
	{
		return 2;	// SUCCESS: No repetition
	}


	// Repeat lseek
	file_seek = 0;
	for (rep_index = 0; rep_index < rep_num; rep_index++)
	{
		// Delay lseek
		if (rep_index > 0)
		{
			// Sleep for the requested number of seconds
			sleep(rep_delay);
		}


		// Seek into file
		if (lseek(file_desc,
			  io_req.offset,
			  SEEK_SET) == -1)
		{
			// Set lseek errno
			*lseek_errno = errno;


			// Write to console (WARNING)
		}
		else
		{
			// Set file seek
			file_seek = 1;


			// Break repetition loop
			break;
		}
	}


	// Check for seek into file
	if (file_seek != 1)
	{
		return -2;	// ERROR: Couldn't seek into file
	}


	return 1;	// SUCCESS
}


int fld_io_read(fld_int32_t file_desc,		// -- IN
	        char *data,			// -- IN
	        fld_uint64_t size,		// -- IN
                fld_uint32_t rep_num,		// -- IN
                fld_uint32_t rep_delay,		// -- IN
	        fld_uint64_t *size_read,	// -- OUT
	        fld_int32_t *read_errno)	// -- OUT
{
	fld_uint32_t file_read;
	fld_uint32_t rep_index;
	fld_uint64_t bytes_read;


	// NOTE: Arguments


	// IN: Check for file descriptor
	if (file_desc < 1)
	{
		return 0;	// ERROR: No file descriptor
	}


	// IN: Check for data
	if (data == NULL)
	{
		return -1;	// ERROR: No data
	}


	// OUT: Check for size read
	if (size_read == NULL)
	{
		return -2;	// ERROR: No size read
	}


	// OUT: Check for read errno
	if (read_errno == NULL)
	{
		return -3;	// ERROR: No read errno
	}


	// NOTE: Processing


	// Check for read size
	if (size == 0)
	{
		// Set size read
		*size_read = 0;


		return 2;	// SUCCESS: No read size
	}


	// Check for repetition number
	if (rep_num == 0)
	{
		// Set size read
		*size_read = 0;


		return 3;	// SUCCESS: No repetition
	}


	// Repeat read
	file_read = 0;
	for (rep_index = 0; rep_index < rep_num; rep_index++)
	{
		// Delay read
		if (rep_index > 0)
		{
			// Sleep for the requested number of seconds
			sleep(rep_delay);
		}


		// Read file
		bytes_read = read(file_desc,
				  data,
				  size);
		if (bytes_read == -1)
		{
			// Set read errno
			*read_errno = errno;


			// Write to console (WARNING)
		}
		else
		{
			// Set file read
			file_read = 1;


			// Break repetition loop
			break;
		}
	}


	// Check for read from file
	if (file_read != 1)
	{
		// Set size read
		*size_read = 0;


		return -4;	// ERROR: Couldn't read file
	}


	// Set size read
	*size_read = bytes_read;


	return 1;	// SUCCESS
}


int fld_io_write(fld_int32_t file_desc,		// -- IN
		 char *data,			// -- IN
		 fld_uint64_t size,		// -- IN
		 fld_uint32_t rep_num,		// -- IN
		 fld_uint32_t rep_delay,	// -- IN
        	 fld_uint64_t *size_write,	// -- OUT
		 fld_int32_t *write_errno)	// -- OUT
{
	fld_uint32_t file_write;
	fld_uint32_t rep_index;
	fld_uint64_t bytes_write;


	// NOTE: Arguments


	// IN: Check for file descriptor
	if (file_desc < 1)
	{
		return 0;	// ERROR: No file descriptor
	}


	// IN: Check for data
	if (data == NULL)
	{
		return -1;	// ERROR: No data
	}


	// OUT: Check for size write
	if (size_write == NULL)
	{
		return -2;	// ERROR: No size write
	}


	// OUT: Check for write errno
	if (write_errno == NULL)
	{
		return -3;	// ERROR: No write errno
	}


	// NOTE: Processing


	// Check for write size
	if (size == 0)
	{
		// Set size write
		*size_write = 0;


		return 2;	// SUCCESS: No write size
	}


	// Check for repetition number
	if (rep_num == 0)
	{
		// Set size write
		*size_write = 0;


		return 3;	// SUCCESS: No repetition
	}


	// Repeat write
	file_write = 0;
	for (rep_index = 0; rep_index < rep_num; rep_index++)
	{
		// Delay write
		if (rep_index > 0)
		{
			// Sleep for the requested number of seconds
			sleep(rep_delay);
		}


		// Write file
		bytes_write = write(file_desc,
				    data,
				    size);
		if (bytes_write == -1)
		{
			// Set write errno
			*write_errno = errno;


			// Write to console (WARNING)
		}
		else
		{
			// Set file write
			file_write = 1;


			// Break repetition loop
			break;
		}
	}


	// Check for write from file
	if (file_write != 1)
	{
		// Set size write
		*size_write = 0;


		return -4;	// ERROR: Couldn't write file
	}


	// Set size write
	*size_write = bytes_write;


	return 1;	// SUCCESS
}


int fld_io_close(fld_int32_t file_desc,		// -- IN
	         fld_uint32_t rep_num,		// -- IN
	         fld_uint32_t rep_delay,	// -- IN
		 fld_int32_t *close_errno)	// -- OUT
{
	fld_uint32_t file_close;
	fld_uint32_t rep_index;


	// NOTE: Arguments


	// IN: Check for file descriptor
	if (file_desc < 1)
	{
		return 0;	// ERROR: No file descriptor
	}


	// OUT: Check for close errno
	if (close_errno == NULL)
	{
		return -1;	// ERROR: No close errno
	}


	// NOTE: Processing


	// Check for repetition number
	if (rep_num == 0)
	{
		return 2;	// SUCCESS: No repetition
	}


	// Repeat close
	file_close = 0;
	for (rep_index = 0; rep_index < rep_num; rep_index++)
	{
		// Delay close
		if (rep_index > 0)
		{
			// Sleep for the requested number of seconds
			sleep(rep_delay);
		}


		// Close file
		if (close(file_desc) == -1)
		{
			// Set close errno
			*close_errno = errno;


			// Write to console (WARNING)
		}
		else
		{
			// Set file close
			file_close = 1;


			// Break repetition loop
			break;
		}
	}


	// Check for file close
	if (file_close != 1)
	{
		return -2;	// ERROR: Couldn't close file
	}


	return 1;	// SUCCESS
}


// NOTE: I/O transfer
//
//  An I/O transfer consists of a sequence of three library calls:
//
//   - open()
//   - ioctl()
//   - close()


int fld_io_transfer(char *open_path,		// -- IN
		    fld_int32_t open_flags,	// -- IN
		    mode_t open_modes,		// -- IN
		    enum fld_ioctl_e ioctl_cmd,	// -- IN
		    void *ioctl_arg,		// -- IN
		    fld_uint32_t *error_stage,	// -- OUT
		    fld_int32_t *errno_stage,	// -- OUT
		    fld_int32_t *return_value)	// -- OUT
{
	fld_int32_t fd;
	//enum fld_ioctl_e cmd_masked;


	// NOTE: Arguments


	// IN: Check for open path
	if (open_path == NULL)
	{
		// Set return value
		if (return_value != NULL)
		{
			*return_value = 0;
		}


		return 0;	// ERROR: No open path
	}


	// IN: Check for valid ioctl command
	if ((ioctl_cmd < FLD_IOCTL_FIRST) ||
	    (ioctl_cmd > FLD_IOCTL_LAST))
	{
		// Set return value
		if (return_value != NULL)
		{
			*return_value = -1;
		}


		return -1;	// ERROR: Invalid ioctl command
	}


	// OUT: Check for stage error
	if (error_stage == NULL)
	{
		// Set return value
		if (return_value != NULL)
		{
			*return_value = -2;
		}


		return -2;	// ERROR: No stage error
	}


	// OUT: Check for stage errno
	if (errno_stage == NULL)
	{
		// Set return value
		if (return_value != NULL)
		{
			*return_value = -3;
		}


		return -3;	// ERROR: No stage errno
	}


	// OUT: Check for return value
	if (return_value == NULL)
	{
		return -4;	// ERROR: No return value
	}


	// NOTE: Processing


	// Open /dev/fl
	fd = open("/dev/fl", open_flags, open_modes);
	if (fd == -1)
	{
		//FLD_DEBUG_UPRINTF(("[fld_io_transfer] ERROR: Couldn't open /dev/fl (%u - %s)\n", errno, strerror(errno)))


		// Set stage error
		*error_stage = 0;


		// Set stage errno
		*errno_stage = errno;


		return -5;	// ERROR: Couldn't open /dev/fl
	}


	// Ioctl /dev/fl
	if (ioctl(fd, ioctl_cmd, ioctl_arg) == -1)
	{
		// Set stage error
		*error_stage = 1;


		// Set stage errno
		*errno_stage = errno;


		// Close /dev/fl
		if (close(fd) == -1)
		{
			return -6;	// ERROR: Couldn't ioctl nor close /dev/fl
		}


		return -7;	// ERROR: Couldn't ioctl /dev/fl
	}


	// Close /dev/fl
	if (close(fd) == -1)
	{
		// Set stage error
		*error_stage = 2;


		// Set stage errno
		*errno_stage = errno;


		return -8;	// ERROR: Couldn't close /dev/fl
	}


	return 1;	// SUCCESS
}


int fld_io_alloc_transfer_buffer(void)
{
	// NOTE: Processing


	// Calculate maximum I/O transfer size
	FLD_MAX_TRANSFER_SIZE = (fld_maxdmasz - 1) << page_shift;


	FLD_DEBUG_UPRINTF(("[fld_io_alloc_transfer_buffer] page_shift = %u\n", page_shift))
	FLD_DEBUG_UPRINTF(("[fld_io_alloc_transfer_buffer] FLD_MAX_TRANSFER_SIZE = %u\n", FLD_MAX_TRANSFER_SIZE))


        // NOTE: Maximum transfer size
        //
        //  The maximum transfer size depends on the value of the maxdmasz
        //  kernel variable. By default, maxdmasz is 1025 for IRIX32 kernels
        //  and 257 for IRIX64 kernels. Its value specifies number of pages.
        //
        //  Filesystem code for XFS, por example, calculates the maximum
        //  transfer size using the formula:
        //
        //    max_transfer = (maxdmasz - 1) << page_shift (12 for IRIX32,
        //                                                 14 for IRIX64)
        //
        //  This yields a maximum transfer size of 4MiB for IRIX32 kernels
        //  and 16MiB for IRIX64 with the default maxdmasz value of 1025 (IRIX32)
        //  and 257 (IRIX64), respectively.
        //
        //  File looper uses this formula to preallocate the I/O transfer buffer
        //  in fld.
        //
        //  Nonetheless, maxdmasz can be set using systune, so its value must
        //  be checked for on each transfer. If the value changes, two cases
        //  must be checked for:
        //
        //   * maxdmasz was increased -> I/O transfer buffer grows
        //   * maxdmasz was decreased -> I/O transfer buffer shrinks
        //
        //  Only the first case must be handled.


	// Allocate I/O transfer buffer
	io_data = (char *) malloc(FLD_MAX_TRANSFER_SIZE);
	if (io_data == NULL)
	{
		FLD_DEBUG_UPRINTF(("[fld_io_alloc_transfer_buffer] ERROR: malloc(FLD_MAX_TRANSFER_SIZE)\n"))


		// Write to console
		syslog(LOG_ALERT,
		       "ALERT: Couldn't allocate I/O transfer buffer (%u bytes)\n",
		       FLD_MAX_TRANSFER_SIZE);


		return -1;	// ERROR: Couldn't allocate I/O transfer buffer
	}


	FLD_DEBUG_UPRINTF(("[fld_io_alloc_transfer_buffer] io_data = %p\n", io_data))


	return 1;	// SUCCESS
}


int fld_io_process(void)
{
	fld_int32_t afd;
	fld_int32_t ret;
	fld_int32_t open_errno;
	fld_int32_t lseek_errno;
	fld_int32_t read_errno;
	fld_int32_t write_errno;
	fld_int32_t close_errno;
	fld_int32_t errno_stage;
	fld_uint32_t error_stage;
	fld_uint32_t break_loop;
	fld_uint32_t control_index;


#if defined(FLD_DEBUG)


	char *fld_io_string[] =
	{
		"FLD_IO_READ",
		"FLD_IO_WRITE"
	};
	char *error_stage_string[] =
	{
		"open",
		"ioctl",
		"close"
	};


#endif	// defined(FLD_DEBUG)


	// NOTE: Processing


	// NOTE: I/O loop


	//FLD_DEBUG_UPRINTF(("[fld_io_process] I/O LOOP START\n"))


	// Loop indefinitely
	break_loop = 0;
	while (FLD_LOOP_FOREVER)
	{
		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_REQ START\n"))


		// Transfer FLD_IOCTL_REQ
		if (fld_io_transfer("/dev/fl",
				    O_RDONLY,
				    0,
				    FLD_IOCTL_REQ,
				    &io_req,
				    &error_stage,
				    &errno_stage,
				    &ret) < 1)
		{
			FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: %s (%i - %u - %s)\n", error_stage_string[error_stage], ret, errno, strerror(errno)))
			FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_REQ ERROR START\n"))


			// Print I/O request
			//fld_ioreq_print(&io_req);


			FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_REQ ERROR STOP\n"))


			// NOTE: Process errors


			// Switch out errors
			switch (error_stage)
			{
				case 1:

					// NOTE: ioctl error
					//
					//  FLD_IOCTL_REQ may return any of the following
					//  errno values:
					//
					//   ENODATA  - No more I/O requests pending
					//   EFAULT   - Couldn't write back I/O request or
					//              couldn't copy over write buffer from
					//              kernel address space
					//   ESUCCESS - I/O request successfully received


					if (errno_stage == ENODATA)
					{
						// Set loop break
						break_loop = 1;
					}
					else if (errno_stage == EFAULT)
					{
						// Write to console (WARNING)
						syslog(LOG_WARNING,
						       "WARNING: Couldn't copy out data (FLD_IOCTL_REQ)\n");


						// Set loop break
						break_loop = 1;
					}


					break;

				case 0:

					// NOTE: open error


					// Write to console (ERROR)
					syslog(LOG_ERR,
					       "ERROR: Couldn't open /dev/fl (FLD_IOCTL_REQ)\n");


					// Set loop break
					break_loop = 1;


					break;

				default:

					// NOTE: close error
					//
					//  Continue as much as possible.


					// Write to console (WARNING)
					syslog(LOG_WARNING,
					       "WARNING: Couldn't close /dev/fl (FLD_IOCTL_REQ)\n");


					break;
			}


			// Check for loop break
			if (break_loop == 1)
			{
				// Break request loop
				break;
			}
		}


		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_REQ STOP\n"))


		// Print I/O request
		//fld_ioreq_print(&io_req);


		// NOTE: Check for correct I/O request control value


		if (io_req.control == FLD_IO_READ)
		{
			control_index = 0;
		}
		else if (io_req.control == FLD_IO_WRITE)
		{
			control_index = 1;
		}
		else
		{
			// NOTE: I/O request control error


			// Write to console (ERROR)
			syslog(LOG_ERR,
			       "ERROR: I/O request control error (FLD_IOCTL_REQ)\n");


			// Break request loop
			break;
		}


		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: I/O START (%s)\n", fld_io_string[control_index]))


		// NOTE: Process I/O request


		if (control_index == 0)
		{
			// NOTE: Read I/O


			// Open attached file
			if (fld_io_open(io_req.file,
				        O_RDONLY,
				        0,
				        3,
				        10,
				        &afd,
					&open_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: open(%s, O_RDONLY): %u - %s\n", io_req.file, open_errno, strerror(open_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't open attached file (FLD_IO_READ)\n");


				// Break request loop
				break;
			}


			// Seek to requested offset
			if (fld_io_lseek(afd,
					 io_req.offset,
					 SEEK_SET,
					 3,
					 10,
					 &lseek_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: lseek(SEEK_SET): %u - %s\n", lseek_errno, strerror(lseek_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't seek into attached file (FLD_IO_READ)\n");


				// Break request loop
				break;
			}


			// Read in requested data
			if (fld_io_read(afd,
				        io_data,
				        io_req.size,
				        3,
				        10,
				        &io_data_size,
				        &read_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: read(): %u - %s\n", read_errno, strerror(read_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't read from attached file (FLD_IO_READ)\n");


				// Break request loop
				break;
			}


			// Close attached file
			if (fld_io_close(afd,
					 3,
					 10,
					 &close_errno) < 1)
			{
				// NOTE: close error
				//
				//  Continue as much as possible.


				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: close(): %u - %s\n", close_errno, strerror(close_errno)))


				// Write to console (WARNING)
				syslog(LOG_WARNING,
				       "WARNING: Couldn't close attached file (FLD_IO_READ)\n");
			}


			// NOTE: Finish I/O request


			// Set I/O response control field
			if (io_data_size == io_req.size)
			{
				io_req.control = FLD_IO_READ_OK;
			}
			else if (io_data_size < io_req.size)
			{
				io_req.control = FLD_IO_READ_INCMPL;
			}
			else if (io_data_size == 0)
			{
				io_req.control = FLD_IO_READ_EOF;
			}
			else
			{
				io_req.control = FLD_IO_READ_NOK;
			}


			// Set I/O response data fields
			io_req.data = io_data;
			io_req.data_size = io_data_size;
		}
		else if (control_index == 1)
		{
			// NOTE: Write I/O


			// Open attached file
			if (fld_io_open(io_req.file,
				        O_WRONLY | O_SYNC,
				        0,
				        3,
				        10,
				        &afd,
					&open_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: open(%s, O_WRONLY | O_SYNC): %u - %s\n", io_req.file, open_errno, strerror(open_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't open attached file (FLD_IO_WRITE)\n");


				// Break request loop
				break;
			}


			// Seek to requested offset
			if (fld_io_lseek(afd,
					 io_req.offset,
					 SEEK_SET,
					 3,
					 10,
					 &lseek_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: lseek(SEEK_SET): %u - %s\n", lseek_errno, strerror(lseek_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't seek into attached file (FLD_IO_WRITE)\n");


				// Break request loop
				break;
			}


			if (io_req.is_format == 1)
			{
				// NOTE: Format write


				// Reset I/O data buffer
				bzero((void *) io_data,
				      io_req.data_size);
			}


			// Write out requested data
			if (fld_io_write(afd,
					 io_data,
					 io_req.data_size,
					 3,
					 10,
					 &io_data_size,
					 &write_errno) < 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: write(): %u - %s\n", write_errno, strerror(write_errno)))


				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't write to attached file (FLD_IO_READ)\n");


				// Break request loop
				break;
			}


			// Close attached file
			if (fld_io_close(afd,
					 3,
					 10,
					 &close_errno) < 1)
			{
				// NOTE: close error
				//
				//  Continue as much as possible.


				FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: close(): %u - %s\n", close_errno, strerror(close_errno)))


				// Write to console (WARNING)
				syslog(LOG_WARNING,
				       "WARNING: Couldn't close attached file (FLD_IO_WRITE)\n");
			}


			// NOTE: Finish I/O request


			// Set I/O response control field
			if (io_data_size == io_req.data_size)
			{
				/*
				// TEMP
				if (io_req.is_format == 1)
				{
					io_req.control = FLD_IO_WRITE_EOF;
				}
				else
				{
					io_req.control = FLD_IO_WRITE_OK;
				}
				*/


				FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: >>> FLD_IO_WRITE_OK\n"))


				io_req.control = FLD_IO_WRITE_OK;
			}
			else if (io_data_size < io_req.data_size)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: >>> FLD_IO_WRITE_INCMPL\n"))


				io_req.control = FLD_IO_WRITE_INCMPL;
			}
			else if (io_data_size == 0)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: >>> FLD_IO_WRITE_EOF\n"))


				io_req.control = FLD_IO_WRITE_EOF;
			}
			else
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: >>> FLD_IO_WRITE_NOK\n"))


				io_req.control = FLD_IO_WRITE_NOK;
			}
		}


		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: I/O STOP (%s)\n", fld_io_string[control_index]))
		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_FIN START\n"))


		// Transfer FLD_IOCTL_FIN
		ret = 0;
		if (fld_io_transfer("/dev/fl",
				    O_RDONLY,
				    0,
				    FLD_IOCTL_FIN,
				    &io_req,
				    &error_stage,
				    &errno_stage,
				    &ret) < 1)
		{
			FLD_DEBUG_UPRINTF(("[fld_io_process] ERROR: %s (%i - %u - %s)\n", error_stage_string[error_stage], ret, errno, strerror(errno)))
			//FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_FIN ERROR START\n"))


			// Print I/O request
			//fld_ioreq_print(&io_req);


			//FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_FIN ERROR STOP\n"))


			// NOTE: Process errors


			// Switch out errors
			switch (error_stage)
			{
				case 1:

					// NOTE: ioctl error
					//
					//  FLD_IOCTL_FIN may return any of the following
					//  errno values:
					//
					//   ENODATA  - I/O response successgully sent, no more
					//              I/O requests pending
					//   EFAULT   - Couldn't copy in I/O response or
					//              couldn't copy in I/O response data
					//   EINVAL   - Unknown ioctl command
					//   ESUCCESS - I/O response successfully sent, more
					//              I/O requests pending


					if (errno_stage == ENODATA)
					{
						// Set loop break
						break_loop = 1;
					}
					else if (errno_stage == EFAULT)
					{
						// Write to console (ERROR)
						syslog(LOG_ERR,
						       "ERROR: Couldn't copy in data (FLD_IOCTL_FIN)\n");


						// Set loop break
						break_loop = 1;
					}


					break;

				case 0:

					// NOTE: open error


					// Write to console (ERROR)
					syslog(LOG_ERR,
					       "ERROR: Couldn't open /dev/fl (FLD_IOCTL_FIN)\n");


					// Set loop break
					break_loop = 1;


					break;

				default:

					// NOTE: close error
					//
					//  Continue as much as possible.


					// Write to console (WARNING)
					syslog(LOG_WARNING,
					       "WARNING: Couldn't close /dev/fl (FLD_IOCTL_FIN)\n");


					break;
			}


			// Check for loop break
			if (break_loop == 1)
			{
				FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_FIN STOP\n"))


				// Break request loop
				break;
			}
		}


		FLD_DEBUG_UPRINTF(("[fld_io_process] INFO: FLD_IOCTL_FIN STOP\n"))
	}


	//FLD_DEBUG_UPRINTF(("[fld_io_process] I/O LOOP STOP\n"))


	return 1;	// SUCCESS
}


int fld_io_start(void)
{
	fld_int32_t return_value;
	fld_int32_t errno_stage;
	fld_uint32_t error_stage;
	struct fld_io_start_args_s io_start_args;


	// NOTE: Processing


	// Check for started daemon
	if (daemon_started == 1)
	{
		FLD_DEBUG_UPRINTF(("[fld_io_start] FLD_IOCTL_STOP START\n"))


		// NOTE: Started daemon
		//
		//  If this daemon was already started, stop
		//  the daemon first before restarting it.


		// Transfer FLD_IOCTL_STOP
		if (fld_io_transfer("/dev/fl",
				    O_RDONLY,
				    0,
				    FLD_IOCTL_STOP,
				    NULL,
				    &error_stage,
				    &errno_stage,
				    &return_value) < 1)
		{
			// NOTE: Stopped daemon
			//
			//  Stopping a daemon can fail because another
			//  daemon got started before this transfer occured.
			//
			//  It can be safely ignored.


			FLD_DEBUG_UPRINTF(("[fld_io_start] FLD_IOCTL_STOP return_value = %i\n", return_value))


			// Write to console (INFO)
			syslog(LOG_INFO,
			       "INFO: Couldn't transfer FLD_IOCTL_STOP\n");
		}


		FLD_DEBUG_UPRINTF(("[fld_io_start] INFO: FLD_IOCTL_STOP STOP\n"))
	}


	// NOTE: Start daemon


	// Reset daemon started
	daemon_started = 0;


	FLD_DEBUG_UPRINTF(("[fld_io_start] INFO: FLD_IOCTL_START START\n"))


	// Transfer FLD_IOCTL_START
	if (fld_io_transfer("/dev/fl",
			    O_RDONLY,
			    0,
			    FLD_IOCTL_START,
			    &io_start_args,
			    &error_stage,
			    &errno_stage,
			    &return_value) < 1)
	{
		FLD_DEBUG_UPRINTF(("[fld_io_start] INFO: FLD_IOCTL_START return_value = %i\n", return_value))


		// Write to console (ERROR)
		syslog(LOG_ERR,
		       "ERROR: Couldn't transfer FLD_IOCTL_START\n");


		return 0;	// ERROR: Couldn't transfer FLD_IOCTL_START
	}


	FLD_DEBUG_UPRINTF(("[fld_io_start] INFO: FLD_IOCTL_START STOP\n"))


	// NOTE: I/O transfer buffer
	//
	//  Allocation of the I/O transfer buffer occurs
	//  only once when executing this function for the
	//  first time.
	//
	//  Subsequent invocations skip the allocation code.


	// Check for I/O transfer buffer
	if (io_data == NULL)
	{
		FLD_DEBUG_UPRINTF(("[fld_io_start] Allocate I/O transfer buffer START\n"))


		// Set initial maxdmasz
		fld_maxdmasz = io_start_args.maxdmasz;


		FLD_DEBUG_UPRINTF(("[fld_io_start] io_start_args.maxdmasz = %i\n",
				   io_start_args.maxdmasz))


		FLD_DEBUG_UPRINTF(("[fld_io_start] fld_maxdmasz = %i\n",
				   fld_maxdmasz))


		// Allocate I/O transfer buffer
		if (fld_io_alloc_transfer_buffer() < 1)
		{
			// Write to console (ERROR)
			syslog(LOG_ERR,
			       "ERROR: Couldn't allocate I/O transfer buffer\n");


			// Reset start args I/O data
			io_start_args.io_data = NULL;


			return -1;	// ERROR: Couldn't allocate I/O transfer buffer
		}
		else
		{
			io_start_args.io_data = io_data;
		}


		FLD_DEBUG_UPRINTF(("[fld_io_start] Allocate I/O transfer buffer STOP\n"))
	}
	else
	{
		// Set I/O transfer buffer
		io_start_args.io_data = io_data;


		FLD_DEBUG_UPRINTF(("[fld_io_start] Skipping: I/O transfer buffer allocation\n"))
	}


	FLD_DEBUG_UPRINTF(("[fld_io_start] FLD_IOCTL_START_ARGS START\n"))


	// Transfer FLD_IOCTL_START_ARGS
	if (fld_io_transfer("/dev/fl",
			    O_RDONLY,
			    0,
			    FLD_IOCTL_START_ARGS,
			    &io_start_args,
			    &error_stage,
			    &errno_stage,
			    &return_value) < 1)
	{
		// Check for ioctl error
		if (error_stage == 1)
		{
			// NOTE: ioctl error


			// Check for ENODATA
			if (errno_stage != ENODATA)
			{
				// Write to console (ERROR)
				syslog(LOG_ERR,
				       "ERROR: Couldn't transfer FLD_IOCTL_START_ARGS (-2)\n");


				return -2;	// ERROR: Couldn't transfer FLD_IOCTL_START_ARGS
			}


			// NOTE: ENODATA
			//
			//  An errno of ENODATA indicates the daemon started
			//  successfully and there is no pending data to be
			//  processed.


			// Set daemon started
			daemon_started = 1;
		}
		else
		{
			// NOTE: open or close error


			// Write to console (ERROR)
			syslog(LOG_ERR,
			       "ERROR: Couldn't transfer FLD_IOCTL_START_ARGS (-3)\n");


			return -3;	// ERROR: Couldn't transfer FLD_IOCTL_START_ARGS
		}
	}
	else
	{
		// NOTE: ESUCCESS
		//
		//  An errno of ESUCCESS indicates the daemon started
		//  successfully and there is pending data to be
		//  processed.


		// Set daemon started
		daemon_started = 1;


		// Send self-signal
		kill(getpid(), SIGURG);
	}


	FLD_DEBUG_UPRINTF(("[fld_io_start] FLD_IOCTL_START_ARGS STOP\n"))


	return 1;	// SUCCESS
}


void fld_sigurg_handler(int signal_number,
			int signal_code,
			struct sigcontext *signal_context)
{
	// NOTE: Processing


	//FLD_DEBUG_UPRINTF(("[fld_sigurg_handler] SIGURG START\n"))


	// NOTE: Daemon started switch
	//
	//  A race condition exists after registering
	//  a new disposition for SIGURG and before
	//  the daemon completed both start ioctls:
	//
	//   FLD_IOCTL_START and FLD_IOCTL_START_ARGS
	//
	//  A switch is used to allow SIGURG to be handled
	//  only when both start ioctls have been
	//  successfully completed.


	// Check for started daemon switch
	if (daemon_started == 1)
	{
		// Process I/O
		fld_io_process();
	}


	//FLD_DEBUG_UPRINTF(("[fld_sigurg_handler] SIGURG STOP\n"))
}


void fld_sighup_handler(int signal_number,
			int signal_code,
			struct sigcontext *signal_context)
{
	// NOTE: Processing


	FLD_DEBUG_UPRINTF(("[fld_sighup_handler] SIGHUP START\n"))


	// Start daemon
	fld_io_start();


	FLD_DEBUG_UPRINTF(("[fld_sighup_handler] SIGHUP STOP\n"))
}


int main(int argc, char *argv[])
{
	struct sigaction sigact;


#if !defined(FLD_DEBUG)


	fld_int32_t daemon_flags;


#endif	// !defined(FLD_DEBUG)


#if FLD_OS_NAME_V_OSF1
#       if FLD_OS_VERSION_V_5_1
#               if FLD_OS_RELEASE_V_B


#               endif
#       endif
#elif FLD_OS_NAME_V_IRIX
#       if FLD_OS_VERSION_V_6_5
#               if FLD_OS_RELEASE_V_30


	fld_uint32_t kern_pointers;


#               endif
#       endif
#endif


	// NOTE: Initialization


#if FLD_OS_NAME_V_OSF1
#       if FLD_OS_VERSION_V_5_1
#               if FLD_OS_RELEASE_V_B


#               endif
#       endif
#elif FLD_OS_NAME_V_IRIX
#       if FLD_OS_VERSION_V_6_5
#               if FLD_OS_RELEASE_V_30


	// Get kernel memory model
	kern_pointers = sysconf(_SC_KERN_POINTERS);
	if (kern_pointers == -1)
	{
		kern_pointers = 32;
	}


	// Get max transfer size
	if (kern_pointers == 32)
	{
		// Default page shift
		page_shift = 12;
	}
	else
	{
		// Default page shift
		page_shift = 14;
	}


	FLD_DEBUG_UPRINTF(("[main] Kernel pointer width = %u\n", kern_pointers))
	FLD_DEBUG_UPRINTF(("[main] Page shift = %u\n", page_shift))


#               endif
#       endif
#endif


	// Reset I/O transfer buffer
	io_data = NULL;


	// Reset daemon started
	daemon_started = 0;


	// NOTE: Processing


	// Open syslog
	openlog(FLD_DAEMON_NAME,
		LOG_PID |
		LOG_CONS,
		LOG_DAEMON);


#if !defined(FLD_DEBUG)


	// Daemonize
	daemon_flags = 0;
	if (_daemonize(daemon_flags,
		       4,
		       -1,
		       -1) == -1)
	{
		// Write to console (ALERT)
		syslog(LOG_ALERT,
		       "ALERT: Couldn't daemonize\n");


		// Close log
		closelog();


		return -1;	// ERROR: Couldn't daemonize fld
	}


#endif	// !defined(FLD_DEBUG)


	// Clean /dev/loop
	// TODO: fld_dir_clean();


	// Fill sigaction structure
	sigact.sa_flags = 0;
	sigact.sa_sighandler.sa_handler1 = fld_sigurg_handler;
	sigemptyset(&sigact.sa_mask);


	// Register SIGURG handler
	if (sigaction(SIGURG,
		      &sigact,
		      NULL) == -1)
	{
		FLD_DEBUG_UPRINTF(("[main] ERROR: signal(SIGURG)\n"))


		// Write to console (ALERT)
		syslog(LOG_ALERT,
		       "ALERT: Couldn't register SIGURG handler\n");


		// Close log
		closelog();


		return -2;	// ERROR: Couldn't register SIGURG handler
	}


	// Fill sigaction structure
	sigact.sa_flags = 0;
	sigact.sa_sighandler.sa_handler1 = fld_sighup_handler;
	sigemptyset(&sigact.sa_mask);


	// Register SIGHUP handler
	if (sigaction(SIGHUP,
		      &sigact,
		      NULL) == -1)
	{
		FLD_DEBUG_UPRINTF(("[main] ERROR: signal(SIGHUP)\n"))


		// Write to console (ALERT)
		syslog(LOG_ALERT,
		       "ALERT: Couldn't register SIGHUP handler\n");


		// Close log
		closelog();


		return -3;	// ERROR: Couldn't register SIGHUP handler
	}


	// Send self-signal for starting the daemon
	kill(getpid(), SIGHUP);


	FLD_DEBUG_UPRINTF(("[main] SLEEP LOOP START\n"))


	// Write to console (INFO)
	syslog(LOG_INFO,
	       "INFO: Daemon started\n");


	// Sleep loop
	while (FLD_LOOP_FOREVER)
	{
		sleep(1 * 60 * 60);
	}


	// Close log
	closelog();


	FLD_DEBUG_UPRINTF(("[main] SLEEP LOOP STOP\n"))


	return 1;	// SUCCESS
}


