#!/usr/bin/perl


use strict;
use IO::Handle;
use Data::Dumper;
use List::Util "first";
use File::Basename;
use Fcntl;


# NOTE: Usage
#
#
#  ./rpp.pl [-io] file0 ... fileN [ -db database0 ... databaseN ]
#
#
#  Tag format:
#
#   %global_namespace:functional_namespace:return_name[MODES,ARG]% - Functional return names
#   %global_namespace:return_name[MODES,ARG]% - Global return names
#
#   %GLOB:<name>:START% - Start a global namespace called <name>
#   %GLOB:<name>:STOP% - Stop a global namespace called <name>
#
#   %FUNC:<name>:START% - Start a functional namespace called <name>
#   %FUNC:<name>:STOP% - Stop a functional namespace called <name>
#
#   %MACRO:<name>:<value>% - Set a macro to the specified value
#   %MACRO:<name>:% - Unset a macro
#
#
#  MODES:
#
#   i - input return tag (i.e add to current database, substitute value and increment the counter specified by p or n mode), allows ARG to specify a static return value
#   o - output return tag (i.e substitute with corresponding value)
#   p - use positive counter for value substitution
#   n - use negative counter for value substitution
#   t - test format for output return tags. Only together with 'o' mode. Outputs a tag in the format: 918289:0:1. Used for tests.
#   g - group return tags by number given in ARG, for sequential processing of return tags by group number. Used for grouping global return tags.
#   m - return tag includes a MACRO to be evaluated: if result is 1, return tag is processed, otherwise return tag is ignored.


# NOTE: Data structures
#
#
#  %GLOBS - global hash containing all global namespaces (indexed by global namespace name)
#  %GLOB  - hash containing functional namespaces for a given global namespace (indexed by functional namespace name)
#  %FUNC  - hash containing return tag names for a given functional namespace (indexed by return tag name)
#  %TAG   - hash containing a tag for a given return tag name (indexed by logical return tag fields: i, o, ..., t, ARG)
#
#


# NOTE: Debugging
#
#  To turn debugging messages on, set DBG to 1.


my $DBG = 0;


# NOTE: Ignore output
#
#  See NOTE in main subroutine for more information.


my $IGNORE_OUTPUT = 0;


my %GLOBS = ();


my @DB_GLOBS = ();


# NOTE: Global namespaces per file
#
#  This variables stores the names of global namespaces defined
#  in each file parsed and is available only in phase FOURTH.
#
#  It is used to implement the -io switch in process_tag().


my @GLOBS_IN_FILE;


# NOTE: Test hashes
#
#  For each return tag with mode 't', its global namespace name
#  is saved and its hash value calculated.
#
#  Both are stored in the TEST_HASHES hash. This hash is used to
#  detect collisions for different global namespace names that yield
#  the same hash value.


my %TEST_HASHES;


sub error_and_exit
{
	print STDOUT "ERROR: Couldn't open file\n";
	exit(0);
}


sub error_and_exit_with_msg
{
	my $MSG = shift;
	print STDOUT "ERROR: $MSG\n";
	exit(0);
}


sub dbg
{
	my $MSG = shift;


	if ($DBG)
	{
		print STDOUT "$MSG";
	}
}


sub print_hash
{
	my $KEY;
	my $HASH_REF = shift;


	# NOTE: Logically sorted hash printer
	#
	#  Print a logically sorted hash in a human readable way.


	print "TYPE   => ${$HASH_REF}{TYPE}\n";
	print "GLOB   => ${$HASH_REF}{GLOB}\n";
	print "FUNC   => ${$HASH_REF}{FUNC}\n";
	print "NAME   => ${$HASH_REF}{NAME}\n";
	print "I      => ${$HASH_REF}{I}\n";
	print "O      => ${$HASH_REF}{O}\n";
	print "P      => ${$HASH_REF}{P}\n";
	print "N      => ${$HASH_REF}{N}\n";
	print "T      => ${$HASH_REF}{T}\n";
	print "G      => ${$HASH_REF}{G}\n";
	print "ARG    => ${$HASH_REF}{ARG}\n";
	print "NUMBER => ${$HASH_REF}{NUMBER}\n";
	print "MATCH  => ${$HASH_REF}{MATCH}\n";
}


sub next_line
{
	my $FH_REF = shift;
	my $LINES_REF = ${$FH_REF}[0];
	my $INDEX_REF = ${$FH_REF}[1];


	return ${$LINES_REF}[${$INDEX_REF}++];
}


sub cur_line
{
	my $FH_REF = shift;
	my $LINES_REF = ${$FH_REF}[0];
	my $INDEX_REF = ${$FH_REF}[1];


	return ${$LINES_REF}[${$INDEX_REF} - 1];
}


sub cur_line_del
{
	my $FH_REF = shift;
	my $LINES_REF = ${$FH_REF}[0];
	my $INDEX_REF = ${$FH_REF}[1];
	my $MODIFIED_REF = ${$FH_REF}[2];


	# Update index
	${$INDEX_REF}--;


	# Delete current line
	splice @{$LINES_REF}, ${$INDEX_REF}, 1;


	# Set modification state
	${$MODIFIED_REF} = 1;
}


sub cur_line_blank
{
	my $FH_REF = shift;
	my $LINES_REF = ${$FH_REF}[0];
	my $INDEX_REF = ${$FH_REF}[1];
	my $MODIFIED_REF = ${$FH_REF}[2];


	# Update index
	#${$INDEX_REF}--;


	# Blank current line
	${$LINES_REF}[${$INDEX_REF} - 1] = "\n";


	# Delete current line
	#splice @{$LINES_REF}, ${$INDEX_REF}, 1;


	# Set modification state
	${$MODIFIED_REF} = 1;
}


sub insert_tag_sorted
{
	my $RET_REF = shift;
	my $RET_TAGS_REF = shift;
	my $REF;
	my $INDEX;


	# Check for empty tags array
	if ((scalar @{$RET_TAGS_REF}) == 0)
	{
		push @{$RET_TAGS_REF}, $RET_REF;
	}
	else
	{
		# NOTE: At least one return tag reference in tags array


		# Loop through array
		$INDEX = 0;
		foreach $REF (@{$RET_TAGS_REF})
		{
			# Look for first lower return number
			if (${$REF}{NUMBER} < ${$RET_REF}{NUMBER})
			{
				last;
			}


			# Increment INDEX
			$INDEX++;
		}


		# Insert new return tag reference
		splice @{$RET_TAGS_REF}, $INDEX, 0, ($RET_REF);
	}
}


sub high_to_low_sort_keys
{
	my $FUNC_REF = shift;
	my $RET_REF;
	my $KEY;
	my @KEYS;
	my @RET_TAGS_SORTED;
	my @KEYS_SORTED;


	# Get keys of functional namespace
	@KEYS = keys %{$FUNC_REF};


	#print "[high_to_low_sort_keys] KEYS LENGTH = ", scalar @KEYS, "\n";


	# Sort keys by return number from higher to lower
	@RET_TAGS_SORTED = ();
	while ((scalar @KEYS) > 0)
	{
		# Pick keys from the front
		$KEY = shift @KEYS;


		#print "[high_to_low_sort_keys] KEY = $KEY\n";


		# Skip __STATE__
		if ($KEY eq '__STATE__')
		{
			#print "[high_to_low_sort_keys] STATEEEEEEEEEEEEE = $KEY\n";
			next;
		}


		# Insert into sorted keys array
		insert_tag_sorted ${$FUNC_REF}{$KEY}, \@RET_TAGS_SORTED;


		#print "[high_to_low_sort_keys] RET_TAGS_SORTED = (", join(', ', @RET_TAGS_SORTED), ")\n";
	}


	# Get sorted keys
	@KEYS_SORTED = ();
	foreach $RET_REF (@RET_TAGS_SORTED)
	{
		push @KEYS_SORTED, ${$RET_REF}{NAME};
	}


	#print "[high_to_low_sort_keys] KEYS_SORTED LENGTH = ", scalar @KEYS_SORTED, "\n";


	return @KEYS_SORTED;
}


sub alphanumeric_sort_macros
{
	my $MACRO_HASH_REF = shift;
	my $RET_REF;
	my $KEY;
	my @KEYS;
	my @RET_TAGS_SORTED;
	my @KEYS_SORTED;


	# Get macros
	@KEYS_SORTED = sort keys %{$MACRO_HASH_REF};


	#print "[alphanumeric_sort_macros] KEYS_SORTED LENGTH = ", scalar @KEYS_SORTED, "\n";


	return @KEYS_SORTED;
}


sub gen_db
{
	my $GLOB_NAME = shift;
	my $GLOB_REF;
	my $GLOB_INDEX;
	my $GLOB_LENGTH;
	my $FUNC_NAME;
	my $FUNC_REF;
	my $FUNC_INDEX;
	my $FUNC_LENGTH;
	my $MACRO_NAME;
	my $MACRO_VALUE;
	my $MACRO_HASH_REF;
	my $RET_REF;
	my $RET_NAME;
	my $NAME_PAD;
	my $FH;
	my @KEYS_SORTED;


	# Get GLOB_REF
	$GLOB_REF = $GLOBS{$GLOB_NAME};


	# Check for global namespace
	if (!defined($GLOB_REF))
	{
		error_and_exit_with_msg "Global namespace \"$GLOB_NAME\" does not exist in GLOBS";
	}


	# NOTE: Create new database file
	#
	#  Return tags are sorted by return number, stored in the NUMBER key,
	#  from higher to lower.


	#open $FH, ">$GLOB_NAME.rpp.db" or error_and_exit_with_msg "Couldn't create database file \"$GLOB_NAME.rpp.db\"";
	sysopen($FH, "$GLOB_NAME.rpp.db", O_WRONLY | O_CREAT | O_TRUNC | O_SYNC);


	#print $FH "my \$GLOB = {\n\n";
	syswrite($FH, "my \$GLOB = {\n\n", length("my \$GLOB = {\n\n"));


	# Get length of global namespace
	$GLOB_LENGTH = scalar(keys %{$GLOB_REF});


	# Loop through functional namespaces
	$GLOB_INDEX = 0;
	foreach $FUNC_NAME (sort keys %{$GLOB_REF})
	{
		# Check for global __STATE__
		if ($FUNC_NAME eq '__STATE__')
		{
			# NOTE: Process macros


			#print "Found global __STATE__\n";


			# Get macro hash reference
			$MACRO_HASH_REF = $GLOB_REF->{__STATE__}{MACROS};


			#print $FH "\t   '__STATE__' => {\n\n";
			#print $FH "\t\t   'MACROS' => {\n\n";
			syswrite($FH, "\t   '__STATE__' => {\n\n", length("\t   '__STATE__' => {\n\n"));
			syswrite($FH, "\t\t   'MACROS' => {\n\n", length("\t\t   'MACROS' => {\n\n"));


			# Get sorted keys
			@KEYS_SORTED = alphanumeric_sort_macros($MACRO_HASH_REF);


			# Get length of functional namespace
			$FUNC_LENGTH = scalar(@KEYS_SORTED);


			#print "[gen_db] FUNC_LENGTH = $FUNC_LENGTH\n";


			# Loop through macro names
			$FUNC_INDEX = 0;
			foreach $MACRO_NAME (@KEYS_SORTED)
			{
				$MACRO_VALUE = ${$MACRO_HASH_REF}{$MACRO_NAME};


				# Check for numeric macro
				if ($MACRO_VALUE !~ m/^[0-9]+/)
				{
					$MACRO_VALUE = "'" . $MACRO_VALUE . "'";
				}


				#print "MACRO_NAME = $MACRO_NAME\n";


				my $TEMP = "\t\t\t'$MACRO_NAME' => $MACRO_VALUE";
				#print $FH "\t\t\t'" . '\' . "$MACRO_NAME' => $MACRO_VALUE";
				#print $FH $TEMP;
				syswrite($FH, $TEMP, length($TEMP));


				# Get padding
				$NAME_PAD = ' ' x (length($RET_NAME) - 1);


				# Check for last return name
				if ($FUNC_INDEX < ($FUNC_LENGTH - 1))
				{
					#print $FH ",";
					syswrite($FH, ',', 1);
				}


				#print $FH "\n";
				syswrite($FH, "\n", 1);


				$FUNC_INDEX++;
			}


			# Get padding
			$NAME_PAD = ' ' x (length('__STATE__') - 1);


			#print $FH "\n\t\t       $NAME_PAD}\n\n";
			#print $FH "\t\t  $NAME_PAD}";
			syswrite($FH, "\n\t\t       $NAME_PAD}\n\n", length("\n\t\t       $NAME_PAD}\n\n"));
			syswrite($FH, "\t\t  $NAME_PAD}", length("\t\t  $NAME_PAD}"));


			# Check for last functional namespace
			if ($GLOB_INDEX < ($GLOB_LENGTH - 1))
			{
				#print $FH ",";
				syswrite($FH, ',', 1);
			}


			#print $FH "\n\n";
			syswrite($FH, "\n\n", 2);
		}
		else
		{
			# Get FUNC_REF
			$FUNC_REF = $GLOB_REF->{$FUNC_NAME};


			#print $FH "\t   '$FUNC_NAME' => {\n\n";
			syswrite($FH, "\t   '$FUNC_NAME' => {\n\n", length("\t   '$FUNC_NAME' => {\n\n"));


			# Get sorted keys
			@KEYS_SORTED = high_to_low_sort_keys($FUNC_REF);
	
	
			# Get length of functional namespace
			#$FUNC_LENGTH = scalar(keys %{$FUNC_REF}) - 1;
			$FUNC_LENGTH = scalar(@KEYS_SORTED);
	
	
			#print "[gen_db] FUNC_LENGTH = $FUNC_LENGTH\n";
	
	
			# Loop through return names
			$FUNC_INDEX = 0;
			foreach $RET_NAME (@KEYS_SORTED)
			{
				#print $FH "\t\t'$RET_NAME' => {\n\n";
				syswrite($FH, "\t\t'$RET_NAME' => {\n\n", length("\t\t'$RET_NAME' => {\n\n"));
	
	
				# Get RET_REF
				$RET_REF = $FUNC_REF->{$RET_NAME};
	
	
				# NOTE: Directly print out return tag fields
	
	
				#print $FH "\t\t\tTYPE   => '$RET_REF->{TYPE}',\n";
				syswrite($FH, "\t\t\tTYPE   => '$RET_REF->{TYPE}',\n", length("\t\t\tTYPE   => '$RET_REF->{TYPE}',\n"));
				#print $FH "\t\t\tGLOB   => '$RET_REF->{GLOB}',\n";
				syswrite($FH, "\t\t\tGLOB   => '$RET_REF->{GLOB}',\n", length("\t\t\tGLOB   => '$RET_REF->{GLOB}',\n"));
				#print $FH "\t\t\tFUNC   => '$RET_REF->{FUNC}',\n";
				syswrite($FH, "\t\t\tFUNC   => '$RET_REF->{FUNC}',\n", length("\t\t\tFUNC   => '$RET_REF->{FUNC}',\n"));
				#print $FH "\t\t\tNAME   => '$RET_REF->{NAME}',\n";
				syswrite($FH, "\t\t\tNAME   => '$RET_REF->{NAME}',\n", length("\t\t\tNAME   => '$RET_REF->{NAME}',\n"));
				#print $FH "\t\t\tI      => $RET_REF->{I},\n";
				syswrite($FH, "\t\t\tI      => $RET_REF->{I},\n", length("\t\t\tI      => $RET_REF->{I},\n"));
				#print $FH "\t\t\tO      => $RET_REF->{O},\n";
				syswrite($FH, "\t\t\tO      => $RET_REF->{O},\n", length("\t\t\tO      => $RET_REF->{O},\n"));
				#print $FH "\t\t\tP      => $RET_REF->{P},\n";
				syswrite($FH, "\t\t\tP      => $RET_REF->{P},\n", length("\t\t\tP      => $RET_REF->{P},\n"));
				#print $FH "\t\t\tN      => $RET_REF->{N},\n";
				syswrite($FH, "\t\t\tN      => $RET_REF->{N},\n", length("\t\t\tN      => $RET_REF->{N},\n"));
				#print $FH "\t\t\tT      => $RET_REF->{T},\n";
				syswrite($FH, "\t\t\tT      => $RET_REF->{T},\n", length("\t\t\tT      => $RET_REF->{T},\n"));
				#print $FH "\t\t\tG      => $RET_REF->{G},\n";
				syswrite($FH, "\t\t\tG      => $RET_REF->{G},\n", length("\t\t\tG      => $RET_REF->{G},\n"));
	
				if ($RET_REF->{ARG} == '')
				{
					#print $FH "\t\t\tARG    => '$RET_REF->{ARG}',\n";
					syswrite($FH, "\t\t\tARG    => '$RET_REF->{ARG}',\n", length("\t\t\tARG    => '$RET_REF->{ARG}',\n"));
				}
				else
				{
					#print $FH "\t\t\tARG    => $RET_REF->{ARG},\n";
					syswrite($FH, "\t\t\tARG    => $RET_REF->{ARG},\n", length("\t\t\tARG    => $RET_REF->{ARG},\n"));
				}
	
	
				#print $FH "\t\t\tNUMBER => $RET_REF->{NUMBER},\n";
				syswrite($FH, "\t\t\tNUMBER => $RET_REF->{NUMBER},\n", length("\t\t\tNUMBER => $RET_REF->{NUMBER},\n"));
				#print $FH "\t\t\tMATCH  => '$RET_REF->{MATCH}'\n\n";
				syswrite($FH, "\t\t\tMATCH  => '$RET_REF->{MATCH}'\n\n", length("\t\t\tMATCH  => '$RET_REF->{MATCH}'\n\n"));
	
	
				# Get padding
				$NAME_PAD = ' ' x (length($RET_NAME) - 1);
	
	
				#print $FH "\t\t       $NAME_PAD}";
				syswrite($FH, "\t\t       $NAME_PAD}", length("\t\t       $NAME_PAD}"));
	
	
				# Check for last return name
				if ($FUNC_INDEX < ($FUNC_LENGTH - 1))
				{
					#print $FH ",";
					syswrite($FH, ",", 1);
				}
	
	
				#print $FH "\n\n";
				syswrite($FH, "\n\n", 2);
	
	
				$FUNC_INDEX++;
			}
	
	
			# Get padding
			$NAME_PAD = ' ' x (length($FUNC_NAME) - 1);
	
	
			#print $FH "\t          $NAME_PAD}";
			syswrite($FH, "\t          $NAME_PAD}", length("\t          $NAME_PAD}"));
	
	
			# Check for last functional namespace
			if ($GLOB_INDEX < ($GLOB_LENGTH - 1))
			{
				#print $FH ",";
				syswrite($FH, ",", 1);
			}
	
	
			#print $FH "\n\n";
			syswrite($FH, "\n\n", 2);
	
	
			$GLOB_INDEX++;
		}
	}


	#print $FH "           };\n";
	syswrite($FH, "           };\n", length("           };\n"));


	# Flush buffers
	$FH->autoflush;


	# Close file
	close $FH;
}


sub add_ret_to_args
{
	my $RET_REF = shift;
	my $ARGS_REF = shift;
	my $ARGS_PN_REF;
	my $ARGS_ARG;
	my $ARGS_RET_REF;
	my $ARGS_ELEM_REF;
	my $ARG;
	my $INDEX;
	my $LENGTH;
	my $VALUE;


	# NOTE: Sorted array
	#
	#  ARGS is a hash with arrays for positive ARG values and negative ARG values.
	#
	#  Each array consists of ARGS elements, where each element contains an ARG value
	#  and a reference to the corresponding return tag hash.
	#
	#  ARGS elements are ordered from lower to higher ARG values.


	# Get ARG
	$ARG = ${$RET_REF}{ARG};


	dbg "[add_ret_to_args] Insert ARG = $ARG\n";


	# Get ARGS_PN_REF
	if ($ARG > 0)
	{
		$ARGS_PN_REF = ${$ARGS_REF}{P};
	}
	else
	{
		$ARGS_PN_REF = ${$ARGS_REF}{N};
	}


	dbg "[add_ret_to_args] ARGS (BEFORE) = \"" . join(', ', map {${$_}[0]} @{$ARGS_PN_REF} ) . "\"\n";


	# Get length of ARGS
	$LENGTH = @{$ARGS_PN_REF};


	if ($LENGTH == 0)
	{
		push @{$ARGS_PN_REF}, [$ARG, $RET_REF];
	}
	else
	{
		# Loop through ARGS
		$INDEX = 0;
		foreach $ARGS_ELEM_REF (@{$ARGS_PN_REF})
		{
			# Unpack ARGS element
			$ARGS_ARG = ${$ARGS_ELEM_REF}[0];
			$ARGS_RET_REF = ${$ARGS_ELEM_REF}[1];


			# Check for duplicate value
			if ($ARG == $ARGS_ARG)
			{
				# Check for same return tag hash
				if ($RET_REF != $ARGS_RET_REF)
				{
					dbg "RETURN TAG = ${$ARGS_RET_REF}{MATCH}\n";
					dbg "NEW RETURN TAG = ${$RET_REF}{MATCH}\n";
					error_and_exit_with_msg "Duplicate return value for different return tags: $ARG";
				}
			}


			# Look for first value greater than ARG
			if ($ARG < $ARGS_ARG)
			{
				# Insert ARG
				splice @{$ARGS_PN_REF}, $INDEX, 0, [$ARG, $RET_REF];


				dbg "[add_ret_to_args] ARGS (AFTER) = \"" . join(', ', map {${$_}[0]} @{$ARGS_PN_REF} ) . "\"\n";


				return;
			}


			$INDEX++;
		}


		# Insert at the end
		push @{$ARGS_PN_REF}, [$ARG, $RET_REF];
	}


	dbg "[add_ret_to_args] ARGS (AFTER) = \"" . join(', ', map {${$_}[0]} @{$ARGS_PN_REF} ) . "\"\n";
}


sub next_counter_value
{
	my $COUNTER = shift;
	my $ARGS_ARRAY_REF = shift;
	my $RET_NUM_REF = shift;
	my $POSITIVE = ($COUNTER > 0);
	my $ARGS_ARG;
	my $ARGS_ELEM_REF;
	my $INDEX;
	my $LENGTH;
	my $REAL_INDEX;
	my $FIRST_ELEM_REF;
	my $FIRST_ARG;
	my $LAST_ELEM_REF;
	my $LAST_ARG;


	# NOTE: Sorted arrays
	#
	#  ARGS arrays are sorted from lower to higher numbers.


	dbg "[next_counter_value] COUNTER = $COUNTER\n";
	dbg "[next_counter_value] ARGS_ARRAY_REF = $ARGS_ARRAY_REF\n";
	dbg "[next_counter_value] RET_NUM_REF = $RET_NUM_REF\n";


	# Get length of ARGS
	$LENGTH = @{$ARGS_ARRAY_REF};


	dbg "[next_counter_value] ARGS LENGTH = $LENGTH\n";


	# Check ARGS length
	if ($LENGTH == 0)
	{
		${$RET_NUM_REF} = $COUNTER;
	}
	else
	{
		dbg "[next_counter_value] ARGS = (", join(', ', map {${$_}[0]} @{$ARGS_ARRAY_REF}), ")\n";


		# Get first value
		$FIRST_ELEM_REF = ${$ARGS_ARRAY_REF}[0];
		$FIRST_ARG = ${$FIRST_ELEM_REF}[0];


		# Get last value
		$LAST_ELEM_REF = ${$ARGS_ARRAY_REF}[-1];
		$LAST_ARG = ${$LAST_ELEM_REF}[0];


		# NOTE: Quick checks
		#
		#  Quick checks pick both extremes of array and compare
		#  against them.
		#
		#  Most of the time, return values exceed one of these
		#  extremes, which makes this algorithm very fast.


		# Check first value
		if ($COUNTER < $FIRST_ARG)
		{
			${$RET_NUM_REF} = $COUNTER;


			return;
		}


		# Check last value
		if ($COUNTER > $LAST_ARG)
		{
			${$RET_NUM_REF} = $COUNTER;


			return;
		}


		# Set starting INDEX
		$INDEX = $LENGTH - 1;
		if ($POSITIVE)
		{
			$INDEX = 0;
		}


		# Loop through ARGS
		for ($INDEX = 0; $INDEX < $LENGTH; $INDEX++)
		{
			# Get real INDEX
			if ($POSITIVE)
			{
				$REAL_INDEX = $INDEX;
			}
			else
			{
				$REAL_INDEX = $LENGTH - 1 - $INDEX;
			}


			# Get ARGS_ELEM_REF
			$ARGS_ELEM_REF = ${$ARGS_ARRAY_REF}[$REAL_INDEX];


			# Get ARG
			$ARGS_ARG = ${$ARGS_ELEM_REF}[0];


			dbg "[next_counter_value] ARGS_ARG = $ARGS_ARG\n";


			# Check ARG
			if ($COUNTER == $ARGS_ARG)
			{
				if ($POSITIVE)
				{
					# Increment
					$COUNTER++;
				}
				else
				{
					# Decrement
					$COUNTER--;
				}


				# NOTE: Sorted array
				#
				#  Since values in the ARGS array are sorted, comparison
				#  can continue normally with the next iteration.
			}
		}


		# Set return number
		${$RET_NUM_REF} = $COUNTER;
	}
}


sub get_test_glob_hash
{
        my $GLOB_NAME = shift;
        my $GLOB_HASH_REF = shift;
        my @GLOB_NAME_ASCII;
        my $GLOB_HASH;
        my $VALUE;
        my $INDEX;
        my $ASCII_LENGTH;
        my $ASCII_NUMS;
        my $ASCII_DIFF;
        my $TEMP_HASH;
        my $FACTOR;


        # NOTE: Glob hash
        #
        #  The hash algorithm consists in a loop that 'pastes' the ASCII
        #  value of each consecutive character in the global namespace name
        #  to each other until:
        #
        #   1. The highest number that is lower than the maximum possible value is reached (INT_MAX = 2147483647)
        #
        #  Should the global namespace name be too short, then it starts again at the beginning of the string.
        #
        #  The next step repeats this loop using the following algorithm:
        #
        #   1. Pick the next character in global namespace name
        #   2. Add to hash number (with possible integer overflow)
        #   3. Get modulus of character's ASCII value with 10
        #   4. Multiply hash number with the number from step 3 (with possible integer overflow)
        #
        #  Integer overflow follows a bounce-back model, not a circular model.


        $GLOB_HASH = "";
        @GLOB_NAME_ASCII = map {ord($_)} (split '', "$GLOB_NAME");


        # Get number of ASCII values
        $ASCII_LENGTH = scalar @GLOB_NAME_ASCII;


        # Loop through ASCII values
        #foreach $VALUE (@GLOB_NAME_ASCII)
        $ASCII_NUMS = 0;
        for ($INDEX = 0; $INDEX < $ASCII_LENGTH;)
        {
                #print chr($VALUE) . " --> $VALUE\n";


                # Get next value
                $VALUE = $GLOB_NAME_ASCII[$INDEX];


                #print "HEHE VALUE = $VALUE\n";


                # Add next character value
                $TEMP_HASH = "$GLOB_HASH$VALUE";


                # NOTE: Check for highest lowest possible hash value


                # Check for number of digits
                if (length("$TEMP_HASH") > 10)
                {
                        last;
                }


                # Check for value
                if (length("$TEMP_HASH") == 10)
                {
                        if ($TEMP_HASH > 2147483647)
                        {
                                last;
                        }
                }


                # NOTE: Temporary hash contains a number from 0 - 2147483647


                $GLOB_HASH = $TEMP_HASH;


                # Check for last ASCII value
                if ($INDEX == ($ASCII_LENGTH - 1))
                {
                        # NOTE: Global namespace name is too short


                        # Start again
                        $INDEX = 0;
                }
                else
                {
                        $INDEX++;
                }


                # Count number of real loops
                $ASCII_NUMS++;
        }


        # NOTE: Add and multiply


        #print "ASCII_NUMS = $ASCII_NUMS\n";
        #print "ASCII_LENGTH = $ASCII_LENGTH\n";


        # Extend ASCII array if too short
        if ($ASCII_NUMS > $ASCII_LENGTH)
        {
                #print "VORHER ", @GLOB_NAME_ASCII;
                #print "\n";


                # Get difference
                $ASCII_DIFF = $ASCII_NUMS - $ASCII_LENGTH;


                # Repeat the last value
                splice @GLOB_NAME_ASCII, $ASCII_LENGTH, $ASCII_DIFF, map {$GLOB_NAME_ASCII[$#GLOB_NAME_ASCII]} 0..($ASCII_DIFF - 1);


                #print "NACHHER ", @GLOB_NAME_ASCII;
                #print "\n";
        }


        # Loop through ASCII values
        foreach $VALUE (@GLOB_NAME_ASCII)
        {
                # NOTE: Add, with overflow handling
                #
                #  On overflow, bounce-back (subtract excess from maximum value).


                if (($GLOB_HASH + $VALUE) > 2147483647)
                {
                        $GLOB_HASH -= $VALUE;
                }
                else
                {
                        $GLOB_HASH += $VALUE;
                }


                # NOTE: Multiply, with overflow handling


                # Get factor
                $FACTOR = $VALUE % 10;


                # NOTE: Multiplication
                #
                #  For better understanding, multiplication is done using successive additions.


                for ($INDEX = 0; $INDEX < $FACTOR; $INDEX++)
                {
                        # NOTE: Overflow
                        #
                        #  On overflow, start from the beginning.


                        # Check overflow
                        if (($GLOB_HASH + $GLOB_HASH) > 2147483647)
                        {
                                # Get excess, and start again
                                $GLOB_HASH = (2 * $GLOB_HASH) - 2147483647;
                        }
                        else
                        {
                                $GLOB_HASH += $GLOB_HASH;
                        }
                }
        }


	${$GLOB_HASH_REF} = $GLOB_HASH;
}


sub get_test_func_index
{
        my $GLOB_REF = shift;
        my $FUNC_NAME = shift;
        my $FUNC_INDEX_REF = shift;
	my @FUNC_NAMES;


	# NOTE: Functional namespace index
	#
	#  The functional namespace index is obtained from the list
	#  of alphabetically sorted functional namespace names.


	# Get sorted functional namespaces
	@FUNC_NAMES = sort keys %{$GLOB_REF};


	# Get index of requested functional namespace
	${$FUNC_INDEX_REF} = first {$FUNC_NAMES[$_] eq $FUNC_NAME} 0..$#FUNC_NAMES;
}


sub process_tag_groups
{
	my $GLOB_REF = shift;
	my $GROUP_NAME;
	my $RET_NUM;
	my $RET_REF;
	my $ARGS_REF;
	my $STATE_REF;
	my $COUNTER_REF;
	my $ARGS_ARRAY_REF;
	my $GROUP_ARRAY_REF;
	my $COUNTER;
	my $POSITIVE;


	dbg "[process_tag_groups] GLOB_REF = $GLOB_REF\n";


	# Get ARGS reference
	$ARGS_REF = $GLOB_REF->{'*'}{__STATE__}{ARGS};


	dbg "[process_tag_groups] ARGS_REF = $ARGS_REF\n";


	# Get COUNTER reference
	$COUNTER_REF = $GLOB_REF->{'*'}{__STATE__}{COUNTER};


	dbg "[process_tag_groups] COUNTER_REF = $COUNTER_REF\n";


	dbg "[process_tag_groups] Before {'*'}\n";
	dbg Dumper ${$GLOB_REF}{'*'};


	foreach $GROUP_NAME (sort keys %{$GLOB_REF->{'*'}{__STATE__}{GROUPS}})
	{
		# Get group array reference
		$GROUP_ARRAY_REF = $GLOB_REF->{'*'}{__STATE__}{GROUPS}{$GROUP_NAME};


		dbg "[process_tag_groups] GROUP_NAME = $GROUP_NAME\n";
		dbg "[process_tag_groups] GROUP_ARRAY_REF = $GROUP_ARRAY_REF\n";


		foreach $RET_REF (@{$GROUP_ARRAY_REF})
		{
			dbg "[process_tag_groups] RET_REF = $RET_REF\n";
			dbg "[process_tag_groups] RET{MATCH} = ${$RET_REF}{MATCH}\n";


			# NOTE: Global return tags
			#
			#  In phase 2, functional output return tags are saved and substituted.
			#
			#  In particular, no group processing is necessary.


			# NOTE: Get next return number


			# Get sign of return number
			$POSITIVE = ${$RET_REF}{P};


			# Get counter for return tag
			if ($POSITIVE)
			{
				$COUNTER = ${$COUNTER_REF}{P};
				$ARGS_ARRAY_REF = ${$ARGS_REF}{P};


				dbg "[process_tag_groups] POSITIVE COUNTER = $COUNTER\n";
			}
			else
			{
				$COUNTER = ${$COUNTER_REF}{N};
				$ARGS_ARRAY_REF = ${$ARGS_REF}{N};


				dbg "[process_tag_groups] NEGATIVE COUNTER = $COUNTER\n";
			}


			# Get next value for counter
			$RET_NUM = undef;
			next_counter_value($COUNTER, $ARGS_ARRAY_REF, \$RET_NUM);


			dbg "[process_tag_groups] RETURN NUMBER = $RET_NUM\n";


			# NOTE: Set return number for return tag


			${$RET_REF}{NUMBER} = $RET_NUM;


			# Update state counter
			if ($POSITIVE)
			{
				${$COUNTER_REF}{P} = $RET_NUM + 1;


				dbg "[process_tag_groups] NEXT RETURN NUMBER = " . ($RET_NUM + 1) . "\n";
			}
			else
			{
				${$COUNTER_REF}{N} = $RET_NUM - 1;


				dbg "[process_tag_groups] NEXT RETURN NUMBER = " . ($RET_NUM - 1) . "\n";
			}
		}
	}


	dbg "[process_tag_groups] After {'*'}\n";
	dbg Dumper ${$GLOB_REF}{'*'};
}


sub process_tag
{
	my $PHASE = shift;
	my $FH_REF = shift;
	my $GLOB_NAME = shift;
	my $FUNC_NAME = shift;
	my $RET_REF = shift;
	my $RET_REF_I;
	my $RET_NAME;
	my $GLOB_NAME_IN_FILE;
	my $GLOB_NAME_IS_IN_FILE;
	my $GLOB_REF;
	my $FUNC_REF;
	my $LINES_REF = ${$FH_REF}[0];
	my $INDEX = ${${$FH_REF}[1]};
	my $MODIFIED_REF = ${$FH_REF}[2];
	my $COUNTER;
	my $COUNTER_REF;
	my $POSITIVE;
	my $STAR_REF;
	my $STATE_REF;
	my $ARGS_REF;
	my $ARGS_ARRAY_REF;
	my $GLOB_HASH;
	my $FUNC_INDEX;
	my $RET_NUM;
	my $ARG;
	my $MATCH;
	my $PREFIX;
	my $GROUP_NAME;
	my $GROUPS_REF;
	my $GROUP_ARRAY_REF;


	# NOTE: Return tag
	#
	#  Return tag is already validated.


	# Get GLOB_REF
	$GLOB_REF = $GLOBS{$GLOB_NAME};


	# Get FUNC_REF
	$FUNC_REF = ${$GLOB_REF}{$FUNC_NAME};


	if (($PHASE eq 'FIRST') ||
            ($PHASE eq 'SECOND'))
	{
		# NOTE: Return tag type
		#
		#  Global return tags modify global state,
		#  functional return tags modify functional state.
		#
		#  In phase FIRST, the ARG value is added to the ARGS
		#  key in the corresponding state (global or functional).


		if (${$RET_REF}{TYPE} eq 'F')
		{
			# Get functional state reference
			$STATE_REF = ${$FUNC_REF}{__STATE__};
		}
		else
		{
			# Get * functional namespace
			$STAR_REF = ${$GLOB_REF}{'*'};


			# Get global state reference
			$STATE_REF = ${$STAR_REF}{__STATE__};
		}
	}
	else
	{
		# NOTE: Default type
		#
		#  Upon entering this subroutine, GLOB_NAME and FUNC_NAME
		#  refer to the current parsing state's global and functional
		#  namespaces.
		#
		#  Nonetheless, the return tag may refer to a global return tag,
		#  which have a different functional namespace name, '*',
		#  and a different FUNC_REF.


		# NOTE: Global namespace
		#
		#  Global namespaces are implemented as a functional namespace
		#  with name '*'.


		# Check for global return tag
		if (${$RET_REF}{TYPE} eq 'G')
		{
			$FUNC_NAME = '*';
			$FUNC_REF = ${$GLOB_REF}{$FUNC_NAME};
		}
	}


	# NOTE: Process return tag by phase


	if ($PHASE eq 'FIRST')
	{
		# NOTE: Add ARG value to ARGS array


		# Get ARG value
		$ARG = ${$RET_REF}{ARG};


		# Add into sorted ARGS array
		add_ret_to_args $RET_REF, ${$STATE_REF}{ARGS};


		$RET_NUM = $ARG;


		# NOTE: Set return number for return tag


		${$RET_REF}{NUMBER} = $RET_NUM;
	}
	elsif ($PHASE eq 'SECOND')
	{
		# Check for global return tags
		if (${$RET_REF}{TYPE} eq 'G')
		{
			# NOTE: Global return tags
			#
			#  In phase 2, global input return tags are grouped by number through the
			#  'g' mode.


			# NOTE: Add to group number


			dbg "[process_tag] GLOBAL INPUT WITHOUT ARG = \"${$RET_REF}{MATCH}\"\n";


			# Get GROUPS reference
			$GROUPS_REF = ${$STATE_REF}{GROUPS};


			# Get group number as string
			if (${$RET_REF}{G})
			{
				$GROUP_NAME = "${$RET_REF}{ARG}";
			}
			else
			{
				$GROUP_NAME = 'LAST';
			}


			# Check for new group
			if (${$GROUPS_REF}{$GROUP_NAME} == undef)
			{
				# NOTE: Create new group


				${$GROUPS_REF}{$GROUP_NAME} = [];
			}


			# Get group array
			$GROUP_ARRAY_REF = ${$GROUPS_REF}{$GROUP_NAME};


			# Add group
			push @{$GROUP_ARRAY_REF}, $RET_REF;


			return;
		}
		else
		{
			# NOTE: Functional return tags
			#
			#  In phase 2, functional input return tags are saved and substituted.
			#
			#  In particular, no group processing is done.


			# NOTE: Get next return number


			# Get ARGS reference
			$ARGS_REF = ${$STATE_REF}{ARGS};


			# Get COUNTER reference
			$COUNTER_REF = ${$STATE_REF}{COUNTER};


			# Get sign of return number
			$POSITIVE = ${$RET_REF}{P};


			# Get counter for return tag
			if ($POSITIVE)
			{
				$COUNTER = ${$COUNTER_REF}{P};
				$ARGS_ARRAY_REF = ${$ARGS_REF}{P};


				dbg "[process_tag] POSITIVE COUNTER = $COUNTER\n";
			}
			else
			{
				$COUNTER = ${$COUNTER_REF}{N};
				$ARGS_ARRAY_REF = ${$ARGS_REF}{N};


				dbg "[process_tag] NEGATIVE COUNTER = $COUNTER\n";
			}


			# Get next value for counter
			$RET_NUM = undef;
			next_counter_value($COUNTER, $ARGS_ARRAY_REF, \$RET_NUM);


			dbg "[process_tag] RETURN NUMBER = $RET_NUM\n";


			# NOTE: Set return number for return tag


			${$RET_REF}{NUMBER} = $RET_NUM;


			# Update state counter
			if ($POSITIVE)
			{
				${$COUNTER_REF}{P} = $RET_NUM + 1;


				dbg "[process_tag] NEXT RETURN NUMBER = " . ($RET_NUM + 1) . "\n";
			}
			else
			{
				${$COUNTER_REF}{N} = $RET_NUM - 1;


				dbg "[process_tag] NEXT RETURN NUMBER = " . ($RET_NUM - 1) . "\n";
			}
		}
	}
	elsif ($PHASE eq 'THIRD')
	{
		# NOTE: Substitute return numbers of global input return tags (optionally grouped)


		# Get return number
		$RET_NUM = ${$RET_REF}{NUMBER};
	}
	else
	{
		# NOTE: Fourth phase
		#
		#  Return tags with 'o' mode can be defined in any global and functional
		#  namespace, not just the current namespaces.


		dbg "[process_tag] FOURTH PHASE\n";


		# Get GLOB_NAME for return tag
		$GLOB_NAME = ${$RET_REF}{GLOB};


		# Get FUNC_NAME for return tag
		$FUNC_NAME = ${$RET_REF}{FUNC};


		# Get NAME for return tag
		$RET_NAME = ${$RET_REF}{NAME};


		# NOTE: Check if return tag is in database


		# Get GLOB_REF
		$GLOB_REF = $GLOBS{$GLOB_NAME};


		# Get FUNC_REF
		$FUNC_REF = ${$GLOB_REF}{$FUNC_NAME};


		# Get corresponding input return tag
		$RET_REF_I = ${$FUNC_REF}{$RET_NAME};


		# NOTE: Check for correct return tag


		# Check for return tag name in FUNC
		if ($RET_REF_I == undef)
		{
			# NOTE: No input return tag
			#
			#  An output return tag has no corresponding input return tag.
			#
			#  This may happen when circular dependencies are in place, or
			#  when the input return tag was not defined at all.
			#
			#  The case of circular dependencies is checked for here.


			if ($IGNORE_OUTPUT)
			{
				print "[process_tag] GLOB_NAME = $GLOB_NAME\n";


				# Check for global namespace of output return tag in current file
				$GLOB_NAME_IS_IN_FILE = 0;
				foreach $GLOB_NAME_IN_FILE (@GLOBS_IN_FILE)
				{
					print "[process_tag] GLOBA_NAME_IN_FILE = $GLOB_NAME_IN_FILE\n";


					if ($GLOB_NAME_IN_FILE eq $GLOB_NAME)
					{
						$GLOB_NAME_IS_IN_FILE = 1;
					}
				}

				if (!$GLOB_NAME_IS_IN_FILE)
				{
					# NOTE: Output return tag is from external file


					print "[process_tag] Silently return\n";


					return;
				}
			}


			error_and_exit_with_msg "Return tag named \"$RET_NAME\" does not exist in $GLOB_NAME:$FUNC_NAME namespace";
		}


		# Check for correct return number
		if (length("${$RET_REF_I}{NUMBER}") == 0)
		{
			# NOTE: No return number
			#
			#  When an input return tag has no number in this phase,
			#  just skip silently the return tag for the user to figure out
			#  that this tag wasn't substituted.


			return;
		}


		# Get counter type for output return tag
		$POSITIVE = ${$RET_REF}{P};


		# Check for correct counter type
		if ($POSITIVE)
		{
			if (${$RET_REF_I}{P} == 0)
			{
				print "INPUT RETURN TAG: ${$RET_REF_I}{MATCH}\n";
				print "OUTPUT RETURN TAG: ${$RET_REF}{MATCH}\n";
				error_and_exit_with_msg "Output return tag has 'p' mode, but input return tag has 'n' mode";
			}
		}
		else
		{
			if (${$RET_REF_I}{N} == 0)
			{
				print "INPUT RETURN TAG: ${$RET_REF_I}{MATCH}\n";
				print "OUTPUT RETURN TAG: ${$RET_REF}{MATCH}\n";
				error_and_exit_with_msg "Output return tag has 'n' mode, but input return tag has 'p' mode";
			}
		}


		# Get return number
		$RET_NUM = ${$RET_REF_I}{NUMBER};
	}


	# NOTE: Substitute return tag with return number


	# NOTE: Current line index
	#
	#  The index of the current line is one less than the current index,
	#  because it gets post-incremented after each line is read.


	$INDEX--;


	dbg "[process_tag] LINE = \"${$LINES_REF}[$INDEX]\"\n";


	# Get matching string of return tag
	$MATCH = ${$RET_REF}{MATCH};


	dbg "[process_tag] MATCH = \"$MATCH\"\n";


	# NOTE: Test format
	#
	#  The 't' mode outputs a return tag in the following format:
	#
	#   <global namespace hash>:functional namespace index:return tag index
	#
	#  Where:
	#
	#   global namespace hash: a unique hash calculated for a global namespace
	#                          from its string representation.
	#
	#   functional namespace index: a number that uniquely identifies a functional
	#                               namespace in a global namespace.
	#
	#   return tag index: a number that uniquely identifies a return tag in a
	#                     functional namespace.
	#
	#
	#  The output uniquely identifies a return tag using integer values, creating
	#  a mapping between a return tag and an integer triplet (as shown in the format
	#  above).
	#
	#  This triplet is used by user and kernel code to unequivocally identify tests
	#  to be executed.


	# Check for 't' mode
	if (${$RET_REF}{T})
	{
		# NOTE: Test mode
		#
		#  Entered only in phase FOURTH.


		# NOTE: Create triplet
		#
		#  1. Get global namespace hash
		#  2. Get functional namespace index
		#  3. Substitute return tag with triplet


		# Get global namespace hash
		get_test_glob_hash $GLOB_NAME, \$GLOB_HASH;


		# Get functional namespace index
		get_test_func_index $GLOB_REF, $FUNC_NAME, \$FUNC_INDEX;


		# NOTE: Return tag index
		#
		#  The return number is used as the return tag index.


		# Substitute return tag with triplet
		${$LINES_REF}[$INDEX] =~ s/\Q$MATCH\E/$GLOB_HASH:$FUNC_INDEX:$RET_NUM/;
	}
	else
	{
		# Substitute return tag with mathematical sign
		${$LINES_REF}[$INDEX] =~ s/\Q$MATCH\E/$RET_NUM/;
	}


	# Set modification state for file handle
	${$MODIFIED_REF} = 1;


	dbg "[process_tag] RET_NUM = \"$RET_NUM\"\n";
	dbg "[process_tag] MODIFIED LINE = \"${$LINES_REF}[$INDEX]\"\n";
}


sub get_ret_tags_in_line
{
	my $PHASE = shift;
	my $GLOB_NAME = shift;
	my $FUNC_NAME = shift;
	my $LINE = shift;
	my $RET_REF = shift;
	my $GLOB_REF;
	my $GROUP_ARRAY_REF;
	my $GROUP_ARRAY_LINDEX;
	my $GROUP_NAME;
	my $HAS_FUNC;
	my $HAS_FUNC_REL;
	my $HAS_GLOB;
	my $WHICH_TAG;
	my $RET_TAG_G;
	my $MATCH_INDEX;
	my $GLOB_COLLISION;
	my $TEST_GLOB;
	my $GLOB_HASH;
	my $TYPE;
	my $MATCH;
	my $INDEX;
	my $GLOB;
	my $FUNC;
	my $RET;
	my $MODES;
	my $ARG;
	my $MACRO;
	my $FR_MATCH;
	my $FR_INDEX;
	my $G_MATCH;
	my $G_INDEX;
	my $I;
	my $O;
	my $A;
	my $D;
	my $P;
	my $N;
	my $T;
	my $G;
	my $M;
	my $MACRO_DEFINITIONS;
	my $MACRO_RESULT;
	my $MACRO_NAME;
	my $MACRO_VALUE;
	my $MACRO_HASH_REF;
	my $IFCOND;


	# NOTE: $RET_REF
	#
	#  $RET_REF is a reference to an ARRAY where the parsed return tags
	#  are returned.


	# NOTE: Return tag types
	#
	#  There are two different return tag types:
	#
	#   1. Functional return tags
	#
	#    1.1 %global_namespace:functional_namespace:return_name[MODES,ARG,MACRO]% - Absolute addressing
	#    1.2 %return_name[MODES,ARG,MACRO]% - Relative addressing
	#
	#   2. Global return tags
	#
	#    %global_namespace:*:return_name[MODES,ARG,MACRO]%


	while (1)
	{
		my %RET_TAG;


		dbg "[get_ret_tags_in_line] LINE = \"$LINE\"\n";


		# Look for functional return tag (absolute addressing)
		$HAS_FUNC = ($LINE =~ /%([a-zA-Z_][a-zA-Z0-9_]*?):([a-zA-Z_][a-zA-Z0-9_]*?):([a-zA-Z_][a-zA-Z0-9_]*?)\[(.*?),[ \t]*(.*?)[ \t]*,[ \t]*(.*?)[ \t]*\]%/);

		$INDEX = -1;
		$WHICH_TAG = -1;
		if ($HAS_FUNC)
		{
			$TYPE ='F';
			$MATCH = $&;
			$INDEX = index($LINE, $MATCH);
			$GLOB = $1;
			$FUNC = $2;
			$RET = $3;
			$MODES = $4;
			$ARG = $5;
			$MACRO = $6;


			$WHICH_TAG = 0;
		}


		# Look for functional return tag (relative addressing)
		$HAS_FUNC_REL = ($LINE =~ /%([a-zA-Z_][a-zA-Z0-9_]*?)\[(.*?),[ \t]*(.*?)[ \t]*,[ \t]*(.*?)[ \t]*\]%/);

		if ($HAS_FUNC_REL)
		{
			$FR_MATCH = $&;
			$FR_INDEX = index($LINE, $FR_MATCH);


			if (($INDEX == -1) ||
			    ($FR_INDEX < $INDEX))
			{
				$TYPE = 'F';
				$MATCH = $FR_MATCH;
				$INDEX = $FR_INDEX;
				$GLOB = $GLOB_NAME;
				$FUNC = $FUNC_NAME;
				$RET = $1;
				$MODES = $2;
				$ARG = $3;
				$MACRO = $4;


				$WHICH_TAG = 1;
			}
		}


		# Look for global return tag (always absolute addressing)
		$HAS_GLOB = ($LINE =~ /%([a-zA-Z_][a-zA-Z0-9_]*?):\*:([a-zA-Z_][a-zA-Z0-9_]*?)\[(.*?),[ \t]*(.*?)[ \t]*,[ \t]*(.*?)[ \t]*\]%/);

		if ($HAS_GLOB)
		{
			$G_MATCH = $&;
			$G_INDEX = index($LINE, $G_MATCH);


			if (($INDEX == -1) ||
                            ($G_INDEX < $INDEX))
			{
				$TYPE = 'G';
                                $MATCH = $G_MATCH;
                                $INDEX = $G_INDEX;
                                $GLOB = $1;
                                $FUNC = '*';
                                $RET = $2;
                                $MODES = $3;
                                $ARG = $4;
                                $MACRO = $5;


				$WHICH_TAG = 2;
                        }
		}


		# NOTE: EOT
		#
		#  End of Tags


		if ($WHICH_TAG != -1)
		{
			dbg "[get_ret_tags_in_line] WHICH_TAG = $WHICH_TAG\n";
		}
		else
		{
			last;
		}


		# NOTE: Modes
		#
		#  Each available mode can be specified only once.
		#
		#  Modes can be specified in any order.
		#
		#  Each parsing phase processes only specific modes.


		# NOTE: Check for valid modes


		# Check for empty modes string
		if ($MODES eq "")
		{
			$LINE =~ s/\n*$//;
			dbg "LINE: $LINE\n";
			dbg "TAG: $MATCH\n"; 
			error_and_exit_with_msg "No modes found. At least one 'i' or 'o' mode must be specified. Valid modes are: iopntgm";
		}


		# Check for iopntgm modes
		if ($MODES =~ m/[^iopntgm]+/)
		{
			$LINE =~ s/\n*$//;
			dbg "LINE: $LINE\n";
			dbg "TAG: $MATCH\n"; 
			error_and_exit_with_msg "Unknown modes found. Valid modes are: iopntgm";
		}


		# Check for number of occurences
		if ((($I = () = ($MODES =~ m/i/g)) > 1) ||
                    (($O = () = ($MODES =~ m/o/g)) > 1) ||
                    (($P = () = ($MODES =~ m/p/g)) > 1) ||
                    (($N = () = ($MODES =~ m/n/g)) > 1) ||
                    (($T = () = ($MODES =~ m/t/g)) > 1) ||
                    (($G = () = ($MODES =~ m/g/g)) > 1) ||
                    (($M = () = ($MODES =~ m/m/g)) > 1))
		{
			$LINE =~ s/\n*$//;
			dbg "LINE: $LINE\n";
			dbg "TAG: $MATCH\n"; 
			error_and_exit_with_msg "Modes can be specified only once. Valid modes are: iopntgm";
		}


		# NOTE: Check for valid mode combinations for all phases


		# Check for 'i' and 'o' mutual exclusion
		if ($I && $O)
		{
			$LINE =~ s/\n*$//;
			dbg "LINE: $LINE\n";
			dbg "RETURN TAG: $MATCH\n"; 
			error_and_exit_with_msg "Only one of 'i' or 'o' mode may be specified, but not both. Valid modes are: iopntgm";
		}


		# NOTE: Check for valid mode combinations by phase


		if ($PHASE eq 'FIRST')
		{
			# NOTE: First phase
			#
			#  Process 'i' mode with ARG.
			#
			#  ARG specifies a static return value for the return tag.
			#
			#  Process 't' mode with 'o' mode tags.


			# Check for 'i' mode with ARG
			if (!$I || (length("$ARG") == 0))
			{
				# Check for 't' mode with 'o' mode
				if (!$T || !$O)
				{
					# Remove matching string from line
					$LINE =~ s/\Q$MATCH\E//;


					next;
				}


				# NOTE: Tag with 't' and 'o' modes
				#
				#  Just check for hash collisions and save global namespace names for
				#  phase FOURTH.
				#
				#  This subroutine should not return any tags with these modes.


				# Check for global namespace in TEST_HASHES
				if (!$TEST_HASHES{$GLOB})
				{
					# Calculate hash
					get_test_glob_hash $GLOB, \$GLOB_HASH;


					# Check for collisions
					$GLOB_COLLISION = undef;
					foreach $TEST_GLOB (keys %TEST_HASHES)
					{
						if ($TEST_HASHES{$TEST_GLOB} == $GLOB_HASH)
						{
							$GLOB_COLLISION = $TEST_GLOB;
							last;
						}
					}

					if (defined($GLOB_COLLISION))
					{
						$LINE =~ s/\n*$//;
                                		print "LINE: $LINE\n";
                                		print "RETURN TAG: $MATCH\n";
                                		error_and_exit_with_msg "Hash collision for new global namespace \"$GLOB\" with global namespace \"$GLOB_COLLISION\" for hash value $GLOB_HASH";
					}


					# Add global namespace name
					$TEST_HASHES{$GLOB} = $GLOB_HASH;


					#print "[get_ret_tags_in_line] Adding \"$GLOB\" with HASH = $GLOB_HASH to test hashes\n";
				}


				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				# Process next return tag
				next;
			}


			# Check for 'g' mode
			if ($G)
			{
				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}


			# NOTE: Global or functional return tag with 'i' mode and ARG, and without 'g' mode


			# Check for 'p' and 'n' mutual exclusion
			if ($P && $N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "Only one of 'p' or 'n' mode may be specified, but not both. Valid modes are: iopntgm";
			}


			# Check for at least one of 'p' or 'n'
			if (!$P && !$N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "At least one of either 'p' or 'n' mode must be specified.";
			}


			# Check for decimal value in ARG
			if (($ARG !~ m/^[-]?[1-9][0-9]{0,9}$/) &&
                            ($ARG != 0))
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "ARG for 'i' mode must be decimal value in the range of a signed integer: -2147483648 - 2147483647";
			}


			dbg "[get_ret_in_line] ARG = \"$ARG\"\n";


			# Check for valid number
			if (($ARG < -2147483648) ||
                            ($ARG > 2147483647))
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "ARG for 'i' mode must be decimal value in the range of a signed integer: -2147483648 - 2147483647";
			}


			# NOTE: Check for valid number by counter type


			# Check for positive number
			if ($P && ($ARG < 1))
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "Sign mismatch for ARG. Mode 'p' values must be in range of a positive signed integer: 1 - 2147483647";
			}


			# Check for negative number
			if ($N && ($ARG > 0))
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "Sign mismatch for ARG. Mode 'n' values must be in range of a negative signed integer: -2147483648 - 0";
			}


			# NOTE: Result
			#
			#  A global or functional return tag with 'i' mode and valid ARG,
			#  with either 'p' or 'n' mode.
		}
		elsif ($PHASE eq 'SECOND')
		{
			# NOTE: Second phase
			#
			#  Process 'i' mode without ARG.
			#
			#  Return tags of this kind use the state counter for assigning a return value.


			# Check for return tag type
			if ($TYPE eq 'F')
			{
				# Check for 'i' mode without ARG
				if (!$I || ($ARG != undef))
				{
					# Remove matching string from line
					$LINE =~ s/\Q$MATCH\E//;


					next;
				}
			}
			else
			{
				# Check for 'i' mode
				if (!$I)
				{
					# Remove matching string from line
					$LINE =~ s/\Q$MATCH\E//;


					next;
				}
			}


			# NOTE: Global or functional return tag with 'i' mode and without ARG


			# Check for 'p' and 'n' mutual exclusion
			if ($P && $N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "Only one of 'p' or 'n' mode may be specified, but not both. Valid modes are: iopntgm";
			}


			# Check for at least one of 'p' or 'n'
			if (!$P && !$N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "At least one of either 'p' or 'n' mode must be specified.";
			}


			# NOTE: Result
			#
			#  A functional return tag with 'i' mode and without ARG,
			#  with either 'p' or 'n' mode.
			#
			#  A global return tag with 'i' mode and optional ARG (set with 'g' mode),
			#  with either 'p' or 'n' mode;


			dbg "[get_ret_tags_in_line] MATCH = \"$MATCH\"\n";
		}
		elsif ($PHASE eq 'THIRD')
		{
			# NOTE: Second phase
			#
			#  Process global return tags with 'i' mode and optionally with 'g' mode.
			#
			#  Return numbers are already set in process_tag_groups called at the end
			#  of phase SECOND.


			# Check for type
			if ($TYPE ne 'G')
			{
				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}


			# Check for 'i' mode
			if (!$I)
			{
				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}


			# NOTE: Lookup return number in GROUPS


			# Get return tag group name
			$GROUP_NAME = "$ARG";

			if (length($GROUP_NAME) == 0)
			{
				$GROUP_NAME = 'LAST';
			}


			# Get reference to global namespace
			$GLOB_REF = $GLOBS{$GLOB_NAME};


			# Get group array
			$GROUP_ARRAY_REF = $GLOB_REF->{'*'}{__STATE__}{GROUPS}{$GROUP_NAME};


			# Get group array length
			$GROUP_ARRAY_LINDEX = (scalar @{$GROUP_ARRAY_REF}) - 1;


			# Look for same return tag match
			$MATCH_INDEX = first {"${${$GROUP_ARRAY_REF}[$_]}{MATCH}" eq "$MATCH"} 0..$GROUP_ARRAY_LINDEX;

			if (length("$MATCH_INDEX") == 0)
			{
				# NOTE: Return tag not found in groups


				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}


			# NOTE: Return number
			#
			#  In the third phase, return numbers for global return tags are
			#  set from the groups hashes in __STATE__. See process_tag_groups.


			# Get return tag from groups
			$RET_TAG_G = ${$GROUP_ARRAY_REF}[$MATCH_INDEX];


			# Get return number
			$RET_TAG{NUMBER} = ${$RET_TAG_G}{NUMBER};
		}
		else
		{
			# NOTE: Fourth phase
			#
			#  Process 'o' mode return tags.


			# Check for 'o' mode without ARG
			if (!$O || ($ARG != undef))
			{
				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}


			# NOTE: Global or functional return tag with 'o' mode and without ARG


			# Check for 'p' and 'n' mutual exclusion
			if ($P && $N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "Only one of 'p' or 'n' mode may be specified, but not both. Valid modes are: iopntgm";
			}


			# Check for at least one of 'p' or 'n'
			if (!$P && !$N)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "RETURN TAG: $MATCH\n"; 
				error_and_exit_with_msg "At least one of either 'p' or 'n' mode must be specified.";
			}


			# NOTE: Result
			#
			#  A global or functional return tag with 'o' mode and without ARG,
			#  with either 'p' or 'n' mode.


			dbg "[get_ret_tags_in_line] MATCH = \"$MATCH\"\n";
		}


		# NOTE: Check for valid return tag


		# Check global namespace only for modes not equal to 'o'
		if (!$O)
		{
			# Check for correct global namespace
			if ($GLOB ne $GLOB_NAME)
			{
				$LINE =~ s/\n*$//;
				dbg "LINE: $LINE\n";
				dbg "TAG: $MATCH\n"; 
				error_and_exit_with_msg "Mismatch of global namespaces. Current global namespace is \"$GLOB_NAME\", but return tag has \"$GLOB\"";
			}
		}


		# Check functional namespace only if not in phase FOURTH
		if ($PHASE ne 'FOURTH')
		{
			# Check functional namespace only for functional type
			if ($TYPE eq 'F')
			{
				# Check for correct functional namespace
				if ($FUNC ne $FUNC_NAME)
				{
					$LINE =~ s/\n*$//;
					dbg "LINE: $LINE\n";
					dbg "TAG: $MATCH\n"; 
					error_and_exit_with_msg "Mismatch of functional namespaces. Current functional namespace is \"$FUNC_NAME\", but return tag has \"$FUNC\"";
				}
			}
		}


		# NOTE: Process macros
		#
		#  Macro processing is done as the last step.


		if ($M)
		{
			#print "[get_ret_tags_in_line] Return tag = $MATCH\n";
			#print "[get_ret_tags_in_line] Condition = $MACRO\n";


			# Check for macro definitions
			if (!defined($MACRO_DEFINITIONS))
			{
				# NOTE: Build macro definitions


				# Get reference to global namespace
				$GLOB_REF = $GLOBS{$GLOB_NAME};


				# Get reference to macro hash
				$MACRO_HASH_REF = ${$GLOB_REF}{__STATE__}{MACROS};


				# Loop through macros
				$MACRO_DEFINITIONS = "";
				foreach $MACRO_NAME (keys(%{$MACRO_HASH_REF}))
				{
					# Get macro value
					$MACRO_VALUE = ${$MACRO_HASH_REF}{$MACRO_NAME};


					$MACRO_DEFINITIONS .= "my \$$MACRO_NAME = $MACRO_VALUE;\n";
				}
			}


			# NOTE: Evaluate macro
			#
			#  If result is 1, return tag is pushed into result array. Otherwise,
			#  return tag is ignored.


                        # Evaluate as simple if
                        $IFCOND = "$MACRO_DEFINITIONS\n\
if ($MACRO)\
{\
        \t\$MACRO_RESULT = 1;\
}\
else\
{\
        \t\$MACRO_RESULT = 0;\
}";


                        {
                                #print "IFCOND = '\n$IFCOND'\n";


                                $MACRO_RESULT = 0;
                                eval $IFCOND;
                        }


                        #print "\$MACRO_RESULT = $MACRO_RESULT\n";


			if ($MACRO_RESULT == 0)
			{
				# NOTE: Ignore return tag


				# Remove matching string from line
				$LINE =~ s/\Q$MATCH\E//;


				next;
			}
		}


		# NOTE: Set return tag hash


		# Fill return tag hash
		$RET_TAG{TYPE} = $TYPE;
		$RET_TAG{MATCH} = $MATCH;
		$RET_TAG{GLOB} = $GLOB;
		$RET_TAG{FUNC} = $FUNC;
		$RET_TAG{NAME} = $RET;
		$RET_TAG{I} = $I;
		$RET_TAG{O} = $O;
		$RET_TAG{P} = $P;
		$RET_TAG{N} = $N;
		$RET_TAG{T} = $T;
		$RET_TAG{G} = $G;


		# Skip third phase for default return number
		if ($PHASE ne 'THIRD')
		{
			$RET_TAG{NUMBER} = undef;
		}


		$RET_TAG{ARG} = $ARG;


		# Remove matching string from line
		$LINE =~ s/\Q$MATCH\E//;


		# Push into result array
		push @{$RET_REF}, \%RET_TAG;
	}
}


sub parse_func
{
	my $PHASE = shift;
	my $FH_REF = shift;
	my $GLOB_NAME = shift;
	my $CUR_NAME = shift;
	my $START_NAME;
	my $STOP_NAME;
	my $FUNC_NAME;
	my $RET_TAG;
	my $RET_NUM;
	my $RET_REF;
	my $RET_NAME;
	my $RET_INDEX;
	my $GLOB_REF;
	my $FUNC_REF;
	my $LINE;
	my $TAG;


	my @RET_TAGS;


	my %RET_TAG;


	# Get hash references
	$GLOB_REF = $GLOBS{$GLOB_NAME};


	while ($LINE = next_line($FH_REF))
	{
		# NOTE: Return tags
		#
		#  Functional namespaces may contain only return tags.
		#
		#  Multiple return tags in the same line are allowed.


		#dbg "[parse_func] PARSE LINE = \"$LINE\"\n";


		# Get all return tags in the same line
		@RET_TAGS = ();
		get_ret_tags_in_line($PHASE, $GLOB_NAME, $CUR_NAME, $LINE, \@RET_TAGS);


		# NOTE: @RET_TAGS
		#
		#  @RET_TAGS contains all return tags, as hashes,
		#  found in the same line.


		#dbg "[parse_func] RET_TAGS = (", join(', ', @RET_TAGS), ")\n";

		foreach $TAG (@RET_TAGS)
		{
			dbg "[parse_func] RETURN TAG = ${$TAG}{NAME}\n";
		}


		# Process return tags sequentially
		for $RET_REF (@RET_TAGS)
		{
			# NOTE: $RET_REF
			#
			#  $RET_REF is a reference to a return tag hash with the following keys:
			#
			#   TYPE => F for Functional, G for Global
			#   MATCH => Matching string, i.e. %...:...:...[MODES,ARG]%
			#   GLOB => Global namespace name
			#   FUNC => Functional namespace name (or * if global namespace return tag)
			#   NAME => Return tag name
			#   I => 0 or 1 if mode i was specified
			#   O => 0 or 1 if mode o was specified
			#   P => 0 or 1 if mode p was specified
			#   N => 0 or 1 if mode n was specified
			#   T => 0 or 1 if mode t was specified
			#   G => 0 or 1 if mode g was specified
			#   ARG => Mode argument (or undef if not specified)


			# Set FUNC_NAME
			$FUNC_NAME = $CUR_NAME;


			if ((($PHASE eq 'FIRST') && (${$RET_REF}{I})) ||
                            ($PHASE eq 'SECOND'))
			{
				# NOTE: Add to GLOBS


				# Check for functional return tag
				if (${$RET_REF}{TYPE} eq 'F')
				{
					$FUNC_REF = ${$GLOB_REF}{$CUR_NAME};
					#$FUNC_NAME = $CUR_NAME;


					dbg "[parse_func] KEYS = ", scalar(keys(%{$FUNC_REF})), "\n";
				}
				else
				{
					# NOTE: Global return tags
					#
					#  Global return tags are stored in the special key '*',
					#  which is not allowed as a functional namespace name, nor
					#  any other global return tag name.


					$FUNC_REF = ${$GLOB_REF}{'*'};
					$FUNC_NAME = '*';
				}


				# Get current total number of return tags
				$RET_NUM = (scalar(keys(%{$FUNC_REF})) - 1) / 2;


				# NOTE: Add as return tag name


				# Get return tag name
				$RET_NAME = ${$RET_REF}{NAME};


				# Check for duplicate return tag name
				if (exists(${$FUNC_REF}{$RET_NAME}))
				{
                                	error_and_exit_with_msg "Duplicate return tag name \"$RET_NAME\" in functional namespace \"$CUR_NAME\", in global namespace \"$GLOB_NAME\".";
				}


				# Add as name
				${$FUNC_REF}{$RET_NAME} = $RET_REF;


				#dbg "[parse_func] $GLOB_NAME.$FUNC_NAME.$RET_NAME:\n";
				#print_hash $RET_REF;


				# NOTE: Add as return tag index


				# Get next return tag index
				#$RET_INDEX = $RET_NUM;


				# Add as index
				#${$FUNC_REF}{$RET_INDEX} = $RET_REF;


				#dbg "[parse_func] $GLOB_NAME.$FUNC_NAME.$RET_INDEX:\n";
				dbg "[parse_func] $GLOB_NAME.$FUNC_NAME.$RET_NAME:\n";
				#print_hash $RET_REF;
			}


			# NOTE: Process return tag


			# Process return tags for all three phases
			process_tag $PHASE, $FH_REF, $GLOB_NAME, $FUNC_NAME, $RET_REF;


			# TEMP: Show modified line
			dbg "MODIFIED LINE = \"", cur_line($FH_REF), "\"\n";
		}


		# Check for found return tags
		if (@RET_TAGS > 0)
		{
			next;
		}


		# NOTE: Functional namespaces
		#
		#  A functional namespace can not be nested.
		#
		#  FUNC START tags are checked for too.


		# Check for FUNC START tag
		if ($LINE =~ /%FUNC:(.*?):START%[ \t]*$/)
		{
			# Check functional namespace name
                        $START_NAME = $1;
                        if ($START_NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
                        {
                                error_and_exit_with_msg "Invalid functional namespace name \"$START_NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and name must start with a letter.";
                        }


			error_and_exit_with_msg "Unexpected FUNC START tag. Expected FUNC STOP tag. Current functional namespace is \"$CUR_NAME\"";
		}


		# Check for FUNC STOP tag
		if ($LINE =~ /%FUNC:(.*?):STOP%[ \t]*$/)
		{
			# Check functional namespace name
                        $STOP_NAME = $1;
                        if ($STOP_NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
                        {
                                error_and_exit_with_msg "Invalid functional namespace name \"$STOP_NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and name must start with a letter.";
                        }


			# Check for current global namespace
			if ($STOP_NAME ne $CUR_NAME)
			{
				error_and_exit_with_msg "Unexpected FUNC STOP tag. Current functional namespace is \"$CUR_NAME\"";
			}


			if ($PHASE eq 'FOURTH')
			{
				# NOTE: Remove GLOB and FUNC in FOURTH phase


				#cur_line_del $FH_REF;
				cur_line_blank $FH_REF;
			}


			dbg "[parse_glob] FUNC STOP \"$STOP_NAME\"\n";


			return 2;	# SUCCESS: FUNC STOP encountered
		}
	}


	# NOTE: EOF
	#
	#  The current functional namespace has no corresponding FUNC STOP tag.


	error_and_exit_with_msg "Missing FUNC STOP tag for current functional namespace \"$CUR_NAME\"";
}


sub add_macro
{
	my $GLOB_NAME = shift;
	my $MACRO_NAME = shift;
	my $MACRO_VALUE = shift;
	my $GLOB_REF;
	my $MACRO_HASH_REF;


	# Get reference to target global namespace
	$GLOB_REF = $GLOBS{$GLOB_NAME};


	# Get reference to macro hash
	$MACRO_HASH_REF = ${$GLOB_REF}{__STATE__}{MACROS};


	# Check for number
	if ($MACRO_VALUE =~ m/^[0-9]+/)
	{
		#print "Macro is a number\n";
		${$MACRO_HASH_REF}{$MACRO_NAME} = $MACRO_VALUE;
	}
	else
	{
		#print "Macro is a string\n";
		${$MACRO_HASH_REF}{$MACRO_NAME} = "$MACRO_VALUE";
	}
}


sub parse_glob
{
	my $FILE = shift;
	my $PHASE = shift;
	my $FH_REF = shift;
	my $CUR_NAME = shift;
	my $NEW_NAME;
	my $STOP_NAME;
	my $FUNC_NAME;
	my $RET_REF;
	my $GLOB_REF;
	my $GLOB_NAME;
	my $GROUP_NAME;
	my $GROUP_ARRAY_REF;
	my $DB_FILE;
	my $LINE;
	my $MACRO_NAME;
	my $MACRO_VALUE;
	my $MACRO_HASH_REF;


	dbg "[parse_glob] START\n";


	while ($LINE = next_line($FH_REF))
	{
		# NOTE: Nesting
		#
		#  Global namespaces, in contrast to functional namespaces,
		#  can be nested.
		#
		#  For each new GLOB START tag, invoke a new parse_glob subroutine.


		# NOTE: START tag
		#
		#  GLOB START tags are only allowed for new global namespaces.


		#dbg "[parse_glob] LINE in $PHASE = \"$LINE\"\n";


		# Check for GLOB START tag
		if ($LINE =~ /%GLOB:(.*?):START%[ \t]*$/)
		{
			# Check global namespace name
			$NEW_NAME = $1;
			if ($NEW_NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
			{
				error_and_exit_with_msg "Invalid global namespace name \"$NEW_NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and name must start with a letter.";
			}


			# Check for current global namespace name
			if ($NEW_NAME eq $CUR_NAME)
			{
				error_and_exit_with_msg "Invalid global namespace named \"$NEW_NAME\". Namespace already started. Expected new global namespace or STOP tag";
			}


			# NOTE: Create new global namespace


			# NOTE: Namespace creation
			#
			#  Global namespaces are created only once:
			#
			#   1. from a database file
			#
			#   2. during the first parsing phase
			#
			#  Since database files are parsed first, always check for global namespaces
			#  before creating a new one.


			if ($PHASE eq 'FIRST')
			{
				# Check for first GLOB START
				if (!exists($GLOBS{$NEW_NAME}))
				{
					$GLOBS{$NEW_NAME} =	{
							__STATE__ =>	{
									MODIFIED => 0,
									MACROS => {}
									},

							'*' =>		{
									__STATE__ =>	{
											COUNTER =>	{
													P => 1,
													N => 0 
													},
											ARGS =>		{
													P => [],
													N => []
													},
											GROUPS =>	{
													LAST => []
													}
											}
									}
							};


					dbg "[parse_file] GLOB START \"$NEW_NAME\"\n";
				}
			}
			elsif ($PHASE eq 'FOURTH')
			{
				# NOTE: Remove GLOB and FUNC in FOURTH phase


				#cur_line_del $FH_REF;
				cur_line_blank $FH_REF;
			}


			parse_glob($FILE, $PHASE, $FH_REF, $NEW_NAME);


			next;
		}


		# NOTE: STOP tag
		#
		#  GLOB STOP tags are only allowed for the current global namespace.


		# Check for GLOB STOP tag
		if ($LINE =~ /%GLOB:(.*?):STOP%[ \t]*$/)
		{
			# Check global namespace name
			$STOP_NAME = $1;
			if ($STOP_NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
			{
				error_and_exit_with_msg "Invalid global namespace name \"$STOP_NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and name must start with a letter.";
			}


			# Check for current global namespace
			if ($STOP_NAME ne $CUR_NAME)
			{
				error_and_exit_with_msg "Unexpected GLOB STOP tag. Current global namespace is \"$CUR_NAME\"";
			}


			if ($PHASE eq 'FOURTH')
			{
				# NOTE: Remove GLOB and FUNC in FOURTH phase


				#cur_line_del $FH_REF;
				cur_line_blank $FH_REF;
			}


			if ($PHASE eq 'SECOND')
			{
				# NOTE: Groups processing
				#
				#  After the second parsing phase, global input return tags
				#  are grouped by number.
				#
				#  These groups need to be processed sequentially from lower
				#  to higher to assign return numbers.
				#
				#  Return tag groups are processed for each global namespace, since
				#  each of them can have its own set of global return tags.


				# Get global namespace
				$GLOB_REF = $GLOBS{$CUR_NAME};


				dbg "[parse_glob] process_tag_groups for $CUR_NAME\n";
				dbg Dumper $GLOB_REF;


				process_tag_groups $GLOB_REF;


			}


			dbg "[parse_glob] GLOB STOP \"$CUR_NAME\"\n";


			return 2;	# SUCCESS: GLOB STOP encountered
		}


		# NOTE: Functional namespaces


		# Check for FUNC START tag
		if ($LINE =~ /%FUNC:(.*?):START%[ \t]*$/)
		{
			# Check functional namespace name
			$FUNC_NAME = $1;
			if ($FUNC_NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
			{
				error_and_exit_with_msg "Invalid functional namespace name \"$FUNC_NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and name must start with a letter.";
			}


			# NOTE: Functional namespace found


			$GLOB_REF = $GLOBS{$CUR_NAME};


			if ($PHASE eq 'FIRST')
			{
				# Check for functional namespace in current global namespace
				if (exists(${$GLOB_REF}{$FUNC_NAME}))
				{
					error_and_exit_with_msg "Unexpected FUNC START for namespace \"$FUNC_NAME\". Namespace already exists in global namespace \"$CUR_NAME\"";
				}
			}


			# NOTE: Namespace creation
			#
			#  Functional namespaces are created only once during the
			#  first parsing phase.


			if ($PHASE eq 'FIRST')
			{
				# Create new functional namespace
				${$GLOB_REF}{$FUNC_NAME} =	{
								__STATE__ =>	{
										COUNTER =>	{
												P => 1,
												N => 0 
												},
										ARGS =>		{
												P => [],
												N => []
												}
										}
								};


				dbg "[parse_glob] FUNC START \"$FUNC_NAME\"\n";


				# NOTE: Global namespace modification
				#
				#  Current global namespace has a new functional namespace,
				#  which is considered a modification.


				$GLOBS{$CUR_NAME}{__STATE__}{MODIFIED} = 1;
			}


			if ($PHASE eq 'FOURTH')
			{
				# NOTE: Remove GLOB and FUNC in FOURTH phase


				#cur_line_del $FH_REF;
				cur_line_blank $FH_REF;
			}


			parse_func($PHASE, $FH_REF, $CUR_NAME, $FUNC_NAME);
		}


		# NOTE: Imports


		# Check for IMPORT tag
		if ($LINE =~ /%IMPORT:GLOB:(.*?):MACRO:(.*?)%[ \t]*$/)
		{
			# NOTE: Import tag
			#
			# Import tags only in phase FIRST.


			$GLOB_NAME = $1;
			$MACRO_NAME = $2;


			if ($PHASE eq 'FIRST')
			{
				#print "[parse_glob] IMPORT:GLOB:\"$GLOB_NAME\":MACRO:\"$MACRO_NAME\"\n";


				# Check for preceding '$'
				if (($MACRO_NAME =~ s/^\$//) != 1)
				{
					error_and_exit_with_msg "Macro name must start with \$\n";
				}


				#print "IMPORT MACRO: $MACRO_NAME\n";


				# NOTE: Import macro


				# Get reference to foreign global namespace
				$GLOB_REF = $GLOBS{$GLOB_NAME};


				# Add macro to current global namespace
				add_macro($CUR_NAME, $MACRO_NAME, ${$GLOB_REF}{__STATE__}{MACROS}{$MACRO_NAME});


				# NOTE: Global namespace modification
				#
				#  Current global namespace has a new functional namespace,
				#  which is considered a modification.


				$GLOBS{$CUR_NAME}{__STATE__}{MODIFIED} = 1;
			}


			#print "IMPORT\n";
		}


		# NOTE: Macros


		# Check for MACRO SET tag
		if ($LINE =~ /%MACRO:(\$[_a-zA-Z0-9]+):(.+?)%[ \t]*$/)
		{
			$MACRO_NAME = $1;
			$MACRO_VALUE = $2;


			# Check for preceding '$'
			if (($MACRO_NAME =~ s/^\$//) != 1)
			{
				error_and_exit_with_msg "Macro name must start with \$\n";
			}


			#print "MACRO SET: $MACRO_NAME\n";


			# Add macro to current global namespace
			add_macro($CUR_NAME, $MACRO_NAME, $MACRO_VALUE);


			#dbg Dumper $MACRO_HASH_REF;


			# NOTE: Global namespace modification
			#
			#  Current global namespace has a new functional namespace,
			#  which is considered a modification.


			$GLOBS{$CUR_NAME}{__STATE__}{MODIFIED} = 1;
		}


		# Check for MACRO UNSET tag
		if ($LINE =~ /%MACRO:(\$[_a-zA-Z0-9]+):%[ \t]*$/)
		{
			$MACRO_NAME = $1;


			# Check for preceding '$'
			if (($MACRO_NAME =~ s/^\$//) != 1)
			{
				error_and_exit_with_msg "Macro name must start with \$\n";
			}


			#print "MACRO UNSET: $MACRO_NAME\n";


			# Get reference to global namespace
			$GLOB_REF = $GLOBS{$CUR_NAME};


			# Get reference to macro hash
			$MACRO_HASH_REF = ${$GLOB_REF}{__STATE__}{MACROS};


			# Check if macro exists
			if (!exists(${$MACRO_HASH_REF}{$MACRO_NAME}))
			{
				error_and_exit_with_msg "Macro \"$MACRO_NAME\" does not exist";
			}


			# Delete macro
			delete(${$MACRO_HASH_REF}{$MACRO_NAME});


			#dbg Dumper $MACRO_HASH_REF;


			# NOTE: Global namespace modification
			#
			#  Current global namespace has a new functional namespace,
			#  which is considered a modification.


			$GLOBS{$CUR_NAME}{__STATE__}{MODIFIED} = 1;
		}
	}


	# NOTE: EOF
	#
	#  The current global namespace has no corresponding GLOB STOP tag.


	error_and_exit_with_msg "Missing GLOB STOP tag for current global namespace \"$CUR_NAME\"";


	return 1;	# SUCCESS
}


sub parse_file
{
	my $FILE = shift;
	my $FH;
	my $BH;
	my $OH;
	my @FH;
	my $NAME;
	my $LINE;
	my @LINES;
	my $INDEX;
	my $PHASE;
	my $DB_FILE;
	my $BACKUP;
	my $GLOB_NAME;
	my $MODIFIED;


	# Open input file
	open $FH, "+<$FILE" or error_and_exit_with_msg "Couldn't open input file \"$FILE\"";


	# Read in whole input file
	@LINES = <$FH>;


	# Close input file
	close $FH;


	# NOTE: Pseudo file handle
	#
	#  File handles are simulated using an ARRAY with three
	#  elements:
	#
	#   (\@LINES, \$INDEX, \$MODIFIED)
	#
	#  All elements are passed by reference, so all subroutines
	#  share the same view of the current line index and modified switch.


	$INDEX = 0;
	$MODIFIED = 0;
	@FH = (\@LINES, \$INDEX, \$MODIFIED);


	# NOTE: Backup
	#
	#  Always create a backup of the input file that is going
	#  to be parsed and possibly modified.


	# Get local time
	my ($sec, $min, $hour, $mday, $mon, $year, $null, $null, $null) = localtime();


	# Create backup file name
	$BACKUP = "$FILE.${mday}_" . ($mon + 1) . "_" . ($year + 1900) . "_${hour}_${min}_${sec}";


	# Open backup file
	open $BH, "+>$BACKUP" or error_and_exit_with_msg "Couldn't open backup file \"$BACKUP\"";


	# Write backup file
	print $BH @LINES;


	# Flush buffers
	$BH->autoflush;


	# Close backup file
	close $BH;


	# NOTE: Parsing phases
	#
	#  The return preprocessor parses each file in four phases:
	#
	#   1. All 'i' mode tags with ARG are processed: saved and substituted
	#
	#      All 't' mode tags with mode 'o' are processed: saved and hashed.
	#
	#      All IMPORT tags are processed.
	#
	#   2. All functional 'i' mode tags without ARG are processed: saved and substituted
	#
	#      All global 'i' mode tags without ARG are processed: saved and substituted in core
	#      by group
	#
	#   3. All global 'i' mode tags without ARG are processed: substituted
	#
	#   4. All 'o' mode tags are processed: saved 'i' mode values from previous phases
	#      and/or database files are substituted
	#
	#      All GLOB, FUNC, MACRO and IMPORT tags are removed.


	# Loop through parsing phase
	@GLOBS_IN_FILE = ();
	for $PHASE ('FIRST', 'SECOND', 'THIRD', 'FOURTH')
	{
		dbg "[parse_file] ################################### $PHASE ############################################\n";


		# Loop through lines
		while ($LINE = next_line(\@FH))
		{
			dbg "[parse_file] LINE in $PHASE = \"$LINE\"\n";


			# Check for GLOB START tag
			if ($LINE =~ /%GLOB:(.*?):START%[ \t]*$/)
			{
				# Check global namespace name
				$NAME = $1;
				if ($NAME !~ /^[a-zA-Z_][a-zA-Z0-9_]*$/)
				{
					error_and_exit_with_msg "Invalid global namespace name \"$NAME\". Allowed characters are: a-z, A-Z, 0-9, _, and must start with a letter.";
				}


				# NOTE: Create new global namespace


				# NOTE: Namespace creation
				#
				#  Global namespaces are created only once:
				#
				#   1. from a database file
				#
				#   2. during the first parsing phase of any input files
				#
				#  but can be entered in any file to add functional namespaces, return tags,
				#  or macros to a global namespace.


				if ($PHASE eq 'FIRST')
				{
					# Check for first GLOB START
					if (!exists($GLOBS{$NAME}))
					{
						$GLOBS{$NAME} =	{
								__STATE__ =>	{
										MODIFIED => 0,
										MACROS => {}
										},

								'*' =>		{
										__STATE__ =>	{
												COUNTER =>	{
														P => 1,
														N => 0 
														},
												ARGS =>		{
														P => [],
														N => []
														},
												GROUPS =>	{
														LAST => []
														}
												}
										}
								};


						dbg "[parse_file] GLOB START \"$NAME\"\n";
					}
				}
				elsif ($PHASE eq 'FOURTH')
				{
					# NOTE: Remove GLOB and FUNC in FOURTH phase


					#cur_line_del \@FH;
					cur_line_blank \@FH;
				}


				# Parse global namespace
				parse_glob $FILE, $PHASE, \@FH, $NAME;
			}
		}


		# Reset lines index for next phase
		$INDEX = 0;
	}


	# NOTE: Generate database files (one or more)


	#dbg Dumper \%GLOBS;
	#print Dumper \%GLOBS;


	# Loop through global namespaces
	foreach $GLOB_NAME (sort keys %GLOBS)
	{
		#dbg "$GLOB_NAME MODIFIED = $GLOBS{$GLOB_NAME}{__STATE__}{MODIFIED}\n";


		# Check for modified global namespace
		if ($GLOBS{$GLOB_NAME}{__STATE__}{MODIFIED})
		{
			gen_db $GLOB_NAME;


			# Reset modification switch
			$GLOBS{$GLOB_NAME}{__STATE__}{MODIFIED} = 0;
		}
	}


	# NOTE: Generate output file (only one)


	#dbg "[parse_func] MODIFIED = $MODIFIED\n";


	# Check for modified file content
	if ($MODIFIED)
	{
		# NOTE: Output file
		#
		#  The output file is generated from the modified input file after
		#  substituting all return tags with corresponding numerical values.
		#
		#  This modified content is stored in @LINES.
		#
		#  The output filename is composed of the input filename, $FILE,
		#  with a suffixed '.rpp':
		#
		#   # ./rpp.pl myfile.c
		#
		#    -> generates output file myfile.c.rpp


		# Open output file
		open $OH, ">$FILE.rpp" or error_and_exit_with_msg "Couldn't open output file \"$FILE.rpp\"";


		# Write content to output file
		print $OH @LINES;


		# Flush buffers
		$OH->autoflush;


		# Close output file
		close $OH;
	}
}


sub parse_db
{
	my $FILE_NAME = shift;
	my $GLOB_REF;
	my $GLOB_NAME;
	my $CODE;
	my $FH;


	# Open database file
	#open($FH, "<", "$FILE_NAME") or error_and_exit_with_msg "Couldn't open database file $FILE_NAME";
	sysopen($FH, "$FILE_NAME", O_RDONLY);


	# Read in GLOB hash
	{
		local $/;
		$CODE = <$FH>;
	}


	#dbg "====================================================\n";
	#dbg Dumper \%GLOBS;
	#dbg "====================================================\n";


	# Evaluate code
	$GLOB_REF = eval($CODE);


	dbg "GLOB_REF = \"$GLOB_REF\"\n";


	# Check result
	if (!defined($GLOB_REF))
	{
		error_and_exit_with_msg "Couldn't parse in database file \"$FILE_NAME\". Error was = $@";
	}


	#dbg "[parse_db] GLOB_REF = $GLOB_REF\n";


	# Get global namespace name
	$GLOB_NAME = basename($FILE_NAME);
	$GLOB_NAME =~ s/\.rpp\.db//;


	#dbg "[parse_db] GLOB_NAME = $GLOB_NAME\n";


	# Check for duplicate in GLOBS
	if ($GLOBS{$GLOB_NAME} != undef)
	{
		error_and_exit_with_msg "Global namespace \"$GLOB_NAME\" from database \"$FILE_NAME\" already exists in GLOBS";
	}


	# Add global namespace to GLOBS
	$GLOBS{$GLOB_NAME} = $GLOB_REF;


	# Reset modification switch
	$GLOBS{$GLOB_NAME}{__STATE__}{MODIFIED} = 0;


	#dbg Dumper $GLOB_REF;


	# Close database file
	close $FH;


	# Mark GLOB_NAME as having a database source
	push @DB_GLOBS, $GLOB_NAME;
}


sub print_syntax_and_exit
{
	print "Syntax: perl rpp.pl [-io] file0 ... fileN [-db database0 ... databaseN]\n";
	exit(0);
}


sub main
{
	my $ARGC;
	my $FILE;
	my $DB_INDEX;
	my $DB_FILE;
	

	my @FILES;
	my @DBS = ();


	# Syntax:
	#
	# perl rpp.pl [-io] file0 ... fileN [-db database0 ... databaseN]


	# Get number of arguments
	$ARGC = scalar @ARGV;


	#dbg "ARGV = (", join(', ', @ARGV), ")\n";


	# Check for correct number of arguments
	if ($ARGC == 0)
	{
		dbg "ERROR: Wrong number of arguments\n";
		print_syntax_and_exit;
	}


	# NOTE: At least one argument on command line


	# Check for databases
	if ($ARGV[0] eq '-db')
	{
		dbg "ERROR: At least one file is required\n";
		print_syntax_and_exit;
	}


	# Check for ignore output switch
	if ($ARGV[0] eq '-io')
	{
		# NOTE: Ignore output
		#
		#  This switch allows rpp to skip unresolved output return tags in phase FOURTH. By default,
		#  an unresolved output return tag would cause a parsing error.
		#
		#  This allows for circular dependencies to be resolved:
		#
		#   For example:
		#
		#	If a.c depends on b.c and viceversa, the procedure for processing
		#	both files, would be:
		#
		#	 rpp.pl -io a.c (generates a.rpp.db, ignoring unresolved output return tags from b.c)
		#	 rpp.pl b.c -db a.rpp.db (generates b.rpp.db with all output return tags from a.c)
		#	 rpp.pl a.c (regenerates same a.rpp.db but now output tags from b.c are available,
		#	             no -io switch required this time)
		#
		#  For this switch to allow rpp to keep catching local unresolved return tag errors, only those
		#  unresolved return tags are silently skipped (or ignored), whose namespace is:
		#
		#   1. Absolutely addressed
		#   2. Are external to the currently parsed file


		$IGNORE_OUTPUT = 1;


		# Remove switch from ARGV
		splice @ARGV, 0, 1;


		# Recheck number of arguments
		if ($ARGC == 1)
		{
			dbg "ERROR: Wrong number of arguments\n";
			print_syntax_and_exit;
		}
	}


	# NOTE: At least one argument that's not '-db', nor '-io'


	# Check for -db switch
	$DB_INDEX = first {$ARGV[$_] eq '-db'} 0..$#ARGV;

	if (length("$DB_INDEX") > 0)
	{
		# NOTE: At least the second element in ARGV


		#dbg "-db at $DB_INDEX\n";


		# Get database files
		@DBS = @ARGV[($DB_INDEX + 1)..$#ARGV];


		#dbg "DB FILES = (", join(', ', @DBS), ")\n";


		# Check for at least one database file
		if (scalar @DBS == 0)
		{
			dbg "ERROR: -db switch requires at least one filename\n";
			print_syntax_and_exit;
		}


		# Decrement for slice below
		$DB_INDEX--;
	}
	else
	{
		$DB_INDEX = $#ARGV;
	}


	# Get input files
	@FILES = @ARGV[0..$DB_INDEX];


	# NOTE: Input files in @FILES (at least one file), database files in @DBS, if specified


	dbg "FILES = ", join(', ', @FILES), "\n";
	dbg "DBS = ", join(', ', @DBS), "\n";


	dbg "Parsing database files\n";


	# Parse in database files
	foreach $DB_FILE (@DBS)
	{
		dbg "$DB_FILE...";


		parse_db $DB_FILE;


		dbg "OK\n";
	}


	#dbg Dumper \%GLOBS;


	dbg "Parsing input files\n";


	# Parse in input files
	foreach $FILE (@FILES)
	{
		dbg "$FILE...";


		parse_file $FILE;


		dbg "OK\n";
	}
}


main();


