/******************************************************************************
 * fl.c  Driver for fl device
 *             06-March-2021
 *
 * The /dev/fl device driver shows you how
 * to write a block device driver that can be
 * statically or dynamically configured into the
 * kernel.
 ******************************************************************************/
/******************************************************************************
 *        FEATURES OF THE /dev/fl
 ******************************************************************************
 *
 * The following list highlights some features of
 * the /dev/fl driver:
 *
 * o Single binary module
 *
 *   The /dev/fl driver is written to
 *   produce a single binary module (the
 *   executable image).  This single binary
 *   module has a file extension of .mod.  You
 *   compile a device driver as a single binary
 *   module so that you can statically or
 *   dynamically configure it into a IRIX
 *   UNIX kernel at any point in time.
 *
 *
 * o Dynamic configuration
 *
 * The /dev/fl driver's single binary module
 * (the fl.o file) can be dynamically
 * configured into a IRIX UNIX kernel (/unix)
 * when a user makes a request (by using the
 * ml utility) at single user or multiuser
 * time.
 ******************************************************************************/
/******************************************************************************
 *    INTERFACES that /dev/fl IMPLEMENTS
 ******************************************************************************
 *
 * o fl_open
 * o fl_close
 * o fl_read
 * o fl_write
 * o read_label
 * o fl_size
 * o fl_size64
 * o fl_strategy
 * o fl_ioctl
 *
 ******************************************************************************/
/******************************************************************************
 *            Include Files Section
 ******************************************************************************/

/******************************************************************************
 * Common driver header files
 ******************************************************************************/


// NOTE: OS header


#include "fl_conf_id.h"


// NOTE: System header files


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#if FL_CPU_BITS_V_64


#include "behavior.h"


#endif	// FL_CPU_BITS_V_64


#include <sys/param.h>
#include <sys/systm.h>
#include <sys/open.h>
#include <sys/ioctl.h>
#include <sys/proc.h>
#include <sys/map.h>
#include <sys/buf.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/mbuf.h>
#include <sys/fcntl.h>
#include <sys/dkio.h>
#include <sys/dvh.h>
#include <sys/scsi.h>
#include <sys/iobuf.h>
#include <sys/dksc.h>
#include <sys/sysmacros.h>
#include <sys/syslog.h>
#include <sys/vnode.h>
#include <sys/time.h>
#include <sys/cred.h>
#include <sys/ksynch.h>
#include <sys/stat.h>
#include <sys/uthread.h>
#include <sys/kmem.h>
#include <sys/ddi.h>
#include <sys/sema.h>
#include <sys/ptimers.h>
#include <sys/profil.h>
#include <sys/invent.h>
#include <sys/driver.h>
#include <sys/cmn_err.h>
#include <sys/mload.h>
#include <sys/var.h>
#include <sys/resource.h>
#include <sys/dvh.h>
#include <sys/fs/spec_lsnode.h>
#include <limits.h>


#               endif
#       endif
#endif


// NOTE: Local header files


// NOTE: Kelog definitions
//
//  Because of IRIX's single module requirement,
//  the following declarations and definitions must be
//  made out of place.


#include "fl.h"
#include "fld.h"
#include "fl_list.h"
#include "fl_test.h"
//#include "fl_check.h"
#include "fl_mem.h"
#include "fl_spin.h"
#include "fl_sem.h"
#include "fl_file.h"
#include "fl_proc.h"
#include "fl_print.h"


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#include "efs.h"
#include "spec_csnode.h"


#               endif
#       endif
#endif


/******************************************************************************
 * Public prototypes
 ******************************************************************************/


%GLOB:fl_c:START%


// Import macros from fl_conf_id.h
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_NAME_V_OSF1%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_NAME_V_IRIX%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_VERSION_V_5_1%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_VERSION_V_6_5%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_RELEASE_V_B%
//  %IMPORT:GLOB:fl_conf_id:MACRO:$FL_OS_RELEASE_V_30%


#if defined(FL_TESTS_F_LOAD_MOD_N_CHECK_CPU) ||\
    defined(FL_TESTS_F_FL_CONFIGURE_N_TREE) ||\
    defined(FL_TESTS_F_FL_OPEN_N_CHECK_CPU) ||\
    defined(FL_TESTS_F_FL_WRITE_N_SCS) ||\
    defined(FL_TESTS_F_FL_WRITE_N_CHECK_CPU)


#define FL_TESTS_CHECKS


#endif


/*
#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


struct install_cd_s
{
	// Is attached switch
	fl_uint32_t is_attached;


	// Partition offset (in bytes)
	fl_uint64_t part_offset;


	// Partition size (in bytes)
	fl_uint64_t part_size;
};


#               endif
#       endif
#endif


struct fl_disk_s
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	struct volume_header vh;


#               endif
#       endif
#endif


};


struct loop_entry_s
{
	fl_uint32_t chr_is_inuse;
        fl_list_t *pid_list_chr;
	fl_uint32_t blk_is_inuse;
        fl_list_t *pid_list_blk;
        fl_uint32_t first_open;
        struct file_s dev;
        struct file_s file;
        struct fl_packet_s *packet;
        struct fl_spin_s *lock;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	char *sector;
	fl_uint32_t is_exclusive;
	fl_uint32_t format_is_finished;
	fl_uint32_t format_error;
	struct install_cd_s install_cd;
        struct cred cred;
        struct file_s dev_chr;
	struct uio *uio;
	struct iovec *iov;
	fl_sem_t *format_sem;


#               endif
#       endif
#endif


	struct fl_disk_s disk;


};


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: NENTS
//
//  This macro is not defined anywhere in a
//  fully loaded IRIX development system.
//
//  Since the DIOCRDEFECTS ioctl command is defined and
//  needs this value, the NENTS macro is defined to
//  the value 0 (zero).
//
//  This writes back only the first defhdr structure
//  to user space which is enough to comply with
//  DIOCRDEFECTS (see man dksc).


struct defect_list {
	struct defect_header defhdr;


	// Not defined due to missing NENTS system macro
	//struct defect_entry defentry[NENTS];
};


#               endif
#       endif
#endif


*/


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: IRIX kernel modules
//
//  In IRIX, kernel modules are not compiled separately and merged into
//  a final single object file. Instead, the compiler expects a single
//  code file, *.c, which contains the whole code.
//
//  For ease of maintenance, the code is kept in physically separated file,
//  and merged here using the C preprocessor to create the single code file.


#include "fl_print.c"
#include "fl_mem.c"
#include "fl_spin.c"
#include "fl_list.c"
#include "fl_sem.c"
#include "fl_file.c.rpp"
#include "fl_proc.c"
//#include "fl_scsi.c"


#               endif
#       endif
#endif


/**************************** Check definitions *******************************/


#ifdef FL_CHECKS


// Correctors for SCS
//FL_CHECK_SCS_CORR(fl_write, 132, 195, 65, 82)


// Checksum for SCS
//static fl_uint32_t fl_check_sum = 0;
//static fl_uint32_t fl_check_scs_sum_fl_write = 0;


// Checksum entry for log
struct loop_entry_s checksum_entry;
 
 
// Check list
//TODO:fl_list_t *check_list;

 
#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B
 
 
#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30
 
 
// CPU check variable
static fl_int32_t cpu;


// Check lock switch
static fl_uint32_t fl_check_lock_switch = 0; 
 
 
// Check error message
static char fl_check_err_msg[100];
 
 
#define FL_CHECK_LOCK_SWITCH(RET) \
if (fl_check_lock_switch == 1)\
{\
	cmn_err(CE_WARN, "%s\n", fl_check_err_msg);\
\
\
	return RET;\
}
 
 
#define FL_CHECK_LOCK_SWITCH_NOCMN(RET) \
if (fl_check_lock_switch == 1)\
{\
        return RET;\
}
 
 
#               endif
#       endif
#endif


#else
 
 
#define FL_CHECK_LOCK_SWITCH(RET)
#define FL_CHECK_LOCK_SWITCH_NOCMN(RET)


#endif	// FL_CHECKS


/**************************** Test definitions ********************************/

#if defined(FL_TESTS) || defined(FL_CHECKS_SCS_SUM)


struct test_req_s
{
	char func[300];
	char name[300];
};


#endif


#ifdef FL_TESTS


char fl_test_func[300];
char fl_test_name[300];
char fl_test_err[300];
char fl_test_suc[300];


fl_int32_t fl_test_ret = 0;
fl_int32_t fl_test_notify = 0;
fl_uint64_t fl_test_counter = 0;
fl_uint64_t fl_test_ctrl_sem = 1;


// Sleeper lock
struct fl_spin_s *sleeper_lock;
struct fl_spin_s *sleeep_lock;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// Sleeper semaphore
sema_t sleeper_sem;


// Sleeper counter
fl_int32_t fl_test_sleeper;


// Sleep event address
fl_uint32_t fl_test_wait;


#               endif
#       endif
#endif


static fl_uint32_t fl_test_passes = 0;


#endif


/******************************************************************************
 * The /dev/fl driver's register_major_number
 * interface uses the following #defines when
 * obtaining a major number.
 ******************************************************************************/

#define NO_DEV -1


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


/******************************************************************************
 * Global variables
 ******************************************************************************/


static struct dsf_control_s dsf_control;


// Loop list
static fl_list_t *loop_list;


// TODO: Create dynamically and page aligned for I/O control
// User request buffer
//static char req_buffer[FL_PACKET_REQ_MAX];
static char req_buffer[_PAGESZ];


// TODO: Rename to req_packet once refactored
// User request packet
struct fl_packet_s packet;


// Active device bitmask
static char dev_bitmask[8192];


// Ctrl semaphore
static fl_sem_t *ctrl_sem;


// Lock for loop list
struct fl_spin_s *loop_list_lock;


// Lock for fl_open
struct fl_spin_s *fl_open_lock;


// NOTE: Daemon lock
//
//  The daemon lock is used to serialize access
//  to the following daemon state variables.


// Lock for daemon
struct fl_spin_s *daemon_lock;


// Daemon pid
pid_t daemon_pid;


// Started daemon pid
pid_t daemon_started;


// Daemon state
enum fl_dst_e daemon_state;


// Daemon ref
void *daemon_ref;


// Daemon data in user space
static char *daemon_data;


// Daemon pending I/O requests counter
static fl_uint32_t daemon_pending;


// I/O list
static fl_list_t *io_list;


// Kernel page shift
static fl_uint32_t page_shift;


// Kernel maxdmasz
static fl_int32_t fl_maxdmasz;


// Maximum transfer size in bytes per I/O
static fl_uint32_t FL_MAX_TRANSFER_SIZE;


// NOTE: Reset switch
//
//  In case of unexpected semaphore values causing
//  internal semaphore value overflows, deny further access
//  to fl_open() and ask user to reload file looper
//  kernel module in order to reset the module's internal
//  state to a known, consistent state.


static fl_uint32_t reset_switch;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: Mandatory global variables
//
//  In IRIX 6.5, the following two global variables
//  are mandatory for loading dynamically loadable
//  kernel modules.


int fl_devflag = D_MP;
char *fl_mversion = M_VERSION;


// Lock variable
static fl_uint32_t lock_switch;


// Driver init switch
static fl_uint32_t fl_driver_init;


// I/O mutex
static mutex_t *io_mutex;


// Serialization mutex
static mutex_t *fl_mutex;


// Control device vertex
static vertex_hdl_t devfl_vertex;


// Control device file
static struct file_s devfl_file;


// Device directory
static char devloop_string[10];


// Device directory file
static struct file_s devloop_file;


#               endif
#       endif
#endif


/******************************************************************************
 * The following list describes the use of the
 * fl_is_dynamic, callback_return_status,
 * and num_fl variables:
 *
 *  o fl_is_dynamic
 *
 *    The fl_is_dynamic variable will be
 *    used to control any differences in tasks
 *    performed by the statically or dynamically
 *    configured /dev/fl driver.
 *    Implementing a device driver to handle a
 *    static or dynamic configuration request
 *    gives customers maximum flexibility in how
 *    they configure the driver into the kernel.
 *
 *  o callback_return_status
 *
 *    The callback_return_status variable is used
 *    by the callback_register_major_number
 *    interface to determine if a previous failure
 *    has occurred in statically configuring the
 *    /dev/fl driver.
 *
 *  o num_fl
 *
 *    The register_major_number interface uses the
 *    num_fl variable to perform a check if
 *    there are devices present.
 ******************************************************************************/

#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30





#               endif
#       endif
#endif


/******************************************************************************
 * The following code sets the members of the
 * the dsent structure to appropriate  values
 * for the /dev/fl driver.  The address of
 * this structure is passed to the devsw_add
 * interface, which registers the I/O services
 * interfaces in the dsent table and reserves a
 * major number for the /dev/fl driver.
 ******************************************************************************/

#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#               endif
#       endif
#endif


// NOTE: Major numbers
//
//  Major numbers are treated differently in different platforms.
//
//  Retrieving the major number is necessary for dynamically
//  creating new device files in /dev/loop. See fl_dev_new().


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: Major number
//
//  The major number is retrieved through a variable in
//  /var/sysgen/master.d/fl containing the driver's major
//  number.
//
//  Internally, majnum is preserved as the interface to
//  the major number.


// Device driver major number
extern major_t fl_majnum;
major_t majnum;


// NOTE: Systune variables
//
//  The following systune variables are defined
//  for file looper:
//
//   - fl_install_cd_recognition
//
//	Check for SGI installation CDs upon attachment (fl -a).
//
//      Possible values:
//
//       0 = Don't check for SGI installation CDs
//       1 = Check for SGI installation CDs
//
//
//   - fl_check_vh_checksum
//
//	Check for correct volume header checksum when checking
//	for SGI installation CD.
//
//      Possible values:
//
//       0 = Don't check for correct volume header checksum
//       1 = Check for correct volume header checksum
//
//
//  Per man mtune, systune variables are by default of type
//  int.


extern int fl_install_cd_recognition;
extern int fl_check_vh_checksum;


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


/******************************************************************************
 *        Private check function definitions
 ******************************************************************************/

#ifdef FL_CHECKS


// NOTE: Check source code
//
//  Source code for the implemented checks is provided in corresponding
//  fl_check_[cpu|scs|tree].c files.


//#include "fl_check_cpu.c"
//#include "fl_check_scs.c"


#endif


/******************************************************************************
 *        Private function definitions
 ******************************************************************************/


%FUNC:fl_entry_new:START%


int fl_entry_new(struct loop_entry_s **loop_entry)	// -- OUT
{
	struct loop_entry_s *loop_entry_1;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// OUT: Check for loop entry
	if (loop_entry == NULL)
	{
		return %NO_LOOP_ENTRY[in,0,]%;	// ERROR: No loop entry
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// Loop entry
	*loop_entry = NULL;


	// NOTE: Processing


        // Allocate loop entry
	fl_mem_alloc((void **) &loop_entry_1,
                     sizeof(struct loop_entry_s),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_entry_new",
                     "Allocate loop entry");


	// Create lock for loop entry
	fl_spin_new(&loop_entry_1->lock);


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Create semaphore for formatting I/O control
	fl_sem_init(&loop_entry_1->format_sem,
		    0);


	// Create CHR pid list
	fl_list_new(&loop_entry_1->pid_list_chr);


        // Allocate user I/O
	fl_mem_alloc((void **) &loop_entry_1->uio,
                     sizeof(struct uio),
                     FL_MEM_ZERO | FL_MEM_WAITOK,
                     "fl_entry_new",
                     "Allocate user I/O");


        // Allocate I/O vector
	fl_mem_alloc((void **) &loop_entry_1->iov,
                     sizeof(struct iovec),
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_entry_new",
                     "Allocate I/O vector");


        // Allocate I/O sector
	fl_mem_alloc((void **) &loop_entry_1->sector,
                     FL_FILE_BLOCK,
                     FL_MEM_ZERO |
		     FL_MEM_WAITOK,
                     "fl_entry_new",
                     "Allocate I/O sector");


#               endif
#       endif
#endif


	// Create BLK pid list
        fl_list_new(&loop_entry_1->pid_list_blk);


	// Set loop entry
	*loop_entry = loop_entry_1;


	return %CREATE_LOOP_ENTRY_SUCCESS[ip,1,]%;	// SUCCESS
}


%FUNC:fl_entry_new:STOP%


int fl_entry_get_by_dev(dev_t dev,				// -- IN
			struct loop_entry_s **loop_entry_dev,	// -- OUT
			struct fl_args_s *args)			// -- IN
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	fl_uint32_t is_found;
	struct loop_entry_s *loop_entry;


	// NOTE: Arguments


	// OUT: Check for loop entry
	if (loop_entry_dev == NULL)
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = 0;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return 0;	// ERROR: No loop entry
	}


	// IN: Check for args
	if (args == NULL)
	{
		return -1;	// ERROR: No args
	}


	// NOTE: Initialization


	// Loop entry
	*loop_entry_dev = NULL;


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Device vertex
	//
	//  The device number dev contains a vertex device
	//  number.


	// Get device info
	loop_entry = (struct loop_entry_s *) device_info_get(dev);


	// Check for loop entry
	is_found = 0;
	if (loop_entry != NULL)
	{
		// Set device loop entry
		*loop_entry_dev = loop_entry;


		// Set found switch
		is_found = 1;
	}


#               endif
#       endif
#endif


	// Check if loop entry was found
	if (is_found != 1)
	{
                FL_DEBUG_UPRINTF((\
				  "[fl_entry_get_by_dev] ERROR: No loop entry for dev = %u\n",
				  dev))


		// Check for return value
		if (args->ret.var != NULL)
		{
			*args->ret.var = -1;
		}


		// Check for reset on return
		if (args->ret.reset == 1)
		{
			bzero((void *) args,
			      sizeof(struct fl_args_s));
		}


		return -1;	// ERROR: Couldn't find loop entry
	}


	// Check for return value
	if (args->ret.var != NULL)
	{
		*args->ret.var = 1;
	}


	// Check for reset on return
	if (args->ret.reset == 1)
	{
		bzero((void *) args,
		      sizeof(struct fl_args_s));
	}


	return 1;	// SUCCESS
}


int fl_node_get_by_dev_name(char *dev_name,			// -- IN
                            fl_list_node_t **loop_node_dev,	// -- OUT
                            fl_uint64_t *loop_index_dev)	// -- OUT
{
	fl_int32_t ret_compare;
        fl_uint32_t was_found;
        fl_uint64_t loop_index;
        fl_uint64_t loop_length;
        fl_list_node_t *loop_node;
	struct loop_entry_s *loop_entry;


        // NOTE: Arguments


	// IN: Check for device name
	if (dev_name == NULL)
	{
		return 0;	// ERROR: No device name
	}


	// OUT: Check for loop node
        if (loop_node_dev == NULL)
        {
                return -1;       // ERROR: No loop node
        }


	// OUT: Check for loop index
	if (loop_index_dev == NULL)
	{
		return -2;	// ERROR: No loop index
	}


	// NOTE: Initialization


	// Device loop node
        *loop_node_dev = NULL;


	// Device loop index
	*loop_index_dev = 0;


	// NOTE: Processing


        // Get length of loop list
        fl_list_length(loop_list, &loop_length);


        // Traverse loop list
        was_found = 0;
        for (loop_index = 0;
	     loop_index < loop_length;
	     loop_index++)
        {
                // Get next node
                fl_list_get_node(loop_list, &loop_node, loop_index);


                // Get loop entry
                loop_entry = (struct loop_entry_s *) loop_node->data;


		FL_DEBUG_UPRINTF((\
				  "[fl_node_get_by_dev_name] INFO: loop_entry->dev.name = \"%s\"\n",
				  loop_entry->dev.name))


		FL_DEBUG_UPRINTF((\
				  "[fl_node_get_by_dev_name] INFO: dev_name = \"%s\"\n",
				  dev_name))


                // Compare device names
		if (fl_dev_compare(loop_entry->dev.name,
				   dev_name,
				   &ret_compare) >= 1)
                {
                        // Set is found variable
                        was_found = 1;


                        // Set result variables
                        *loop_node_dev = loop_node;
			*loop_index_dev = loop_index;


                        break;
                }
        }


        // Check if loop node was found
        if (was_found != 1)
        {
                return -3;      // ERROR: Couldn't find loop node
        }


        return 1;       // SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


int fl_dev_create_loop_dir(void)
{
	fl_int32_t ret_lookup;
	fl_uint32_t is_found;
	fl_uint32_t was_created;
	struct cred cred;


	// NOTE: Initialization


	// Device loop file
	fl_file_init(&devloop_file);


	// Credentials
	bzero((void *) &cred,
	      sizeof(struct cred));


	// NOTE: Processing


	// Set filename
	devloop_file.name = devloop_string;


	// Set basename
	devloop_file.basename = &devloop_string[5];


	// Lookup /dev/loop
	is_found = 0;
	ret_lookup = 0;
	//if (FL_TEST_COND((fl_test_ret == 1), ||)
	if (fl_file_lookup(&devloop_file,
                           NO_FOLLOW,
                           &is_found,
			   &ret_lookup) < 1)
	{
		// Log warning
		cmn_err(CE_WARN, "^Couldn't lookup /dev/loop. Please reload file looper to try again.\n");


		return 0;	// ERROR: Couldn't lookup /dev/loop
	}


	// Check for /dev/loop
	if (is_found != 1)
	{
		// NOTE: No /dev/loop


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_create_loop_dir] INFO: /dev/loop not found\n"))


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_create_loop_dir] INFO: Creating /dev/loop\n"))


		// Create /dev/loop
		was_created = 0;
		if (fl_file_create(&devloop_file,
				   VDIR,
				   0,
				   S_IRUSR |
				   S_IRGRP,
				   &cred,
				   &was_created) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_create_loop_dir] ERROR: Couldn't create /dev/loop\n"))


			// Log warning
			cmn_err(CE_WARN, "^Couldn't create /dev/loop. Please reload file looper to try again.\n");


			return -1;	// ERROR: Couldn't create /dev/loop
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_create_loop_dir] INFO: Created /dev/loop\n"))
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_dev_create_loop_dir] INFO: Found /dev/loop\n"))
	}


	return 1;	// SUCCESS
}


int fl_dev_compare(char *blk_name,	// -- IN
		   char *dev_name,	// -- IN
		   fl_int32_t *ret)	// -- OUT
{


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	char blk_name_canon[PATH_MAX];


#               endif
#       endif
#endif


	// NOTE: Arguments


	// IN: Check for block device name
	if (blk_name == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = 0;
		}


		return 0;	// ERROR: No block device name
	}


	// IN: Check for user device name
	if (dev_name == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -1;
		}


		return -1;	// ERROR: No user device name
	}


	// OUT: Check for return value
	if (ret == NULL)
	{
		return -2;	// ERROR: No return value
	}


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_compare] INFO: blk_name = \"%s\"\n",
			  blk_name))


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_compare] INFO: dev_name = \"%s\"\n",
			  dev_name))


	// NOTE: Block device name vs. user device name
	//
	//  The block device name is the name of the
	//  symbolic link in /dev/loop of an attached file.
	//
	//  The user device name is the name of the hardware
	//  graph based device file.


	// Build up canonicalized block device name
	blk_name_canon[0] = '\0';
	sprintf(blk_name_canon,
		"/hw/loop/%s/block",
		&blk_name[10]);


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_compare] INFO: blk_name_canon = \"%s\"\n",
			  blk_name_canon))


	// Compare device names
	if (strncmp(blk_name_canon,
		    dev_name,
		    PATH_MAX - 1) != 0)
	{
		// Set return value
		*ret = -4;


		return -4;	// ERROR: Device names are not equal
	}


#               endif
#       endif
#endif


	// Set return value
	*ret = 1;


	return 1;	// SUCCESS
}


int fl_dev_map(dev_t dev_vertex,			// -- IN
	       fl_uint32_t dev_type,			// -- IN
	       struct loop_entry_s **loop_entry,	// -- OUT
	       dev_t *dev,				// -- OUT
	       fl_uint32_t *major,			// -- OUT
	       fl_uint32_t *minor,			// -- OUT
	       struct fl_args_s *args)			// -- IN
{
	fl_uint32_t dev_minor;
	struct loop_entry_s *loop_entry_1;


	// NOTE: Arguments


	// IN: Check for valid device type
	if ((dev_type != OTYP_CHR) &&
	    (dev_type != OTYP_BLK))
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = 0;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return 0;	// ERROR: Invalid device type
	}


	// OUT: Check for loop entry
	if (loop_entry == NULL)
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = -1;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return -1;	// ERROR: No loop entry
	}


	// OUT: Check for device
	if (dev == NULL)
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = -2;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return -2;	// ERROR: No device
	}


	// OUT: Check for major number
	if (major == NULL)
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = -3;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return -3;	// ERROR: No major number
	}


	// OUT: Check for minor number
	if (minor == NULL)
	{
		if (args != NULL)
		{
			// Check for return value
			if (args->ret.var != NULL)
			{
				*args->ret.var = -4;
			}


			// Check for reset on return
			if (args->ret.reset == 1)
			{
				bzero((void *) args,
				      sizeof(struct fl_args_s));
			}
		}


		return -4;	// ERROR: No minor number
	}


	// IN: Check for args
	if (args == NULL)
	{
		return -5;	// ERROR: No args
	}


	// NOTE: Initialization


	// Device special file
	*dev = makedev(0, 0);


	// Device major number
	*major = 0;


	// Device minor number
	*minor = 0;


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_map] INFO: START\n"))


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_map] INFO: dev_vertex = %u ?? devfl_vertex = %u\n",
			  dev_vertex,
			  devfl_vertex))


	// Check for control device
	if (dev_vertex == devfl_vertex)
	{
                // Release lock for loop list
                fl_spin_unlock(loop_list_lock);


		// Check for return value
		if (args->ret.var != NULL)
		{
			*args->ret.var = 2;
		}


		// Check for reset on return
		if (args->ret.reset == 1)
		{
			bzero((void *) args,
			      sizeof(struct fl_args_s));
		}


		return 2;	// SUCCESS: Control device
	}


	// Get device info
	loop_entry_1 = (struct loop_entry_s *) device_info_get(dev_vertex);
	if (loop_entry_1 == NULL)
	{
                // Release lock for loop list
                fl_spin_unlock(loop_list_lock);


		// Check for return value
		if (args->ret.var != NULL)
		{
			*args->ret.var = -5;
		}


		// Check for reset on return
		if (args->ret.reset == 1)
		{
			bzero((void *) args,
			      sizeof(struct fl_args_s));
		}


		return -5;	// ERROR: No device info
	}


	// Set loop entry
	*loop_entry = loop_entry_1;


	if (dev_type == OTYP_CHR)
	{
		// NOTE: CHR device


		// Set minor number device file
		dev_minor = loop_entry_1->dev_chr.minor;


		// Set device number
		*dev = loop_entry_1->dev_chr.num;
	}
	else
	{
		// NOTE: BLK device


		// Set minor number device file
		dev_minor = loop_entry_1->dev.minor;


		// Set device number
		*dev = loop_entry_1->dev.num;
	}


	// Set minor device number
	*minor = dev_minor;


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_map] INFO: Mapped vertex %u to dev %u major %u and minor %u and loop entry %p\n",
			  dev_vertex,
			  *dev,
			  *major,
			  *minor,
			  loop_entry_1))


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	// Check for return value
	if (args->ret.var != NULL)
	{
		*args->ret.var = 1;
	}


	// Check for reset on return
	if (args->ret.reset == 1)
	{
		bzero((void *) args,
		      sizeof(struct fl_args_s));
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_map] INFO: STOP\n"))


	return 1;	// SUCCESS
}


int fl_dev_check_minor_is_attached(fl_uint32_t minor)	// -- IN
{
        fl_uint32_t byte;
        fl_uint32_t byte_index;
        fl_uint32_t bit_index;


	// NOTE: Arguments


	// IN: Check correct range for minor
	if ((minor <= 1) ||
	    (minor > ((8192 * 8) - 1)))
	{
		return 0;	// ERROR: Minor number is out of range
	}


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_check_minor_is_attached] INFO: START\n"))


	// Calculate byte offset
	byte_index = minor / 8;


	// Get byte
	byte = dev_bitmask[byte_index];


	// Get bit offset
	bit_index = minor % 8;


	// Check bit
	if ((byte & (char) (1 << bit_index)) == 0)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_dev_check_minor_is_attached] INFO: Minor number is not attached = %u\n",
				  minor))


		return -1;	// ERROR: Minor number is not attached
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_check_minor_is_attached] INFO: Minor number is attached = %u\n",
			  minor))


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_check_minor_is_attached] INFO: STOP\n"))


	return 1;	// SUCCESS
}


int fl_dev_is_attached(char *dev,			// -- IN
		       fl_uint32_t *is_attached)	// -- OUT
{
	fl_uint32_t is_attached_loc;
	fl_uint64_t list_index;
	fl_uint64_t list_length;
	fl_list_node_t *list_node;
	struct loop_entry_s *loop_entry_att;


	// NOTE: Arguments


	// IN: Check for device name
	if (dev == NULL)
	{
		return 0;	// ERROR: No device name
	}


	// OUT: Check for is attached switch
	if (is_attached == NULL)
	{
		return -1;	// ERROR: No is attached switch
	}


	// NOTE: Processing


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	// Get length of loop list
	fl_list_length(loop_list, &list_length);


	//FL_DEBUG_UPRINTF((\
			    //"[fl_dev_is_attached] list_length = %llu\n", list_length))


	// Traverse loop list
	is_attached_loc = 0;
	for (list_index = 0;
	     list_index < list_length;
	     list_index++)
	{
		// Get next node
		fl_list_get_node(loop_list,
				 &list_node,
				 list_index);


		// Get attached loop entry
		loop_entry_att = (struct loop_entry_s *) list_node->data;


		// Look for device
		if (strcmp(dev,
			   loop_entry_att->dev.name) == 0)
		{
			is_attached_loc = 1;
			break;
		}
	}


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	// Set is attached switch
	*is_attached = is_attached_loc;


	return 1;	// SUCCESS
}


int fl_dev_get_lminor(fl_uint32_t *lminor,	// -- OUT
		      fl_uint32_t *is_full)	// -- OUT
{
        fl_uint32_t is_found;
        fl_uint32_t byte;
        fl_uint64_t byte_index;
        fl_uint64_t bit_index;


	// NOTE: Arguments


        // OUT: Check for lowest minor
        if (lminor == NULL)
        {
                return 0;       // ERROR: No lowest minor
        }


	// OUT: Check for is full switch
        if (is_full == NULL)
        {
                return -1;      // ERROR: No is full switch
        }


	// NOTE: Processing


        // Traverse device bytes
        is_found = 0;
        for (byte_index = 0;
	     byte_index < 8192;
	     byte_index++)
        {
                // Get target byte
                byte = dev_bitmask[byte_index];


                // Traverse device bits
                for (bit_index = 0;
		     bit_index < 8;
		     bit_index++)
                {
                        // Check for free device minor number
                        if ((byte & (char) (1 << bit_index)) == 0)
                        {
                                // Set is found
                                is_found = 1;

                                break;
                        }
                }


                // Check for found minor number
                if (is_found == 1)
                {
                        break;
                }
        }


        // Check loop end condition
        *is_full = 1;
        *lminor = 0;
        if (is_found == 1)
        {
                // Calculate lowest minor number
                *lminor = (byte_index * 8) + bit_index;


                // Set result variable
                *is_full = 0;
        }


        return 1;       // SUCCESS
}


int fl_dev_set_lminor(fl_uint32_t lminor,	// -- IN
		      fl_uint32_t value)	// -- IN
{
        fl_uint32_t byte;
        fl_uint32_t byte_index;
        fl_uint32_t bit_index;


	// NOTE: Arguments


	// IN: Check for valid value
	if ((value != 0) &&
	    (value != 1))
	{
		return 0;	// ERROR: Invalid value
	}


        //FL_DEBUG_UPRINTF(("[fl_dev_set_lminor] INFO: lminor = %u value = %u\n", lminor, value))


	// NOTE: Processing


        // Calculate byte index
        byte_index = lminor / 8;


        // Calculate bit index
        bit_index = lminor - (byte_index * 8);


        // Get byte
        byte = dev_bitmask[byte_index];


        // Modify byte
        if (value == 1)
        {
                //FL_DEBUG_PRINTF(("[fl_dev_set_lminor] INFO: byte = %u\n", byte))


                byte |= (char) 1 << bit_index;


                //FL_DEBUG_PRINTF(("[fl_dev_set_lminor] INFO: byte = %u\n", byte))
        }
        else
        {
                byte &= (char) ~(1 << bit_index);
        }


        // Set byte
        dev_bitmask[byte_index] = byte;


        return 1;       // SUCCESS
}


%FUNC:fl_dev_check_path:START%


int fl_dev_check_path(char *dev)	// -- IN
{
	fl_uint8_t chr;
	fl_uint32_t char_index;
	fl_uint32_t char_length;
	fl_uint32_t dev_name_index;


	// NOTE: Arguments


	// IN: Check for device name
	if (dev == NULL)
	{
		return %DEVNAME_NOT_FOUND[in,,]%;	// ERROR: No device name
	}


	// NOTE: Processing


	// NOTE: Device pathname
	//
	//  File looper stores its device files in the /dev/loop directory. This path
	//  is canonicalized in user space by the fl user command.
	//
	//  Depending on the platform, this path may be a real path or a symbolic
	//  link.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Canonical pathname
	//
	//  On IRIX, the canonicalized pathname of /dev/loop is
	//
	//    /dev/loop


        // Check directory prefix
        if ((dev[0] != '/') ||
            (dev[1] != 'd') ||
            (dev[2] != 'e') ||
            (dev[3] != 'v') ||
            (dev[4] != '/') ||
            (dev[5] != 'l') ||
            (dev[6] != 'o') ||
            (dev[7] != 'o') ||
            (dev[8] != 'p') ||
            (dev[9] != '/'))


#		endif
#	endif
#endif

        {
		dsf_control.ret = %fl_c:*:WRONG_DIR_PREFIX[in,,]%;


		return %WRONG_DIRECTORY_PREFIX[in,,]%;	// ERROR: Wrong directory prefix
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Set device name starting index
	dev_name_index = 10;


#               endif
#       endif
#endif


	// Check for empty device name
	if (dev[dev_name_index] == '\0')
	{
		dsf_control.ret = %fl_c:*:EMPTY_BLK_DEV_NAME[in,,]%;


		return %EMPTY_BLK_DEVICE_NAME[in,,]%;	// ERROR: Empty BLK device name
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: CHR device
	//
	//  On IRIX, both device types are necessary, CHR and BLK.
	//
	//  CHR device files typically are prepended a lower case 'r'
	//  to distinguish them from their BLK counterpart.
	//
	//  To avoid device file clobbering, a device name to be attached
	//  may never start with a lower case 'r'.
	//
	//  Clobbering may result when, for example, a second
	//  attachment tries to attach a device file starting with 'r',
	//  after a first attachment that created this same file as its CHR
	//  device. The new BLK device, if leading 'r's would be allowed,
	//  would overwrite the CHR device file of the first attachment:
	//
	//  First attachment:
	//
	//   # fl -a /dev/loop/myiso /path/to/data.iso
	//   # ls /dev/loop
	//   myiso rmyiso
	//   #
	//
	//  Second attachment:
	//
	//   # fl -a /dev/loop/rmyiso /path/to/data.iso
	//
	//   The CHR device file from the first attachment would get overwritten
	//   with the new BLK device file from this second attachment.


	// Check for leading 'r'
	if (dev[dev_name_index] == 'r')
	{
		dsf_control.ret = %fl_c:*:NOT_START_R_LETTER_BLK_DEV_NAME[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %DEVICE_NAME_NOT_START_WITH_R[in,,]%;	// ERROR: Device names may not start with the letter 'r'
	}


#               endif
#       endif
#endif


	// NOTE: Allowed characters
	//
	//  -: 45
	//  .: 46
	//  0: 48
        //  ...
        //  9: 57
	//  A: 65
	//  ...
	//  Z: 90
	//  _: 95
	//  a: 97
	//  ...
	//  z: 122
	//
	//  Device name must start with a character, dot, dash,
	//  or underscore.


	// Get length of device name
	char_length = 0;
	while (char_length < PATH_MAX)
	{
		// Check for NULL byte
		if (dev[char_length] == '\0')
		{
			break;
		}


		// Increment length
		char_length++;
	}


	// NOTE: CHR device name
	//
	//  The corresponding CHR device file name has a prepended
	//  letter 'r' to the name of the BLK device file name.
	//
	//  This extra letter must be accounted for when checking for
	//  BLK device name overflow.
	//
	//  See PATH_MAX - 1 below.


	// Check for device name overflow
	if (char_length >= (PATH_MAX - 1))
	{
		dsf_control.ret = %fl_c:*:DEVICE_NAME_OVERFLOW[in,,]%;


		return %DEVICE_NAME_OVERFLOW[in,,]%;	// ERROR: Device name overflow
	}


	// Check for valid characters
	for (char_index = dev_name_index;
	     char_index < char_length;
	     char_index++)
	{
		// Get next character
		chr = dev[char_index];


		// Compare
		if ((chr == '-') ||
		    (chr == '.') ||
		    (chr == '_') ||
		    ((chr >= 48) && (chr <= 57)) ||
		    ((chr >= 65) && (chr <= 90)) ||
		    ((chr >= 97) && (chr <= 122)))
		{
			continue;
		}


		// NOTE: Device pathname contains invalid characters


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_check_path] INFO: Invalid characters = \"%s\" (%u) at index = %u is %u\n",
				  dev,
				  char_length,
				  char_index,
				  chr))


		dsf_control.ret = %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_CHAR[in,,]%;


		return %BLK_DEV_PATHNAME_HAS_INVALID_CHAR[in,,]%; // ERROR: Device name has at least one invalid character
	}


	// NOTE: Device name contains only valid characters


	// Check for valid first character
	chr = dev[dev_name_index];
	if ((chr >= 48) &&
	    (chr <= 57))
	{
		// NOTE: First character is a number


		dsf_control.ret = %fl_c:*:BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR[in,,]%;


		return %BLK_DEV_PATHNAME_HAS_INVALID_FIRST_CHAR[in,,]%;	// ERROR: Device name starts with an invalid character
	}


	return 1;	// SUCCESS
}


%FUNC:fl_dev_check_path:STOP%


// NOTE: fl_dev_new()
//
//  Creates a new device special file entry in /dev/loop.
//
//  By default, only a BLK device is created. Some platforms require
//  a corresponding CHR device to be created as well. On these
//  platforms, both device files are created.


%FUNC:fl_dev_new:START%


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_dev_new(struct loop_entry_s *loop_entry)	// -- IN
{
	char *symlink_target;
	char hwg_bdev_name[PATH_MAX];
	char hwg_cdev_name[PATH_MAX];
	fl_int32_t ret_remove;
	fl_uint32_t lminor;
	fl_uint32_t lminor_chr;
	fl_uint32_t is_full;
	fl_uint32_t is_removed;
	fl_uint32_t was_created;
	fl_uint32_t blk_is_removed;
	fl_uint32_t chr_is_removed;
	struct vattr va;
	struct cred cred;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return %NO_LOOP_ENTRY[inm,0, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: No loop entry
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// Initialize creation switch
	loop_entry->dev.is_created = 0;


	// Initialize vnode attributes
	bzero((void *) &va,
	      sizeof(struct vattr));


	// NOTE: Credentials
        //
        //  Credentials are set to zero (root)
        //  to allow for creation of any file in
        //  /dev/loop .


        // Set caller credentials
        bzero(&cred,
	      sizeof(struct cred));


	// NOTE: Processing


	// NOTE: Extend hardware graph
	//
	//  Each new device created through file looper, is added
	//  at this connection point:
	//
	//   /hw/loop/{device_name}/char  - for character device
	//   /hw/loop/{device_name}/block - for block device
	//
	//  where device_name is the name given to fl command
	//  for the device (i.e. the component after /dev/loop/).
	//
	//  In addition, the corresponding symbolic links from /dev/loop
	//  to /hw are created.
	//
	//
	//  For example:
	//
	//   # fl -a /dev/loop/mydev /usr/people/max/myfile.iso
	//
	//  would create:
	//
	//   /hw/loop/mydev/char
	//   /hw/loop/mydev/block
	//   /dev/loop/rmydev -> /hw/loop/mydev/char
	//   /dev/loop/mydev -> /hw/loop/mydev/block


	// NOTE: BLK device file
	//
	//  Get lowest free minor number for BLK device.


	fl_dev_get_lminor(&lminor,
			  &is_full);


	// OLD: FL_TEST_F_main_N_BITFIELD_IS_FULL(0, 1)
	// TODO: FL_TEST_F_main_N_BITFIELD_IS_FULL_BLK(0, 1)


	// Check if device bitfield is full
	if (FL_TEST_COND((fl_test_ret == 1), ||)
	    (is_full == 1))
	{
		dsf_control.ret = %fl_c:*:BITFIELD_IS_FULL_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %BITFIELD_IS_FULL_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Device bitfield is full
	}


	// NOTE: CHR device file
	//
	//  Get next minor number for CHR device.


	lminor_chr = lminor + 1;


	// Create hardware graph block device
	if (fl_hwgraph_new_dev(loop_entry,
		               FL_HWGRAPH_BLK,
			       hwg_bdev_name) < 1)
	{
		dsf_control.ret = %fl_c:*:NOT_CREATE_HWGRAPH_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %NOT_CREATE_HWGRAPH_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create hardware graph BLK device
	}


	// Create hardware graph char device
	if (fl_hwgraph_new_dev(loop_entry,
		               FL_HWGRAPH_CHR,
			       hwg_cdev_name) < 1)
	{
		// Remove BLK device file
		if (fl_hwgraph_remove_dev_by_path(hwg_bdev_name,
					          FL_HWGRAPH_CHR) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove BLK device file \"%s\"\n",
					  hwg_bdev_name))


			dsf_control.ret = %fl_c:*:NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_HWGRAPH_CHR_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create hardware graph CHR device, nor remove hardware graph BLK device
		}


		dsf_control.ret = %fl_c:*:NOT_CREATE_HWGRAPH_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


	 	return %NOT_CREATE_HWGRAPH_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create hardware graph CHR device
	}


	// Build symbolic link target to BLK device file
	symlink_target = loop_entry->dev.symlink_target;
	symlink_target[0] = '\0';
	strcat(symlink_target,
	       "/hw/loop/");
	strcat(symlink_target,
	       loop_entry->dev.basename);
	strcat(symlink_target,
	       "/block");


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_new] INFO: Create symbolic link from \"%s\" to \"%s\"\n",
			  loop_entry->dev.name,
			  symlink_target))


	// Create symbolic link to BLK device file
	was_created = 0;
	if (fl_file_create(&loop_entry->dev,
			   VLNK,
			   0,
			   S_IRUSR |
			   S_IRGRP,
			   &loop_entry->cred,
			   &was_created) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_dev_new] ERROR: Couldn't create symbolic link from \"%s\" to \"%s\"\n",
				  loop_entry->dev.name,
				  symlink_target))


		// Remove CHR device file
		blk_is_removed = 1;
		if (fl_hwgraph_remove_dev_by_path(hwg_cdev_name,
					          FL_HWGRAPH_BLK) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove CHR device file \"%s\"\n",
					  hwg_cdev_name))


			blk_is_removed = 0;
		}


		// Remove BLK device file
		chr_is_removed = 1;
		if (fl_hwgraph_remove_dev_by_path(hwg_bdev_name,
					          FL_HWGRAPH_CHR) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove BLK device file \"%s\"\n",
					  hwg_bdev_name))


			chr_is_removed = 0;
		}


		// Check for removed switches
		if ((blk_is_removed == 0) &&
		    (chr_is_removed == 1))
		{
			// NOTE: BLK not removed and CHR removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to BLK device, nor remove BLK device
		}
		else if ((blk_is_removed == 1) &&
			 (chr_is_removed == 0))
		{
			// NOTE: BLK removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to BLK device, nor remove CHR device
		}
		else if ((blk_is_removed == 0) &&
			 (chr_is_removed == 0))
		{
			// NOTE: BLK not removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_BLK_NOT_REMOVE_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to BLK device, nor remove CHR and BLK devices
		}
		else
		{
			// NOTE: BLK removed and CHR removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;
		}


		return %NOT_CREATE_SYMLINK_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to BLK device file
	}


	// Build symbolic link target to CHR device file
	symlink_target = loop_entry->dev_chr.symlink_target;
	symlink_target[0] = '\0';
	strcat(symlink_target,
	       "/hw/loop/");
	strcat(symlink_target,
	       loop_entry->dev.basename);
	strcat(symlink_target,
	       "/char");


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_new] INFO: Create symbolic link from \"%s\" to \"%s\"\n",
			  loop_entry->dev_chr.name,
			  symlink_target))


	// Create symbolic link to CHR device file
	was_created = 0;
	if (fl_file_create(&loop_entry->dev_chr,
			   VLNK,
			   0,
			   S_IRUSR |
			   S_IRGRP,
			   &loop_entry->cred,
			   &was_created) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_dev_new] ERROR: Couldn't create symbolic link from \"%s\" to \"%s\"\n",
				  loop_entry->dev_chr.name,
				  symlink_target))


		// Remove symbolic link to BLK device file
		is_removed = 0;
		ret_remove = 0;
		if (fl_file_remove(&loop_entry->dev,
                                   &is_removed,
				   &ret_remove) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove symbolic link \"%s\" (-> \"%s\")\n",
					  loop_entry->dev.name,
					  loop_entry->dev_chr.name))
		}


		// Remove CHR device file
		blk_is_removed = 1;
		if (fl_hwgraph_remove_dev_by_path(hwg_cdev_name,
					          FL_HWGRAPH_BLK) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove CHR device file \"%s\"\n",
					  hwg_cdev_name))


			blk_is_removed = 0;
		}


		// Remove BLK device file
		chr_is_removed = 1;
		if (fl_hwgraph_remove_dev_by_path(hwg_bdev_name,
					          FL_HWGRAPH_CHR) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_dev_new] WARNING: Couldn't remove BLK device file \"%s\"\n",
					  hwg_bdev_name))


			chr_is_removed = 0;
		}


		// Check for removed switches
		if ((is_removed == 0) &&
		    (blk_is_removed == 0) &&
		    (chr_is_removed == 1))
		{
			// NOTE: Symlink not removed and BLK not removed and CHR removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link and BLK device
		}
		else if ((is_removed == 0) &&
			 (blk_is_removed == 1) &&
			 (chr_is_removed == 0))
		{
			// NOTE: Symlink not removed and BLK removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link and CHR device
		}
		else if ((is_removed == 0) &&
			 (blk_is_removed == 0) &&
			 (chr_is_removed == 0))
		{
			// NOTE: Symlink not removed and BLK not removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_SYMLINK_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove symbolic link, BLK or CHR devices
		}
		else if ((is_removed == 1) &&
		    	 (blk_is_removed == 0) &&
		    	 (chr_is_removed == 1))
		{
			// NOTE: Symlink removed and BLK not removed and CHR removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove BLK device
		}
		else if ((is_removed == 1) &&
			 (blk_is_removed == 1) &&
			 (chr_is_removed == 0))
		{
			// NOTE: Symlink removed and BLK removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove CHR device
		}
		else if ((is_removed == 1) &&
			 (blk_is_removed == 0) &&
			 (chr_is_removed == 0))
		{
			// NOTE: Symlink removed and BLK not removed and CHR not removed


			dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CREATE_SYMLINK_TO_CHR_NOT_REMOVE_BLK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device, nor remove BLK and CHR devices
		}


		dsf_control.ret = %fl_c:*:NOT_CREATE_SYMLINK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %NOT_CREATE_SYMLINK_CHR[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't create symbolic link to CHR device
	}


	// Allocate minor number for symbolic link to
	// BLK device file.
	fl_dev_set_lminor(lminor, 1);


	// Allocate minor number for symbolic link to
	// CHR device file.
	fl_dev_set_lminor(lminor_chr, 1);


	// Set BLK device minor number
	loop_entry->dev.minor = lminor;


	// Set BLK device number
	loop_entry->dev.num = makedev(0, lminor);


	// Set CHR device minor number
	loop_entry->dev_chr.minor = lminor_chr;


	// Set CHR device number
	loop_entry->dev_chr.num = makedev(0, lminor_chr);


	// Set device created switch
	loop_entry->dev.is_created = 1;


	return %DEV_NEW_SUCCESS[ipm,1, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// SUCCESS
}


#               endif
#       endif
#endif


%FUNC:fl_dev_new:STOP%


int fl_dev_get_size(dev_t dev,			// -- IN
                    fl_uint64_t *dev_size,	// -- OUT
                    fl_uint32_t *was_found,	// -- OUT
                    fl_uint32_t *is_valid)	// -- OUT
{
	struct loop_entry_s *loop_entry;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30



#               endif
#       endif
#endif


	// NOTE: Arguments


	// OUT: Check for device size
	if (dev_size == NULL)
	{
		return 0;	// ERROR: No device size
	}


	// OUT: Check for was found switch
	if (was_found == NULL)
	{
		return -1;	// ERROR: No was found switch
	}


	// OUT: Check for is valid switch
	if (is_valid == NULL)
	{
		return -2;	// ERROR: No was found switch
	}


	// NOTE: Processing


        // Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Get device info
	loop_entry = (struct loop_entry_s *) device_info_get(dev);


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_get_size] INFO: dev_vertex = %d major = %d minor = %d\n",
			  dev,
			  getemajor(dev),
			  geteminor(dev)))


	// Check for loop entry
        *dev_size = 0;
	*was_found = 0;
	*is_valid = 0;
	if (loop_entry != NULL)
	{
		// Set found switch
		*was_found = 1;


		// Check for installation CD
		if (loop_entry->install_cd.is_attached == 1)
		{
			// Set device size
			*dev_size = loop_entry->install_cd.part_size;


			// Set is valid switch
			*is_valid = 1;
		}
		else
		{
			if (loop_entry->file.vap->va_size >= 0)
			{
				// Set device size
				*dev_size = loop_entry->file.vap->va_size;


				// Set is valid switch
				*is_valid = 1;
			}
		}
	}


#               endif
#       endif
#endif


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	return 1;	// SUCCESS
}


long long fl_dev_get_size_in_blocks(dev_t dev)	// -- IN
{
	fl_uint32_t is_valid;
	fl_uint32_t was_found;
	fl_uint64_t dev_size_bytes;
	fl_int64_t dev_size_blocks;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	// NOTE: Processing


	// NOTE: Proc
	//
	//  A process structure is not needed since this entry point
	//  is primarily called from kernel code.


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_get_size] INFO: START\n"))


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	FL_DEBUG_LOCK_SWITCH(-1)
	FL_CHECK_LOCK_SWITCH(-1)


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_get_size] INFO: vertex device major = %u minor = %u type = %s pid = %llu\n",
			  getemajor(dev),
			  geteminor(dev),
			  "BLK",
			  dbgpid))


#               endif
#       endif
#endif


	// Acquire serialization lock
	FL_SERIAL_LOCK();


	// Get device size
	is_valid = 0;
	was_found = 0;
	dev_size_bytes = 0;
	fl_dev_get_size(dev,
			&dev_size_bytes,
			&was_found,
			&is_valid);


	// Check for found device
	if ((was_found == 1) &&
	    (is_valid == 1))
	{
		// NOTE: Device size
		//
		//  The cast from unsigned long long (dev_size_bytes) to
		//  signed long long (dev_size_blocks) is valid because
		//  the size stored in dev_size_bytes is already validated
		//  to be a positive number, and thus is less than or equal
		//  to 2^63.


		// Calculate size in blks
		dev_size_blocks = (fl_int64_t) (dev_size_bytes / FL_FILE_BLOCK);


		// Adjust block number
		if ((dev_size_bytes % FL_FILE_BLOCK) > 0)
		{
			dev_size_blocks++;
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_get_size] INFO: Device size (in blocks) = %llu\n",
				  dev_size_blocks))


		FL_DEBUG_UPRINTF((\
				  "[fl_dev_get_size] INFO: STOP\n"))


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		return dev_size_blocks;
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_dev_get_size] INFO: STOP\n"))


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return -1;
}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


%FUNC:fl_vh_check_install_cd:START%


int fl_vh_check_install_cd(struct loop_entry_s *loop_entry,	// -- IN
			   fl_uint32_t *is_install_cd)		// -- OUT
{
	fl_uint32_t is_closed;
	fl_uint64_t bytes_read;
	fl_uint64_t part_size;
	fl_uint64_t part_offset;
	struct uio *uio;
	struct iovec *iov;
	struct volume_header *vh;


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return %NO_LOOP_ENTRY[in,0,]%;	// ERROR: No loop entry
	}


	// OUT: Check for is install cd switch
	if (is_install_cd == NULL)
	{
		return %NO_IS_INSTALL_CD_SWITCH[in,-1,]%;	// ERROR: No is install cd switch
	}


	// NOTE: Initialization


	// SGI volume header
	bzero((void *) loop_entry->sector,
	      FL_FILE_BLOCK);


	// User I/O
	bzero((void *) loop_entry->uio,
	      sizeof(struct uio));


	// NOTE: Processing


	// TODO: FL_TEST_F_main_N_NOT_OPEN_FILE(0, 1)


	// Open attached file
	loop_entry->file.is_open = 0;
	if (FL_TEST_COND((fl_test_ret == 1), ||)
            (fl_file_open(&loop_entry->file,
                          FREAD,
                          &loop_entry->cred,
                          &loop_entry->file.is_open) < 1))
	{
                // TODO: FL_TEST_F_main_N_NOT_OPEN_FILE(1, -6)


                //dsf_control.ret = % fl_c:*:NOT_OPEN_FILE_FOR_VH_CHECK[in,,]%;


                // TODO: return % NOT_OPEN_FILE[in,,]%;   // ERROR: Couldn't open file


		return %NOT_OPEN_FILE_FOR_VH_CHECK[in,,]%;	// ERROR: Couldn't open attached file
	}


        // Set I/O vector
	iov = loop_entry->iov;
	iov->iov_base = (void *) loop_entry->sector;
	iov->iov_len = FL_FILE_BLOCK;


	// Set up uio
	uio = loop_entry->uio;
        uio->uio_segflg = UIO_SYSSPACE;
        uio->uio_iov = iov;
        uio->uio_iovcnt = 1;
        uio->uio_offset = 0;


	// NOTE: Signed type
	//
	//  The uio_resid field has a type ssize_t, which is a signed
	//  int (32 bit kernels) or long (64 bit kernels) type.
	//
	//  Whenever a value is assigned to this field, quantities need
	//  to be validated for overflows, resulting in negative uio_resid
	//  values.


        uio->uio_resid = (ssize_t) iov->iov_len;
        uio->uio_fmode = FREAD  |
			 FRSYNC |
			 FDSYNC;


	// Read volume header
	bytes_read = 0;
	if (fl_file_read(&loop_entry->file,
                         IO_RSYNC |
			 IO_DSYNC,
                         &loop_entry->cred,
                         uio,
                         &bytes_read) < 1)
	{
                FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] ERROR: Couldn't read volume header\n"))


		return %NOT_READ_FILE_FOR_VH_CHECK[in,,]%;	// ERROR: Couldn't read volume header
	}


	// Close file
	is_closed = 0;
	if (fl_file_close(&loop_entry->file,
			  &loop_entry->cred,
			  &is_closed) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] ERROR: Couldn't close attached file\n"))


		return %NOT_CLOSE_FILE_FOR_VH_CHECK[in,,]%;	// ERROR: Couldn't close attached file
	}


	// Set volume header
	vh = (struct volume_header *) loop_entry->sector;


	// DEBUG: Print volume header
	//fl_print_vh(vh);


	// Reset is install cd switch
	*is_install_cd = 0;


	// Check for valid volume header
	if ((fl_check_vh_checksum == 0) ||
	    (is_vh(vh) == 1))
	{
		// NOTE: SGI volume header


		FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] INFO: is_vh(vh) = %d\n",
				  is_vh(vh)))


		FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] INFO: vh_checksum(vh) = %d\n",
				  vh_checksum(vh)))


		FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] INFO: Found SGI volume header on attached file\n"))


		// NOTE: SGI installation CDs
		//
		//  SGI installation CDs are characterized for having
		//  partition number 7 set to type 5 (NOT_SUPPORTED).


		// Check for install CD
		if (vh->vh_pt[7].pt_type == 5)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_vh_check_install_cd] INFO: Found SGI installation CD image\n"))


			// Set is install cd switch
			*is_install_cd = 1;


			// Get EFS partition size
			part_size = vh->vh_pt[7].pt_nblks;


			// Get EFS partition offset
			part_offset = vh->vh_pt[7].pt_firstlbn;


			// Check attached file size
			if (loop_entry->file.size.sectors < 3)
			{
				return %FILE_SIZE_TOO_SMALL_FOR_VH_CHECK[in,,]%;	// ERROR: Attached file size too small
			}


			// Check partition offset
			if (part_offset > (loop_entry->file.size.sectors - 3))
			{
				return %PARTITION_OFFSET_OVERFLOW_IN_VH_CHECK[in,,]%;	// ERROR: Partition offset overflow
			}


			// Check partition size
			if ((part_offset + part_size) > loop_entry->file.size.sectors)
			{
				return %PARTITION_SIZE_OVERFLOW_IN_VH_CHECK[in,,]%;	// ERROR: Partition size overflow
			}


			// Set installation CD values
			loop_entry->install_cd.is_attached = 1;
			loop_entry->install_cd.part_size = part_size * FL_FILE_BLOCK;
			loop_entry->install_cd.part_offset = part_offset * FL_FILE_BLOCK;


			// NOTE: Volume header
			//
			//  This loop entry will point to a logical disk partition
			//  inside an SGI volume.
			//
			//  The volume header for a logical partition is always the
			//  volume header of the containing volume. This applies to
			//  any partition with any partition type.
			//
			//  In particular, an I/O control request of DIOCGETVH should
			//  always return the volume header of the volume.


			// Update volume header	
			bcopy((void *) vh,
			      (void *) &loop_entry->disk.vh,
			      sizeof(struct volume_header));


			FL_DEBUG_UPRINTF((\
					  "[fl_vh_check_install_cd] INFO: Partition size = %u\n",
					  loop_entry->install_cd.part_size))


			FL_DEBUG_UPRINTF((\
					  "[fl_vh_check_install_cd] INFO: Partition offset = %u\n",
					  loop_entry->install_cd.part_offset))
		}
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_vh_check_install_cd] INFO: SGI volume header not found on attached file\n"))


		return %NOT_FOUND_VH_IN_FILE[in,,]%;	// ERROR: Couldn't find SGI volume header in attached file
	}


	return %VH_FOUND_IN_FILE[ip,1,]%;	// SUCCESS
}


%FUNC:fl_vh_check_install_cd:STOP%
%FUNC:fl_loop_new:START%


int fl_loop_new(struct loop_entry_s *loop_entry,	// -- IN
		fl_int32_t *ret)			// -- OUT
{
	fl_int32_t ret_lookup;
	fl_int32_t ret_remove;
	fl_uint32_t is_attached;
	fl_uint32_t is_found;
	fl_uint32_t is_removed;
	fl_uint32_t is_install_cd;
	struct vnode *devvp;
	struct volume_header *vh;

	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Loop entry
	if (loop_entry == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = %NO_LOOP_ENTRY[in,0,]%;
		}


		return %NO_LOOP_ENTRY[on,,]%;	// ERROR: No loop entry
	}


	// OUT: Return value
	if (ret == NULL)
	{
		return %NO_RETURN_VALUE[in,,]%;	// ERROR: No return value
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: START\n"))


	// Check if BLK device is already attached
	is_attached = 0;
	fl_dev_is_attached(loop_entry->dev.name,
			   &is_attached);


	FL_TEST_F_main_N_BLK_DEV_ALREADY_ATTACHED(0, 1)


        if (is_attached == 1)
        {
		FL_TEST_F_main_N_BLK_DEV_ALREADY_ATTACHED(1, -2)


		dsf_control.ret = %fl_c:*:BLK_DEV_ALREADY_ATTACHED[in,,]%;


		// Set return value
		*ret = %BLK_DEV_ALREADY_ATTACHED[in,,]%;


                return %BLK_DEV_ALREADY_ATTACHED[on,,]%;      // ERROR: BLK device is already attached
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: fl_file_lookup(file) START\n"))


	FL_TEST_F_main_N_NOT_LOOKUP_FILENAME(0, 1)


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Do file name lookup
	is_found = 0;
	ret_lookup = 0;
	if (FL_TEST_COND((fl_test_ret == 1), ||)
	    (fl_file_lookup(&loop_entry->file,
                            NO_FOLLOW,
                            &is_found,
			    &ret_lookup) < 1))


#               endif
#       endif
#endif


	{
		FL_TEST_F_main_N_NOT_LOOKUP_FILENAME(1, -3)


		dsf_control.ret = %fl_c:*:NOT_LOOKUP_FILENAME[in,,]%;


		// Set return value
		*ret = %NOT_LOOKUP_FILENAME[in,,]%;


		return %NOT_LOOKUP_FILENAME[on,,]%;	// ERROR: Couldn't lookup file name
	}


	if (is_found != 1)
	{
		dsf_control.ret = %fl_c:*:FILE_NOT_FOUND[in,,]%;


		// Set return value
		*ret = %FILE_NOT_FOUND[in,,]%;


		return %FILE_NOT_FOUND[on,,]%;	// ERROR: Couldn't find file
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: fl_file_lookup(file) STOP\n"))
        FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: fl_file_stat(file) START\n"))


	FL_TEST_F_main_N_NOT_FILE_STAT(0, 1)


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        // Stat file
	if (FL_TEST_COND((fl_test_ret == 1), ||)
            (fl_file_stat(&loop_entry->file, 0) < 1))


#               endif
#       endif
#endif


        {
		FL_TEST_F_main_N_NOT_FILE_STAT(1, -5)


		dsf_control.ret = %fl_c:*:NOT_FILE_STAT[in,,]%;


		// Set return value
		*ret = %NOT_FILE_STAT[in,,]%;


                return %NOT_FILE_STAT[on,,]%;	// ERROR: Couldn't stat file
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Check for unsigned size in bytes
	if (loop_entry->file.vap->va_size < 0LL)
	{
		dsf_control.ret = %fl_c:*:FILE_NEGATIVE_SIZE[in,,]%;


		// Set return value
		*ret = %FILE_NEGATIVE_SIZE[in,,]%;


                return %FILE_NEGATIVE_SIZE[on,,]%;	// ERROR: File has negative size
	}


	// Set file size
	loop_entry->file.size.bytes = (fl_uint64_t) loop_entry->file.vap->va_size;
	loop_entry->file.size.sectors = (loop_entry->file.size.bytes % FL_FILE_BLOCK) != 0 ?
					(loop_entry->file.size.bytes / FL_FILE_BLOCK) + 1 :
					(loop_entry->file.size.bytes / FL_FILE_BLOCK);


#               endif
#       endif
#endif


	// Fill default volume header
	vh = &loop_entry->disk.vh;
	vh->vh_magic = VHMAGIC;
	vh->vh_dp.dp_secbytes = FL_FILE_BLOCK;
	vh->vh_dp._dp_heads = 1;
	vh->vh_dp._dp_sect = 32;
	vh->vh_csum = 0;


	// Set volume header partition
	vh->vh_pt[8].pt_nblks = 1;
	vh->vh_pt[8].pt_firstlbn = 0;
	vh->vh_pt[8].pt_type = PTYPE_VOLHDR;


	// Set entire volume partition
	vh->vh_pt[10].pt_nblks = loop_entry->file.size.sectors;
	vh->vh_pt[10].pt_firstlbn = 1;
	vh->vh_pt[10].pt_type = PTYPE_VOLUME;


	// Calculate checksum
	vh->vh_csum = vh_checksum(vh);


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_file_stat(file) STOP\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_dev_check_path(dev) START\n"))


        // NOTE: Check device pathname
        //
        //  For security reasons, device pathname checking must be done
        //  here, it can't be delegated to the user space command
        //  that sent the request packet.
        //
        //  A malicious user could decode the ufl<->fl.mod protocol
        //  and send any device pathname that would be created as
        //  a BLK device file anywhere in the file system.
	//


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30



	// NOTE: IRIX hardware graph
	//
	//  Although devices are created as entries in the hardware
	//  graph, the same naming rules still apply, since all entries
	//  in /hw/loop will be accessible from /dev/loop through
	//  symbolic links:
	//
	//   /dev/loop/mydev  -> /hw/loop/mydev/block
	//   /dev/loop/rmydev -> /hw/loop/mydev/char


#		endif
#	endif
#endif


        if (fl_dev_check_path(loop_entry->dev.name) < 1)
        {
		// Set return value
		*ret = %DEV_NAME_INVALID[in,,]%;


                return %DEV_NAME_INVALID[on,,]%;      // ERROR: Invalid device pathname
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_dev_check_path(dev) STOP\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_file_lookup(dev) START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] BLK\n"))


	FL_TEST_F_main_N_NOT_BLK_DEV_LOOKUP(0, 1)


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: IRIX symlink to /hw/loop
	//
	//  In IRIX, a symlink is stored in /dev/loop, pointing
	//  to the real device special file in the hardware graph
	//  under /hw/loop/{device_name}/{char, block}.


	// Do BLK symlink lookup
	is_found = 0;
	ret_lookup = 0;
        if (FL_TEST_COND((fl_test_ret == 1), ||)
	    (fl_file_lookup(&loop_entry->dev,
                            NO_FOLLOW,
                            &is_found,
			    &ret_lookup) < 1))


#               endif
#       endif
#endif


	{
		FL_TEST_F_main_N_NOT_BLK_DEV_LOOKUP(1, -8)


		dsf_control.ret = %fl_c:*:NOT_BLK_DEV_LOOKUP[in,,]%;


		// Set return value
		*ret = %NOT_BLK_DEV_LOOKUP[in,,]%;


		return %NOT_BLK_DEV_LOOKUP[on,,]%;	// ERROR: Couldn't lookup BLK device file
	}


	if (is_found == 1)
	{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		FL_DEBUG_UPRINTF((\
				  "[fl_loop_new] INFO: Symlink to BLK device exists\n"))


		// NOTE: BLK device name exists
		//
		//  If it is a symbolic link, assume that it has been
		//  created by file looper.


		// Check for symbolic link to device file
		devvp = loop_entry->dev.vp;
		if (devvp->v_type != VLNK)
		{
			dsf_control.ret = %fl_c:*:BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[in,,]%;


			// Set return value
			*ret = %BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[in,,]%;


			return %BLK_DEV_NAME_NOT_SYMLINK_DEV_FILE[on,,]%;	// ERROR: Device file name is not a symbolic link to a BLK device file
		}


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


		FL_TEST_F_main_N_NOT_BLK_DEV_FILE_DELETE(0, 1)


		// Delete device file
		is_removed = 0;
		ret_remove = 0;
		if (FL_TEST_COND((fl_test_ret == 1), ||)
		    (fl_file_remove(&loop_entry->dev,


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


                                    &is_removed,
				    &ret_remove) < 1))
		{
			FL_TEST_F_main_N_NOT_BLK_DEV_FILE_DELETE(1, -11)


			dsf_control.ret = %fl_c:*:NOT_BLK_DEV_FILE_DELETE[in,,]%;


			// Set return value
			*ret = %NOT_BLK_DEV_FILE_DELETE[in,,]%;


			return %NOT_BLK_DEV_FILE_DELETE[on,,]%;	// ERROR: Couldn't delete BLK device file
		}


		// NOTE: BLK device file deleted
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] CHR\n"))


	// NOTE: CHR device
	//
	//  IRIX requires a CHR device in addition to the BLK device.


	FL_TEST_F_main_N_NOT_CHR_DEV_LOOKUP(0, 1)


	// Do CHR device name lookup
	is_found = 0;
	ret_lookup = 0;
	if (FL_TEST_COND((fl_test_ret == 1), ||)
            (fl_file_lookup(&loop_entry->dev_chr,
                            NO_FOLLOW,
                            &is_found,
			    &ret_lookup) < 1))
	{
		FL_TEST_F_main_N_NOT_CHR_DEV_LOOKUP(1, -8)


		dsf_control.ret = %fl_c:*:NOT_CHR_DEV_LOOKUP[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		// Set return value
		*ret = %NOT_CHR_DEV_LOOKUP[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %NOT_CHR_DEV_LOOKUP[onm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't lookup CHR device
	}


	if (is_found == 1)
	{
		// NOTE: CHR device name exists
		//
		//  If it is a symbolic link, assume that it has been
		//  created by file looper.


		FL_DEBUG_UPRINTF((\
				  "[fl_loop_new] INFO: Symlink to CHR device exists\n"))


		// Check for symbolic link to device file
		devvp = loop_entry->dev_chr.vp;
		if (devvp->v_type != VLNK)
		{
			dsf_control.ret = %fl_c:*:CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			// Set return value
			*ret = %CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %CHR_DEV_NAME_NOT_SYMLINK_DEV_FILE[onm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Device file name is not a CHR device file
		}


		FL_TEST_F_main_N_NOT_CHR_DEV_FILE_DELETE(0, 1)


		// Delete symbolic link to device file
		is_removed = 0;
		ret_remove = 0;
		if (FL_TEST_COND((fl_test_ret == 1), ||)
		    (fl_file_remove(&loop_entry->dev_chr,
                                    &is_removed,
				    &ret_remove) < 1))
		{
			FL_TEST_F_main_N_NOT_CHR_DEV_FILE_DELETE(1, -11)


			dsf_control.ret = %fl_c:*:NOT_CHR_DEV_FILE_DELETE[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			// Set return value
			*ret = %NOT_CHR_DEV_FILE_DELETE[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


			return %NOT_CHR_DEV_FILE_DELETE[onm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Couldn't delete CHR device file
		}
	}


#               endif
#       endif
#endif


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_file_lookup(dev) STOP\n"))


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: fl_vh_check_install_cd(loop_entry) START\n"))


	// Check for install CD recognition
	if (fl_install_cd_recognition == 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_loop_new] INFO: fl_vh_check_install_cd(loop_entry) START\n"))


		// Check for install CD
		switch (fl_vh_check_install_cd(loop_entry,
					       &is_install_cd))
		{
			case %fl_c:fl_vh_check_install_cd:NOT_OPEN_FILE_FOR_VH_CHECK[on,,]%:

				dsf_control.ret = %fl_c:*:NOT_OPEN_FILE_FOR_VH_CHECK[in,,]%;


				// Set return value
				*ret = %NOT_OPEN_FILE_FOR_VH_CHECK[in,,]%;


				return %NOT_OPEN_FILE_FOR_VH_CHECK[on,,]%;	// ERROR: Couldn't open attached file for volume header check

			case %fl_c:fl_vh_check_install_cd:NOT_READ_FILE_FOR_VH_CHECK[on,,]%:

				dsf_control.ret = %fl_c:*:NOT_READ_FILE_FOR_VH_CHECK[in,,]%;


				// Set return value
				*ret = %NOT_READ_FILE_FOR_VH_CHECK[in,,]%;


				return %NOT_READ_FILE_FOR_VH_CHECK[on,,]%;	// ERROR: Couldn't read from attached file for volume header check

			case %fl_c:fl_vh_check_install_cd:NOT_CLOSE_FILE_FOR_VH_CHECK[on,,]%:

				dsf_control.ret = %fl_c:*:NOT_CLOSE_FILE_FOR_VH_CHECK[in,,]%;


				// Set return value
				*ret = %NOT_CLOSE_FILE_FOR_VH_CHECK[in,,]%;


				return %NOT_CLOSE_FILE_FOR_VH_CHECK[on,,]%;	// ERROR: Couldn't close attached file in volume header check

			default:

				// NOTE: Allowed return values
				//
				//  All other return values, positive or negative,
				//  are allowed.


				break;
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_loop_new] INFO: fl_vh_check_install_cd(loop_entry) STOP\n"))
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] INFO: fl_vh_check_install_cd(loop_entry) STOP\n"))


#               endif
#       endif
#endif


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_dev_new(dev) START\n"))


	// Create new device
	if (fl_dev_new(loop_entry) < 1)
	{
		// Set return value
		*ret = %NOT_CREATE_DEVS[in,,]%;


		return %NOT_CREATE_DEVS[on,,]%;	// ERROR: Couldn't create devices
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_loop_new] fl_dev_new(dev) STOP\n"))


	// NOTE: Ready
	//
	//  The file is attached to a new device file and is now
	//  ready to accept I/O requests.


	// Set return value
	*ret = 1;


	return 1;
}


%FUNC:fl_loop_new:STOP%
%FUNC:fl_loop_add:START%


int fl_loop_add(struct loop_entry_s *loop_entry)	// -- IN
{
	fl_list_node_t *list_node;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return 0;	// ERROR: No loop entry
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	// Create new list node
	fl_list_new_node(loop_list, &list_node);


	// Set data field
	list_node->data = (void *) loop_entry;


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	// Add list node to loop list
	fl_list_push(loop_list, list_node);


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	dsf_control.ret = %fl_c:*:ATTACH_SUCCESS[ip,1,]%;


	return 1;	// SUCCESS
}


%FUNC:fl_loop_add:STOP%
%FUNC:fl_vnode_get_usecount:START%


int fl_vnode_get_usecount(struct vnode *vp,	// -- IN
			  fl_int32_t *usecount,	// -- OUT
			  fl_int32_t *ret)	// -- OUT
{
	// NOTE: Arguments


	// IN: Check for vnode
	if (vp == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = 0;
		}


		return 0;	// ERROR: No vnode
	}


	// OUT: Check for vnode usecount
	if (usecount == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -1;
		}


		return -1;	// ERROR: No vnode usecount
	}


	// OUT: Check for return value
	if (ret == NULL)
	{
		return -2;	// ERROR: No return value
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Set vnode usecount
	*usecount = vp->v_count;


#               endif
#       endif
#endif


	// Set return value
	*ret = 1;


	return 1;	// SUCCESS
}


%FUNC:fl_vnode_get_usecount:STOP%


int fl_entry_rel(struct loop_entry_s *loop_entry,	// -- IN
		 fl_uint32_t *dev_is_removed)		// -- OUT
{
	fl_int32_t ret_lookup;
	fl_int32_t ret_remove;
	fl_uint32_t is_closed;
	fl_uint32_t is_found;


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return 0;	// ERROR: No loop entry
	}


	// OUT: Check for device is removed switch
	if (dev_is_removed == NULL)
	{
		return -1;	// ERROR: No device is removed switch
	}


	// NOTE: Processing


	// NOTE: Release loop entry resources
	//
	//  Release all allocated resources
	//  in the loop entry except the ret field,
	//  which is used to indicate any errors to
	//  the fl userland utility in the fl_read()
	//  entry point.


	// Check for created device file
	if (loop_entry->dev.is_created == 1)
	{
		// Check for device vnode
		if (loop_entry->dev.vp == NULL)
		{
			// NOTE: Lookup device vnode
			//
			//  There's a path where the device file is created
			//  but no vnode is returned. Get device vnode.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			is_found = 0;
			ret_lookup = 0;
        		if (fl_file_lookup(&loop_entry->dev,
                                           NO_FOLLOW,
                           		   &is_found,
					   &ret_lookup) < 1)


#               endif
#       endif
#endif


			{
				FL_DEBUG_UPRINTF((\
						  "[fl_entry_rel] INFO: fl_file_lookup.ret = %d\n",
						  ret_lookup))
			}
		}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Lookup CHR device
		is_found = 0;
		ret_lookup = 0;
		fl_file_lookup(&loop_entry->dev_chr,
                               NO_FOLLOW,
                               &is_found,
			       &ret_lookup);


		FL_DEBUG_UPRINTF((\
				  "[fl_entry_rel] INFO: fl_file_lookup(dev_chr) = %d\n",
				  ret_lookup))


#               endif
#       endif
#endif


	}


	FL_DEBUG_UPRINTF((\
			  "[fl_entry_rel] dev.vp = %p\n",
			  loop_entry->dev.vp))


	// Release device resources
	if (loop_entry->dev.vp != NULL)
	{
		// NOTE: Remove device file
		//
		//  There are multiple paths that lead here, with
		//  and without created device file. Check is required.


		FL_DEBUG_UPRINTF((\
				  "[fl_entry_rel] INFO: dev.is_created = %d\n",
				  loop_entry->dev.is_created))


		// Check for created device file
		*dev_is_removed = 0;
		if (loop_entry->dev.is_created == 1)
		{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			// Remove BLK device file
			fl_file_remove(&loop_entry->dev,
				       dev_is_removed,
				       &ret_remove);


#               endif
#       endif
#endif


			FL_DEBUG_UPRINTF((\
					  "[fl_entry_rel] INFO: fl_file_remove(BLK) = %d dev_is_removed = %d\n",
					  ret_remove,
					  *dev_is_removed))


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			// Remove CHR device file
			ret_remove = 0;
			fl_file_remove(&loop_entry->dev_chr,
				       dev_is_removed,
				       &ret_remove);


			FL_DEBUG_UPRINTF((\
					  "[fl_entry_rel] INFO: fl_file_remove(CHR) = %d dev_is_removed = %d\n",
					  ret_remove,
					  *dev_is_removed))


#               endif
#       endif
#endif


		}
	}


	// NOTE: Release file resources


	FL_DEBUG_UPRINTF((\
			  "[fl_entry_rel] INFO: file.vp = %p\n",
			  loop_entry->file.vp))


	if (loop_entry->file.vp != NULL)
	{
		// NOTE: Close file if open
		//
		//  There are multiple paths that lead here, with and
		//  without opened file. Check is required.


                FL_DEBUG_UPRINTF((\
				  "[fl_entry_rel] INFO: file.is_open = %d\n",
				  loop_entry->file.is_open))


		// Check for open file
		is_closed = !loop_entry->file.is_open;
		if (loop_entry->file.is_open == 1)
		{
                        // Close file
                        if (fl_file_close(&loop_entry->file,
                                          &loop_entry->cred,
                                          &is_closed) < 1)
                        {
                                FL_DEBUG_UPRINTF((\
						  "[fl_entry_rel] ERROR: Couldn't close file\n"))
                        }
                        else
                        {
                                FL_DEBUG_UPRINTF((\
						  "[fl_entry_rel] INFO: File closed\n"))
                        }
                }
        }


	// Check for vnode attributes
	if (loop_entry->file.vap != NULL)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_entry_rel] INFO: Free vnode attributes\n"))


		// Free vnode attributes
		fl_mem_free(loop_entry->file.vap);


		// Reset vnode attributes
		loop_entry->file.vap = NULL;
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// Reset packet
	//bzero((void *) &loop_entry->packet,
	      //sizeof(struct fl_packet_s));


	FL_DEBUG_UPRINTF((\
			  "[fl_entry_rel] INFO: loop_entry->pid_list_blk = %p\n",
			  loop_entry->pid_list_blk))


	if (loop_entry->pid_list_blk != NULL)
	{
		fl_list_null(&loop_entry->pid_list_blk);
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: CHR device
	//
	//  Free CHR device related resources.


	FL_DEBUG_UPRINTF((\
			  "[fl_entry_rel] INFO: loop_entry->pid_list_chr = %p\n",
			  loop_entry->pid_list_chr))


	if (loop_entry->pid_list_chr != NULL)
	{
		fl_list_null(&loop_entry->pid_list_chr);
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_entry_rel] INFO: loop_entry->dev_chr.name = %p\n",
			  loop_entry->dev_chr.name))


        if (loop_entry->dev_chr.name != NULL)
        {
		// Free CHR device name
                fl_mem_free(loop_entry->dev_chr.name);


		// Reset CHR device file name
		loop_entry->dev_chr.name = NULL;
        }


	// NOTE: I/O resources
	//
	//  Free I/O related resources.


	// Free uio
	fl_mem_free(loop_entry->uio);
	loop_entry->uio = NULL;


	// Free iov
	fl_mem_free(loop_entry->iov);
	loop_entry->iov = NULL;


	// Free sector
	fl_mem_free(loop_entry->sector);
	loop_entry->sector = NULL;


	// Deinitialize lock
	fl_spin_free(loop_entry->lock);


	// Free semaphore for formatting I/O control
	fl_sem_deinit(&loop_entry->format_sem);


#               endif
#       endif
#endif


	return 1;	// SUCCESS
}


int fl_pid_check_active(struct loop_entry_s *loop_entry,	// -- IN
			fl_uint32_t pid_list)			// -- IN
{
	fl_uint64_t pid_length;
	fl_uint64_t pid_length_chr;
	fl_uint64_t pid_length_blk;


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return 0;	// ERROR: No loop entry
	}


	// IN: Check for pid list
	if ((pid_list != FL_PID_CHR) &&
	    (pid_list != FL_PID_BLK) &&
            (pid_list != FL_PID_CHR_BLK))
	{
		return -1;	// ERROR: Invalid pid list
	}


	// NOTE: Processing


	pid_length = 0;
	if (pid_list == FL_PID_CHR)
	{
		// Get length of pid list
		fl_list_length(loop_entry->pid_list_chr,
			       &pid_length);


		FL_DEBUG_UPRINTF((\
				  "[fl_pid_check_active] INFO: pid_length = %llu\n",
				  pid_length))
	}
	else if (pid_list == FL_PID_BLK)
	{
		// Get length of pid list
		fl_list_length(loop_entry->pid_list_blk,
			       &pid_length);


		FL_DEBUG_UPRINTF((\
				  "[fl_pid_check_active] INFO: pid_length = %llu\n",
				  pid_length))
	}
	else
	{
		// Get length of pid list
		fl_list_length(loop_entry->pid_list_chr,
			       &pid_length_chr);


		FL_DEBUG_UPRINTF((\
				  "[fl_pid_check_active] INFO: pid_length_chr = %llu\n",
				  pid_length_chr))


		fl_list_length(loop_entry->pid_list_blk,
			       &pid_length_blk);


		FL_DEBUG_UPRINTF((\
				  "[fl_pid_check_active] INFO: pid_length_blk = %llu\n",
				  pid_length_blk))


		// Set final pid length
		if ((pid_length_chr > 0) ||
                    (pid_length_blk > 0))
		{
			pid_length = 1;
		}
	}


	// Check length
	if (pid_length == 0)
	{
		return -1;	// ERROR: No active processes
	}


	return 1;	// SUCCESS
}


int fl_user_check(uid_t uid,			// -- IN
		  struct fl_args_s *args)	// -- IN
{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	struct cred cred;


#               endif
#       endif
#endif


	// NOTE: Arguments


	// IN: Check for args
	if (args == NULL)
	{
		return 0;	// ERROR: No args
	}


	// NOTE: Processing


	// Get user creds for process
	fl_proc_get_cred(&cred);


	// Check for uid
	if (cred.cr_uid != uid)
	{
		// Check for return value
		if (args->ret.var != NULL)
		{
			*args->ret.var = -1;
		}


		// Check for reset on return
		if (args->ret.reset == 1)
		{
			bzero((void *) args,
			      sizeof(struct fl_args_s));
		}


		return -1;	// ERROR: User is not uid
	}


	// Check for return value
	if (args->ret.var != NULL)
	{
		*args->ret.var = 1;
	}


	// Check for reset on return
	if (args->ret.reset == 1)
	{
		bzero((void *) args,
		      sizeof(struct fl_args_s));
	}


	return 1;	// SUCCESS
}


%FUNC:fl_cmd_attach_with_entry:START%


int fl_cmd_attach_with_entry(struct loop_entry_s *loop_entry)	// -- IN
{
	fl_int32_t ret;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return %NO_LOOP_ENTRY[in,0,]%;	// ERROR: No loop entry
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	// Get process credentials
	fl_proc_get_cred(&loop_entry->cred);


	FL_TEST_F_main_N_NOT_ALLOC_VATTR(0, 1)


	// Allocate vnode attributes
	fl_mem_alloc((void **) &loop_entry->file.vap,
                     sizeof(struct vattr),
                     FL_MEM_WAITOK,
                     "fl_cmd_attach",
                     "Allocate vnode attributes");


	// Link loop entry to dsf control session
	dsf_control.loop_entry = loop_entry;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


        // Create new loop
        if (fl_loop_new(loop_entry,
			&ret) < 1)
        {
                FL_DEBUG_UPRINTF((\
				  "[fl_cmd_attach_with_entry] INFO: fl_loop_new() = %d error\n",
				  ret))


                return %NOT_NEW_LOOP_ENTRY[in,,]%;      // ERROR: Couldn't create new loop entry
        }


        // Add new loop entry
	fl_loop_add(loop_entry);


        //loop_print(loop_list);


        return 1;       // SUCCESS
}


%FUNC:fl_cmd_attach_with_entry:STOP%
%FUNC:fl_cmd_attach:START%


int fl_cmd_attach(fl_int32_t *ret)	// -- OUT
{
	fl_uint32_t dev_is_removed;
	struct loop_entry_s *loop_entry;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// OUT: Check for return value
	if (ret == NULL)
	{
		return 0;	// ERROR: No return value
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	// Create new loop entry
	fl_entry_new(&loop_entry);


	// Populate loop entry
	bcopy((void *) &packet,
	      (void *) &loop_entry->packet,
	      sizeof(struct fl_packet_s));
	loop_entry->dev.name = loop_entry->packet.dev;
	loop_entry->file.name = loop_entry->packet.file;
	loop_entry->first_open = 1;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Device name length
        //
        //  It is assumed that the length of the device name
        //  is less than PATH_MAX.
        //
        //  This is necessary to build up the CHR device file
        //  name with a prepended 'r' (for raw).


	loop_entry->dev.basename = loop_entry->dev.name + 10;


	// NOTE: Allocate CHR device filename
	//
	//  Allocation requires extra space for the prepended 'r',
	//  and the terminating NULL byte:
	//
	//   dev_chr.name = strlen(...) + 1 + 1


	FL_TEST_F_main_N_NOT_ALLOC_CHR_DEV_NAME(0, 1)


	// Allocate CHR device file name
	fl_mem_alloc((void **) &loop_entry->dev_chr.name,
                     strlen(loop_entry->dev.name) + 2,
                     FL_MEM_ZERO | FL_MEM_WAITOK,
                     "fl_cmd_attach",
                     "Allocate CHR device file name");


	// Copy dirname of BLK device name
	bcopy(loop_entry->dev.name,
              loop_entry->dev_chr.name,
              10);


	// Prepend 'r'
	loop_entry->dev_chr.name[10] = 'r';


	// Add BLK basename
	bcopy(loop_entry->dev.basename,
              loop_entry->dev_chr.name + 11,
              strlen(loop_entry->dev.basename));


	// Set CHR basename
	loop_entry->dev_chr.basename = loop_entry->dev_chr.name + 10;


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_attach] loop_entry->dev_chr.name = \"%s\"\n",
			  loop_entry->dev_chr.name))


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_attach] loop_entry->dev_chr.basename = \"%s\"\n",
			  loop_entry->dev_chr.basename))


#               endif
#       endif
#endif


	// Attach to new loop entry
	if (fl_cmd_attach_with_entry(loop_entry) < 1)
	{
		// Release loop entry
		fl_entry_rel(loop_entry,
			     &dev_is_removed);


		// Free loop entry
		fl_mem_free(loop_entry);


		// Set return value
		*ret = %NOT_ATTACH_TO_LOOP_ENTRY[in,,]%;


		return %NOT_ATTACH_TO_LOOP_ENTRY[on,,]%;	// ERROR: Couldn't attach to new loop entry
	}


        return 1;	// SUCCESS
}


%FUNC:fl_cmd_attach:STOP%
%FUNC:fl_cmd_detach:START%


int fl_cmd_detach()
{
	fl_uint32_t dev_is_removed;
	fl_uint64_t loop_index;
	fl_list_node_t *loop_node;
	struct loop_entry_s *loop_entry;


	// NOTE: Processing


	// Acquire lock for fl_open
	fl_spin_lock(fl_open_lock);


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	// Get loop node for device name
	loop_node = NULL;
	loop_index = 0;
	//if (fl_node_get_by_dev_name(dsf_control.packet_resp.dev,
	if (fl_node_get_by_dev_name(packet.dev,
                                    &loop_node,
                                    &loop_index) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_detach] ERROR: Couldn't detach, BLK device file not found\n"))


		// Release lock for loop list
		fl_spin_unlock(loop_list_lock);


		// Release lock for fl_open
		fl_spin_unlock(fl_open_lock);


		dsf_control.ret = %fl_c:*:BLK_DEV_NAME_NOT_FOUND[in,,]%;


		return %BLK_DEV_NAME_NOT_FOUND[in,0,]%;	// ERROR: BLK device name not found
	}


	// Get loop entry
	loop_entry = (struct loop_entry_s *) loop_node->data;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Check for active CHR processes
	if (fl_pid_check_active(loop_entry,
				FL_PID_CHR) == 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_detach] ERROR: Couldn't detach loop because of active processes on CHR device\n"))


		// Release lock for loop list
		fl_spin_unlock(loop_list_lock);


		// Release lock for fl_open
		fl_spin_unlock(fl_open_lock);


		dsf_control.ret = %fl_c:*:CHR_DEV_HAS_ACTIVE_PROCS[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;


		return %CHR_DEV_HAS_ACTIVE_PROCS[inm,, $FL_OS_NAME_V_IRIX && $FL_OS_VERSION_V_6_5 && $FL_OS_RELEASE_V_30]%;	// ERROR: Active processes on CHR device
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_detach] INFO: No active processes using (CHR) %s\n",
			  packet.dev))


#		endif
#	endif
#endif


	// Check for active BLK processes
	if (fl_pid_check_active(loop_entry,
				FL_PID_BLK) == 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_detach] ERROR: Couldn't detach loop because of active processes on BLK device\n"))


		// Release lock for loop list
		fl_spin_unlock(loop_list_lock);


		// Release lock for fl_open
		fl_spin_unlock(fl_open_lock);


		dsf_control.ret = %fl_c:*:BLK_DEV_HAS_ACTIVE_PROCS[in,,]%;


		return %BLK_DEV_HAS_ACTIVE_PROCS[in,,]%;	// ERROR: Active processes on BLK device
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_detach] No active processes using (BLK) \"%s\"\n",
			  packet.dev))


	// NOTE: There are no processes with the associated CHR/BLK device file open


	// Unlink loop node
	fl_list_del(loop_list,
		    &loop_node,
		    loop_index);


	// Release minor number
	fl_dev_set_lminor(loop_entry->dev.minor,
			  0);


	// Release locks
	fl_spin_unlock(loop_list_lock);
	fl_spin_unlock(fl_open_lock);


	// NOTE: Clean up loop entry


	// Free loop node
	fl_list_free_node(loop_list,
			  &loop_node);


	// Release loop entry
	dev_is_removed = 0;
	fl_entry_rel(loop_entry,
		     &dev_is_removed);


	// NOTE: Deinitialization


	// Free loop entry
	fl_mem_free(loop_entry);


	dsf_control.ret = %fl_c:*:DETACH_SUCCESS[ip,2,]%;


	return %DETACH_SUCCESS[ip,1,]%;	// SUCCESS
}


%FUNC:fl_cmd_detach:STOP%


int fl_entry_serialize(struct loop_entry_s *loop_entry,	// -- IN
		       char **data_pointer)		// -- OUT
{
	char *dp;
	char *fp;
	fl_uint16_t field_index;
	fl_uint64_t dev_size;
	struct fl_packet_s *packet;


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return 0;	// ERROR: No loop entry
	}


	// OUT: Check for data pointer
	if (data_pointer == NULL)
	{
		return -1;	// ERROR: No data pointer
	}


	// NOTE: Processing


	// NOTE: Copy successive fields (fp) to data (dp)


	// Set packet pointer
	packet = &loop_entry->packet;


	// Set packet data pointer
	dp = *data_pointer;


        // Copy device file name size
        fp = (char *) &packet->dev_size;
        for (field_index = 0;
	     field_index < 2;
	     field_index++)
	{
                *dp++ = *fp++;
	}


        // Copy device file name
        fp = packet->dev;
        for (field_index = 0;
	     field_index < packet->dev_size;
	     field_index++)
	{
                *dp++ = *fp++;
	}


        // Copy file name size
        fp = (char *) &packet->file_size;
        for (field_index = 0;
	     field_index < 2;
	     field_index++)
	{
                *dp++ = *fp++;
	}


        // Copy file name
        fp = packet->file;
        for (field_index = 0;
	     field_index < packet->file_size;
	     field_index++)
	{
                *dp++ = *fp++;
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Loop list lock
	//
	//  The following code must be semantically in sync
	//  with fl_size64().
	//
	//  Calling fl_size64() directly will cause instant
	//  deadlock on loop list lock.


	// Reset device size
	dev_size = 0;


	// Check for installation CD
	if (loop_entry->install_cd.is_attached == 1)
	{
		// Set device size
		dev_size = loop_entry->install_cd.part_size;
	}
	else
	{
		if (loop_entry->file.vap->va_size >= 0)
		{
			// Set device size
			dev_size = loop_entry->file.vap->va_size;
		}
	}


	// Set field pointer
	fp = (char *) &dev_size;


#               endif
#       endif
#endif


	// Copy file size in bytes
	for (field_index = 0;
	     field_index < 8;
	     field_index++)
	{
                *dp++ = *fp++;
	}


	// Set result variables
	*data_pointer = dp;


	return 1;	// SUCCESS
}


int fl_packet_new_resp()
{
	char *dp;
	char *fp;
	fl_uint32_t field_index;
	struct fl_packet_s *packet;


	// NOTE: Processing


	// Set packet pointer
	packet = &dsf_control.packet_resp;


	// Switch commands
	switch (packet->cmd)
	{
		case FL_CMD_ATTACH:
		case FL_CMD_DETACH:

			// NOTE: On Tru64, pointers are 8 bytes long, use data pointer as buffer


        		// Set data pointer
        		dp = (char *) &packet->data;


        		// Set field pointer to command
        		fp = (char *) &packet->cmd;


        		// Copy to packet
        		for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
                		*dp++ = *fp++;
			}


			// Set field pointer to return code
			fp = (char *) &dsf_control.ret;


			// Copy to packet
			for (field_index = 0;
			     field_index < 4;
			     field_index++)
			{
				*dp++ = *fp++;
			}


			break;

		case FL_CMD_LIST:

			// NOTE: List packet data
			//
			//  Packet data for the list command is allocated and set in
			//  fl_cmd_list(). This includes command code, length of list,
			//  and return code.


			break;

		default:

			break;
	}


	return 1;	// SUCCESS
}


%FUNC:fl_cmd_list:START%


int fl_cmd_list(void)
{
	char *dp;
	char *fp;
	fl_uint32_t cmd;
	fl_uint32_t list_size;
	fl_uint32_t entry_size;
	fl_uint32_t field_index;
	fl_uint64_t list_length;
	fl_uint64_t list_index;
	fl_uint32_t list_length_uint32;
	fl_list_node_t *list_node;
	struct loop_entry_s *loop_entry;


	FL_TEST_TYPE_DECL


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


        // NOTE: Response packet format
        //
        //  A response packet is composed of the following fields, sequentially
        //  laid out in a byte stream:
        //
        //  Command:                            C (4 bytes - unsigned int)
        //  Command data:                       D (? bytes - unsigned int)
        //--------------------------------------------------------------------------------------
        //  Command: FL_CMD_LIST		C  (4 bytes - unsigned int)
        //  Command data:                       D  (? bytes - unsigned int)
        //    Length of list:                   L  (4 bytes - unsigned int)
        //    0. Device file name size:         DS (2 bytes - unsigned short)
        //    0. Device file name:              DN (? bytes - char[])
        //    0. File name size:                FS (2 bytes - unsigned short)
        //    0. File name:                     FN (? bytes - char[])
        //    0. File size:                     S  (8 bytes - unsigned long long)
        //    ...
        //    L - 1. Device file name size:     DS (2 bytes - unsigned short)
        //    L - 1. Device file name:          DN (? bytes - char[])
        //    L - 1. File name size:            FS (2 bytes - unsigned short)
        //    L - 1. File name:                 FN (? bytes - char[])
        //    L - 1. File size:                 S  (8 bytes - unsigned long long)
        //
        //  All multibyte data values are copied in host endianness:
	//
	//    OSF/1: little endian
	//    IRIX:  big endian
	//
	//  On each invocation, at most a page size of bytes, _PAGESZ, is filled
	//  with available loop entries. The resulting buffer is simply called
	//  a listing.
	//
	//  The internal state is stored in the global variable dsf_control,
	//  specifically in the field list_index. This field contains the index
	//  of the next loop entry to be added to the next listing. Initially,
	//  this field contains the value 0 (zero).


	// Acquire lock of loop list
	fl_spin_lock(loop_list_lock);


	// NOTE: List size
	//
	//  This variable is initialized to 4 + 4 to account for the command
	//  code and list length.


	list_size = 4 +	// Command code size
		    4;	// List length size


	// Set data to request buffer
	dp = req_buffer;


	// Copy command code to data
	cmd = FL_CMD_LIST;
	fp = (char *) &cmd;
	for (field_index = 0;
	     field_index < 4;
	     field_index++)
	{
		*dp++ = *fp++;
	}


	// Get length of loop list
	fl_list_length(loop_list,
		       &list_length);


	// Check for list length
        if (list_length == 0)
        {
		// Set list size
		dsf_control.list_size = list_size;


		// Set return code
		dsf_control.ret = %fl_c:*:NO_LOOP_ENTRIES[ip,3,]%;


		// Copy list length to data
		list_length_uint32 = 0;
		fp = (char *) &list_length_uint32;
		for (field_index = 0;
		     field_index < 4;
		     field_index++)
		{
			*dp++ = *fp++;
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_list] INFO: Number of loop entries = %u list_size = %llu\n",
				  list_length_uint32,
				  list_size))


		// Release lock of loop list
		fl_spin_unlock(loop_list_lock);


		return 2;       // SUCCESS: Couldn't find loop entries
	}
	else
	{
		// NOTE: Double traversal of loop list
		//
		//  In the first pass, the number of valid entries and total
		//  size are calculated.
		//
		//  In the second pass, the actual data fields are copied to the packet
		//  data.
		//
		//  The maximum data size for one listing is _PAGESZ.
		//
		//  If there is more data available, successive calls to this function
		//  will return subsequent listings until all list entries have been
		//  returned.
		//
		//  Each invocation starts at the last index processed plus one, stored
		//  in dsf_control.list_index.


		// NOTE: Filter loop entries (first pass)


		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_list] INFO: First pass\n"))


		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_list] INFO: list_index = %llu\n",
				  dsf_control.list_index))


		FL_DEBUG_UPRINTF((\
				  "[fl_cmd_list] INFO: list_length = %llu\n",
				  list_length))


		list_length_uint32 = 0;
        	for (list_index = dsf_control.list_index;
		     list_index < list_length;
		     list_index++)
        	{
                	// Get next list node
                	fl_list_get_node(loop_list,
					 &list_node,
					 list_index);


                	// Set loop entry
                	loop_entry = (struct loop_entry_s *) list_node->data;


			// Get entry data size
			entry_size = 2 +
                             	     loop_entry->packet.dev_size +
                             	     2 +
                             	     loop_entry->packet.file_size +
                             	     8;


			// Check max data size
			if (entry_size > (_PAGESZ - list_size))
			{
				// NOTE: Truncate listing
				//
				//  There is more data available than can fit into
				//  the current listing (which is _PAGESZ bytes).


				// Set return code
				dsf_control.ret = %fl_c:*:LIST_TRUNCATED[ip,4,]%;


				break;
			}


			// Add entry to total list size
			list_size += entry_size;


			// Increment entry counter
			list_length_uint32++;
		}


		// Set list size
		dsf_control.list_size = list_size;
	}


	// NOTE: Byte order (endianness)
	//
	//  Cast from fl_uint64_t to fl_uint32_t for correct handling of
	//  endianness.
	//
	//  Overflow is taken care of by the max data size check in first pass
	//  for loop, see above.


	// Copy list length to data
	fp = (char *) &list_length_uint32;
	for (field_index = 0;
	     field_index < 4;
	     field_index++)
	{
		*dp++ = *fp++;
	}


        FL_DEBUG_UPRINTF((\
			  "[fl_cmd_list] INFO: Number of loop entries = %u list_size = %llu\n",
			  list_length_uint32,
			  list_size))


	// Reset list length
	//list_length = (fl_uint64_t) list_index;
	list_length = dsf_control.list_index +
		      (fl_uint64_t) list_length_uint32;


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_list] INFO: Second pass\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_list] INFO: list_index = %llu\n",
			  dsf_control.list_index))


	FL_DEBUG_UPRINTF((\
			  "[fl_cmd_list] INFO: list_length = %llu\n",
			  list_length))


	// Traverse loop list (second pass)
	for (list_index = dsf_control.list_index;
	     list_index < list_length;
	     list_index++)
	{
		// Get next list node
		fl_list_get_node(loop_list,
				 &list_node,
				 list_index);


		// Set loop entry
		loop_entry = (struct loop_entry_s *) list_node->data;


		// Copy loop entry fields to packet data
		fl_entry_serialize(loop_entry,
				   &dp);
	}


	// Set list index
	dsf_control.list_index = list_index;


	// Release lock of loop list
	fl_spin_unlock(loop_list_lock);


	// Check for truncated return code
	if (dsf_control.ret == %fl_c:*:LIST_TRUNCATED[op,,]%)
        {
                return 3;       // SUCCESS: Listing truncated
        }


	// Set default return code to SUCCESS
	dsf_control.ret = %fl_c:*:LIST_SUCCESS[ip,5,]%;


	return 1;	// SUCCESS
}


%FUNC:fl_cmd_list:STOP%


int fl_session_reset(struct fl_proc_s *fl_proc)	// -- IN
{
	fl_uint64_t value_before;
	fl_uint64_t value_after;
	struct dsf_control_s *dsf_control;
	struct fl_args_s args;


	FL_DEBUG_UPRINTF((\
			  "[fl_session_reset] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_session_reset] INFO: fl_proc = %p\n",
			  fl_proc))


	// NOTE: Arguments


	// IN: Check for fl proc
	if (fl_proc == NULL)
	{
		return 0;	// ERROR: No fl proc
	}


	// NOTE: Initialization


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


	// Check for correct state
	if (fl_proc->reset == FL_PROC_RESET)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_session_reset] INFO: FL_PROC_RESET\n"))


		return -1;	// ERROR: Fl proc already reset
	}


	// Check for dsf_control
	if (fl_proc->ctrl != NULL)
	{
		// Get dsf_control
		dsf_control = fl_proc->ctrl;


		// Reset dsf_control
		dsf_control->packet = NULL;
		dsf_control->loop_entry = NULL;
		dsf_control->list_index = 0;
		dsf_control->ret = 0;


		// Check for allocated response packet
		if (dsf_control->packet_resp.data != NULL)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_session_reset] INFO: packet_resp.cmd = %d\n",
					  dsf_control->packet_resp.cmd))


			// Check for list command
			if (dsf_control->packet_resp.cmd == FL_CMD_LIST)
			{
				// Free only if not using data_list quick buffer
				if (dsf_control->packet_resp.data != dsf_control->packet_resp.data_list)
				{
					FL_DEBUG_UPRINTF((\
							  "[fl_session_reset] INFO: FL_CMD_LIST uses dynamically allocated buffer\n"))


					// Free list packet
					fl_mem_free(dsf_control->packet_resp.data);
				}
				else
				{
					FL_DEBUG_UPRINTF((\
							  "[fl_session_reset] INFO: FL_CMD_LIST uses quick buffer\n"))
				}
			}


			// Reset packet data
			dsf_control->packet_resp.data = NULL;
		}


		// Reset response packet data size
		dsf_control->packet_resp.size = 0;


		// Reset fl proc ctrl
		fl_proc->ctrl = NULL;
	}


	// Check semaphore value
	fl_sem_get_value(ctrl_sem, &value_before);


#ifdef FL_TESTS


	fl_test_ctrl_sem = value_before;


#endif


	// Increment semaphore for ctrl
	if (fl_sem_incr(ctrl_sem,
			1,
			FL_SEM_SWITCH_RETURN_ON_SLEEP,
			&args) < 1)
	{
		// NOTE: Tripwire for semaphore overflow
		//
		//  In case of an externally induced semaphore value
		//  overflow, deny further access to the driver and
		//  ask user to reload module.
		//
		//  A reload will reset file looper's internal state,
		//  effectively avoiding deadlocks, kernel panics, and
		//  other unexpected behavior.


		reset_switch = 1;
	}


	// Check semaphore value
        fl_sem_get_value(ctrl_sem,
			 &value_after);


	FL_DEBUG_UPRINTF((\
			  "[fl_session_reset] INFO: ctrl_sem.value = %llu -> %llu\n",
			  value_before,
			  value_after))


#ifdef FL_TESTS


	fl_test_ctrl_sem = value_after;


	if (value_after > 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_session_reset] INFO: fl_test_ctrl_sem = %llu > 1\n",
				  value_after))
	}


#endif


	// Set fl proc reset flag
	fl_proc->reset = FL_PROC_RESET;


        FL_DEBUG_UPRINTF((\
			  "[fl_session_reset] INFO: STOP\n"))


	return 1;	// SUCCESS
}


int fl_packet_read(char *data,			// -- IN
		   struct fl_packet_s *packet)	// -- IN
{
	char *dp;
	char *fp;
	fl_uint16_t field_index;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Check for data
	if (data == NULL)
	{
		return 0;	// ERROR: No data
	}


	// IN: Check for packet
	if (packet == NULL)
	{
		return -1;	// ERROR: No packet
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


        // NOTE: Request packet format
        //
        //  A request packet is composed of the following fields, sequentially
        //  laid out in a byte stream:
        //
        //  Command:                    C (4 bytes - unsigned int)
        //  Command data:               ? (? bytes - unsigned int)
        //----------------------------------------------------------------------------------
        //  Command: FL_CMD_ATTACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //    File name size:           FS (2 bytes - unsigned short)
        //    File name:                FN (? bytes - char[])
        //
        //  Command: FL_CMD_DETACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //
        //  Command: FL_CMD_LIST
        //  Command data:               D (0 bytes - unsigned int)
        //
        //  All multibyte data values are copied in host endianness.


	// Set data pointer
	dp = data;


	// Initialize past command field
	dp += 4;


	FL_DEBUG_UPRINTF((\
			  "[fl_packet_read] INFO: packet->cmd = %d\n",
			  packet->cmd))


	// Switch commands
	switch (packet->cmd)
	{
		case FL_CMD_ATTACH:

			// Copy device file name size
			fp = (char *) &packet->dev_size;
			for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->dev_size = %d\n",
					  packet->dev_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_DEV_NAME_AT(0, 1)


			// Allocate device file name
			fl_mem_alloc((void **) &packet->dev,
                                     (fl_uint3264_t) packet->dev_size,
                                     FL_MEM_WAITOK,
                                     "packet_read",
                                     "Allocate device file name (FL_CMD_ATTACH)");


			// Copy device file name
			fp = packet->dev;
			for (field_index = 0;
			     field_index < packet->dev_size;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->dev = \"%s\" strlen() = %llu\n",
					  packet->dev,
					  strlen(packet->dev)))


			// Copy file name size
			fp = (char *) &packet->file_size;
			for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->file_size = %d\n",
					  packet->file_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_FILE_NAME_AT(0, 1)


			// Allocate file name
			fl_mem_alloc((void **) &packet->file,
                                     (fl_uint3264_t) packet->file_size,
                                     FL_MEM_WAITOK,
                                     "packet_read",
                                     "Allocate file name (FL_CMD_ATTACH)");


			// Copy file name
			fp = packet->file;
        		for (field_index = 0;
			     field_index < packet->file_size;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->file = \"%s\" strlen() = %llu\n",
					  packet->file,
					  strlen(packet->file)))


			break;

		case FL_CMD_DETACH:

                        // Copy device file name size
                        fp = (char *) &packet->dev_size;
                        for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                                *fp++ = *dp++;
			}


                        FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->dev_size = %d\n",
					  packet->dev_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_BLK_DEV_NAME_DE(0, 1)


                        // Allocate device file name
			fl_mem_alloc((void **) &packet->dev,
                                     (fl_uint3264_t) packet->dev_size,
                                     FL_MEM_WAITOK,
                                     "packet_read",
                                     "Allocate device file name (FL_CMD_DETACH)");


                        // Copy device file name
                        fp = packet->dev;
                        for (field_index = 0;
			     field_index < packet->dev_size;
			     field_index++)
			{
                                *fp++ = *dp++;
			}


                        FL_DEBUG_UPRINTF((\
					  "[fl_packet_read] INFO: packet->dev = %s\n",
					  packet->dev))


			break;

		case FL_CMD_LIST:

			break;

		default:

			break;
	}


	return 1;	// SUCCESS
}


int fl_packet_parse(char *buffer,		// -- IN
		    fl_uint32_t cmd,		// -- IN
		    struct fl_packet_s *packet)	// -- IN
{
	char *dp;
	char *fp;
	fl_uint16_t field_index;
	fl_uint32_t buffer_cmd;


	FL_TEST_TYPE_DECL


	// NOTE: Arguments


	// IN: Check for buffer
	if (buffer == NULL)
	{
		return 0;	// ERROR: No buffer
	}


	// IN: Check for valid command
	if ((cmd != FL_CMD_ATTACH) &&
	    (cmd != FL_CMD_DETACH) &&
	    (cmd != FL_CMD_LIST))
	{
		return -1;	// ERROR: Invalid command
	}


	// IN: Check for packet
	if (packet == NULL)
	{
		return -2;	// ERROR: No packet
	}


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// NOTE: Processing


        // NOTE: Request packet format
        //
        //  A request packet is composed of the following fields, sequentially
        //  laid out in a byte stream:
        //
        //  Command:                    C (4 bytes - unsigned int)
        //  Command data:               ? (? bytes - unsigned int)
        //----------------------------------------------------------------------------------
        //  Command: FL_CMD_ATTACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //    File name size:           FS (2 bytes - unsigned short)
        //    File name:                FN (? bytes - char[])
        //
        //  Command: FL_CMD_DETACH
        //  Command data:               D  (? bytes - unsigned int)
        //    Device file name size:    DS (2 bytes - unsigned short)
        //    Device file name:         DN (? bytes - char[])
        //
        //  Command: FL_CMD_LIST
        //  Command data:               D (0 bytes - unsigned int)
        //
        //  All multibyte data values are copied in host endianness.


	// Set data pointer
	dp = req_buffer;


	// Get buffer command
	buffer_cmd = *((fl_uint32_t *) dp);


	// Check for correct command
	if (buffer_cmd != cmd)
	{
		return -3;	// ERROR: Incorrect command
	}


	// Set packet command
	packet->cmd = buffer_cmd;


	FL_DEBUG_UPRINTF((\
			  "[fl_packet_parse] INFO: packet->cmd = %d\n",
			  packet->cmd))


	// Move to next field
	dp += 4;


	// Switch commands
	switch (packet->cmd)
	{
		case FL_CMD_ATTACH:

			// Copy device file name size
			fp = (char *) &packet->dev_size;
			for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->dev_size = %d\n",
					  packet->dev_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_DEV_NAME_AT(0, 1)


			// Copy device file name
			fp = packet->dev;
			for (field_index = 0;
			     field_index < packet->dev_size;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->dev = \"%s\" strlen() = %llu\n",
					  packet->dev,
					  strlen(packet->dev)))


			// Copy file name size
			fp = (char *) &packet->file_size;
			for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->file_size = %d\n",
					  packet->file_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_FILE_NAME_AT(0, 1)


			// Copy file name
			fp = packet->file;
        		for (field_index = 0;
			     field_index < packet->file_size;
			     field_index++)
			{
                		*fp++ = *dp++;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->file = \"%s\" strlen() = %llu\n",
					  packet->file,
					  strlen(packet->file)))


			break;

		case FL_CMD_DETACH:

                        // Copy device file name size
                        fp = (char *) &packet->dev_size;
                        for (field_index = 0;
			     field_index < 2;
			     field_index++)
			{
                                *fp++ = *dp++;
			}


                        FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->dev_size = %d\n",
					  packet->dev_size))


			FL_TEST_F_fl_packet_read_N_NOT_ALLOC_BLK_DEV_NAME_DE(0, 1)


                        // Copy device file name
                        fp = packet->dev;
                        for (field_index = 0;
			     field_index < packet->dev_size;
			     field_index++)
			{
                                *fp++ = *dp++;
			}


                        FL_DEBUG_UPRINTF((\
					  "[fl_packet_parse] INFO: packet->dev = %s\n",
					  packet->dev))


			break;

		case FL_CMD_LIST:

			break;

		default:

			break;
	}


	return 1;	// SUCCESS
}


int fl_data_get_cmd(char *data,		// -- IN
		    fl_uint32_t *cmd)	// -- OUT
{
	char *dp;
	char *fp;
	fl_uint64_t field_index;


	// NOTE: Arguments


	// IN: Check for data
	if (data == NULL)
	{
		return 0;	// ERROR: No data
	}


	// OUT: Check for command
	if (cmd == NULL)
	{
		return -1;	// ERROR: No command
	}


	// NOTE: Initialization


        // Set data pointer
        dp = data;


	// NOTE: Processing


        // Copy command
        fp = (char *) cmd;
        for (field_index = 0;
	     field_index < 4;
	     field_index++)
	{
        	*fp++ = *dp++;
	}


	return 1;	// SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


int fl_open_new_pid(dev_t dev,			// -- IN
		    fl_uint32_t pid_type)	// -- IN
{
	pid_t pid;
	fl_list_t *pid_list;
	fl_list_node_t *pid_node;
	struct loop_entry_s *loop_entry;
	struct fl_args_s args;


	FL_TEST_TYPE_DECL


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30





#               endif
#       endif
#endif


	// NOTE: Arguments


	// IN: Check for valid pid types
	if ((pid_type != FL_PID_CHR) &&
	    (pid_type != FL_PID_BLK))
	{
		return 0;	// ERROR: Invalid pid type
	}


	// NOTE: Initialization


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	// Get pid of current process
	fl_proc_get_data(FL_DATA_PID,
			 &pid);


	FL_DEBUG_UPRINTF((\
			  "[fl_open_new_pid] INFO: pid = %d\n",
			  pid))


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY(FL_TEST_STATE_BEFORE,\
						 1,\
						 &args)


	// Get loop entry
	loop_entry = NULL;
	//if (FL_TEST_COND((fl_test_ret == 1), ||)
        if (fl_entry_get_by_dev(dev,
				&loop_entry,
				&args) < 1)
	{
		// NOTE: Loop entry not found
		//
		//  The loop entry for this device was not found. This happens when the
		//  user detached the BLK device in the time passed between execution of
		//  functions
		//
		//  fl_dev_check_minor_is_attached(minor)
		//
		//  and
		//
		//  fl_spin_lock(loop_list_lock)
		//
		//  See fl_cmd_detach() for details.


		// Release lock for loop list
		fl_spin_unlock(loop_list_lock);


		FL_DEBUG_UPRINTF(("[fl_open_new_pid] INFO: return -1\n"))


		FL_TEST("fl_open",\
			"NO_LOOP_ENTRY")
		return -1;      // ERROR: Loop entry not found
	}


	// Acquire lock for loop entry
	fl_spin_lock(loop_entry->lock);


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: Exclusive access
	//
	//  Implementation of DIOCFORMAT I/O control command
	//  requires exclusive access to a device.


	FL_TEST_F_fl_open_N_IS_IN_EXCLUSIVE_USE(FL_TEST_STATE_BEFORE,\
						1,\
						FL_TEST_NO_ARGS)


	if (loop_entry->is_exclusive == 1)
	{
		// NOTE: Exclusive access


		// Release lock for loop entry
		fl_spin_unlock(loop_entry->lock);


		// Release lock for loop list
		fl_spin_unlock(loop_list_lock);


		FL_DEBUG_UPRINTF(("[fl_open_new_pid] INFO: return -2\n"))


		return -2;      // ERROR: Device is being used in exclusive access mode
	}


#               endif
#       endif
#endif


	// Set pid list
	if (pid_type == FL_PID_CHR)
	{
		pid_list = loop_entry->pid_list_chr;
	}
	else
	{
		pid_list = loop_entry->pid_list_blk;
	}


	// Create new pid node
	pid_node = NULL;
	fl_list_new_node(pid_list,
			 &pid_node);


	// Set data
	pid_node->data = (void *) pid;


	// Add pid node to loop entry
	fl_list_push(pid_list,
		     pid_node);


	// Mark device in use
	if (pid_type == FL_PID_CHR)
	{
		loop_entry->chr_is_inuse = 1;
	}
	else
	{
		loop_entry->blk_is_inuse = 1;
	}


	// NOTE: First open
	//
	//  Some platforms require initializations to be done on a BLK
	//  device file when opened. Other functions expect certain
	//  data structures to be set and fl_open() is a good place
	//  for this.
	//
	//  These initializations are done only once for each newly opened
	//  BLK device file.


	if (pid_type == FL_PID_BLK)
	{
		// Check for first open
		if (loop_entry->first_open == 1)
		{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


			// Reset first open flag
			loop_entry->first_open = 0;
		}
	}


	// Release lock for loop entry
	fl_spin_unlock(loop_entry->lock);


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	FL_TEST_F_fl_open_N_SIMPLE_OPEN_ATTACHED_BLK_DEV(FL_TEST_STATE_BEFORE,\
							 0,\
							 FL_TEST_NO_ARGS)


	FL_TEST("fl_open",\
		"SIMPLE_OPEN_ATTACHED_BLK_DEV")
	return 1;	// SUCCESS
}


int fl_close_empty_pids(dev_t dev,		// -- IN
			fl_uint32_t pid_type)	// -- IN
{
	fl_list_t *pid_list;
	fl_uint64_t pid_length;
	struct loop_entry_s *loop_entry;
	struct fl_args_s args;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


	// NOTE: Arguments


	// IN: Check for valid pid type
	if ((pid_type != FL_PID_CHR) &&
	    (pid_type != FL_PID_BLK))
	{
		return 0;	// ERROR: Invalid pid type
	}


	// NOTE: Initialization


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


	// Get current pid process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Acquire lock for loop list
	fl_spin_lock(loop_list_lock);


	// NOTE: Always succeeds
	//
	//  This call always succeeds, since there can be no close without a prior open.


	// Get loop entry for dev
	fl_entry_get_by_dev(dev,
			    &loop_entry,
			    &args);


	if (pid_type == FL_PID_CHR)
	{
		// Set pid list
		pid_list = loop_entry->pid_list_chr;
	}
	else
	{
		// Set pid list
		pid_list = loop_entry->pid_list_blk;
	}


	// Acquire lock for loop entry
	fl_spin_lock(loop_entry->lock);


	// Get length of CHR pid list
	fl_list_length(pid_list,
		       &pid_length);


	FL_DEBUG_UPRINTF((\
			  "[fl_close_empty_pids] INFO: Pid length = %llu (pid=%llu)\n",
			  pid_length,
			  dbgpid))


	// Empty pid list
	fl_list_empty(pid_list);


	if (pid_type == FL_PID_CHR)
	{
		// Mark device not in use
		loop_entry->chr_is_inuse = 0;
	}
	else
	{
		// Mark device not in use
		loop_entry->blk_is_inuse = 0;
	}


	// Release lock for loop entry
	fl_spin_unlock(loop_entry->lock);


	// Release lock for loop list
	fl_spin_unlock(loop_list_lock);


	return 1;	// SUCCESS
}


int fl_close_empty_proc(pid_t pid)	// -- IN
{
	fl_uint64_t proc_counter;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


	// NOTE: Processing


	// Get current pid process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Acquire lock for proc list
	fl_proc_lock();


	// Loop through proc list
	proc_counter = 0;
	while (fl_proc_shift_unlocked(&proc) >= 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_close_empty_proc] INFO: proc = %p proc->pid = %u (pid=%llu)\n",
				  proc,
				  proc->pid,
				  dbgpid))


		// Get fl proc
		fl_proc = (struct fl_proc_s *) proc->priv;


		// Check for current pid
		if (proc->pid == pid)
		{
			// Reset session
			fl_session_reset(fl_proc);
		}


		// Free proc
		fl_mem_free(proc);


		// Free fl proc
		fl_mem_free(fl_proc);


		// Increment proc counter
		proc_counter++;
	}


	// Release lock for proc list
	fl_proc_unlock();


	FL_DEBUG_UPRINTF((\
			  "[fl_close_empty_proc] INFO: Process counter = %llu (pid=%llu)\n",
			  proc_counter,
			  dbgpid))


	return 1;	// SUCCESS
}


int fl_int_compare_ranges(fl_uint64_t uint_num64,	// -- IN
                          fl_uint32_t max_index,	// -- IN
                          fl_uint32_t *int_digits)	// -- OUT
{
        fl_uint32_t int_index;
        fl_uint64_t int_dec;


	// NOTE: Arguments


        // IN: Check for valid max index
        if ((max_index != 10) &&
            (max_index != 20))
        {
                return 0;      // ERROR: Invalid integer width
        }


	// OUT: Check for integer digits
        if (int_digits == NULL)
        {
                return -1;      // ERROR: No integer digits
        }


	// NOTE: Processing


        // NOTE: Convert to unsigned


        // Compare ranges
        *int_digits = 0;
        int_dec = 10;
        for (int_index = 0;
	     int_index < max_index;
	     int_index++)
        {
                (*int_digits)++;


                // Continue until number is higher, or end of loop
                if (int_dec <= uint_num64)
                {
                        int_dec *= 10;
                        continue;
                }


                break;
        }


        return 1;       // SUCCESS
}


int fl_int_get_digits(fl_uint32_t int_sign,	// -- IN
                      void *int_num,		// -- OUT
                      fl_uint32_t int_width,	// -- IN
                      fl_uint32_t *int_digits)	// -- OUT
{
        fl_int32_t int_num32;
        fl_int64_t int_num64;
        fl_uint64_t uint_num64;


	// NOTE: Arguments

        
        // IN: Check for valid integer sign
        if ((int_sign != 0) &&
            (int_sign != 1))
        {       
                return 0;       // ERROR: No integer sign
        }


	// OUT: Check for integer number
        if (int_num == NULL)
        {       
                return -1;      // ERROR: No integer number
        }


	// IN: Check for valid integer width
        if ((int_width != 32) &&
            (int_width != 64))
        {       
                return -2;      // ERROR: Invalid integer width
        }


	// OUT: Check for integer digits
        if (int_digits == NULL)
        {       
                return -3;      // ERROR: No integer digits
        }


	// NOTE: Processing

        
        // NOTE: Integer ranges
        //  
        //  int                 ==>              -2147483648 - 2147483647 (10)
        //  unsigned int        ==>                        0 - 4294967295 (10)
        //  long long           ==>     -9223372036854775808 - 9223372036854775807  (19)
        //  unsigned long long  ==>                        0 - 18446744073709551615 (20)

        
        *int_digits = 0; 
        if (int_width == 32)
        {       
                if (int_sign == 1)
                {       
                        int_num32 = *((fl_int32_t *) int_num);

                        
                        if (int_num32 == -2147483648)
                        {       
                                *int_digits = 10;
                        }
                        else
                        {       
                                if (int_num32 < 0)
                                {       
                                        uint_num64 = (fl_uint64_t) (int_num32 * -1);
                                }
                                else
                                {       
                                        uint_num64 = (fl_uint64_t) int_num32;
                                }

                                
                                // Compare ranges
                                fl_int_compare_ranges(uint_num64,
                                                      10,
                                                      int_digits);
                        }
                }
                else
                {
                        uint_num64 = (fl_uint64_t) *((fl_uint32_t *) int_num);


                        // Compare ranges
                        fl_int_compare_ranges(uint_num64,
                                              10,
                                              int_digits);
                }
        }
        else
        {
                if (int_sign == 1)
                {
                        int_num64 = *((fl_int64_t *) int_num);


                        if (int_num64 == -9223372036854775808)
                        {
                                *int_digits = 19;
                        }
                        else
                        {
                                if (int_num64 < 0)
                                {
                                        uint_num64 = (fl_uint64_t) (int_num64 * -1);
                                }
                                else
                                {
                                        uint_num64 = (fl_uint64_t) int_num64;
                                }


                                // Compare ranges
                                fl_int_compare_ranges(uint_num64,
                                                      20,
                                                      int_digits);
                        }
                }
                else
                {
                        uint_num64 = *((fl_uint64_t *) int_num);


                        // Compare ranges
                        fl_int_compare_ranges(uint_num64,
                                              20,
                                              int_digits);
                }
        }


        return 1;       // SUCCESS
}


int fl_ioctl_del_daemon(void)
{
	fl_uint32_t daemon_found;
	fl_uint64_t proc_index;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;


	// NOTE: Processing


	// Acquire lock for proc list
	fl_proc_lock();


	// Loop through proc list
	proc_index = 0;
	daemon_found = 0;
	while (fl_proc_get_index_unlocked(&proc,
					  proc_index) >= 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl_del_daemon] INFO: proc = %p proc->pid = %u\n",
				  proc,
				  proc->pid))


		// Get fl proc
		fl_proc = (struct fl_proc_s *) proc->priv;


		// Check for daemon process
		if (fl_proc->state == FL_ST_DAEMON)
		{
			// NOTE: Daemon process
			//
			//  Only one other daemon process, at most,
			//  is in the global proc_list (fl_proc.c).


			// Set daemon found
			daemon_found = 1;


			break;
		}


		// Increment proc index
		proc_index++;
	}


	// Check for daemon process
	if (daemon_found == 1)
	{
		// Delete proc node
		fl_proc_del_index_unlocked(&proc,
					   proc_index);


		// Free proc
		fl_mem_free(proc);


		// Free fl proc
		fl_mem_free(fl_proc);


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl_del_daemon] INFO: Daemon process deleted at %llu\n",
				  proc_index))
	}


	// Release lock for proc list
	fl_proc_unlock();


	return 1;	// SUCCESS
}


int fl_strategy_alloc_io_node(enum fld_io_e io_control,		// -- IN
			      buf_t *io_bp,			// -- IN
			      fl_uint64_t io_offset,		// -- IN
			      fl_uint64_t io_size,		// -- IN
			      dev_t io_dev,			// -- IN
			      char *file,			// -- IN
			      char *io_data,			// -- IN
			      fl_uint64_t io_data_size,		// -- IN
			      fl_uint32_t is_format,		// -- IN
			      fl_list_node_t **io_node,		// -- OUT
			      struct fld_io_req_s **io_req)	// -- OUT
{
	struct fld_io_req_s *io_req_1;


	// NOTE: Arguments


	// IN: Check for valid I/O control
	if ((io_control != FLD_IO_READ) &&
	    (io_control != FLD_IO_WRITE))
	{
		return 0;	// ERROR: Invalid I/O control
	}


	// IN: Check for buffer
	if (io_bp == NULL)
	{
		return -1;	// ERROR: No buffer
	}


	// IN: Check for file name
	if (file == NULL)
	{
		return -2;	// ERROR: No file name
	}


	// IN: Check for data
	if (io_data == NULL)
	{
		return -3;	// ERROR: No data
	}


	// IN: Check for is format switch
	if ((is_format != 0) &&
	    (is_format != 1))
	{
		return -4;	// ERROR: Invalid is format switch
	}


	// OUT: Check for I/O node
	if (io_node == NULL)
	{
		return -5;	// ERROR: No I/O node
	}


	// OUT: Check for I/O request
	if (io_req == NULL)
	{
		return -6;	// ERROR: No I/O request
	}


	// NOTE: Processing


	// Allocate I/O request struct
	fl_mem_alloc((void **) &io_req_1,
                     sizeof(struct fld_io_req_s),
                     FL_MEM_ZERO | FL_MEM_WAITOK,
                     "fl_strategy_alloc_io_node",
                     "Allocate I/O request struct");


	// Populate I/O request
	io_req_1->control = io_control;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->control = %u\n",
			  io_req_1->control))


	io_req_1->bp = io_bp;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->bp = %p\n",
			  io_req_1->bp))


	io_req_1->offset = io_offset;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->offset = %llu\n",
			  io_req_1->offset))


	io_req_1->size = io_size;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->size = %llu\n",
			  io_req_1->size))


	io_req_1->dev = io_dev;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->dev = %u\n",
			  io_req_1->dev))


	io_req_1->file[0] = '\0';
	strcat(io_req_1->file,
	       file);


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_alloc_io_node] INFO: io_req_1->file = \"%s\"\n",
			  io_req_1->file))


	// Check for write I/O
	if (io_req_1->control == FLD_IO_WRITE)
	{
		// Set data pointer
		io_req_1->data = io_data;


		FL_DEBUG_UPRINTF((\
				  "[fl_strategy_alloc_io_node] INFO: io_req_1->data = %p\n",
				  io_req_1->data))


		// Set data size
		io_req_1->data_size = io_data_size;


		FL_DEBUG_UPRINTF((\
				  "[fl_strategy_alloc_io_node] INFO: io_req_1->data_size = %llu\n",
				  io_req_1->data_size))
	}


	// Set is format switch field
	io_req_1->is_format = is_format;


	// Create new I/O list node
	fl_list_new_node(io_list,
			 io_node);


	// Set data field
	(*io_node)->data = (void *) io_req_1;


	// Set I/O request
	*io_req = io_req_1;


	return 1;	// SUCCESS
}


int fl_strategy_enqueue_io(enum fld_io_e io_control,		// -- IN
			   buf_t *io_bp,			// -- IN
			   fl_uint64_t io_offset,		// -- IN
			   fl_uint64_t io_size,			// -- IN
			   dev_t io_dev,			// -- IN
			   char *file,				// -- IN
			   char *io_data,			// -- IN
			   fl_uint64_t io_data_size,		// -- IN
			   fl_uint32_t is_format,		// -- IN
			   struct fld_io_req_s **io_req,	// -- OUT
			   fl_int32_t *ret)			// -- OUT
{
	fl_list_node_t *io_node;


	// NOTE: Arguments


	// IN: Check for valid I/O control
	if ((io_control != FLD_IO_READ) &&
	    (io_control != FLD_IO_WRITE))
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = 0;
		}


		return 0;	// ERROR: Invalid I/O control
	}


	// IN: Check for buffer
	if (io_bp == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -1;
		}


		return -1;	// ERROR: No buffer
	}


	// IN: Check for file name
	if (file == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -2;
		}


		return -2;	// ERROR: No file name
	}


	// IN: Check for data
	if (io_data == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -3;
		}


		return -3;	// ERROR: No data
	}


	// IN: Check for is format switch
	if ((is_format != 0) &&
	    (is_format != 1))
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -4;
		}


		return -4;	// ERROR: Invalid is format switch
	}


	// OUT: Check for I/O request
	if (io_req == NULL)
	{
		// Conditionally set return value
		if (ret != NULL)
		{
			*ret = -5;
		}


		return -5;	// ERROR: No I/O request
	}


	// OUT: Check for return value
	if (ret == NULL)
	{
		return -6;	// ERROR: No return value
	}


	// NOTE: Processing


	// Allocate I/O node
	fl_strategy_alloc_io_node(io_control,
				  io_bp,
				  io_offset,
				  io_size,
				  io_dev,
				  file,
				  io_data,
				  io_data_size,
				  is_format,
				  &io_node,
				  io_req);


	// Enqueue I/O node
	fl_list_push(io_list,
		     io_node);


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_enqueue_io] INFO: I/O node enqueued: %llu bytes\n",
			  io_size))


	// Check for daemon state
	if (daemon_state == FL_DST_IDLE)
	{
		// Send signal to daemon
		if (proc_signal(daemon_ref,
				SIGURG) == -1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_strategy_enqueue_io] ERROR: Couldn't send SIGURG to daemon\n"))
		}
		else
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_strategy_enqueue_io] INFO: Signal SIGURG sent\n"))
		}
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_strategy_enqueue_io] INFO: daemon_state != FL_DST_IDLE = %u = %s\n",
				  daemon_state,
				  (daemon_state == FL_DST_NULL) ? "NULL" : "REQ"))
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_enqueue_io] INFO: daemon_state = %u = %s\n",
			  daemon_state,
			  (daemon_state == FL_DST_NULL) ? "NULL" : "REQ"))


	// NOTE: Pending I/O
	//
	//  The strategy routine must increment the pending I/O
	//  request counter whenever it enqueues a new I/O request.
	//
	//  This counter ensures that the daemon will start reading
	//  pending I/O requests when:
	//
	//   1. Entering for the first time
	//   2. Entering on any subsequent time (after a fld crash,
	//                                       signal, ...)
	//
	//  The counter must always be set, regardless of the daemon
	//  state.


	// Increment pending I/O request counter
	daemon_pending++;


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy_enqueue_io] INFO: daemon_pending = %u\n",
			  daemon_pending))


	// Set return value
	*ret = 1;


	return 1;	// SUCCESS
}


int fl_strategy_dequeue_io(struct fld_io_req_s **io_req)	// -- OUT
{
	fl_list_node_t *io_node;


	// NOTE: Arguments


	// OUT: Check for I/O request
	if (io_req == NULL)
	{
		return 0;	// ERROR: No I/O request
	}


	// NOTE Processing


	// Dequeue I/O node
	if (fl_list_shift(io_list,
			  &io_node) < 1)
	{
		return -1;	// ERROR: I/O list is empty
	}


	// Get I/O request
	*io_req = (struct fld_io_req_s *) io_node->data;


	// Free I/O node
	fl_list_free_node(io_list,
			  &io_node);


	return 1;	// SUCCESS
}


int fl_strategy_requeue_io(void)	// -- IN
{
	fl_list_node_t *io_node;


	// NOTE: Processing


	// Dequeue I/O node
	if (fl_list_shift(io_list,
			  &io_node) < 1)
	{
		return 0;	// ERROR: I/O list is empty
	}


	// Requeue I/O node
	fl_list_push(io_list,
		     io_node);


	return 1;	// SUCCESS
}


int fl_strategy_peek_io(struct fld_io_req_s **io_req)	// -- OUT
{
	fl_list_node_t *io_node;


	// NOTE: Arguments


	// OUT: Check for I/O request
	if (io_req == NULL)
	{
		return 0;	// ERROR: No I/O request
	}


	// NOTE: Processing


	// Peek at first I/O node
	if (fl_list_get_node(io_list,
			     &io_node,
			     0) < 1)
	{
		return -1;	// ERROR: I/O list is empty
	}


	// Get I/O request
	*io_req = (struct fld_io_req_s *) io_node->data;


	return 1;	// SUCCESS
}


%FUNC:fl_hwgraph_add_edges:START%


int fl_hwgraph_add_edges(char **path_comp,		// -- OUT
			 vertex_hdl_t *vertex_comp,	// -- OUT
			 vertex_hdl_t *vertex_comp_new)	// -- OUT
{
        // NOTE: Add edges to /hw
        //
        //  Add edges in path_comp, a NULL terminated array of strings,
	//  to /hw sequentially.
        //
        //  All vertices from /hw to the last vertex created
        //  are stored in vertex_comp.
        //
        //  All newly created vertices are stored in vertex_comp_new.
        //
        //  On return, the hardware graph path is created, as specified
        //  in the path components array, path_comp.
	//
	//  If a path component already exists (i.e., a label and
	//  corresponding target vertex already were created), they are
	//  skipped and the path creation is continued.


        char *comp;
        char hw_path[PATH_MAX];
        unsigned int add_edge;
        unsigned int check_exists;
        unsigned int comp_index;
        vertex_hdl_t vertex_current;
        vertex_hdl_t vertex_target;
        vertex_hdl_t vertex_temp;


        // NOTE: Arguments


	// OUT: Check for path components
        if (path_comp == NULL)
        {
                return 0;       // ERROR: No path components
        }


	// OUT: Check for empty path components
        if (path_comp[0] == NULL)
        {
                return -1;      // ERROR: Path components are empty
        }


	// OUT: Check for vertex components
        if (vertex_comp == NULL)
        {
                return -2;      // ERROR: No vertex components
        }


	// OUT: Check for new vertex components
        if (vertex_comp_new == NULL)
        {
                return -3;      // ERROR: No new vertex components
        }


	// NOTE: Processing


        // Check for /hw component
        if (strcmp(path_comp[0],
		   "hw") != 0)
        {
                return -4;      // ERROR: No /hw component
        }


#if defined(FL_DEBUG)


        for (comp_index = 0;
	     path_comp[comp_index] != NULL;
	     comp_index++)
        {
                FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_add_edges] INFO: path_comp[%u] = %s\n",
				  comp_index,
				  path_comp[comp_index]))
        }


#endif	// defined(FL_DEBUG)


        // Initialize hardware path string
        hw_path[0] = '\0';


        add_edge = 0;
        check_exists = 1;
        vertex_current = hwgraph_root;
        for (comp_index = 0;
	     path_comp[comp_index] != NULL;
	     comp_index++)
        {
                // Convenience pointer
                comp = path_comp[comp_index];


                // Check for pathname overflow
                if ((strlen(hw_path) + strlen(comp)) > (PATH_MAX - 1))
                {
                        return -5;      // ERROR: Pathname overflow
                }


                // Add component to hardware path
                strcat(hw_path,
		       "/");
                strcat(hw_path,
		       comp);


                FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_add_edges] INFO: hw_path = %s\n",
				  hw_path))


                if (check_exists == 1)
                {
                        // NOTE: Check if /hw/.../comp exists


                        // Get vertex for hardware path (+1 ref on success)
                        vertex_temp = hwgraph_path_to_vertex(hw_path);


                        if (vertex_temp == GRAPH_VERTEX_NONE)
                        {
                                add_edge = 1;
                        }
                        else
                        {
                                vertex_current = vertex_temp;


                                // Update component vertices
                                vertex_comp[comp_index] = vertex_current;


				// Unref vertex
				if (hwgraph_vertex_unref(vertex_temp) != GRAPH_SUCCESS)
				{
                			FL_DEBUG_UPRINTF((\
							  "[fl_hwgraph_add_edges] ERROR: Couldn't unref vertex for \"%s\"\n",
							  hw_path))
				}
                        }
                }


                // Initialize new component vertices
                vertex_comp_new[comp_index] = GRAPH_VERTEX_NONE;


                // Check for new edge
                if (add_edge == 1)
                {
                        // Create target vertex (+1 ref on success)
                        if (hwgraph_vertex_create(&vertex_target) != GRAPH_SUCCESS)
                        {
                                return -6;      // ERROR: Couldn't create target vertex
                        }


                        FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_add_edges] INFO: Create edge = %s\n",
					  comp))


                        // Create edge
                        if (hwgraph_edge_add(vertex_current,
					     vertex_target,
					     comp) != GRAPH_SUCCESS)
                        {
                                return -7;      // ERROR: Couldn't add edge
                        }


                        // NOTE: Add new component


                        vertex_comp[comp_index] = vertex_target;
                        vertex_comp_new[comp_index] = vertex_target;


                        // Update current vertex
                        vertex_current = vertex_target;


                        // Skip subsequent exists checks
                        check_exists = 0;
                }
        }


#if defined(FL_DEBUG)


        for (comp_index = 0;
	     path_comp[comp_index] != NULL;
	     comp_index++)
        {
                FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_add_edges] INFO: path_comp[%u] = %s\n",
				  comp_index,
				  path_comp[comp_index]))


                FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_add_edges] INFO: vertex_comp[%u] = %u\n",
				  comp_index,
				  vertex_comp[comp_index]))


                FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_add_edges] INFO: vertex_comp_new[%u] = %u\n",
				  comp_index,
				  vertex_comp_new[comp_index]))
        }


#endif	// defined(FL_DEBUG)


        return 1;       // SUCCESS
}


%FUNC:fl_hwgraph_add_edges:STOP%
%FUNC:fl_hwgraph_new_dev:START%


int fl_hwgraph_new_dev(struct loop_entry_s *loop_entry,	// -- IN
		       fl_uint32_t dev_type,		// -- IN
		       char *hwgdev_name)		// -- OUT
{
	char *dev_name;
	char string_hw[3];
	char string_loop[5];
	char hw_path[PATH_MAX];
	char *path_comp[4];
	vertex_hdl_t vertex_dev;
	vertex_hdl_t vertex_char;
	vertex_hdl_t vertex_block;
	vertex_hdl_t vertex_comp[3];
	vertex_hdl_t vertex_comp_new[3];


	// NOTE: Arguments


	// IN: Check for loop entry
	if (loop_entry == NULL)
	{
		return 0;	// ERROR: No loop entry
	}


	// IN: Check for valid device type
	if ((dev_type != FL_HWGRAPH_CHR) &&
	    (dev_type != FL_HWGRAPH_BLK))
	{
		return -1;	// ERROR: Invalid device type
	}


	// OUT: Check for hardware graph device name
	if (hwgdev_name == NULL)
	{
		return -2;	// ERROR: No hardware graph device name
	}


	// NOTE: Processing


	// Get device name
	dev_name = loop_entry->dev.basename;


	// Check for device name
	if (dev_name == NULL)
	{
		return -3;	// ERROR: No device name
	}


	// Check for empty device name
	if (dev_name[0] == '\0')
	{
		return -4;	// ERROR: Empty device name
	}


	// Check device name length
	if (strlen(dev_name) > (PATH_MAX - 1 - 15))
	{
		return -5;	// ERROR: Device name overflow
	}


        // Copy "hw" component
        strcpy(string_hw,
	       "hw");
        string_hw[2] = '\0';


        // Copy "loop" component
        strcpy(string_loop,
	       "loop");
        string_loop[4] = '\0';


        // NOTE: Create path component array


	// Component "/hw"
        path_comp[0] = string_hw;
	vertex_comp[0] = NULL;
	vertex_comp_new[0] = NULL;


	// Component "/loop"
        path_comp[1] = string_loop;
	vertex_comp[1] = NULL;
	vertex_comp_new[1] = NULL;


	// Component "/{dev_name}"
        path_comp[2] = dev_name;
	vertex_comp[2] = NULL;
	vertex_comp_new[2] = NULL;


	// Terminating NULL component
        path_comp[3] = NULL;


	// Create hardware path
	if (fl_hwgraph_add_edges(path_comp,
                                 vertex_comp,
                                 vertex_comp_new) < 1)
        {
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_dev] ERROR: Couldn't create /hw/loop/dev_name path\n"))


		return -6;	// ERROR: Couldn't create /hw/loop/dev_name path
        }


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_dev] SUCCESS: Created /hw/loop/dev_name path\n"))


	// NOTE: Check device file
	//
	//  Check if device file already exists.


	// Create hardware path
	hw_path[0] = '\0';
	strcat(hw_path,
	       "/hw/loop/");
	strcat(hw_path,
	       dev_name);


	// Check for device type
	if (dev_type == FL_HWGRAPH_CHR)
	{
		strcat(hw_path,
		       "/char");
	}
	else
	{
		strcat(hw_path,
		       "/block");
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_dev] INFO: Hardware path to check \"%s\"\n",
			  hw_path))


	// Check for device file (+1 ref on success)
	vertex_dev = hwgraph_path_to_vertex(hw_path);


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_dev] INFO: hwgraph_path_to_vertex() = vertex_dev = %u\n",
			  vertex_dev))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_dev] |-hwgraph_vertex_refcnt(vertex_dev=%u) = %d\n",
			  vertex_dev,
			  hwgraph_vertex_refcnt(vertex_dev)))


	if (vertex_dev != GRAPH_VERTEX_NONE)
	{
		// NOTE: Device vertex already exists
		//
		//  The assumption is made that any device file under
		//  /hw/loop/dev_name is a previously created device file.
		//
		//  The reference counter is at least 1 from the creation.


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_dev] INFO: |-hwgraph_vertex_refcnt(vertex_dev=%u) = %d (already exists)\n",
				  vertex_dev,
				  hwgraph_vertex_refcnt(vertex_dev)))


		// Unref after hwgraph_path_to_vertex()
		hwgraph_vertex_unref(vertex_dev);


		if (dev_type == FL_HWGRAPH_CHR)
		{
			// Set CHR device vertex
			loop_entry->dev_chr.vertex = vertex_dev;
		}
		else
		{
			// Set BLK device vertex
			loop_entry->dev.vertex = vertex_dev;
		}


		// Set device info pointer to loop entry
		device_info_set(vertex_dev,
				loop_entry);


		// Copy hardware graph device name
		strcpy(hwgdev_name,
		       hw_path);


		return 2;	// SUCCESS: Device was created previously
	}


	// NOTE: Create device file


	if (dev_type == FL_HWGRAPH_CHR)
	{
		// Create /hw/loop/dev_name/char device (+1 ref on success)
		if (hwgraph_char_device_add(vertex_comp[2],
					    "char",
					    FL_DRIVER_PREFIX,
					    &vertex_char) != GRAPH_SUCCESS)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_new_dev] ERROR: Couldn't create char device\n"))


			return -8;	// ERROR: Couldn't create char device
		}


		// Set device vertex
		vertex_dev = vertex_char;


		// Set CHR device vertex
		loop_entry->dev_chr.vertex = vertex_dev;


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_dev] SUCCESS: Created /hw/loop/dev_name/char device\n"))
	}
	else
	{
		// Create /hw/loop/dev_name/block device (+1 ref on success)
		if (hwgraph_block_device_add(vertex_comp[2],
					     "block",
					     FL_DRIVER_PREFIX,
					     &vertex_block) != GRAPH_SUCCESS)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_new_dev] ERROR: Couldn't create block device\n"))


			return -9;	// ERROR: Couldn't create block device
		}


		// Set device vertex
		vertex_dev = vertex_block;


		// Set BLK device vertex
		loop_entry->dev.vertex = vertex_dev;


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_dev] SUCCESS: Created /hw/loop/dev_name/block device\n"))
	}


	// Set device info pointer to loop entry
	device_info_set(vertex_dev,
			loop_entry);


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_dev] |-hwgraph_vertex_refcnt(vertex_dev=%u) = %d\n",
			  vertex_dev,
			  hwgraph_vertex_refcnt(vertex_dev)))


	// Copy hardware graph device name
	strcpy(hwgdev_name,
	       hw_path);


	return 1;	// SUCCESS
}


%FUNC:fl_hwgraph_new_dev:STOP%
%FUNC:fl_hwgraph_new_devfl:START%


int fl_hwgraph_new_devfl(void)
{
	char *symlink_target;
	char string_hw[3];
	char string_loop[5];
	char string_devfl[8];
        char *path_comp[3];
	fl_int32_t ret_lookup;
	fl_uint32_t is_found;
	fl_uint32_t was_created;
	vertex_hdl_t vertex_devfl;
        vertex_hdl_t vertex_comp[2];
        vertex_hdl_t vertex_comp_new[2];
	struct vattr va;
	struct cred cred;


	// NOTE: Initialization


	// Control device file
	fl_file_init(&devfl_file);


	// Vnode attributes
	bzero((void *) &va,
	      sizeof(struct vattr));


	// NOTE: Credentials
        //
        //  Credentials are set to zero (root)
        //  to allow for creation of any file in
        //  /dev.


        // Set caller credentials
        bzero((void *) &cred,
	      sizeof(struct cred));


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] INFO: START\n"))


        // Copy "hw" component
        strcpy(string_hw, "hw");
        string_hw[2] = '\0';


        // Copy "loop" component
        strcpy(string_loop, "loop");
        string_loop[4] = '\0';


	// NOTE: Create path component array


	// Component "/hw" as "hw"
        path_comp[0] = string_hw;
	vertex_comp[0] = NULL;
	vertex_comp_new[0] = NULL;


	// Component "/loop" as "loop"
        path_comp[1] = string_loop;
	vertex_comp[1] = NULL;
	vertex_comp_new[1] = NULL;


	// Terminating NULL component
	path_comp[2] = NULL;


	// Add path components
	if (fl_hwgraph_add_edges(path_comp,
                                 vertex_comp,
                                 vertex_comp_new) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] ERROR: Couldn't add /hw/loop path components\n"))


		// Log warning
		cmn_err(CE_WARN, "^Couldn't add /hw/loop path components\n");


		return 0;	// ERROR: Couldn't add /hw/loop path components
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] SUCCESS: Added /hw/loop path components\n"))


	// NOTE: /dev/fl
	//
	//  File looper is controlled through /dev/fl, also called
	//  the control device file, which is a symbolic link to the
	//  character device file /hw/loop/fl:
	//
	//   /dev/fl -> /hw/loop/fl


	// Check for /hw/loop/fl
	vertex_devfl = hwgraph_path_to_vertex("/hw/loop/fl");
	if (vertex_devfl != GRAPH_VERTEX_NONE)
	{
		// NOTE: Loadable vs. registered
		//
		//  Once a DLKM is registered, the following code
		//  path is not necessary anymore.


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: /hw/loop/fl found at vertex = %u\n",
				  vertex_devfl))


		// Unref after hwgraph_path_to_vertex()
		hwgraph_vertex_unref(vertex_devfl);


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: |-hwgraph_vertex_refcnt(vertex_devfl=%u) = %d\n",
				  vertex_devfl,
				  hwgraph_vertex_refcnt(vertex_devfl)))


		// Remove /hw/loop/fl char device
		if (fl_hwgraph_remove_dev(vertex_comp[1],
				          "fl",
					  vertex_devfl,
				          FL_HWGRAPH_CHR) < 1)
		{ 
			// Log warning
			cmn_err(CE_WARN, "^Couldn't remove /hw/loop/fl char device\n");


			return -1;	// ERROR: Couldn't remove /hw/loop/fl char device
		}
	}


	// NOTE: Create /hw/loop/fl


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] INFO: Create /hw/loop/fl\n"))


	// Create /hw/loop/fl
	if (hwgraph_char_device_add(vertex_comp[1],
				    "fl",
				    FL_DRIVER_PREFIX,
				    &vertex_devfl) != GRAPH_SUCCESS)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] ERROR: Couldn't create /hw/loop/fl\n"))


		// Log warning
		cmn_err(CE_WARN, "^Couldn't create /hw/loop/fl\n");


		return -2;	// ERROR: Couldn't create /hw/loop/fl
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] INFO: Created /hw/loop/fl\n"))


	// Set NULL device info on /hw/loop/fl
	device_info_set(vertex_devfl,
			NULL);


	// NOTE: Added to fl_init() because in registered modules
	//       this function does not reach this point.
	// Set global vertex for control file
	//devfl_vertex = vertex_devfl;


	// NOTE: Check for symbolic link /dev/fl


        // Copy "/dev/fl" string
        strcpy(string_devfl, "/dev/fl");
        string_devfl[7] = '\0';


	// Set filename for control device
	devfl_file.name = string_devfl;


	// Set basename for control device
	devfl_file.basename = &devfl_file.name[5];


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] INFO: Lookup /dev/fl\n"))


	// Lookup symbolic link /dev/fl
	is_found = 0;
	ret_lookup = 0;
	//if (FL_TEST_COND((fl_test_ret == 1), ||)
	if (fl_file_lookup(&devfl_file,
			   NO_FOLLOW,
			   &is_found,
			   &ret_lookup) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: Lookup of /dev/fl failed\n"))


		// TODO: FL_TEST_F_main_N_NOT_LOOKUP_FILENAME(1, -3)


		// TODO: return % NOT_LOOKUP_FILENAME[in,,]%;	// ERROR: Couldn't lookup file name


		// Log warning
		cmn_err(CE_WARN, "^Couldn't lookup symbolic link /dev/fl\n");


		return -3;	// ERROR: Couldn't lookup symbolic link /dev/fl
	}


	// Check for symbolic link /dev/fl
	if (is_found != 1)
	{
		// NOTE: Symbolic link /dev/fl not found


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: Symbolic link /dev/fl not found\n"))


		// Build symbolic link target to BLK device file
		symlink_target = devfl_file.symlink_target;
		symlink_target[0] = '\0';
		strcat(symlink_target,
		       "/hw/loop/fl");


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: Create symbolic link from \"%s\" to \"%s\"\n",
				  devfl_file.name,
				  symlink_target))


		// Create symbolic link to BLK device file
		was_created = 0;
		if (fl_file_create(&devfl_file,
				   VLNK,
				   0,
				   S_IRUSR |
				   S_IRGRP,
				   &cred,
				   &was_created) < 1)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_new_devfl] ERROR: Couldn't create symbolic link from \"%s\" to \"%s\"\n",
					  devfl_file.name,
					  symlink_target))


			return -4;	// ERROR: Couldn't create symbolic link /dev/fl
		}


		FL_DEBUG_UPRINTF((\
				 "[fl_hwgraph_new_devfl] INFO: Symbolic link /dev/fl -> /hw/loop/fl created\n"))
	}
	else
	{
		// NOTE: Directory entry /dev/fl found


		// Check for symbolic link
		if (devfl_file.vp->v_type != VLNK)
		{
			// Log error
			cmn_err(CE_ALERT, "^File /dev/fl already exists and is not a symbolic link\n");


			return -5;	// ERROR: File /dev/fl already exists and is not a symbolic link
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_new_devfl] INFO: Symbolic link /dev/fl found\n"))
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_new_devfl] INFO: STOP\n"))


	return 1;	// SUCCESS
}


%FUNC:fl_hwgraph_new_devfl:STOP%
%FUNC:fl_hwgraph_remove_dev_by_path:START%


int fl_hwgraph_remove_dev_by_path(char *dev_path,	// -- IN
				  fl_uint32_t dev_type)	// -- IN
{
	char **path_comp;
	char path_buffer[PATH_MAX];
	fl_uint32_t path_length;
	fl_uint32_t comp_index;
	fl_uint32_t comp_num;
	vertex_hdl_t vertex_source;
	vertex_hdl_t vertex_dest;


	// NOTE: Arguments


	// IN: Check for device path
	if (dev_path == NULL)
	{
		return 0;	// ERROR: No device path
	}


	// IN: Check for empty device path
	if (dev_path[0] == '\0')
	{
		return -1;	// ERROR: Empty device path
	}


	// IN: Check for valid device type
	if ((dev_type != FL_HWGRAPH_CHR) &&
	    (dev_type != FL_HWGRAPH_BLK))
	{
		return -2;	// ERROR: Invalid device type
	}


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev_by_path] INFO: START\n"))


	// Split device pathname
	if (fl_file_split_path(dev_path,
			       &path_comp,
			       &comp_num) < 1)
	{
		return -3;	// ERROR: Couldn't split device path
	}


	// Check for minimum number of components
	if (comp_num < 2)
	{
		// Free split buffer
		fl_mem_free(path_comp[0]);


		// Free components array
		fl_mem_free(path_comp);


		return -4;	// ERROR: Not enough components
	}


	// Check for /hw root
	if ((path_comp[0][0] != 'h') ||
	    (path_comp[0][1] != 'w') ||
	    (path_comp[0][2] != '\0'))
	{
		// Free split buffer
		fl_mem_free(path_comp[0]);


		// Free components array
		fl_mem_free(path_comp);


		return -5;	// ERROR: Pathname root /hw not found
	}


	// Get pathname of source vertex
	path_length = 0;
	for (comp_index = 0;
	     comp_index < (comp_num - 1);
	     comp_index++)
	{
		// NOTE: '/'
		//
		//  Each component is preceded by a slash, of length
		//  1.


		path_buffer[path_length] = '/';


		// Update pathname length
		path_length++;


		// Copy component
		strcpy(&path_buffer[path_length],
		       path_comp[comp_index]);


		// Update pathname length
		path_length += strlen(path_comp[comp_index]);


		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_remove_dev_by_path] INFO: Component \"%s\"\n",
				  path_comp[comp_index]))
	}


	// Get source vertex
	vertex_source = hwgraph_path_to_vertex(path_buffer);
	if (vertex_source == GRAPH_VERTEX_NONE)
	{
		// Free split buffer
		fl_mem_free(path_comp[0]);


		// Free components array
		fl_mem_free(path_comp);


		return -6;	// ERROR: Couldn't get source vertex
	}


	// Get destination vertex
	vertex_dest = hwgraph_path_to_vertex(dev_path);
	if (vertex_dest == GRAPH_VERTEX_NONE)
	{
		// Free split buffer
		fl_mem_free(path_comp[0]);


		// Free components array
		fl_mem_free(path_comp);


		return -7;	// ERROR: Couldn't get destination vertex
	}


	// NOTE: References
	//
	//  If a device path exists, file looper already
	//  has a reference to it.
	//
	//  Unreference both vertices after hwgraph_path_to_vertex()


	// Unref source vertex after hwgraph_path_to_vertex()
	hwgraph_vertex_unref(vertex_source);


	// Unref destination vertex after hwgraph_path_to_vertex()
	hwgraph_vertex_unref(vertex_dest);


	if (dev_type == FL_HWGRAPH_CHR)
	{
		// Remove char device
		if (fl_hwgraph_remove_dev(vertex_source,
				          path_comp[comp_num - 1],
				          vertex_dest,
				          FL_HWGRAPH_CHR) < 1)
		{
			// Free split buffer
			fl_mem_free(path_comp[0]);


			// Free components array
			fl_mem_free(path_comp);


			return -8;	// ERROR: Couldn't remove char device
		}
	}
	else
	{
		// Remove block device
		if (fl_hwgraph_remove_dev(vertex_source,
				          path_comp[comp_num - 1],
				          vertex_dest,
				          FL_HWGRAPH_BLK) < 1)
		{
			// Free split buffer
			fl_mem_free(path_comp[0]);


			// Free components array
			fl_mem_free(path_comp);


			return -9;	// ERROR: Couldn't remove block device
		}
	}


	// Free split buffer
	fl_mem_free(path_comp[0]);


	// Free components array
	fl_mem_free(path_comp);


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev_by_path] INFO: STOP\n"))


	return 1;	// SUCCESS
}


%FUNC:fl_hwgraph_remove_dev_by_path:STOP%
%FUNC:fl_hwgraph_remove_dev:START%


int fl_hwgraph_remove_dev(vertex_hdl_t vertex_source,	// -- IN
			  char *edge_name,		// -- IN
			  vertex_hdl_t vertex_dest,	// -- IN
			  fl_uint32_t dev_type)		// -- IN
{
	fl_int32_t ret;


	// NOTE: Arguments


	// IN: Check for source vertex
	if (vertex_source == GRAPH_VERTEX_NONE)
	{
		return 0;	// ERROR: No source vertex
	}


	// IN: Check for edge name
	if (edge_name == NULL)
	{
		return -1;	// ERROR: No edge name
	}


	// IN: Check for destination vertex
	if (vertex_dest == GRAPH_VERTEX_NONE)
	{
		return -2;	// ERROR: No destination vertex
	}


	// IN: Check for valid device type
	if ((dev_type != FL_HWGRAPH_CHR) &&
	    (dev_type != FL_HWGRAPH_BLK))
	{
		return -3;	// ERROR: Invalid device type
	}


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: START\n"))


	// NOTE: Device reference count
	//
	//  Device files created by file looper have always at least
	//  4 references:
	//
	//   1. Origin
	//   2. Edge ("block" or "char")
	//   3. File looper reference (after *_{char,block}_device_add())


	// NOTE: /hw/loop
	//
	//  The assumption is made that any device file under
	//  /hw/loop/dev_name is a previously created device file
	//  that can be safely removed.


	// NOTE: Edge
	//
	//  Remove edge ("block" or "char"). Decrements reference
	//  count by 1.


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: hwgraph_edge_remove(vertex_source=%u, \"%s\") START\n",
			  vertex_source,
			  edge_name))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: |-hwgraph_vertex_refcnt(vertex_source=%u) = %d\n",
			  vertex_source,
			  hwgraph_vertex_refcnt(vertex_source)))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: |-hwgraph_vertex_refcnt(vertex_dest=%u) = %d\n",
			  vertex_dest,
			  hwgraph_vertex_refcnt(vertex_dest)))


	// Remove device edge
	if (hwgraph_edge_remove(vertex_source,
				edge_name,
				NULL) != GRAPH_SUCCESS)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_remove_dev] ERROR: Couldn't remove %s edge for source vertex %u\n",
				  edge_name,
				  vertex_source))


		return -4;	// ERROR: Couldn't remove edge for source vertex
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: hwgraph_edge_remove(vertex_source=%u, \"%s\") STOP\n",
			  vertex_source,
			  edge_name))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: |-hwgraph_vertex_refcnt(vertex_source=%u) = %d\n",
			  vertex_source,
			  hwgraph_vertex_refcnt(vertex_source)))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: |-hwgraph_vertex_refcnt(vertex_dest=%u) = %d\n",
			  vertex_dest,
			  hwgraph_vertex_refcnt(vertex_dest)))


	// NOTE: File looper reference
	//
	//  Unref file looper's reference on the device. Decrements reference
	//  count by 1.


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: hwgraph_vertex_unref(vertex_dest=%u) START\n",
			  vertex_dest))


	// Unref after hwgraph_{block,char}_device_add()
	hwgraph_vertex_unref(vertex_dest);


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: hwgraph_vertex_unref(vertex_dest=%u) STOP\n",
			  vertex_dest))


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: |-hwgraph_vertex_refcnt(vertex_dest=%u) = %d\n",
			  vertex_dest,
			  hwgraph_vertex_refcnt(vertex_dest)))


	// NOTE: Origin
	//
	//  The last reference to the device file is from the so called origin.
	//
	//  With a reference count of 1, hwgraph_vertex_destroy() can be called to
	//  free the device and its ressources. Also decrements reference count by 1.


	// Check for ref count
	if (hwgraph_vertex_refcnt(vertex_dest) == 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_remove_dev] INFO: hwgraph_vertex_destroy(vertex_dest=%u) START\n",
				  vertex_dest))


		ret = hwgraph_vertex_destroy(vertex_dest);
		if (ret != GRAPH_SUCCESS)
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_remove_dev] ERROR: Couldn't remove previous device vertex for source vertex %u ret = %u\n",
					  vertex_source,
					  ret))


			return -5;	// ERROR: Couldn't remove previous device vertex
		}
		else
		{
			FL_DEBUG_UPRINTF((\
					  "[fl_hwgraph_remove_dev] INFO: Previous device vertex removed = %u\n",
					  vertex_source))
		}
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_hwgraph_remove_dev] ERROR: Couldn't remove previous device vertex for source vertex %u. Refcount > 1 (%u)\n",
				  vertex_source,
				  hwgraph_vertex_refcnt(vertex_dest)))


		return -6;	// ERROR: Couldn't remove previous device vertex (reference count greater than 1)
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_hwgraph_remove_dev] INFO: STOP\n"))


	return 1;	// SUCCESS
}


%FUNC:fl_hwgraph_remove_dev:STOP%


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


// NOTE: IRIX Module de-/initialization
//
//  There are 4 module operations for non-registered modules:
//
//   1. Loading
//   2. Unloading
//   3. Registration
//   4. Unregistration
//
//  and 2 module operations for registered modules:
//
//   1. Accessing an unloaded module
//   2. Unloading a loaded module
//
//  The driver entries called in each operation are as follows:
//
//  1. ml ld -v -c ./test.o -p test_ -s 78
//
//  [test_init] init() entry point called
//  [test_start] start() entry point called
//  [test_reg] reg() entry point called
//
//  2. ml unld -v `ml | grep test_ | awk '{print $2}'`
//
//  [test_unreg] unreg() entry point called
//  [test_unload] unload() entry point called
//
//  3. ml reg -v -c ./test.o -p test_ -s 78
//
//  [test_init] init() entry point called
//  [test_start] start() entry point called
//  [test_reg] reg() entry point called
//  [test_unload] unload() entry point called
//
//  4. ml unreg -v `ml | grep test_ | awk '{print $2}'`
//
//  [test_init] init() entry point called
//  [test_start] start() entry point called
//  [test_unreg] unreg() entry point called
//  [test_unload] unload() entry point called
//
//
//  Registered module operations:
//
//  1. Accessing an unloaded module
//
//  [test_init] init() entry point called
//  [test_start] start() entry point called
//
//  2. Unloading a loaded module
//
//  [test_unload] unload() entry point called
//
//
//  Initialization
//  --------------
//
//  There are two cases that must be treated separately:
//
//   1. Non-registered modules
//   2. Registered modules
//
//  1. Non-registered modules
//
//  Initialization takes place after an explicit `ml ld` command
//  issued by a privileged user.
//
//  Any of init(), start() or reg() are eligible for this task.
//
//  2. Registered modules
//
//  Initialization takes place after an implicit load by accessing
//  (i.e. opening) a corresponding device special file for the
//  (registered) module for the first time.
//
//  Any of init() or start() are eligible for this task.
//
//
//  Deinitialization
//  ----------------
//
//  Same as for Initialization:
//
//  1. Non-registered modules
//
//  Deinitialization takes place after an explicit `ml unld` command
//  issued by a privileged user.
//
//  Any of unreg() or unload() are eligible for this task.
//
//  2. Registered modules
//
//  Deinitialization takes place after an explicit `ml unld` command
//  issued by a privileged user.
//
//  Only unload() is eligible for this task.


int fl_serial_lock(void)
{
	//fl_uint3264_t dbgpid;


	// Get pid of current process
	//drv_getparm(PPID,
		      //&dbgpid);


	//FL_DEBUG_UPRINTF((\
			  //"[fl_open] MUTEX_LOCK() (pid=%llu)\n", dbgpid))


	// Acquire serialization lock
	//MUTEX_LOCK(fl_mutex,
		     //-1);


	//FL_DEBUG_UPRINTF((\
			  //"[fl_open] MUTEX_LOCKED() (pid=%llu)\n", dbgpid))


	return 1;	// SUCCESS
}


int fl_serial_unlock(void)
{
	//fl_uint3264_t dbgpid;


	// Get pid of current process
	//drv_getparm(PPID,
		      //&dbgpid);


	//FL_DEBUG_UPRINTF((\
			  //"[fl_open] MUTEX_UNLOCK() (pid=%llu)\n", dbgpid))


	// Release serialization lock
	//MUTEX_UNLOCK(fl_mutex);


	//FL_DEBUG_UPRINTF((\
			  //"[fl_open] MUTEX_UNLOCKED() (pid=%llu)\n", dbgpid))


	return 1;	// SUCCESS
}


int fl_init()
{
	// NOTE: Initialization


	// /dev/loop string
	devloop_string[0] = '\0';
	strcat(devloop_string,
	       "/dev/loop");


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: cpuboard() = %d\n",
			  cpuboard()))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: fl_file_write() = %p\n",
			  fl_file_write))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: scache_linemask = %u\n",
			  scache_linemask))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: EFS superblock = %u\n",
			  sizeof(struct efs)))


#ifdef FL_CHECKS


	//FL_CHECK_SCS_OPEN_LOG()


        // Set global cpu variable
        cpu = cpuboard();


#if defined(FL_TESTS) && defined(FL_TESTS_F_LOAD_MOD_N_CHECK_CPU)


	// Check cpu
	if (1)


#elif defined(FL_TESTS_CHECKS)


	// Check cpu
	if (0)


#else


	// Check cpu
	if (0)

#endif


	{
		// NOTE: Check lock
		//
		//  In IRIX, kernel module initialization does not allow
		//  interruption of the kernel loading and initialization process.
		//
		//  To prevent access in case of a failed CPU check, a global check
		//  lock is used. This lock is checked on entrance into each entry point.


		// Lock driver entry points
		fl_check_lock_switch = 1;


		// Set driver initialization switch
		fl_driver_init = 0;


		// Set check error message
		bzero((void *) fl_check_err_msg, 100);
		sprintf(fl_check_err_msg, "fl: License check failed for %d (ctype)", cpu);


		// Log error
		cmn_err(CE_WARN, "%s\n", fl_check_err_msg);


#ifdef FL_TESTS


		// NOTE: Failed CPU check
		//
		//  Return value and error message are set for test.pl.
		//
		//  There is no faster way to check for a failed CPU check than
		//  accessing the kernel module. Unlike other OSs, in IRIX
		//  the kernel module loading mechanism does not allow to
		//  be interrupted, and therefore always succeeds.


		// Set test return value
		fl_test_ret = EACCES;


		// Set test error message
		strcpy(fl_test_err, "ERROR: License check failed for cpu");


#endif


		return EACCES;  // ERROR: License check failed for cpu
	}


#endif	// FL_CHECKS


	// NOTE: File looper initialization
	//
	//  Initialize libraries and global variables.


	// Set initial maxdmasz
	fl_maxdmasz = maxdmasz;


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: maxdmasz = %u \n",
			  maxdmasz))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: v.v_maxdmasz = %u \n",
			  v.v_maxdmasz))


#if _PAGESZ == 4096


	// Set page shift
	page_shift = 12;


#else


	// Set page shift
	page_shift = 14;


#endif	// _PAGESZ


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: page_shift = %u\n",
			  page_shift))


	// Calculate maximum transfer size
	FL_MAX_TRANSFER_SIZE = (fl_maxdmasz - 1) << page_shift;


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: FL_MAX_TRANSFER_SIZE = %u\n",
			  FL_MAX_TRANSFER_SIZE))


	// Initialize list library
	fl_list_init(NULL);


	// Initialize mem library
	if (fl_mem_init() < 1)
	{
		// Deinitialize list library
		fl_list_deinit();


		return EINVAL;  // ERROR: Couldn't create mem list
	}


/*
#ifdef FL_CHECKS


			// NOTE: IRIX tree check
			//
			//  No tree check on IRIX.


                        // Create check list
                        if (fl_list_new(&check_list) < 1)
                        {
                                // Free mem list
                                fl_list_free(&mem_list);


                                // Terminate lock for mem list
                                simple_lock_terminate(&mem_list_lock);


                                // Deinitialize list library
                                fl_list_deinit();


                                return EINVAL;  // ERROR: Couldn't create check list
                        }


                        // Initialize tree check
                        FL_CHECK_TREE_INIT(system_bus)


                        // Check tree
                        FL_CHECK_TREE_EXEC()


#if defined(FL_TESTS) && defined(FL_TESTS_F_FL_CONFIGURE_N_TREE)


                        // Check for successful tree check
                        if (0)


#elif defined(FL_TESTS_CHECKS)


                        // Check for successful tree check
                        if (1)


#else


                        // Check for successful tree check
                        if (FL_CHECK_TREE_SUCCESS())


#endif


                        {
                        }
                        else
                        {
                                // Deinitialize tree check
                                FL_CHECK_TREE_DEINIT()


                                // Free check list
                                fl_list_free(&check_list);


                                // Free mem list
                                fl_list_free(&mem_list);


                                // Terminate lock for mem list
                                simple_lock_terminate(&mem_list_lock);


                                // Deinitialize list library
                                fl_list_deinit();


                                // Log error
                                log(LOG_ERR, "fl: License check failed (ttype)\n");


                                return EACCES;  // ERROR: License check failed for tree
                        }


                        // Deinitialize tree check
                        FL_CHECK_TREE_DEINIT()


                        // Free check list
                        fl_list_free(&check_list);

#endif
*/


	// Initialize proc library
	if (fl_proc_init() < 1)
	{
		// Deinitialize mem library
		fl_mem_deinit();


		// Deinitialize list library
		fl_list_deinit();


		return EINVAL;  // ERROR: Couldn't create proc list
	}


	// Create loop list
	fl_list_new(&loop_list);


	// Create I/O list
	fl_list_new(&io_list);


	// Create semaphore for ctrl
	fl_sem_init(&ctrl_sem, 1);


	// Set device bitfield for /dev/fl
	fl_dev_set_lminor(0, 1);


	// Set device bitfield for /dev/testfl
	fl_dev_set_lminor(1, 1);


	// Create lock for loop list
       	fl_spin_new(&loop_list_lock);


	// Create lock for fl_open
       	fl_spin_new(&fl_open_lock);


	// Create lock for daemon
	fl_spin_new(&daemon_lock);


	// Allocate serialization mutex
	fl_mutex = MUTEX_ALLOC(MUTEX_DEFAULT,
			       KM_SLEEP,
			       NULL);


	// Allocate I/O list mutex
	io_mutex = MUTEX_ALLOC(MUTEX_DEFAULT,
			       KM_SLEEP,
			       NULL);


#ifdef FL_TESTS


	// Create lock for sleepers
	fl_spin_new(&sleeper_lock);


	// Create lock for sleeep
	fl_spin_new(&sleeep_lock);


	// Semaphore for sleepers
	initnsema(&sleeper_sem,
                  0,
                  "sleeper_sem");


#endif


	// NOTE: Initialize daemon variables


	daemon_pid = 0;
	daemon_state = FL_DST_NULL;
	daemon_ref = NULL;
	daemon_pending = 0;


	// NOTE: Platform specific initializations


	// Set major number
	majnum = fl_majnum;


        // Set driver initialization switch
        fl_driver_init = 1;


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: majnum = %d\n",
			  majnum))


/* NOTE: Only for non-registered modules


	// NOTE: /dev/loop
	//
	//  File looper creates CHR and BLK device files for each
	//  file attached. These files are located in the hardware graph
	//  under /hw/loop, and accessed through symbolic links from
	//  /dev/loop.


	// Create /dev/loop
	fl_dev_create_loop_dir();

	
	// NOTE: /dev/fl
	//
	//  File looper uses a control device, /dev/fl, which is a
	//  symbolic link to /hw/loop/fl.


	// Create control device /dev/fl
	fl_hwgraph_new_devfl();
*/


	// Set vertex for control device
	devfl_vertex = hwgraph_path_to_vertex("/hw/loop/fl");


	FL_DEBUG_UPRINTF((\
                          "[fl_init] INFO: devfl_vertex = %u\n",
                          devfl_vertex))


	FL_DEBUG_UPRINTF((\
			  "[fl_init] INFO: STOP\n"))


	return ESUCCESS;	// SUCCESS
}


int fl_start()
{
	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_start] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_start] INFO: STOP\n"))


	return ESUCCESS;	// SUCCESS
}


int fl_reg()
{
	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_reg] INFO: START\n"))


	// NOTE: /dev/loop
	//
	//  File looper creates CHR and BLK device files for each
	//  file attached. These files are located in the hardware graph
	//  under /hw/loop, and accessed through symbolic links from
	//  /dev/loop.


	// Create /dev/loop
	fl_dev_create_loop_dir();

	
	// NOTE: /dev/fl
	//
	//  File looper uses a control device, /dev/fl, which is a
	//  symbolic link to /hw/loop/fl.


	// Create control device /dev/fl
	fl_hwgraph_new_devfl();


	FL_DEBUG_UPRINTF((\
			  "[fl_reg] INFO: STOP\n"))


	return ESUCCESS;	// SUCCESS
}


int fl_unreg()
{
	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_unreg] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_unreg] INFO: STOP\n"))


	return ESUCCESS;	// SUCCESS
}


int fl_unload()
{
	fl_uint64_t loop_index;
	fl_uint64_t loop_length;
	fl_list_node_t *loop_node;
	fl_uint32_t dev_is_removed;
	struct loop_entry_s *loop_entry;


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_unload] INFO: START\n"))


	// NOTE: Deinitialization
        //
        //  Deinitializations take place only if the kernel module has been
        //  successfully initialized in fl_init().


        if (fl_driver_init != 1)
        {
		return ESUCCESS;        // SUCCESS: Driver is not initialized
        }


	FL_DEBUG_UPRINTF((\
			  "[fl_unload] INFO: Deinitialization\n"))


	FL_DEBUG_LOCK_SWITCH(ESUCCESS)


#ifdef FL_CHECKS


	//FL_CHECK_SCS_CLOSE_LOG()


#endif


	// NOTE: Detach loop entries


	// Get length of loop list
	fl_list_length(loop_list,
		       &loop_length);


	FL_DEBUG_UPRINTF((\
			  "[fl_unload] INFO: Number of loop entries = %llu\n",
			  loop_length))


	// Check number of loop entries
	if (loop_length > 0)
	{
		// Clean active loop entries
		for (loop_index = 0;
		     loop_index < loop_length;
		     loop_index++)
		{
			// Get next node
			fl_list_shift(loop_list,
				      &loop_node);


			// Get loop entry
			loop_entry = (struct loop_entry_s *) loop_node->data;


			// Release loop entry
			fl_entry_rel(loop_entry,
				     &dev_is_removed);


			// Free loop entry
			fl_mem_free(loop_entry);


			// Free loop node
			fl_list_free_node(loop_list,
					  &loop_node);


			FL_DEBUG_UPRINTF((\
					  "[fl_unload] INFO: Cleaned loop entry %d\n",
					  loop_index))
		}
	}


	// NOTE: Daemon process
	//
	//  Check for the case that this module is
	//  unloaded with the only user process entering
	//  being the daemon process, in which case a proc
	//  and fl_proc would still be in the global proc
	//  list (fl_proc.c).
	//
	//  The daemon process does not clean the global proc
	//  list on last close (see fl_close()). This is done
	//  for efficiency.


	// Clean global proc list
	fl_close_empty_proc((pid_t) 0);


	// Free loop list
	fl_list_free(&loop_list);


	// Free I/O list
	fl_list_free(&io_list);


	// Free semaphore for ctrl
	fl_sem_deinit(&ctrl_sem);


	// Deinitialize proc library
	fl_proc_deinit();


	// Free lock for loop list
	fl_spin_free(loop_list_lock);


	// Free lock for fl_open
	fl_spin_free(fl_open_lock);


	// Free lock for daemon
	fl_spin_free(daemon_lock);


#ifdef FL_TESTS


	// Terminate lock for sleepers
	fl_spin_free(sleeper_lock);
	fl_spin_free(sleeep_lock);


#endif


	// Deinitialize mem library
	fl_mem_deinit();


	// NOTE: Liblist
	//
	//  Deinitialization of libmem must be last because all other libraries
	//  rely on it.


	// Deinitialize list library
	fl_list_deinit();


	// Deallocate serialization mutex
	MUTEX_DEALLOC(fl_mutex);


	// Deallocate I/O mutex
	MUTEX_DEALLOC(io_mutex);


	// NOTE: Control device
	//
	//  Deinitialize control device file /dev/fl.


	// Deinitialize control device
	fl_file_deinit(&devfl_file);


	// NOTE: Device directory
	//
	//  Deinitialize device directory /dev/loop.


	// Deinitialize device directory
	fl_file_deinit(&devloop_file);


	// Last check of malloc_count
	FL_DEBUG_UPRINTF((\
			  "[fl_unload] INFO: malloc_count = %d\n",
			  malloc_count))


	FL_DEBUG_UPRINTF((\
			  "[fl_unload] INFO: STOP\n"))


	return ESUCCESS;	// SUCCESS
}


#               endif
#       endif
#endif


/******************************************************************************
 *
 * Name:
 *   fl_open
 *
 * Function:
 *   Makes the device ready to accept requests from the user.
 *
 * Formal parameters:
 *   devp - A dev_t for the target device
 *   flag - Read, write, or read-write flags
 *   fmt  - Constant that indicates block or character mode
 *
 * Modified parameters:
 *   None
 *
 * Implicit input:
 *   None
 *
 * Implicit output:
 *   None
 *
 * Return value:
 *   ESUCCESS - Successful completion
 *   ENODEV   - The minor number does not map to a device
 *
 * Caller:
 *   Kernel code through the switch tables
 *
 * Side effects:
 *   None
 *
 ******************************************************************************/

#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_open(dev_t *devp,	// -- IN
            int oflag,		// -- IN
            int otyp,		// -- IN
            cred_t *crp)	// -- IN


#               endif
#       endif
#endif


{
	//buf_t *bp;
	pid_t pid;
	fl_uint32_t is_locked;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;
	//struct fld_io_req_s *io_req;
	struct fl_args_s args;


#ifdef FL_DEBUG


	fl_uint64_t proc_length;


#endif


	FL_TEST_TYPE_DECL


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	dev_t dev;
	dev_t open_dev;
	fl_int32_t ret;
	fl_uint32_t minor;
	fl_uint32_t major;
	vertex_hdl_t dev_vertex;
	struct loop_entry_s *loop_entry;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	// NOTE: Self-assignments
	//
	//  SGI MIPSPRO C/C++ compilers issue remarks when
	//  variables and/or parameters are declared but not used.
	//
	//  Self-assignments are used as a simple workaround,
	//  since some variables and/or parameters are declared
	//  by API requirement.


	oflag = oflag;
	crp = crp;
	is_locked = is_locked;


	// NOTE: Initialization


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	FL_TEST_TYPE_INIT


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: oflag = 0x%x\n",
			  oflag))


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: otyp = %u\n",
			  otyp))


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// NOTE: IRIX device numbers
	//
	//  In IRIX, device numbers have two different meanings depending on the driver.
	//
	//  If the driver chooses to use the hardware graph, the dev parameter to fl_open()
	//  is a pointer to one of its vertices (i.e. a vertex handle).
	//
	//  If the driver chooses to use traditional UNIX device numbers, the dev parameter
	//  is a pointer to a dev_t type containing the actual device number.
	//
	//  File looper uses the hardware graph for the control device file and device
	//  files created for attached files.


	// NOTE: Hardware graph device numbers
	//
	//  File looper uses internally a device bitfield to
	//  identify each virtual device (each comprising both a BLK
	//  and a CHR device file).
	//
	//  When the hardware graph is being used, the minor number
	//  for the devices can't be chosen by the driver, nor is
	//  their numbering scheme described in the documentation.
	//
	//  For portability and ease of driver logic, each device in
	//  a hardware graph enabled driver, uses a mapping between
	//  vertices and bitfield indices (represented by the minor
	//  number).
	//
	//  The translation is done at the beginning of each entry point.


	// NOTE: /dev/fl
	//
	//  The control device file, /dev/fl, is the only
	//  vertex created by file looper that has a NULL
	//  pointer device info associated to it.


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


        // Set device vertex
        dev_vertex = (vertex_hdl_t) *devp;


	FL_TEST_F_fl_open_N_NOT_MAP_DEVICE(FL_TEST_STATE_BEFORE,\
					   1,\
					   args)


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       otyp,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_TEST_F_fl_open_N_NOT_MAP_DEVICE(FL_TEST_STATE_ERROR,\
						   -1,\
						   args)


		FL_DEBUG_UPRINTF((\
				  "[fl_open] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  otyp == OTYP_CHR ? "CHR" : "BLK",
				  dbgpid))


		FL_DEBUG_UPRINTF((\
				  "[fl_open] ERROR: return ENOEXIST\n"))


		return ENOEXIST;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  otyp == OTYP_CHR ? "CHR" : "BLK",
			  dbgpid))


	// Set open device
	open_dev = dev_vertex;


#               endif
#       endif
#endif


#if defined(FL_TESTS) || defined(FL_CHECKS_SCS_SUM)


	// Test for /dev/testfl (minor = 111)
	if (minor == 111)
	{
		return ESUCCESS;
	}


#endif


	// NOTE: Reset switch check


	FL_TEST_F_fl_open_N_HAS_RESET_SWITCH_ON(FL_TEST_STATE_BEFORE,\
					        1,\
					        FL_TEST_NO_ARGS)


	if (reset_switch == 1)
	{
		// NOTE: Semaphore value overflow
		//
		//  In case of an overflow of the value of the control
		//  semaphore, deny further access and ask user to
		//  reload file looper to reset back its internal state.


		FL_TEST_F_fl_open_N_HAS_RESET_SWITCH_ON(FL_TEST_STATE_BEFORE,\
							-1,\
							FL_TEST_NO_ARGS)


		// Log warning
		cmn_err(CE_WARN, "^Overflow detected for internal semaphore. Please reload file looper's kernel module and try again.\n");


		return EOVERFLOW;	// ERROR: Semaphore value overflow
	}


	FL_DEBUG_LOCK_SWITCH(EBUSY)
	FL_CHECK_LOCK_SWITCH(EACCES)


        FL_TEST_F_main_N_NOT_SUSER_OPEN_DEVFL(0, 0)
        FL_TEST_F_main_N_NO_PERMISSION_CPU_CHECK(0, 0)
        FL_TEST_F_main_N_NOT_OPEN_DEVFL(0, 0)


	// Acquire serialization lock
	FL_SERIAL_LOCK();


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Check for /dev/fl
	if ((otyp == OTYP_CHR) &&
	    (minor == 0))


#               endif
#       endif
#endif


	{
		FL_DEBUG_UPRINTF((\
			  	  "[fl_open] INFO: Daemon process (pid=%llu)\n",
				  dbgpid))


		// NOTE: Daemon process
		//
		//  The daemon process is handled separately from
		//  all other processes, control and I/O processes.


		// Get pid of current process
		fl_proc_get_data(FL_DATA_PID,
				 &pid);

		FL_DEBUG_UPRINTF((\
			  	  "[fl_open] INFO: Daemon lock (pid=%llu)\n",
				  dbgpid))


		// Acquire lock for daemon
		fl_spin_lock(daemon_lock);


		// Check for daemon process
		if (pid == daemon_pid)
		{
			// Release lock for daemon
			fl_spin_unlock(daemon_lock);


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			FL_DEBUG_UPRINTF((\
					  "[fl_open] SUCCESS: Daemon process (pid=%llu)\n",
					  dbgpid))


			return ESUCCESS;	// SUCCESS: Daemon process
		}


		FL_DEBUG_UPRINTF((\
			  	  "[fl_open] INFO: Daemon unlock (pid=%llu)\n",
				  dbgpid))


		// Release lock for daemon
		fl_spin_unlock(daemon_lock);


		// NOTE: root
		//
		//  Only root may access /dev/fl (or an equivalent DSF).
		//
		//  This check can safely be done outside the fl_open lock.


		FL_TEST_F_fl_open_N_NOT_SUPER_USER(FL_TEST_STATE_BEFORE,\
						   1,\
						   args)


		// Check for super user
		//if (FL_TEST_COND((fl_test_ret == 1), ||)
	        if (fl_user_check(0,
				  &args) < 1)
		{
			FL_TEST_F_fl_open_N_NOT_SUPER_USER(FL_TEST_STATE_ERROR,\
							   -1,\
							   args)


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EUSERS;	// ERROR: User must be root
		}


		FL_TEST_F_fl_open_N_OPEN_CONTENTION(FL_TEST_STATE_BEFORE,\
						    0,\
						    FL_TEST_NO_ARGS)


		// NOTE: Start /dev/fl session


		FL_TEST_F_fl_open_N_INTERRUPT_CTRL_SEM(FL_TEST_STATE_BEFORE,\
						       1,\
						       &args)


		// Decrement semaphore for ctrl
		ret = 0;
		args.ret.var = &ret;
		args.ret.reset = 1;
		//ret = fl_sem_decr(ctrl_sem,
		fl_sem_decr(ctrl_sem,
			    1,
			    FL_SEM_SWITCH_ENABLE_INTERRUPT,
			    &args);
		if (ret != 1)
		{
			// NOTE: Interrupted by user


			FL_TEST_F_fl_open_N_INTERRUPT_CTRL_SEM(FL_TEST_STATE_ERROR,\
							       -1,\
							       &args)


			FL_DEBUG_UPRINTF((\
					  "[fl_open] INFO: INTERRUPTED (pid=%llu)\n",
					  dbgpid))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINTR;	// ERROR: Execution interrupted by user
		}


		FL_TEST_F_fl_open_N_OPEN_CONTENTION(FL_TEST_STATE_AFTER,\
						    0,\
						    FL_TEST_NO_ARGS)


		FL_DEBUG_UPRINTF((\
				  "[fl_open] INFO: fl_sem_decr(ctrl_sem) = %d (pid=%llu)\n",
				  ret,
				  dbgpid))


               	//FL_DEBUG_UPRINTF((\
				    //"[fl_open] INFO: Open from /dev/fl after fl_sem_decr()\n"))
               	//FL_DEBUG_UPRINTF((\
				    //"[fl_open] INFO: fl_test_func = %s\n", fl_test_func))
               	//FL_DEBUG_UPRINTF((\
				    //"[fl_open] INFO: fl_test_name = %s\n", fl_test_name))
               	//FL_DEBUG_UPRINTF((\
				    //"[fl_open] INFO: fl_test_type = %d\n", fl_test_type))
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: fl_spin_trylock() (pid=%llu)\n",
			  dbgpid))


	// NOTE: IRIX open lock
	//
	//  In IRIX, access to any device driver entry point is synchronized
	//  by an external semaphore.
	//
	//  The open lock is not used.


	// Acquire lock for fl_open
	is_locked = 0;
	fl_spin_trylock(fl_open_lock,
			&is_locked);


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: fl_spin_trylock().is_locked = %u (pid=%llu)\n",
			  is_locked,
			  dbgpid))


	// NOTE: CHR and BLK devices
	//
	//  This entry point can be called from both types of device files.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Check for CHR device
	if ((otyp == OTYP_CHR) &&
	    (major == 0))


#               endif
#       endif
#endif


	{
		// NOTE: Character device request


		FL_DEBUG_UPRINTF((\
			  	  "[fl_open] INFO: OTYP_CHR && major == 0 (pid=%llu)\n",
				  dbgpid))


		if (minor == 0)
		{
			// NOTE: Request comes from /dev/fl
			//
			//  /dev/fl is the file looper control dsf and is used
			//  to transfer control information between the fl user command
			//  and the fl kernel module.
			//
			//  The name /dev/fl is not mandatory, but rather a
			//  recommended naming convention. Any name may be used as
			//  long as its minor number is 0.


#ifdef FL_CHECKS


#if defined(FL_TESTS) && defined(FL_TESTS_F_FL_OPEN_N_CHECK_CPU)


			// Check cpu
			if (1)


#elif defined(FL_TESTS_CHECKS)


			// Check cpu
			if (0)


#else


                	// Check cpu
			if (0)


#endif


                	{


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


				// Initialize check error message
                                bzero((void *) fl_check_err_msg,
				      100);


                                // Set check error message
                                sprintf(fl_check_err_msg,
					"fl: License check failed for %d (ctype)",
					cpu);


                                // Log error
                                cmn_err(CE_WARN,
					"%s\n",
					fl_check_err_msg);


#               endif
#       endif
#endif


				//FL_TEST_F_fl_open_N_RESET_SWITCH_ON_INCR(FL_TEST_STATE_BEFORE,\
								         //1,\
								         //&args)


				// Increment semaphore for ctrl
				if (fl_sem_incr(ctrl_sem,
						1,
						FL_SEM_SWITCH_RETURN_ON_SLEEP,
						&args) < 1)
				{
					// NOTE: Tripwire for semaphore overflow
					//
					//  In case of an externally induced semaphore value
					//  overflow, deny further access to the driver and
					//  ask user to reload module.
					//
					//  A reload will reset file looper's internal state,
					//  effectively avoiding deadlocks, kernel panics, and
					//  other unexpected behavior.


					reset_switch = 1;


					//FL_TEST_F_fl_open_N_RESET_SWITCH_ON_INCR(FL_TEST_STATE_ERROR,\
										 //-1,\
										 //&args)
				}


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// NOTE: Memory dump
				//
				//  Show a list of all current allocations.


				//fl_mem_dump();


#ifdef FL_TESTS


				// NOTE: Failed CPU check
				//
				//  Return value and error message are set for test.pl.
				//
				//  There is no faster way to check for a failed CPU check than
				//  accessing the kernel module. Unlike other OSs, in IRIX
				//  the kernel module loading mechanism does not allow to
				//  be interrupted, and therefore always succeeds.


				// Set test return value
				fl_test_ret = EACCES;


				// Set test error message
				strcpy(fl_test_err,
				       "ERROR: License check failed for cpu");


#endif


				// Release serialization lock
				FL_SERIAL_UNLOCK();


                        	return EACCES;	// ERROR: License check failed for cpu
                	}


#endif // FL_CHECKS


			// NOTE: Process
			//
			//  Keep a reference to every process that opens /dev/fl
			//  so it can be recognised in subsequent calls to other
			//  entry points.
			//
			//  This allows the following I/O discipline to be
			//  implemented:
			//
			//   Only one process at a time may hold the dsf_control
			//   structure to issue a single command.
			//
			//   A command is issued by invoking, in order, the following
			//   entry points:
			//
			//    fl_write() - to send the command request
			//    fl_read()  - to read back the command output
			//
			//
			//  A process is described by a struct proc_s, defined in
			//  fl_proc.h, using a private (for the fl kernel module),
			//  global (accessible from anywhere in the fl kernel module)
			//  list, stored in fl_proc.c.
			//
			//  Each struct proc_s contains a pointer to private data
			//  that is used in this case to store another structure,
			//  struct fl_proc_s, containing the control bits about the
			//  process state from file looper's view.


			// Create new fl_proc
        		fl_mem_alloc((void **) &fl_proc,
                                     sizeof(struct fl_proc_s),
                                     FL_MEM_ZERO |
				     FL_MEM_WAITOK,
                                     "fl_open",
                                     "Allocate fl proc");


        		// Create new proc
        	   	fl_proc_new(&proc);


			FL_DEBUG_UPRINTF((\
					  "[fl_open] INFO: proc = %p proc->pid = %u (pid=%llu)\n",
					  proc,
					  proc->pid,
					  dbgpid))


			// Populate custom fields
			fl_proc->state = FL_ST_OPEN;
			fl_proc->reset = FL_PROC_NULL;
			fl_proc->ctrl = &dsf_control;


			// Link fl_proc to proc
			proc->priv = fl_proc;


			// Acquire lock for proc list
			fl_proc_lock();


			// Add new proc
			fl_proc_push_unlocked(proc);


#ifdef FL_DEBUG


			fl_proc_get_length_unlocked(&proc_length);


			FL_UPRINTF(("[fl_open] INFO: Length of proc_list = %llu (pid=%llu)\n",
				    proc_length,
				    dbgpid))


#endif


			// Release lock for proc list
			fl_proc_unlock();


			FL_DEBUG_UPRINTF((\
					  "[fl_open] INFO: Open request from /dev/fl (pid=%llu)\n",
					  dbgpid))


			FL_TEST_F_fl_open_N_SIMPLE_OPEN_DEVFL(FL_TEST_STATE_BEFORE,\
							      0,\
							      FL_TEST_NO_ARGS)
			FL_TEST("fl_open",\
				"SIMPLE_OPEN_DEVFL")


			FL_TEST_F_custom_N_PINGPONG(FL_TEST_STATE_BEFORE,\
						    0,\
						    FL_TEST_NO_ARGS)
			FL_TEST("custom",\
				"PINGPONG")
		}
		else
		{
			// NOTE: Request from /dev/loop/r*


			// NOTE: File modes
			//
			//  File looper supports only FREAD and FWRITE
			//  file modes for all its devices.


			/*
			// Check for valid open flags
			if ((oflag & ~(FREAD | FWRITE)) != 0)
			{
				FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_CHR(FL_TEST_STATE_ERROR,\
								           -1,\
								           FL_TEST_NO_ARGS)


				FL_DEBUG_UPRINTF((\
						  "[fl_open] INFO: Invalid flags for CHR device 0x%x (pid=%llu)\n",
						  oflag,
						  dbgpid))


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EINVAL;	// ERROR: Invalid open flags for CHR device
			}
			*/


			// Create new CHR pid
			ret = fl_open_new_pid(open_dev,
					      FL_PID_CHR);


			FL_DEBUG_UPRINTF((\
					  "[fl_open] INFO: fl_open_new_pid ret = %d\n",
					  ret))


			switch (ret)
			{
				case -1:

					FL_DEBUG_UPRINTF((\
							  "[fl_open] INFO: Device (CHR) not found (pid=%llu)\n",
							  dbgpid))


					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY(FL_TEST_STATE_ERROR_CHR,\
										 -1,
										 FL_TEST_NO_ARGS)


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					FL_TEST("fl_open",\
						"NOT_FOUND_LOOP_ENTRY")
					return ENODEV;	// ERROR: Loop entry not found for CHR device

#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


				case -2:

					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					FL_TEST_F_fl_open_N_IS_IN_EXCLUSIVE_USE(FL_TEST_STATE_ERROR_CHR,\
										-1,
										FL_TEST_NO_ARGS)


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return EBUSY;	// ERROR: CHR device is being used in exclusive access mode


#               endif
#       endif
#endif


			}


			FL_DEBUG_UPRINTF((\
					  "[fl_open] INFO: Device (CHR) is attached (pid=%llu)\n",
					  dbgpid))
		}
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        else if ((otyp == OTYP_BLK) &&
		 (major == 0))


#               endif
#       endif
#endif


	{
		// NOTE: Request from /dev/loop/[^r]*


		// NOTE: File modes
		//
		//  File looper supports only FREAD and FWRITE
		//  file modes for all its devices.


		// Check for valid open flags
		if ((oflag & ~(FREAD | FWRITE)) != 0)
		{
			FL_TEST_F_fl_open_N_INVALID_OPEN_FLAGS_BLK(FL_TEST_STATE_ERROR,\
								   -1,\
								   FL_TEST_NO_ARGS)


			// Release lock for fl_open
			fl_spin_unlock(fl_open_lock);


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;	// ERROR: Invalid open flags for BLK device
		}


		// Create new BLK pid
		ret = fl_open_new_pid(open_dev,
				      FL_PID_BLK);
		switch (ret)
		{
			case -1:

				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				FL_DEBUG_UPRINTF((\
						  "[fl_open] INFO: Device (BLK) not found (pid=%llu)\n",
						  dbgpid))


				FL_TEST_F_fl_open_N_NOT_FOUND_LOOP_ENTRY(FL_TEST_STATE_ERROR_BLK,\
									 -1,
									 FL_TEST_NO_ARGS)


				FL_TEST("fl_open",\
					"NOT_FOUND_LOOP_ENTRY")
				return ENODEV;  // ERROR: Loop entry not found for BLK device


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


			case -2:

				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				FL_TEST_F_fl_open_N_IS_IN_EXCLUSIVE_USE(FL_TEST_STATE_ERROR_BLK,\
									-1,
									FL_TEST_NO_ARGS)


				return EBUSY;	// ERROR: BLK device is being used in exclusive access mode


#               endif
#       endif
#endif


		}


		FL_DEBUG_UPRINTF((\
				  "[fl_open] INFO: Device (BLK) is attached (pid=%llu)\n",
				  dbgpid))
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_open] ERROR: UNKNOWN DEVICE major = %u minor = %u type = %s (pid=%llu)\n",
				  major,
				  minor,
				  otyp == OTYP_BLK ? "OTYP_BLK" : otyp == OTYP_CHR ? "OTYP_CHR" : "OTYP_UNKNOWN",
				  dbgpid))
	}


	// Release lock for fl_open
	fl_spin_unlock(fl_open_lock);


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	FL_DEBUG_UPRINTF((\
			  "[fl_open] INFO: STOP (pid=%llu)\n",
			  dbgpid))


	return ESUCCESS;	// SUCCESS
}


/******************************************************************************
 *
 * Name:
 *   fl_close
 *
 * Formal parameters:
 *   dev   - A dev_t for the target device
 *   flag  - Read, write, or read-write flags
 *   fmt   - Constant that indicates character or block mode
 *
 * Modified parameters:
 *   None
 *
 * Implicit input:
 *   None
 *
 * Implicit output:
 *   None
 *
 * Return values:
 *   ESUCCESS - Successful completion
 *   ENODEV   - Device number does not map to a physical device
 *
 * Caller:
 *   Kernel code through the switch tables
 *
 * Side effects:
 *   None
 *
 ******************************************************************************/


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_close(dev_t dev,		// -- IN
             int flag,		// -- IN
             int otyp,		// -- IN
             cred_t *crp)	// -- IN


#               endif
#       endif
#endif


{
	dev_t close_dev;
	pid_t pid;
	struct fl_args_s args;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	fl_uint32_t minor;
	fl_uint32_t major;
	vertex_hdl_t dev_vertex;
	struct loop_entry_s *loop_entry;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	// Self-assignments
	flag = flag;
	crp = crp;


	// NOTE: Initialization


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Set device vertex
	dev_vertex = (vertex_hdl_t) dev;


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       OTYP_CHR,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_close] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  "CHR",
				  dbgpid))


		return ENOEXIST;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_close] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  otyp == OTYP_CHR ? "CHR" : "BLK",
			  dbgpid))


	// Set close device
	close_dev = dev_vertex;


#               endif
#       endif
#endif


        FL_DEBUG_UPRINTF((\
			  "[fl_close] INFO: START (pid=%llu)\n",
			  dbgpid))


	FL_DEBUG_LOCK_SWITCH(EBUSY)


#if defined(FL_TESTS) || defined(FL_CHECKS_SCS_SUM)


	// Test for /dev/testfl (minor = 111)
	if (minor == 111)
	{
		return ESUCCESS;
	}


#endif


	// Acquire serialization lock
	FL_SERIAL_LOCK();


	// NOTE: CHR and BLK devices
	//
	//  This entry point can be called from both types of device files.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Switch through dsfs
        if (otyp == OTYP_CHR)


#               endif
#       endif
#endif


        {
		// NOTE: Character device request


		if (minor == 0)
		{
			// Get pid of current process
			fl_proc_get_data(FL_DATA_PID,
					 &pid);


			// Acquire lock for daemon
			fl_spin_lock(daemon_lock);


			// Check for daemon process
			if (pid == daemon_pid)
			{
				// Release lock for daemon
				fl_spin_unlock(daemon_lock);


				FL_DEBUG_UPRINTF((\
						  "[fl_close] SUCCESS: Daemon process (pid=%llu)\n",
						  dbgpid))


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return ESUCCESS;	// SUCCESS: Daemon process
			}


			// Release lock for daemon
			fl_spin_unlock(daemon_lock);


			// Empty global proc list
			fl_close_empty_proc(pid);
		}
		else
		{
			// NOTE: Last close
			//
			//  Most UNIX systems call the close() entry point only on the
			//  last close.


			fl_close_empty_pids(close_dev,
					    FL_PID_CHR);
		}
	}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


        else if ((otyp == OTYP_BLK) &&
		 (major == 0))


#               endif
#       endif
#endif


        {
		// NOTE: Last close
		//
		//  Most UNIX systems call the close() entry point only on the
		//  last close.


		fl_close_empty_pids(close_dev,
				    FL_PID_BLK);
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_close] INFO: UNKNOWN DEVICE: major = %u minor = %u type = %s pid = %llu\n",
				  major,
				  minor,
				  "CHR",
				  dbgpid))
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_close] INFO: STOP (pid=%llu)\n",
			  dbgpid))


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return ESUCCESS;	// SUCCESS
}


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_read(dev_t dev,		// -- IN
            uio_t *uio,		// -- IN
            cred_t *crp)	// -- IN


#               endif
#       endif
#endif


{
	fl_int32_t ret;
	fl_int32_t fl_read_ret;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;
	struct fl_args_s args;


	FL_TEST_TYPE_DECL


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	fl_uint32_t major;
	fl_uint32_t minor;
	struct loop_entry_s *loop_entry;
	vertex_hdl_t dev_vertex;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	// Self-assignments
	dev = dev;
	crp = crp;


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_read] INFO: START\n"))


	FL_TEST_F_main_N_NO_PERMISSION_PACKET_SIZE(0, 0)
        FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_SIZE_DEVFL(0, 0)
	FL_TEST_F_main_N_NO_PERMISSION_PACKET_DATA(0, 0)
        FL_TEST_F_main_N_NOT_READ_RESPONSE_PACKET_DEVFL(0, 0)


	FL_DEBUG_LOCK_SWITCH(EBUSY)
	FL_CHECK_LOCK_SWITCH(EACCES)


	// Acquire serialization lock
	FL_SERIAL_LOCK();


	FL_DEBUG_UPRINTF((\
			  "[fl_read] INFO: uio->uio_resid = %u uio->uio_offset = %llu\n",
			  uio->uio_resid,
			  uio->uio_offset))


	FL_DEBUG_UPRINTF((\
			  "[fl_read] INFO: uio->uio_iov[0].iov_base = 0x%p uio->uio_iov[0].iov_len = %llu\n",
			  uio->uio_iov[0].iov_base,
			  uio->uio_iov[0].iov_len))


        // NOTE: NULL pointer exception (NPE)
        //
        //  In IRIX, it is allowed to pass a NULL pointer to a driver's read()
	//  entry point.


        // Check for NPE
        if (uio->uio_iov[0].iov_base == NULL)
        {
		// Release serialization lock
		FL_SERIAL_UNLOCK();


                return EFAULT;  // ERROR: Bad address
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Set device vertex
        dev_vertex = (vertex_hdl_t) dev;


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       OTYP_CHR,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_read] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  "CHR",
				  dbgpid))


		return ENOEXIST;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_read] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  "CHR",
			  dbgpid))


#               endif
#       endif
#endif


	// Check for /dev/fl
	if (minor == 0)
	{
		// NOTE: Read from /dev/fl


		// Get proc
		if (fl_proc_get(&proc) < 1)
		{
			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;       // ERROR: Couldn't get proc
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_read] INFO: proc = %p proc->pid = %u\n",
				  proc,
				  proc->pid))


		// Get fl proc
		fl_proc = (struct fl_proc_s *) proc->priv;


		// Check for reset
		if (fl_proc->reset == FL_PROC_RESET)
		{
			FL_TEST_F_custom_N_PINGPONG(FL_TEST_STATE_BEFORE,\
						    0,\
						    FL_TEST_NO_ARGS)


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			FL_TEST("custom",\
				"PINGPONG")
			return EINVAL;  // ERROR: Process is reset
		}


		// Reset session
		fl_session_reset(fl_proc);


		// TODO: FL_TEST_F_fl_write_N_WRONG_SESSION_STATE(0, 2)


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		FL_TEST("fl_read",\
			"NOT_READ_FROM_DEVFL")
		return EINVAL;	// ERROR: Can't read from /dev/fl
	}
	else
	{
		// NOTE: Attached CHR device files
		//
		//  CHR device special files that correspond to attached BLK
		//  devices allow raw access to the BLK device.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		ret = uiophysio((int (*)(struct buf *)) fl_strategy,
				NULL,
			  	loop_entry->dev.vertex,
				B_READ,
				uio);


		FL_DEBUG_UPRINTF((\
				  "[fl_read] INFO: uiophysio() = %d\n",
				  ret))


		switch (ret)
		{
			case ENOSPC:

				return ENOSPC;

			case EFAULT:

				return EFAULT;

			case EAGAIN:

				return EAGAIN;

			default:

				// NOTE: Success


				break;
		}


		// Acquire serialization lock
		FL_SERIAL_LOCK();


		// Set return value
		fl_read_ret = ESUCCESS;


#		endif
#	endif
#endif


	}


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return fl_read_ret;	// SUCCESS
}


%FUNC:fl_write:START%


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


int fl_write(dev_t dev,		// -- IN
             uio_t *uio,	// -- IN
             cred_t *crp)	// -- IN


#               endif
#       endif
#endif


{
	fl_int32_t ret;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;
	struct fl_args_s args;


	FL_TEST_TYPE_DECL
	//FL_CHECK_SCS_VARS(fl_write, 188, 89, 191, 174)


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	fl_uint32_t major;
	fl_uint32_t minor;
	vertex_hdl_t dev_vertex;
	struct loop_entry_s *loop_entry;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	// Self-assignments
	dev = dev;
	crp = crp;


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_write] INFO: START\n"))


	FL_TEST_F_main_N_NO_PERMISSION_WRITE_REQUEST_PACKET(0, 0)
	FL_TEST_F_main_N_NOT_WRITE_REQUEST_PACKET_DEVFL(0, 0)


	FL_DEBUG_LOCK_SWITCH(EBUSY)
	FL_CHECK_LOCK_SWITCH(EACCES)


        FL_DEBUG_UPRINTF((\
			  "[fl_write] INFO: uio->uio_resid = %u uio->uio_offset = %llu\n",
			  uio->uio_resid,
			  uio->uio_offset))


        FL_DEBUG_UPRINTF((\
			  "[fl_write] INFO: uio->uio_iov[0].iov_base = 0x%p uio->uio_iov[0].iov_len = %llu\n",
			  uio->uio_iov[0].iov_base,
			  uio->uio_iov[0].iov_len))


	// Acquire serialization lock
	FL_SERIAL_LOCK();


        // NOTE: NULL pointer exception (NPE)
        //
        //  In IRIX, it is allowed to pass a NULL pointer to a driver's write() entry point.
	//
        //  Check this before continuing.


        // Check for NPE
        if (uio->uio_iov[0].iov_base == NULL)
        {
		// Release serialization lock
		FL_SERIAL_UNLOCK();


                return EFAULT;  // ERROR: Bad address
        }


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#	if FL_OS_VERSION_V_6_5
#		if FL_OS_RELEASE_V_30


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Set device vertex
        dev_vertex = (vertex_hdl_t) dev;


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       OTYP_CHR,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_write] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  "CHR",
				  dbgpid))


		return ENOEXIST;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_write] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  "CHR",
			  dbgpid))


#               endif
#       endif
#endif


	// Check for /dev/fl
	if (minor == 0)
	{
		// NOTE: Write to /dev/fl


		// Get proc
		if (fl_proc_get(&proc) < 1)
		{
			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;       // ERROR: Couldn't get proc
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_write] INFO: proc = %p proc->pid = %u\n",
				  proc,
				  proc->pid))


		// Get fl proc
		fl_proc = (struct fl_proc_s *) proc->priv;


		// Check for reset
		if (fl_proc->reset == FL_PROC_RESET)
		{
			FL_TEST_F_custom_N_PINGPONG(FL_TEST_STATE_BEFORE,\
						    0,\
						    FL_TEST_NO_ARGS)


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			FL_TEST("custom",\
				"PINGPONG")
			return EINVAL;  // ERROR: Process is reset
		}


		// Reset session
		fl_session_reset(fl_proc);


		// TODO: FL_TEST_F_fl_write_N_WRONG_SESSION_STATE(0, 2)


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		FL_TEST("fl_write",\
			"NOT_WRITE_TO_DEVFL")
		return EINVAL;	// ERROR: Can't write to /dev/fl


/* TODO: Include SCS into ioctl()
#ifdef FL_CHECKS


		// NOTE: SCS check


		FL_CHECK_SCS_EXEC(fl_write, 2444)


#if defined(FL_TESTS) && defined(FL_TESTS_F_FL_WRITE_N_SCS)


		// Check SCS
		if (0)


#elif defined(FL_TESTS_CHECKS)


		// Check SCS
		if (1)


#else


		// Check SCS
		if (FL_CHECK_SCS_SUCCESS(fl_write))


#endif


		{
		}
		else
		{
			if (fl_check_sum != 1)
			{
				// NOTE: Clean up session


				// Reset read bytes
				uio->uio_resid = packet_size;


				// Reset session
				fl_session_reset(fl_proc);


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


				// Log error
				cmn_err(CE_WARN,
					"fl: License check failed for %02x-%02x-%02x-%02x (%u-%u-%u-%u) (stype)\n",
					scs_sum_fl_write[0],
				        scs_sum_fl_write[1],
				        scs_sum_fl_write[2],
					scs_sum_fl_write[3],
					scs_sum_fl_write[0],
					scs_sum_fl_write[1],
					scs_sum_fl_write[2],
					scs_sum_fl_write[3]);


#               endif
#       endif
#endif


				//fl_mem_dump();


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EACCES;
			}
		}


#endif	// FL_CHECKS
*/


	}
	else
	{
		// NOTE: Attached CHR device files
		//
		//  CHR device special files that correspond to attached BLK
		//  devices, allow raw access to the BLK device.


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		ret = uiophysio((int (*)(struct buf *)) fl_strategy,
				NULL,
			  	loop_entry->dev.vertex,
				B_WRITE,
				uio);


		FL_DEBUG_UPRINTF((\
				  "[fl_write] INFO: uiophysio() = %d\n",
				  ret))


		switch (ret)
		{
			case ENOSPC:

				return ENOSPC;

			case EFAULT:

				return EFAULT;

			case EAGAIN:

				return EAGAIN;

			default:

				// NOTE: Success


				break;
		}


		// Acquire serialization lock
		FL_SERIAL_LOCK();


#		endif
#	endif
#endif


	}


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return ESUCCESS;	// SUCCESS
}


%FUNC:fl_write:STOP%


/******************************************************************************
 *
 * Name:
 *   fl_size
 *   fl_size64
 *
 * Function:
 *   Obtains the size of the partition.
 *
 * Formal parameters:
 *   dev - A dev_t for the target device
 *
 * Modified parameters:
 *   None
 *
 * Implicit input:
 *   None
 *
 * Implicit output:
 *   None
 *
 * Return values:
 *   Partition size, in blks, for the device on successful completion
 *   -1 - Device number does not map to a physical device or a valid
 *        partition map does not exist for this device
 * Caller:
 *   The kernel through the switch tables.
 *
 * Side effects:
 *   None
 *
 ******************************************************************************/

int fl_size(dev_t dev)	// -- IN
{
	return (int) fl_dev_get_size_in_blocks(dev);
}


long long fl_size64(dev_t dev)	// -- IN
{
	return fl_dev_get_size_in_blocks(dev);
}


/******************************************************************************
 *
 * Name:
 *   fl_ioctl
 *
 * Function:
 *   Handles all nonread and nonwrite I/O requests.
 *
 * Formal parameters:
 *   dev  - A dev_t for the target device
 *   cmd  - The ioctl command
 *   data - Address of a user data buffer
 *   flag - Flags from the file handle
 *
 * Modified parameters:
 *   Some ioctl commands pass data back to the caller through the
 *   data buffer.
 *
 * Implicit input:
 *   None
 *
 * Implicit output:
 *   None
 *
 * Return value:
 *   ENODEV  - Minor number does not map to a physical device
 *   EBADF   - The partition was not open
 *   EINVAL  - Disk label is invalid
 *   EBADF   - Command needs an open context, buf can't find one
 *   EACCESS - User doesn't have the necessary priviledges
 *
 * Caller:
 *   Kernel code through the switch tables
 *
 * Side effects:
 *   None
 *
 ******************************************************************************/

#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


%FUNC:fl_ioctl:START%


int fl_ioctl(dev_t dev,		// -- IN
             int cmd,		// -- IN
             void *arg,		// -- IN
             int mode,		// -- IN
             cred_t *crp,	// -- IN
             int *rvalp)	// -- OUT
{
	buf_t *bp;
	pid_t pid;
	char default_drive_type[28] = {

		"FUJITSU\0"\
		"MAJ9999M"\
		"C\0\0\0\0\0\0\0"\
		"HP00"

	};
	fl_int32_t ret;
	fl_int32_t ret_attach;
	fl_int32_t return_error;
	fl_uint32_t minor;
	fl_uint32_t major;
	fl_uint32_t format_has_error;
	fl_uint32_t dev_size_uint;
	//fl_uint32_t length;
	//fl_uint32_t group;
	//fl_uint32_t cmdlen_err;
	//fl_uint32_t scsi_error;
	fl_uint64_t dev_size_blocks;
	fl_uint64_t pid_length;
	fl_uint64_t dev_size;
	fl_uint64_t io_size;
	fl_uint64_t io_offset;
	fl_uint64_t io_data_size;
	fl_list_t *pid_list;
	fl_list_node_t *pid_node;
	struct proc_s *proc;
	struct fl_proc_s *fl_proc;
	struct loop_entry_s *loop_entry;
	struct dk_ioctl_data defects_req;
	struct defect_list defect_list_resp;
	struct fld_io_req_s *uio_req;
	struct fld_io_req_s *io_req;
	struct fld_io_req_s io_resp;
	struct fld_io_start_args_s io_start_args;
	struct volume_header *vh;
	struct fl_args_s args;
	vertex_hdl_t dev_vertex;


	FL_TEST_TYPE_DECL


#ifdef FL_CHECKS_SCS_SUM


	fl_uint32_t check_sum_value;


#endif


#if defined(FL_TESTS) || defined(FL_CHECKS_SCS_SUM)


	char *string_int;
	fl_uint32_t int_digits;
	struct test_req_s test_req;


#endif


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


	// Self-assignments
	bp = bp;
	dev = dev;
	arg = arg;
	mode = mode;
	crp = crp;
	rvalp = rvalp;


#ifdef FL_TESTS


	fl_test_ret = fl_test_ret;


#endif


	///////////////////////////////// ////
	//                         ///  //  //
	//  NOTE: Initialization   ////   ///
	//                         ///// //
	/////////////////////////////////


	FL_TEST_TYPE_INIT


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NOTE: Processing


	FL_DEBUG_UPRINTF((\
			  "[fl_ioctl] INFO: START\n"))


	FL_DEBUG_UPRINTF((\
			  "[fl_ioctl] INFO: cmd = 0x%x type = CHR\n",
			  cmd))


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Set device vertex
        dev_vertex = (vertex_hdl_t) dev;


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       OTYP_CHR,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  "CHR",
				  dbgpid))


		return ENOEXIST;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_ioctl] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  "CHR",
			  dbgpid))


	// Acquire serialization lock
	FL_SERIAL_LOCK();


#ifdef FL_TESTS


	// NOTE: Test commands
	//
	//  The following commands are used in test compiles.


       	if (cmd == 0x11111111)
       	{
               	// NOTE: Test request
               	//
               	//  Test requests are sent by test.pl to request execution of a specific test.
               	//
               	//  A test request struct is defined to pass in the test function and name to be
               	//  executed. A pointer to this struct is passed in arg.


               	FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: Test request (cmd = %x)\n",
				  cmd))


               	// Initialize test request struct
               	bzero((void *) &test_req,
		      sizeof(struct test_req_s));


               	// Get test request from user space
               	if (copyin(arg,
                           (void *) &test_req,
                           sizeof(struct test_req_s)) != 0)
               	{
			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] ERROR: copyin()\n"))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EFAULT;  // ERROR: Couldn't copy in test request
		}


		// Check for test variable
		if (strcmp("test_notify",
			   test_req.func) == 0)
		{
			// NOTE: test_notify
			//
			// fl_int32_t fl_test_notify;


			// NOTE: Convert string to integer


			string_int = test_req.name;


			// Check for sign
			if (string_int[0] == '-')
			{
				string_int++;
			}


			// Convert to unsigned long long (no overflow checking)
			fl_test_notify = strtoull(string_int,
						  NULL,
						  10);


			// Set sign
			if (string_int != test_req.name)
			{
				fl_test_notify = -fl_test_notify;
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: test_notify = %d\n",
					  fl_test_notify))
		}
		else
		{
			// NOTE: Test parameters


              		// Set test parameters
                	strncpy(fl_test_func,
				test_req.func,
				300);
               		strncpy(fl_test_name,
				test_req.name,
				300);


               		FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: Test function = \"%s\"\n",
					  fl_test_func))


               		FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: Test name = \"%s\"\n",
					  fl_test_name))
		}


		// Release serialization lock
		FL_SERIAL_UNLOCK();


               	return ESUCCESS;        // SUCCESS: Test parameters set
       	}
	else if (cmd == 0x22222222)
	{
               	// NOTE: Test query
               	//
               	//  Test queries are sent by test.pl to request the value of a specific test variable.


               	// Initialize test request struct
               	bzero((void *) &test_req,
		      sizeof(struct test_req_s));


               	// Get test request from user space
               	if (copyin(arg,
                           (void *) &test_req,
                           sizeof(struct test_req_s)) != 0)
               	{
                       	FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] ERROR: copyin()\n"))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


                       	return EFAULT;  // ERROR: Couldn't copy in test request
               	}


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: Test variable = \"%s\"\n",
				  test_req.func))


		// NOTE: Set result in name field buffer
		//
		//  There are two types of results, depending on the type of the
		//  test variable:
		//
		//   1. Int
		//   2. String


		// Check test variable type
		if (strcmp("test_err",
			   test_req.func) == 0)
		{
			// NOTE: String type


			// Set result buffer
			sprintf(test_req.name,
				"%u %s",
				strlen(fl_test_err),
				fl_test_err);
		}
		else if (strcmp("test_suc",
				test_req.func) == 0)
		{
			// NOTE: String type


			// Set result buffer
			sprintf(test_req.name,
				"%u %s",
				strlen(fl_test_suc),
				fl_test_suc);
		}
		else if (strcmp("test_ret", test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(1,
                                          (void *) &fl_test_ret,
                                          32,
                                          &int_digits);


			// Account for sign
			if (fl_test_ret < 0)
			{
				int_digits++;
			}


			sprintf(test_req.name,
				"%u %d",
				int_digits,
				fl_test_ret);
		}
		else if (strcmp("test_notify",
				test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(1,
                                          (void *) &fl_test_notify,
                                          32,
                                          &int_digits);


			// Account for sign
			if (fl_test_notify < 0)
			{
				int_digits++;
			}


			sprintf(test_req.name,
				"%u %d",
				int_digits,
				fl_test_notify);
		}
		else if (strcmp("test_sleeper",
				test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(1,
                                          (void *) &fl_test_sleeper,
                                          32,
                                          &int_digits);


			// Account for sign
			if (fl_test_sleeper < 0)
			{
				int_digits++;
			}


			sprintf(test_req.name,
				"%u %d",
				int_digits,
				fl_test_sleeper);
		}
		else if (strcmp("test_counter",
				test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(0,
                                          (void *) &fl_test_counter,
                                          64,
                                          &int_digits);

			sprintf(test_req.name,
				"%u %llu",
				int_digits,
				fl_test_counter);
		}
		else if (strcmp("test_ctrl_sem",
				test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(0,
                                          (void *) &fl_test_ctrl_sem,
                                          64,
                                          &int_digits);

			sprintf(test_req.name,
				"%u %llu",
				int_digits,
				fl_test_ctrl_sem);
		}
		else if (strcmp("test_malloc_count",
				test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			int_digits = 0;
			fl_int_get_digits(1,
                                          (void *) &malloc_count,
                                          64,
                                          &int_digits);


			// Account for sign
			if (malloc_count < 0)
			{
				int_digits++;
			}


			sprintf(test_req.name,
				"%u %lld",
				int_digits,
				malloc_count);
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: test_req.name = %s\n",
				  test_req.name))


		// Send value of test variable
		if (copyout((void *) test_req.name,
                            (void *) arg,
                            300) != 0)
		{
			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EFAULT;	// ERROR: Couldn't send value of test variable
		}


		// Release serialization lock
		FL_SERIAL_UNLOCK();


               	return ESUCCESS;        // SUCCESS: Test parameter queried
	}
	else if (cmd == 0x33333333)
	{
               	// NOTE: Kernel module operation
               	//
               	//  Module operations are sent by test.pl to start specific tests.


               	// Initialize test request struct
               	bzero((void *) &test_req,
		      sizeof(struct test_req_s));


               	// Get test request from user space
               	if (copyin(arg,
                           (void *) &test_req,
                           sizeof(struct test_req_s)) != 0)
               	{
                       	FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] ERROR: copyin()\n"))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


                       	return EFAULT;  // ERROR: Couldn't copy in test request
               	}


		// NOTE: Process module operation codes


		if (strcmp("101",
			   test_req.func) == 0)
		{
			// NOTE: Wake up one sleeper


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: Wake up one sleeper\n"))


                        // Wake up
			//thread_wakeup_one((unsigned long) &fl_test_wait);
			//vsema(&sleeper_sem);
			//wakeup(&fl_test_wait);
			fl_spin_unlock(sleeep_lock);
		}


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: Test variable = \"%s\"\n",
				  test_req.func))


		// Release serialization lock
		FL_SERIAL_UNLOCK();


               	return ESUCCESS;        // SUCCESS: Test parameter queried
	}


	// NOTE: Unknown command
	//
	//  A command was requested in test mode that is not implemented.
	//
	//  The return statement comes after the next macro block.


#endif


#ifdef FL_CHECKS_SCS_SUM


	// NOTE: Check commands
	//
	//  The following commands are used in check compiles.


	if (cmd == 0x44444444)
	{
		// NOTE: SCS check set
		//
		//  Set SCS check variables.


               	FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: SCS check set (cmd = %x)\n",
				  cmd))


               	// Initialize check request struct
               	bzero((void *) &test_req,
		      sizeof(struct test_req_s));


               	// Get test request from user space
               	if (copyin(arg,
                           (void *) &test_req,
                           sizeof(struct test_req_s)) != 0)
               	{
			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] ERROR: copyin()\n"))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EFAULT;  // ERROR: Couldn't copy in test request
		}


		// Check for SCS check variable
		if (strcmp("check_sum",
			   test_req.func) == 0)
		{
			// NOTE: Convert string to integer
			//
			// *.func: Check variable name
			// *.name: Variable value


			// Set variable value
			string_int = test_req.name;


			// Check for sign
			if (string_int[0] == '-')
			{
				string_int++;
			}


			// Convert to unsigned long long (no overflow checking)
			check_sum_value = strtoull(string_int,
						   NULL,
						   10);


			// Set sign
			if (string_int != test_req.name)
			{
				check_sum_value = -check_sum_value;
			}


			// Check for valid values
			if ((check_sum_value != 0) &&
			    (check_sum_value != 1))
			{
				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EINVAL;	// ERROR: Invalid value for variable
			}


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] check_sum_value = %d\n",
					  check_sum_value))
		}
		else
		{
			// NOTE: Unknown SCS check variable


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;	// ERROR: Unknown SCS check variable
		}


		// Release serialization lock
		FL_SERIAL_UNLOCK();


               	return ESUCCESS;        // SUCCESS: Test parameter queried
	}
	else if (cmd == 0x55555555)
	{
		// NOTE: SCS check query
		//
		//  Set SCS check variables.


               	FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] SCS check query (cmd = %x)\n",
				  cmd))


               	// Initialize check request struct
               	bzero((void *) &test_req,
		      sizeof(struct test_req_s));


               	// Get test request from user space
               	if (copyin(arg,
                           (void *) &test_req,
                           sizeof(struct test_req_s)) != 0)
               	{
			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] ERROR: copyin()\n"))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EFAULT;  // ERROR: Couldn't copy in test request
		}


		// Check for SCS check variable
		if (strcmp("check_scs_sum_fl_write",
			   test_req.func) == 0)
		{
			// NOTE: Int type


			// Set result buffer
			fl_int_get_digits(0,
                                          (void *) &fl_check_scs_sum_fl_write,
                                          32,
                                          &int_digits);


			sprintf(test_req.name,
				"%u %u",
				int_digits,
				fl_check_scs_sum_fl_write);


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: test_req.name = %s\n",
					  test_req.name))


			FL_UPRINTF(("[fl_ioctl] INFO: check_scs_sum_fl_write = %s\n",
				    test_req.name))


			// Send value of test variable
			if (copyout((void *) test_req.name,
                                    (void *) arg,
                                    300) != 0)
			{
				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't send value of test variable
			}
		}
		else
		{
			// NOTE: Unknown SCS check variable


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;	// ERROR: Unknown SCS check variable
		}


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		return ESUCCESS;        // SUCCESS: Test parameter queried
	}


	// NOTE: Unknown command
	//
	//  A command was requested in SCS check mode that is not implemented.


	//return EINVAL;	// ERROR: Unknown command


#endif


	FL_CHECK_LOCK_SWITCH(EACCES)


	// NOTE: Only CHR devices
	//
	//  In IRIX, only CHR devices are allowed to receive ioctl() calls.


	// Check minor number
	if (minor == 0)
	{
		// NOTE: From /dev/fl


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: fl_proc_get_data()\n"))


		// Get pid of current process
		fl_proc_get_data(FL_DATA_PID,
				 &pid);


		FL_DEBUG_UPRINTF((\
				  "[fl_ioctl] INFO: pid = %u\n",
				  pid))


		// Acquire lock for daemon
		fl_spin_lock(daemon_lock);


		// Check for daemon process
		if (pid == daemon_pid)
		{
			// NOTE: Daemon process (/dev/fl)


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] SUCCESS: Daemon process\n"))


			// Switch commands
			switch (cmd)
			{
				case FLD_IOCTL_REQ:
	
	        			FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_REQ\n"))


					// NOTE: Get next I/O
					//
					//  The daemon received a SIGURG. Check for an
					//  empty I/O list.
					//
					//  This can happen when a SIGURG is sent from
					//  another source (non-fl module, i.e. another
					//  user process, another kernel module).


					// Acquire I/O daemon lock
					MUTEX_LOCK(io_mutex,
						   -1);


					// Get next I/O request
					if (fl_strategy_peek_io(&io_req) < 1)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: No data available\n"))


						// Set daemon state to IDLE
						daemon_state = FL_DST_IDLE;


						// Release I/O daemon lock
						MUTEX_UNLOCK(io_mutex);


						// Last check of malloc_count
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: malloc_count = %d\n",
								  malloc_count))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return ENODATA;	// ERROR: No data available
					}


					// Set daemon state to REQ
					daemon_state = FL_DST_REQ;


					// Release I/O daemon lock
					MUTEX_UNLOCK(io_mutex);


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: arg = %p\n", arg))


					// Get pointer to user I/O request
					uio_req = (struct fld_io_req_s *) arg;


					// Write back I/O request
					if (copyout((caddr_t) io_req,
                                    		    (caddr_t) uio_req,
                                    		    sizeof(struct fld_io_req_s)) != 0)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copyout I/O request\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EFAULT;	// ERROR: Couldn't copyout I/O request
					}


					// Check for write I/O
					if (io_req->control == FLD_IO_WRITE)
					{
						// NOTE: Write vs. format
						//
						//  In a write I/O request, the daemon writes data
						//  transferred from kernel space to the attached file.
						//
						//  In contrast, a format operation (not I/O) is a special
						//  write I/O request with no data transferred to user space.
						//
						//  The data, a zero filled buffer, is dynamically generated
						//  by the daemon.


						if (io_req->is_format != 1)
						{
							// Write back I/O request data field
							if (copyout((caddr_t) io_req->data,
								    (caddr_t) daemon_data,
								    io_req->data_size) != 0)
							{
								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] ERROR: Couldn't copyout I/O request %p\n",
										  &uio_req->data))
								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] INFO: FLD_IOCTL_REQ ERROR START\n"))


								// TEMP: Print out I/O request
								//fl_print_ioreq((struct fld_io_req_s *) &io_req);


								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] INFO: FLD_IOCTL_REQ ERROR STOP\n"))


								// Release lock for daemon
								fl_spin_unlock(daemon_lock);


								// Release serialization lock
								FL_SERIAL_UNLOCK();


								return EFAULT;	// ERROR: Couldn't copyout I/O request
							}
						}
					}


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


	        			return ESUCCESS;	// SUCCESS: Got next I/O request

				case FLD_IOCTL_FIN:
	
	        			FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: FLD_IOCTL_FIN START\n"))


					// Reset format has error switch
					format_has_error = ESUCCESS;


					// Copy in I/O response
					if (copyin(arg,
				   		   (void *) &io_resp,
                                    		   sizeof(struct fld_io_req_s)) != 0)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copyin I/O response\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EFAULT;	// ERROR: Couldn't copyin I/O response
					}


					// TEMP: Print out I/O response
					//fl_print_ioreq((struct fld_io_req_s *) &io_resp);


					// Get buf pointer
					bp = (buf_t *) io_resp.bp;


					// NOTE: Finish I/O


					// Check for read I/O
					switch (io_resp.control)
					{
						case FLD_IO_READ_OK:

							// NOTE: Copy in data buffer


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] FLD_IO_READ_OK START\n"))


							// Copy in I/O response data
							if (copyin(io_resp.data,
								   (void *) bp->b_un.b_addr,
								   io_resp.data_size) != 0)
							{
								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] ERROR: Couldn't copyin I/O response data\n"))


								// Release lock for daemon
								fl_spin_unlock(daemon_lock);


								// Release serialization lock
								FL_SERIAL_UNLOCK();


								return EFAULT;	// ERROR: Couldn't copyin I/O response data
							}


							// Finish I/O
							bp->b_resid = 0;
							biodone(bp);


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_READ_OK STOP\n"))


							break;

						case FLD_IO_READ_EOF:

	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_READ_EOF START\n"))


							// Finish I/O
							bp->b_resid = bp->b_bcount;
							bp->b_error = ESUCCESS;
							biodone(bp);


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_READ_EOF STOP\n"))


							break;

						case FLD_IO_READ_NOK:

	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_READ_NOK START\n"))


							// Finish I/O
							bp->b_resid = bp->b_bcount;
							bp->b_flags |= B_ERROR;
							bp->b_error = EIO;
							biodone(bp);


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_READ_NOK STOP\n"))


							break;

						case FLD_IO_WRITE_OK:

	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_OK START\n"))


							if (io_resp.is_format != 1)
							{
								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] INFO: io_resp.is_format != 1\n"))


								// Finish I/O
								bp->b_resid = 0;
								bp->b_error = ESUCCESS;
								biodone(bp);
							}


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_OK STOP\n"))


							break;

						case FLD_IO_WRITE_EOF:

	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_EOF START\n"))


							if (io_resp.is_format != 1)
                                                        {
								// Finish I/O
								bp->b_resid = bp->b_bcount;
								bp->b_error = ESUCCESS;
								biodone(bp);
							}
							else
							{
								// NOTE: NOK is equal to EFAULT


								format_has_error = EFAULT;
							}


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_EOF STOP\n"))


							break;

						case FLD_IO_WRITE_NOK:

	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_NOK START\n"))


							if (io_resp.is_format != 1)
                                                        {
								// Finish I/O
								bp->b_resid = bp->b_bcount;
								bp->b_flags |= B_ERROR;
								bp->b_error = EIO;
								biodone(bp);
							}
							else
							{
								// NOTE: NOK is equal to EIO


								format_has_error = EIO;
							}


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: FLD_IO_WRITE_NOK STOP\n"))


							break;

						default:

							// NOTE: Never reached


	        					FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] ERROR: Unknown FLD_IO_*\n"))


							// Acquire I/O daemon lock
							MUTEX_LOCK(io_mutex,
								   -1);


							// Set daemon state to IDLE
							daemon_state = FL_DST_IDLE;


							// Release I/O daemon lock
							MUTEX_UNLOCK(io_mutex);


							// Release lock for daemon
							fl_spin_unlock(daemon_lock);


							// Release serialization lock
							FL_SERIAL_UNLOCK();


							return EINVAL;	// ERROR: Unknown daemon I/O control
					}


					// Acquire I/O daemon lock
					MUTEX_LOCK(io_mutex,
						   -1);


					if (io_resp.is_format != 1)
					{
						// NOTE: Regular write I/O


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: Regular write I/O\n"))


						// Dequeue I/O request
						fl_strategy_dequeue_io(&io_req);


						// Free I/O request
						fl_mem_free(io_req);


						// Decrement pending I/O request counter
						daemon_pending--;
					}
					else
					{
						// NOTE: Format write I/O


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: Format write I/O\n"))


						// Get current I/O request
						fl_strategy_peek_io(&io_req);


						// Get loop entry for I/O request
						loop_entry = (struct loop_entry_s *) io_req->data;


						// Update disk or partition offset
						io_req->offset = io_req->offset + io_req->data_size;


						// Update total data size
						io_req->size = io_req->size - io_req->data_size;


						if ((io_req->size > 0) &&
						    (loop_entry != NULL) &&
						    (format_has_error == ESUCCESS))
						{
							// NOTE: More data to format


							FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: More data to format\n"))


							if (io_req->size < FL_MAX_TRANSFER_SIZE)
							{
								io_req->data_size = io_req->size;
							}
							else
							{
								io_req->data_size = FL_MAX_TRANSFER_SIZE;
							}


							// Requeue current I/O request
							fl_strategy_requeue_io();
						}
						else
						{
							// NOTE: Finished formatting


							FL_DEBUG_UPRINTF((\
									  "[fl_ioctl] INFO: Finished formatting\n"))


							// Dequeue I/O request
							fl_strategy_dequeue_io(&io_req);


							// Free I/O request
							fl_mem_free(io_req);


							// Decrement pending I/O request counter
							daemon_pending--;


							// NOTE: Wakeup client
							//
							//  The client is waiting on its format semaphore for
							//  the formatting to completely finish.


							if (loop_entry == NULL)
							{
								FL_DEBUG_UPRINTF((\
										  "[fl_ioctl] INFO: More data to format but client pressed Ctrl-C\n"))
							}
							else
							{
								// Acquire lock for loop entry
								fl_spin_lock(loop_entry->lock);


								// NOTE: Format error
								//
								//  An error code is always returned upon finishing
								//  formatting writes, both for successful formatting
								//  and error conditions.


								loop_entry->format_error = format_has_error;


								// NOTE: Wake up client


								// Increment format semaphore
								if (fl_sem_incr((fl_sem_t *) io_req->bp,
										1,
										FL_SEM_SWITCH_RETURN_ON_SLEEP,
										&args) < 1)
								{
									// NOTE: Tripwire for semaphore overflow
									//
									//  In case of an externally induced semaphore value
									//  overflow, deny further access to the driver and
									//  ask user to reload module.
									//
									//  A reload will reset file looper's internal state,
									//  effectively avoiding deadlocks, kernel panics, and
									//  other unexpected behavior.


									reset_switch = 1;
								}


								// Set format is finished switch
								loop_entry->format_is_finished = 1;


								// Release lock for loop entry
								fl_spin_unlock(loop_entry->lock);
							}
						}
					}


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: FLD_IOCTL_FIN: daemon_pending = %u\n",
							  daemon_pending))


					// Check for pending I/O requests
					if (daemon_pending == 0)
					{
						// Set daemon state to IDLE
						daemon_state = FL_DST_IDLE;


						// Release I/O daemon lock
						MUTEX_UNLOCK(io_mutex);


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: FLD_IOCTL_FIN STOP\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return ENODATA;	// SUCCESS: No more I/O requests
					}


					// Release I/O daemon lock
					MUTEX_UNLOCK(io_mutex);


	        			FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: FLD_IOCTL_FIN STOP\n"))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return ESUCCESS;	// SUCCESS: More I/O requests pending

				case FLD_IOCTL_START:
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_START\n"))
	
	
					// NOTE: Daemon process already started
	
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: Daemon process double entry\n"))
	
	
					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return EINVAL;	// ERROR: Daemon process already started
	
				case FLD_IOCTL_START_ARGS:
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_START_ARGS\n"))
	
	
					// NOTE: Daemon process already started
	
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: Daemon process double entry\n"))
	
	
					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return EINVAL;	// ERROR: Daemon process already started

				case FLD_IOCTL_STOP:
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_STOP\n"))


					// Reset daemon pid
					daemon_pid = 0;


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return ESUCCESS;	// SUCCESS

				default:
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] ERROR: Unknown command = %d\n",
							  cmd))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return EINVAL;	// ERROR: Unknown ioctl command
			}
		}
		else
		{
			// NOTE: User process (/dev/fl)


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: fl_proc_get()\n"))


        		// Get proc
        		if (fl_proc_get(&proc) < 1)
        		{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: Couldn't get proc\n"))


				// Release lock for daemon
				fl_spin_unlock(daemon_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


                		//return EINVAL;       // ERROR: Couldn't get proc
				return (-1 * %fl_c:*:DEVFL_IOCTL_PROC_NOT_FOUND[in,,]%) + 257;	// ERROR: Couldn't get proc
        		}


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: proc = %p proc->pid = %u\n",
					  proc,
					  proc->pid))


			// TODO: Call only for some FLD_IOCTL_*'s


			// Get fl_proc
			fl_proc = (struct fl_proc_s *) proc->priv;
	
	
	        	// Check for reset
	        	if (fl_proc->reset == FL_PROC_RESET)
	        	{
	                	FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: FL_PROC_RESET for process\n"))
	
	
				FL_TEST_F_custom_N_PINGPONG(FL_TEST_STATE_BEFORE,\
							    0,\
							    FL_TEST_NO_ARGS)
	
	
				// Release lock for daemon
				fl_spin_unlock(daemon_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


	                	FL_TEST("custom",\
					"PINGPONG")
	                	//return EINVAL;  // ERROR: Process is reset
				return (-1 * %fl_c:*:DEVFL_IOCTL_PROC_IS_RESET[in,,]%) + 257;	// ERROR: Process is reset
	        	}


			// NOTE: Authenticity checks
			//
			//  Check only for the following commands:
			//
			//    - FL_CMD_ATTACH
			//    - FL_CMD_DETACH
			//    - FL_CMD_LIST


			if ((cmd == FL_CMD_ATTACH) ||
			    (cmd == FL_CMD_DETACH) ||
			    (cmd == FL_CMD_LIST))
			{


#ifdef FL_CHECKS


				// NOTE: CPU check


#if defined(FL_TESTS) && defined(FL_TESTS_F_FL_WRITE_N_CHECK_CPU)


				// Check CPU
				if (1)


#elif defined(FL_TESTS_CHECKS)

				// Check CPU
				if (0)


#else


				// Check CPU
				if (0)


#endif


				{
					// NOTE: Clean up session


					// Reset session
					fl_session_reset(fl_proc);


					// NOTE: Check lock (see fl_init())


#if !defined(FL_TESTS)


					// NOTE: fl_ioctl()
					//
					//  Test data is retrieved through the fl_ioctl() interface. If
					//  the next lock is set in test mode, fl_ioctl() can not return
					//  any data.


					// Lock driver entry points
					fl_check_lock_switch = 1;


#endif


					// Reset check error message
					bzero((void *) fl_check_err_msg,
					      100);


					// Set check error message
					sprintf(fl_check_err_msg,
						"fl: License check failed for %d (ctype)",
						cpu);


					// Log error
					cmn_err(CE_WARN,
						"%s\n",
						fl_check_err_msg);


#ifdef FL_TESTS


					// NOTE: Failed CPU check
					//
					//  Return value and error message are set here for test.pl.
					//
					//  Unlike other OSs, in IRIX the kernel module loading mechanism
					//  (entered by calling fl_init()) does not allow to be interrupted and
					//  return an error, and therefore must always succeed.


					// Set test return value
					fl_test_ret = EACCES;


					// Set test error message
					strcpy(fl_test_err,
					       "ERROR: License check failed for cpu");


#endif


					//fl_mem_dump();


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return (-1 * %fl_c:*:CHECK_IOCTL_CPU_EACCESS[in,,]%) + 257;	// ERROR: License check failed for cpu
				}


#endif	// defined(FL_CHECKS)


			}


			// Switch commands
			switch (cmd)
			{
				case FL_CMD_ATTACH:

					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FL_CMD_ATTACH\n"))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// NOTE: Request buffer vs request packet
					//
					//  The data transferred between user and kernel
					//  space is a data buffer called request buffer.
					//
					//  A request packet is a structured representation
					//  of a request buffer and thus, allows direct access
					//  to the different parts of the buffer. It is defined
					//  as a struct fl_packet_s in fl.h.


					// Copy in user request buffer
					if (copyin(arg,
				   		   (void *) req_buffer,
                                    		   FL_PACKET_REQ_ATTACH) != 0)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copy in user request buffer\n"))


						// Release lock for fl_open
						fl_spin_unlock(fl_open_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return 513;	// ERROR: Couldn't copy in user request buffer
					}


					// Reset request packet
					bzero((void *) &packet,
					      sizeof(struct fl_packet_s));


					// Parse user request buffer
					fl_packet_parse(req_buffer,
							FL_CMD_ATTACH,
						        &packet);


					// Print user request packet
					fl_print_packet(&packet);


					// NOTE: Return value vs return code
					//
					//  A [return value] is a term describing an integral value
					//  returned by a function call to indicate status or other
					//  control information (like the result of a computation).
					//
					//  The [return code] is a term used in file looper to refer
					//  to the value returned to the user space program that issued
					//  the I/O control attach command.
					//
					//  This code is stored in the global dsf_control structure
					//  (dsf_control.ret) by the following function call, and
					//  retrieved as return value upon return (see return statement
					//  below for details).
					//
					//  The return values used on UNIX (error codes named EXYZ:
					//  ESUCCESS, EFAULT, EBUSY, EINVAL, ...) are not related to
					//  the return codes used by file looper. Return codes use a
					//  different numbering scheme.


					// Reset return code
					dsf_control.ret = 0;


					// Attach command
					ret_attach = 0;
					if (fl_cmd_attach(&ret_attach) < 1)
					{
						// NOTE: Return code
						//
						//  The fl user command still needs to read in the return code
						//  for the failed command execution.
						//
						//  Silently continue.
					}


					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					// NOTE: Return code encoding
					//
					//  In order to be as efficient as possible and avoid returning
					//  a separate return code through copyout(), the following
					//  encoding is used to return the file looper return code
					//  using standard UNIX return value scheme:
					//
					//   - Positive return codes have an assigned range of
					//     1 - 256, both inclusive
					//
					//   - Negative return codes have an assigned range of
					//     257 - 512, both inclusive
					//
					//   - I/O control return values have an assigned range of
					//     513 - 768, both inclusive


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: dsf_control.ret = %d\n",
							  dsf_control.ret))


					if (dsf_control.ret >= 1)
					{
						return dsf_control.ret;			// SUCCESS
					}
					else
					{
						return (-1 * dsf_control.ret) + 257;	// ERROR
					}

				case FL_CMD_DETACH:

					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FL_CMD_DETACH\n"))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// NOTE: Request buffer vs request packet
					//
					//  The data transferred between user and kernel
					//  space is a data buffer called request buffer.
					//
					//  A request packet is a structured representation
					//  of a request buffer and thus, allows direct access
					//  to the different parts of the buffer. It is defined
					//  as a struct fl_packet_s in fl.h.


					// Copy in user request buffer
					if (copyin(arg,
				   		   (void *) req_buffer,
                                    		   FL_PACKET_REQ_DETACH) != 0)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copy in user request buffer\n"))


						// Release lock for fl_open
						fl_spin_unlock(fl_open_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return 514;	// ERROR: Couldn't copy in user request buffer
					}


					// Reset request packet
					bzero((void *) &packet,
					      sizeof(struct fl_packet_s));


					// Parse user request buffer
					fl_packet_parse(req_buffer,
							FL_CMD_DETACH,
						        &packet);


					// Print user request packet
					fl_print_packet(&packet);


					// NOTE: Return value vs return code
					//
					//  A [return value] is a term describing an integral value
					//  returned by a function call to indicate status or other
					//  control information (like the result of a computation).
					//
					//  The [return code] is a term used in file looper to refer
					//  to the value returned to the user space program that issued
					//  the I/O control attach command.
					//
					//  This code is stored in the global dsf_control structure
					//  (dsf_control.ret) by the following function call, and
					//  retrieved as return value upon return (see return statement
					//  below for details).
					//
					//  The return values used on UNIX (error codes named EXYZ:
					//  ESUCCESS, EFAULT, EBUSY, EINVAL, ...) are not related to
					//  the return codes used by file looper. Return codes use a
					//  different numbering scheme.


					// Reset return code
					dsf_control.ret = 0;


					// Detach command
					if (fl_cmd_detach() < 1)
					{
						// NOTE: Return code
						//
						//  The fl user command still needs to read in the return code
						//  for the failed command execution.
						//
						//  Silently continue.
					}


					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					// NOTE: Return code encoding
					//
					//  In order to be as efficient as possible and avoid returning
					//  a separate return code through copyout(), the following
					//  encoding is used to return the file looper return code
					//  using standard UNIX return value scheme:
					//
					//   - Positive return codes have an assigned range of
					//     1 - 256, both inclusive
					//
					//   - Negative return codes have an assigned range of
					//     257 - 512, both inclusive


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: dsf_control.ret = %d\n",
							  dsf_control.ret))


					if (dsf_control.ret >= 1)
					{
						return dsf_control.ret;			// SUCCESS
					}
					else
					{
						return (-1 * dsf_control.ret) + 257;	// ERROR
					}

				case FL_CMD_LIST:

					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FL_CMD_LIST\n"))
		

					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// NOTE: Return value vs return code
					//
					//  A [return value] is a term describing an integral value
					//  returned by a function call to indicate status or other
					//  control information (like the result of a computation).
					//
					//  The [return code] is a term used in file looper to refer
					//  to the value returned to the user space program that issued
					//  the I/O control attach command.
					//
					//  This code is stored in the global dsf_control structure
					//  (dsf_control.ret) by the following function call, and
					//  retrieved as return value upon return (see return statement
					//  below for details).
					//
					//  The return values used on UNIX (error codes named EXYZ:
					//  ESUCCESS, EFAULT, EBUSY, EINVAL, ...) are not related to
					//  the return codes used by file looper. Return codes use a
					//  different numbering scheme.


					// Reset return code
					dsf_control.ret = 0;


					// List command
					fl_cmd_list();


					// Copy out current list
					if (copyout((caddr_t) &req_buffer,
                                    		    (caddr_t) arg,
                                    		    dsf_control.list_size) != 0)
					{
						// NOTE: Listing state
						//
						//  If the listing can't be copied out to the user,
						//  reset the internal state, so no listing gets lost.
						//
						//  For the user this means that, in the worst case,
						//  the listing starts back at the beginning when the
						//  next list command is issued.


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copy out request buffer\n"))


						// Reset internal list state
						dsf_control.list_index = 0;
						dsf_control.list_size = 0;


						// Release lock for fl_open
						fl_spin_unlock(fl_open_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return 515;	// ERROR: Couldn't copy out request buffer
					}


					// Release lock for fl_open
					fl_spin_unlock(fl_open_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					// NOTE: List return code
					//
					//  The list command only returns positive return codes.
					//
					//  See fl_cmd_list() for details.


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: dsf_control.ret = %d\n",
							  dsf_control.ret))


					return dsf_control.ret;			// SUCCESS

				case FLD_IOCTL_START:
	
					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_START\n"))


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: arg = %p\n",
							  arg))


					// Check for double entry
					if (pid == daemon_started)
					{
						// NOTE: Daemon process already started


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: PROC: Daemon process double entry\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EINVAL;	// ERROR: Daemon process already started
					}
	
	
					//TODO:FL_TEST_F_fl_ioctl_N_FLD_IOCTL_START_NOT_SUPER_USER(0, 1)
	
	
					// Check for super user
					//if (FL_TEST_COND((fl_test_ret == 1), ||)
					if (fl_user_check(0,
							  &args) < 1)
					{
						//TODO:FL_TEST_F_fl_ioctl_N_FLD_IOCTL_START_NOT_SUPER_USER(1, -1)
	
	
						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EPERM;	// ERROR: User must be root
					}


					// NOTE: fld
					//
					//  The daemon process, fld, is used to send
					//  a signal SIGURG, urgent data is available,
					//  whenever there is an I/O request waiting in
					//  the I/O queue.
					//
					//  The daemon responds by entering fl_ioctl() with
					//  command FLD_IOCTL_REQ and fetching the next
					//  available I/O request from the I/O queue.
	
	
					// NOTE: Stale daemon process
					//
					//  To avoid stale daemon processes, look for
					//  an older daemon process in proc_list and
					//  delete it.
					//
					//  This situation arises when different processes
					//  invoke FLD_IOCTL_START sequentially.
					//
					//  There can be at most one (1) other daemon
					//  process in proc_list.
	
	
					// Delete stale daemon
					//fl_ioctl_del_daemon();


					// NOTE: Daemon start arguments
					//
					//  Each daemon process must exchange a start arguments
					//  struct. This struct contains the current maxdmasz
					//  value used in I/O transfers and is necessary for
					//  the daemon process to create a correctly sized
					//  I/O transfer buffer.


					// Set current maxdmasz
					io_start_args.maxdmasz = fl_maxdmasz;


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: arg = %p\n",
							  arg))


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: fl_maxdmasz = %d\n",
							  fl_maxdmasz))


					// Copy out daemon I/O start arguments
					if (copyout((caddr_t) &io_start_args,
                                    		    (caddr_t) arg,
                                    		    sizeof(struct fld_io_start_args_s)) != 0)
					{
						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copyout daemon I/O start arguments\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EFAULT;	// ERROR: Couldn't copyout daemon I/O start arguments
					}


					// Set pid of started daemon
					daemon_started = pid;


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: arg = %p\n",
							  arg))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


	        			return ESUCCESS;	// SUCCESS: Daemon successfully started

				case FLD_IOCTL_START_ARGS:

					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: IOCTL = FLD_IOCTL_START_ARGS\n"))


					// Check for started daemon pid
					if (pid != daemon_started)
					{
						// NOTE: Daemon process not started


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EPERM;	// ERROR: Daemon process not started
					}


					// Copy in daemon I/O start arguments
					if (copyin(arg,
				   		   (void *) &io_start_args,
                                    		   sizeof(struct fld_io_start_args_s)) != 0)
					{
						// Reset started daemon pid
						daemon_started = 0;


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] ERROR: Couldn't copyin daemon I/O start arguments\n"))


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return EFAULT;	// ERROR: Couldn't copyin daemon I/O start arguments
					}


					// Check for daemon I/O transfer buffer
					if (io_start_args.io_data == NULL)
					{
						// Reset started daemon pid
						daemon_started = 0;


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return ENOMEM;	// ERROR: Couldn't allocate daemon I/O transfer buffer
					}


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: io_data = %p\n",
							  io_start_args.io_data))


					// Set daemon pid
					daemon_pid = proc->pid;


					// Check for previous daemon process ref
					if (daemon_ref != NULL)
					{
						// Unref daemon process
						proc_unref(daemon_ref);
					}


					// Set daemon process ref
					daemon_ref = proc_ref();


					// Set daemon state to IDLE
					daemon_state = FL_DST_IDLE;


					// Set daemon process state
					fl_proc->state = FL_ST_DAEMON;


					// Set daemon data in user space
					daemon_data = io_start_args.io_data;


					// Reset started daemon pid
					daemon_started = 0;


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: Proc state set to FL_ST_DAEMON\n"))


					// Reset session
					fl_session_reset(fl_proc);


					// NOTE: Check for I/O requests


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: Checking for pending I/O requests\n"))


					// Acquire I/O daemon lock
					MUTEX_LOCK(io_mutex,
						   -1);


					// Check for pending I/O requests
					if (daemon_pending == 0)
					{
						// NOTE: No pending I/O requests


						FL_DEBUG_UPRINTF((\
								  "[fl_ioctl] INFO: No pending I/O requests\n"))


						// Release I/O daemon lock
						MUTEX_UNLOCK(io_mutex);


						// Release lock for daemon
						fl_spin_unlock(daemon_lock);


						// Release serialization lock
						FL_SERIAL_UNLOCK();


						return ENODATA;	// SUCCESS: Daemon successfully started (no data available)
					}


					// NOTE: Pending I/O requests


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] INFO: Pending I/O requests\n"))


					// Release I/O daemon lock
					MUTEX_UNLOCK(io_mutex);


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


	        			return ESUCCESS;	// SUCCESS: Daemon successfully started (data available)

				default:
	
#ifndef FL_TESTS
	
	
					// NOTE: Tests
					//
					//  Test FL_TEST_F_fl_open_N_NO_LOOP_ENTRY requires this call not to be reset.
	
	
	        			// NOTE: Only one call from /dev/fl is allowed, to minimize contention
	
	
	        			// Reset session
	        			fl_session_reset(fl_proc);
	
	
#endif


					FL_DEBUG_UPRINTF((\
							  "[fl_ioctl] ERROR: Unknown command = %d\n",
							  cmd))


					// Release lock for daemon
					fl_spin_unlock(daemon_lock);


					// Release serialization lock
					FL_SERIAL_UNLOCK();


					return EINVAL;	// ERROR: Unknown ioctl command
			}
		}
	}


#if defined(FL_TESTS)


	// NOTE: Return statement
	//
	//  In test mode, this return statement must come after the preceding macro block.
	//
	//  IRIX requires the code in this macro block to run in test mode too.


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return EINVAL;	// ERROR: Unknown command


#else


	// NOTE: Disk I/O controls
	//
	//  Calls processed by the following switch come from CHR devices
	//  in /dev/loop (i.e. /dev/loop/r*).
	//
	//  These are required by mount, and others. I/O control commands
	//  must be supported for every disk device present in the
	//  system (see man dksc).


	// Switch commands
	switch (cmd)
	{
		case DIOCADDBB:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCADDBB (0x%x)\n",
					  cmd))


			// NOTE: Bad block list
			//
			//  File looper virtual disks don't have (nor need)
			//  a bad block list.


			break;

		case DIOCDRIVETYPE:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCDRIVETYPE (0x%x)\n",
					  cmd))


			// NOTE: Drive type
			//
			//  The drive type is composed of three fields:
			//
			//   - VENDOR IDENTIFICATION (8 bytes)
			//   - PRODUCT IDENTIFICATION (16 bytes)
			//   - PRODUCT REVISION LEVEL (4 bytes)
			//
			//  File looper returns a default value of:
			//
			//   - FUJITSU\0        (7 bytes, last byte is \0)
			//   - MAJ9999MC\0...\0 (9 bytes, rest is \0)
			//   - HP00             (4 bytes)


			// Write back default drive type
			if (copyout((caddr_t) default_drive_type,
                                    (caddr_t) arg,
                                    SCSI_DEVICE_NAME_SIZE) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyout(SCSI_DEVICE_NAME_SIZE)\n"))


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy out drive type
			}


			break;

		case DIOCFORMAT:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCFORMAT (0x%x)\n",
					  cmd))


			//TODO:FL_TEST_F_fl_ioctl_N_DIOCFORMAT_NOT_SUPER_USER(0, 1)


			// Check for super user
			//if (FL_TEST_COND((fl_test_ret == 1), ||)
			if (fl_user_check(0,
					  &args) < 1)
			{
				//TODO:FL_TEST_F_fl_ioctl_N_DIOCFORMAT_NOT_SUPER_USER(1, -1)


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EPERM;	// ERROR: User must be root
			}


			// Acquire lock for fl_open
			fl_spin_lock(fl_open_lock);


			// Acquire lock for loop list
			fl_spin_lock(loop_list_lock);


			// Acquire lock for loop entry
			fl_spin_lock(loop_entry->lock);


			// NOTE: Exclusive access
			//
			//  Formatting of a logical device requires exclusive
			//  access by the invoking process.
			//
			//  Technically, active pid lists for both CHR and BLK devices
			//  only contain the list of processes that were active
			//  before the last process (also in that list) closes
			//  the device file and flushes these lists.
			//
			//  In other words, it is possible for a last process to
			//  invoke the DIOCFORMAT I/O control command and fail,
			//  even though it is the exclusive active process of the
			//  device.
			//
			//  The heuristics implement formatting of a logical device
			//  if it is the only active process in the CHR process list
			//  with no other processes on any of the lists (CHR, BLK).
			//
			//  Formatting a logical disk drive may take some time, it is
			//  not an instant operation. To avoid access from other
			//  processes during a formatting operation, an exclusive switch
			//  is used on the device (in its loop entry). See fl_open() for
			//  more details.


			// NOTE: Check for exclusive access


			// Check for active BLK devices
	                if (loop_entry->blk_is_inuse == 1)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: loop_entry->blk_is_inuse == 1\n"))


				// NOTE: Release locks


				// Release lock for loop entry
				fl_spin_unlock(loop_entry->lock);


				// Release lock for loop list
				fl_spin_unlock(loop_list_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EBUSY;	// ERROR: BLK device in use
			}


			// Get pid of current process
			if (fl_proc_get_data(FL_DATA_PID,
					     &pid) < 1)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: fl_proc_get_data()\n"))


				// NOTE: Release locks


				// Release lock for loop entry
				fl_spin_unlock(loop_entry->lock);


				// Release lock for loop list
				fl_spin_unlock(loop_list_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return ENOPROC;	// ERROR: Couldn't get pid of current process
			}


			// Set pid list
			pid_list = loop_entry->pid_list_chr;


			// Get length of CHR pid list
			fl_list_length(pid_list,
				       &pid_length);
			if (pid_length != 1)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: pid_length != 1\n"))


				// NOTE: Release locks


				// Release lock for loop entry
				fl_spin_unlock(loop_entry->lock);


				// Release lock for loop list
				fl_spin_unlock(loop_list_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EBUSY;	// ERROR: CHR device in use
			}


			// Get pid node
			fl_list_get_node(pid_list,
					 &pid_node,
					 0);


			// Check for current process pid
			if ((fl_uint64_t) pid_node->data != (fl_uint64_t) pid)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: pid_node->data != pid\n"))


				// NOTE: Release locks


				// Release lock for loop entry
				fl_spin_unlock(loop_entry->lock);


				// Release lock for loop list
				fl_spin_unlock(loop_list_lock);


				// Release lock for fl_open
				fl_spin_unlock(fl_open_lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return ENOPROC;	// ERROR: Current process is not active CHR process
			}


			// NOTE: Exclusive access
			//
			//  The current process is the exclusive active process of the
			//  CHR device (and therefore of the BLK device too).


			// Set exclusive switch for device
			loop_entry->is_exclusive = 1;


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: is_exclusive = 1\n"))


			// NOTE: Finished format
			//
			//  The daemon sets the format is finished switch if it finishes
			//  the formatting writes before the client enters to finish it
			//  (for example when interrupted by the user).


			loop_entry->format_is_finished = 0;


			// NOTE: Release locks


			// Release lock for loop entry
			fl_spin_unlock(loop_entry->lock);


			// Release lock for loop list
			fl_spin_unlock(loop_list_lock);


			// Release lock for fl_open
			fl_spin_unlock(fl_open_lock);


			// NOTE: Format
			//
			//  Formatting operation can now be started on the requested
			//  device. Any attempt to access this device through CHR or BLK
			//  device files will return an error. See fl_open() for more
			//  details on the use of the is exclusive switch.
			//
			//  Detaching the device file will return EBUSY in fl_cmd_detach()
			//  because one CHR device file is opened and in use (for this
			//  I/O control invokation).


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCFORMAT FORMAT OK\n"))


			// NOTE: Formatting writes
			//
			//  A formatting I/O control command is implemented
			//  using a series of writes with zero filled buffers.
			//
			//  A special write I/O request is used that allows a series
			//  of writes to be requested without being dequeued.
			//
			//  When all necessary writes are complete, the I/O request
			//  is dequeued.
			//
			//  Technically, the special write I/O request is not static,
			//  but gets requeued to the end of the I/O queue after each
			//  full I/O request cycle to allow other I/O requests to be
			//  handled.


			// Get logical device size in bytes
			dev_size = fl_size64(dev_vertex) * FL_FILE_BLOCK;


			// NOTE: Set I/O request fields


			// Check for installation CD
			io_offset = 0;
			if (loop_entry->install_cd.is_attached == 1)
			{
				// Adjust file offset into partition
				io_offset += loop_entry->install_cd.part_offset;
			}

			io_size = dev_size;

			if (dev_size < FL_MAX_TRANSFER_SIZE)
			{
				io_data_size = dev_size;
			}
			else
			{
				io_data_size = FL_MAX_TRANSFER_SIZE;
			}


			// Acquire I/O daemon lock
			MUTEX_LOCK(io_mutex,
				   -1);


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: fl_strategy_enqueue_io()\n"))


			// Send synchronous I/O request to fld
			fl_strategy_enqueue_io(FLD_IO_WRITE,
				               (buf_t *) loop_entry->format_sem,
					       io_offset,
					       io_size,
					       dev_vertex,
					       loop_entry->file.name,
					       (char *) loop_entry,
					       io_data_size,
					       1,
					       &io_req,
					       &ret);


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: fl_strategy_enqueue_io() = %d\n",
					  ret))


			// Release I/O daemon lock
			MUTEX_UNLOCK(io_mutex);


			// NOTE: Sleep on format semaphore
			//
			//  Each client sleeps on its own format semaphore
			//  stored in the loop entry corresponding to the CHR
			//  device.
			//
			//  When formatting is finished, it is woken up by the
			//  daemon, fld.


			// Decrement format semaphore
			ret = fl_sem_decr(loop_entry->format_sem,
					  1,
					  FL_SEM_SWITCH_ENABLE_INTERRUPT,
					  &args);


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: fl_sem_decr() = %d\n",
					  ret))


			// Acquire I/O daemon lock
			MUTEX_LOCK(io_mutex,
				   -1);


			// Acquire lock for loop entry
			fl_spin_lock(loop_entry->lock);


			if (ret == 3)
			{
				// NOTE: Interrupted by Ctrl-C
				//
				//  The daemon, fld, is signalled through the I/O request
				//  in the I/O list. Whenever it is NULL, the daemon stops
				//  immediately any further formatting writes.


				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] INFO: fl_sem_decr() = 3 (Ctrl-C)\n"))


				// NOTE: Client vs. fld
				//
				//  Either the client enters first, or the daemon.
				//
				//  If the format is finished switch is set to 1,
				//  the daemon already entered and finished the format.
				//
				//  The I/O request is already dequeued at this point.


				if (loop_entry->format_is_finished != 1)
				{
					// NOTE: Client first
					//
					//  The client entered first and finished the formatting.
					//
					//  When the daemon enters the next time, it will
					//  finish the formatting too and proceed to the next
					//  I/O request.


					// Reset loop entry in I/O request
					io_req->data = NULL;
				}


				// NOTE: Return on interrupt
				//
				//  Always return ESUCCESS on interrupting the ioctl() syscall.


				// Set return code
				ret = ESUCCESS;
			}
			else
			{
				// Fill new volume header
				vh = &loop_entry->disk.vh;
				vh->vh_magic = VHMAGIC;
				vh->vh_dp.dp_secbytes = FL_FILE_BLOCK;
				vh->vh_dp._dp_heads = 1;
				vh->vh_dp._dp_sect = 32;
				vh->vh_csum = 0;


				// Set volume header partition
				vh->vh_pt[8].pt_nblks = 1;
				vh->vh_pt[8].pt_firstlbn = 0;
				vh->vh_pt[8].pt_type = PTYPE_VOLHDR;


				// NOTE: Volume size
				//
				//  The volume size must be stored in partition number 9, or
				//  at index 8 of the partition table.
				//
				//  The partition size field (measured in blocks), pt_nblks,
				//  is of type unsigned int. This means that the maximum
				//  number of blocks is 2^32 for the whole volume.
				//
				//  Block sizes for file systems in IRIX may vary from the
				//  standard 512 bytes for EFS to 16kib or more for XFS.
				//
				//  The assumption is made for the default volume header,
				//  that the block size is 512 bytes. Should this be insufficient
				//  to describe the size of the whole volume, a warning is
				//  issued to syslog and the maximum possible value is set.
				//
				//  This could happen when the associated file (and therefore
				//  the resulting logical disk drive) has a size greater than:
				//
				//   2^32 * 512 bytes = 2^32 * 2^9 = 2^41 = 2 TiB


				// NOTE: Size value
				//
				//  The file was already attached, the size value is guaranteed
				//  to be a positive value at this point. See fl_loop_new() for
				//  details.


				// Get logical device size in blocks
				dev_size_blocks = loop_entry->file.size.sectors;


				if (dev_size_blocks > (fl_uint64_t) FL_UINT_MAX_BLOCKS)
				{
					// Log warning
					cmn_err(CE_WARN, "^The volume size is greater than the maximum possible value, setting maximum possible value instead.\n");


					// Set volume size
					vh->vh_pt[10].pt_nblks = FL_UINT_MAX_BLOCKS;
				}
				else
				{
					// Set volume size
					//vh->vh_pt[10].pt_nblks = loop_entry->file.size.sectors;
					vh->vh_pt[10].pt_nblks = (fl_uint32_t) dev_size_blocks;
				}


				vh->vh_pt[10].pt_firstlbn = 1;
				vh->vh_pt[10].pt_type = PTYPE_VOLUME;


				// Calculate checksum
				vh->vh_csum = vh_checksum(vh);


				// Set return code
				ret = loop_entry->format_error;
			}


			// Reset exclusive switch for device
			loop_entry->is_exclusive = 0;


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: is_exclusive = 0\n"))


			// NOTE: Release locks


			// Release lock for loop entry
			fl_spin_unlock(loop_entry->lock);


			// Release I/O daemon lock
			MUTEX_UNLOCK(io_mutex);


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return ret;

		case DIOCGETVH:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCGETVH (0x%x)\n",
					  cmd))


			// NOTE: Volume header
			//
			//  Return the current volume header.


			// Get loop entry for device
			fl_entry_get_by_dev(dev_vertex,
					    &loop_entry,
					    &args);


			// NOTE: Locking
			//
			//  Locking is necessary for serializing I/O controls
			//  among different callers with the same CHR device
			//  opened.


			// Acquire lock for loop entry
			fl_spin_lock(loop_entry->lock);


			// Write back result
			if (copyout((caddr_t) &loop_entry->disk.vh,
                                    (caddr_t) arg,
                                    sizeof(struct volume_header)) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyout(struct volume_header)\n"))


				// Release locks
				fl_spin_unlock(loop_entry->lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy out volume header
			}


			// Release locks
			fl_spin_unlock(loop_entry->lock);


			break;

		case DIOCSETVH:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCSETVH (0x%x)\n",
					  cmd))


			// NOTE: Volume header
			//
			//  Set the current volume header.


			// Get loop entry for device
			fl_entry_get_by_dev(dev_vertex,
					    &loop_entry,
					    &args);


			// NOTE: Locking
			//
			//  Locking is necessary for serializing I/O controls
			//  among different callers with the same CHR device
			//  opened.


			// Acquire lock for loop entry
			fl_spin_lock(loop_entry->lock);


			// Read in new volume header
			if (copyin(arg,
				   (void *) &loop_entry->disk.vh,
				   sizeof(struct volume_header)) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyin(struct volume_header)\n"))


				// Release locks
				fl_spin_unlock(loop_entry->lock);


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy in volume header
			}


			// Release locks
			fl_spin_unlock(loop_entry->lock);


			break;

		case DIOCRDEFECTS:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCRDEFECTS (0x%x)\n",
					  cmd))


//			/*
//			 * This structure is used by the mode select/get ioctls to determine
//			 * where the data is to be sent to/recieved from.
//			 */
//			struct dk_ioctl_data {
//			        caddr_t i_addr;
//			        u_int   i_len;
//			        u_char  i_page;         /* Page code is used only on mode select */
//			};
//
//
//			struct defect_list {
//				struct defect_header defhdr;
//				//struct defect_entry defentry[NENTS];
//			};
//
//
//			struct defect_header {
//			        u_char  reserved0;
//			        u_char  format_bits;
//			        u_char  defect_listlen[2];
//			};
//
//
//			struct defect_entry {
//			        u_char  def_cyl[3];
//			        u_char  def_head;
//			        u_char  def_sector[4];  /* bytes from index, or sector */
//			};


			// Get dk_ioctl_data from user space
			if (copyin(arg,
                                   (void *) &defects_req,
                                   sizeof(struct dk_ioctl_data)) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyin(struct dk_ioctl_data)\n"))


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy in DIOCRDEFECTS request
			}


			// Set to NULL
			bzero((void *) &defect_list_resp,
			      sizeof(struct defect_list));


			// Write back result
			if (copyout((caddr_t) &defect_list_resp,
                                    (caddr_t) defects_req.i_addr,
                                    sizeof(struct defect_list)) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyout(struct defect_list)\n"))


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy out defects list
			}


			break;

		case DIOCREADCAPACITY:

			//  NOTE: DIOCREADCAPACITY
			//
			//   DIOCREADCAPACITY shall return a signed integer with the number
			//   of blocks (512 bytes) of the device.


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCREADCAPACITY (0x%x) from major = %u minor = %u\n",
					  cmd,
					  major,
					  minor))


			// NOTE: The BLK device has the same minor number, loop entry is the same


			// Set default return value
			return_error = ESUCCESS;


			// NOTE: Size value
			//
			//  The file was already attached, the size value must be a
			//  positive value at this point.
			//
			//  See fl_loop_new() for details.


			// Get logical device size in blocks
			dev_size_blocks = (fl_uint64_t) fl_size64(dev_vertex);


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: dev_size_blocks = %lld\n",
					  dev_size_blocks))


			// NOTE: 32 bits vs 64 bits
			//
			//  IRIX supports large files, i.e. files with a size
			//  greater than 2^31 bytes. The documentation specifies
			//  DIOCREADCAPACITY to return an unsigned integer, therefore
			//  not allowing to return size values for large files bigger
			//  than INT_MAX * 512.
			//
			//  In this case, truncate size value and return 2^31 (the
			//  maximum possible value).
			//
			//  Indicate this condition by returning EFAULT.


			// Check for overflow
			if (dev_size_blocks > (fl_int64_t) INT_MAX)
			{
				// NOTE: Truncate device size


				dev_size_blocks = (fl_uint64_t) INT_MAX;


				// Set return value to EFAULT
				return_error = EFAULT;
			}


			// NOTE: Device size
			//
			//  Device size in blocks is greater or equal to zero
			//  and fits in an unsigned integer.
			//
			//  See above checks.


			// Set final device size in blocks
			dev_size_uint = (fl_uint32_t) dev_size_blocks;


			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: arg = %p\n",
					  arg))


			// Write back result
			if (copyout((caddr_t) &dev_size_uint,
                                    (caddr_t) arg,
                                    sizeof(fl_uint32_t)) != 0)
			{
				FL_DEBUG_UPRINTF((\
						  "[fl_ioctl] ERROR: copyout()\n"))


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EFAULT;	// ERROR: Couldn't copy out device size
			}


			return return_error;

		case DIOCSENSE:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCSENSE (0x%x)\n",
					  cmd))


//			struct dk_ioctl_data {
//			        caddr_t i_addr;
//			        u_int   i_len;
//			        u_char  i_page;         // Page code is used only on mode select //
//			};


//			struct mode_sense_data {
//			        u_char  sense_len;
//			        u_char  mediatype;
//			        u_char  wprot:1, reserv0:2, dpofua:1, reserv1:4;
//			        u_char  bd_len;
//			        u_char  block_descrip[8];       // note that this field will NOT be supplied
//			                by some drives, so that dk_pages data may actually start at this
//			                offset;  check the bd_len value! //
//			        union   dk_pages dk_pages;
//			};


			break;

		case DIOCSELECT:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCSELECT (0x%x)\n",
					  cmd))


			//TODO:FL_TEST_F_fl_ioctl_N_DIOCSELECT_NOT_SUPER_USER(0, 1)


			// Check for super user
			//if (FL_TEST_COND((fl_test_ret == 1), ||)
			if (fl_user_check(0,
					  &args) < 1)
			{
				//TODO:FL_TEST_F_fl_ioctl_N_DIOCSELECT_NOT_SUPER_USER(1, -1)


				// Release serialization lock
				FL_SERIAL_UNLOCK();


				return EPERM;	// ERROR: User must be root
			}


			break;

		case DIOCTEST:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCTEST (0x%x)\n",
					  cmd))


			// NOTE: Virtual device tests
			//
			//  Virtual device specific tests can be added here.


			break;

		case DIOCSELFLAGS:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: DIOCSELFLAGS (0x%x)\n",
					  cmd))


			break;

		default:

			FL_DEBUG_UPRINTF((\
					  "[fl_ioctl] INFO: UNKNOWN COMMAND (cmd = %x = %d)\n",
					  cmd,
					  cmd))


			// Release serialization lock
			FL_SERIAL_UNLOCK();


			return EINVAL;	// ERROR: Command not implemented


	}


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	return ESUCCESS;	// SUCCESS


#endif	// FL_TESTS // !FL_TESTS


}


%FUNC:fl_ioctl:STOP%


#               endif  // FL_OS_RELEASE_V_30
#       endif  // FL_OS_VERSION_V_6_5
#endif  // FL_OS_NAME_V_IRIX


/******************************************************************************
 *
 * Name:
 *   fl_strategy
 *
 * Function:
 *   Performs all block-oriented read and write operations.
 *
 * Formal parameters:
 *   bp - Address of a buf structure
 *
 * Modified parameters:
 *   When an error occurs, the routine initializes members of the
 *   buf structure as follows:
 *
 *   b_resid - Number of bytes not transferred
 *   b_flags - B_ERROR
 *   b_error - Error status value
 *
 * Implicit input:
 *   None
 *
 * Implicit output:
 *   None
 *
 * Return value:
 *   None
 *
 * Caller:
 *   The physio routine for character I/O operations
 *   The kernel through the switch tables for block I/O operations
 *
 * Side effects:
 *   None
 *
 ******************************************************************************/


void fl_strategy(struct buf *bp)	// -- IN
{
	dev_t dev;
	dev_t loop_dev;
	fl_int32_t ret;
	fl_int64_t offset;
	fl_uint32_t page_io;
	fl_uint32_t major;
	fl_uint32_t minor;
	fl_uint32_t count;
	fl_int3264_t max_blks;
	fl_uint64_t file_size;
	struct uio uio;
	struct loop_entry_s *loop_entry;
	struct fl_args_s args;
	vertex_hdl_t dev_vertex;


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	void *null_pointer;


#if defined(FL_DEBUG)


	fl_uint3264_t dbgpid;


#endif	// defined(FL_DEBUG)


#               endif
#       endif
#endif


	FL_TEST_TYPE_DECL


	// NOTE: Initialization


	FL_TEST_TYPE_INIT


	// Args
	bzero((void *) &args,
	      sizeof(struct fl_args_s));


	// NULL pointer
	null_pointer = NULL;


	// User I/O
	bzero((void *) &uio,
	      sizeof(struct uio));


	// Set uio fields
        uio.uio_segflg = UIO_SYSSPACE;


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// NOTE: Processing


        FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: START\n"))


        FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: Mode: %s\n",
			  (bp->b_flags & B_READ) == 1 ? "B_READ" : "B_WRITE"))


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#elif FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Get device number
	dev = bp->b_edev;


	// Get pid of current process
	FL_DEBUG_DRV_GETPARM(PPID,\
			     &dbgpid)
	//drv_getparm(PPID,
		    //&dbgpid);


	// Set device vertex
        dev_vertex = (vertex_hdl_t) dev;


	// Map hardware graph to bitfield device number
	if (fl_dev_map(dev_vertex,
		       OTYP_BLK,
		       &loop_entry,
		       &dev,
		       &major,
		       &minor,
		       &args) < 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_strategy] ERROR: Couldn't map device numbers for vertex = %u dev = %u type = %s pid = %llu\n",
				  dev_vertex,
				  dev,
				  "BLK",
				  dbgpid))


                bp->b_resid = bp->b_bcount;
                biodone(bp);


		return;	// ERROR: Couldn't map device numbers
	}


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: dev_vertex = %u loop_entry = %p dev = %u major = %u minor = %u type = %s pid = %llu\n",
			  dev_vertex,
			  loop_entry,
			  dev,
			  major,
			  minor,
			  "CHR",
			  dbgpid))


	// Set loop entry device
	loop_dev = dev_vertex;


#               endif
#       endif
#endif


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	FL_DEBUG_LOCK_SWITCH_STRAT(EIO)


#               endif
#       endif
#endif


#if FL_OS_NAME_V_OSF1
#       if FL_OS_VERSION_V_5_1
#               if FL_OS_RELEASE_V_B


#               endif
#       endif
#endif


	// Acquire serialization lock
	FL_SERIAL_LOCK();


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: Get loop entry for loop_dev = %u and dev = (%d, %d)\n",
			  loop_dev,
			  major,
			  minor))


	// NOTE: Loop list lock
	//
	//  Inside fl_strategy(), locking is not necessary because it
	//  is protected by the pid lists storing at least one
	//  active process.


	// NOTE: Loop entry
	//
	//  This entry point is entered from kernel code (buffer cache, file system, ...)
	//  or through physiock() from fl_read() and fl_write().
	//
	//  Loop entry must exist, since it was called for an active file looper BLK.


	// Get loop entry for device
	fl_entry_get_by_dev(loop_dev,
			    &loop_entry,
			    &args);


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: loop_entry = %p\n",
			  loop_entry))


	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] INFO: Associated device (%u , %u): %s\n",
			    //major,
			    //minor,
			    //loop_entry->dev.name))


	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] INFO: Associated file (%u , %u): %s\n",
			    //major,
			    //minor,
			    //loop_entry->file.name))


	FL_TEST_F_fl_strategy_N_OFFSET_OUT_OF_RANGE(0, 1)


	// NOTE: Check for range
	//
	//  This assumes a b_blkno field type of signed long.
	//
	//  Any strategy function can be called from any driver. Thus,
	//  values in a buf must be sanity checked for:
	//
	//   1. Offset to be accessed, greater or equal to the max size
	//   2. Negative blk number value in buf (b_blkno)
	//   3. Overflow of the resulting offset
	//
	//  Any of these cases return an EOF condition.


	// Get max possible blks in blk offset
	max_blks = LONG_MAX / FL_FILE_BLOCK;


	// Get offset in bytes
	offset = bp->b_blkno * FL_FILE_BLOCK;


	// Get count in bytes
	count = bp->b_bcount;


	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] INFO: max_blks = %lld\n",
			    //max_blks))


	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] INFO: offset = %lld\n",
			    //offset))


	// NOTE: Raw EFS filesystems
	//
	//  When attaching raw EFS filesystems, the size reported by
	//  fl_size{64}() is relative to the logical file consisting of
	//  the virtual volume header and the actual attached file.
	//
	//  Thus, the resulting logical file is larger than the original
	//  attached file.
	//
	//  In any other case, the logical file and the actual attached
	//  file are the same.


	// Get size of logical file
	file_size = loop_entry->file.vap->va_size;


	if (FL_TEST_COND((fl_test_ret == 1), ||)
	    (offset >= file_size) ||
	    (bp->b_blkno < 0) ||
            (bp->b_blkno >= max_blks))
        {
		FL_DEBUG_UPRINTF((\
				  "[fl_strategy] SUCCESS: EOF\n"))


		// NOTE: EOF
		//
		//  For read operations (B_READ), POSIX mandates to return no error.
		//
		//  So just return normally without changing bp->b_resid to signal
		//  an EOF condition to upper layers.


                bp->b_resid = bp->b_bcount;
                biodone(bp);


		FL_TEST_F_fl_strategy_N_OFFSET_OUT_OF_RANGE(1, -4)


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		FL_TEST("fl_strategy", "OFFSET_OUT_OF_RANGE")
		return;	// SUCCESS: EOF
        }


	// Check for paged I/O
	page_io = (bp->b_flags & B_PAGEIO) != 0;
	if (page_io == 1)
	{
		// NOTE: Paged I/O
		//
		//  Map pages to virtual address space for I/O.
		//
		//  Pages are not mapped out in this code path because of a bug
		//  in biomove() which does not handle correctly offsets when
		//  a buf with B_PAGEIO is entered after a bp_mapout().


		bp_mapin(bp);
	}


	// NOTE: Check for NULL buffer address
	//
	//  File looper implements buf I/O using mapped b_un.d_addr pointer to
	//  buffers and page I/O using b_pages (mapped above).
	//
	//  To prevent fault drivers or kernel code to request I/O to
	//  a NULL target (B_WRITE) or source (B_READ) address, check for
	//  any one of these I/O methods and return immediately in any other case.
	//
	//  Non-handling of this situation causes a kernel panic in
	//  subsequent code (fl_file_read()/fl_file_write()).


	if (bp->b_un.b_addr == NULL)
	{
		// Print request
		//fl_print_uio(&uio);
		//fl_print_buf(bp);


		FL_DEBUG_UPRINTF((\
				  "[fl_strategy] SUCCESS: Buffer address NPE\n"))


                bp->b_resid = bp->b_bcount;
                biodone(bp);


		//TODO:FL_TEST_F_fl_strategy_N_BUFADDRESS_IS_NULL(1, -4)


		// Release serialization lock
		FL_SERIAL_UNLOCK();


		FL_TEST("fl_strategy", "BUFADDRESS_IS_NULL")
		return;	// SUCCESS: Buf address NPE
	}


	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] bp->b_un.b_addr = %p\n", bp->b_un.b_addr))
	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] bp->b_pages = %p\n", bp->b_pages))
	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] bp->b_blkno = %u\n", bp->b_blkno))
	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] uio_offset = %lu\n", bp->b_blkno * FL_FILE_BLOCK))
	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] iov.iov_base = %p\n", iov.iov_base))
	//FL_DEBUG_UPRINTF((\
			    //"[fl_strategy] iov.iov_len = %llu\n", iov.iov_len))


	//TODO:FL_TEST_F_fl_strategy_N_READ_MODE_ONLY(0, 0)


	// Check for read operation
	if ((bp->b_flags & B_READ) == 1)
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_strategy] INFO: Reading from attached file\n"))
	}
	else
	{
		FL_DEBUG_UPRINTF((\
				  "[fl_strategy] INFO: Writing to attached file\n"))
	}


#if FL_OS_NAME_V_IRIX
#       if FL_OS_VERSION_V_6_5
#               if FL_OS_RELEASE_V_30


	// Check for installation CD
	if (loop_entry->install_cd.is_attached == 1)
	{
		// NOTE: Installation CD


		// Adjust file offset into partition
		offset += loop_entry->install_cd.part_offset;
	}


#               endif
#       endif
#endif


	// NOTE: Maximum transfer size
	//
	//  The maximum transfer size depends on the value of the maxdmasz
	//  kernel variable. By default, maxdmasz is 1025 for IRIX32 kernels
	//  and 257 for IRIX64 kernels. Its value specifies number of pages.
	//
	//  Filesystem code for XFS, por example, calculates the maximum
	//  transfer size using the formula:
	//
	//    max_transfer = (maxdmasz - 1) << page_shift (12 for IRIX32,
	//                                                 14 for IRIX64)
	//  
	//  This yields a maximum transfer size of 4MiB for IRIX32 kernels
	//  and 16MiB for IRIX64 with the default maxdmasz value of 1025 (IRIX32)
	//  and 257 (IRIX64), respectively.
	//
	//  File looper uses this formula to preallocate the I/O transfer buffer
	//  in fld.
	//
	//  On a sidenote: The maxdmasz kernel variable can't be changed
	//  dynamically, i.e. in a running kernel. A reboot is necessary.


	// Acquire I/O daemon lock
	MUTEX_LOCK(io_mutex,
		   -1);


	// Enqueue I/O request
	fl_strategy_enqueue_io(((bp->b_flags & B_READ) == 1) ? FLD_IO_READ : FLD_IO_WRITE,
			       bp,
			       offset,
			       count,
			       bp->b_edev,
			       loop_entry->file.name,
			       bp->b_un.b_addr,
			       count,
			       0,
			       (struct fld_io_req_s **) &null_pointer,
			       &ret);


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: fl_strategy_enqueue_io() = %d\n",
			  ret))


	// Release I/O daemon lock
	MUTEX_UNLOCK(io_mutex);


	// Release serialization lock
	FL_SERIAL_UNLOCK();


	FL_DEBUG_UPRINTF((\
			  "[fl_strategy] INFO: STOP\n"))


	return;
}


%GLOB:fl_c:STOP%


